// Programm zur Ermittlung von Gütefaktoren bei Algorithmus 2M1
// (1) Es werden 2 Module A und C mit den Gütefaktoren GA und GC zu
//     einem Modul D vereinigt. Berechne den garantierten
//     durchschnittlichen Gütefaktor GD für das Modul D. Das Ergebnis
//     wird als Tabelle (Zeile GA, Spalte GC, Kreuzungspunkt GD)
//     zurückgegeben.
// (2) Es werden 2 Module A und C mit den Gütefaktoren GA und GC zu
//     einem Modul D parallel vereinigt. Dabei ergibt sich der
//     garantierte durchschnittliche Gütefaktor GD für das Modul D.
//     Finde GA = GC, so daß GD = GA ist (Fixpunkt).
// (3) Es werden M Module auf N Prozessoren mit einem zufälligen serien-
//     parallelen Modulabhängigkeitsgraphen (Wahrscheinlichkeit für eine
//     serielle Vereinigung ist VERHAELTNIS) ausgeführt. Berechne den
//     garantierten durchschnittlichen Gütefaktor des Endergebnisses.
//
// Fehler: 1->konnte Ergebnisdatei nicht erstellen / Schreibfehler
//         2->Ungültige Modulanzahl bei (3)

#include <float.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "../prgx/save.h"
#include "../prgx/zufall.h"


#ifdef EXPONENTIAL

// Berechnet eine Umrechnungstabelle um von gleichverteilten Zahlen
// zu exponentialverteilten zu gelangen
void berechne_verteilung( double* zufall )
{
  int i;
  for( i=1; i<SCHRITT; i++ )
  {
    zufall[i] = -log( 1.0 - ((double)(i)) / ((double)(SCHRITT)) );
  }
}

#else // EXPONENTIAL ist nicht gesetzt

// Berechnet eine Umrechnungstabelle um von gleichverteilten Zahlen
// zu normalverteilten zu gelangen
void berechne_verteilung( double* zufall )
{
  double a, d, s, t, x, last;
  int i, j;
  const double w2pi = 2.506628274631; // Wurzel aus 2*PI
  // Integral( q(t+ALPHA)dt, -ALPHA, 0 )
  t = 0;
  for( j=1; j<INTEG; j++ )
  {
    a = ((double)(ALPHA)) * ((double)(j)) / ((double)(INTEG));
    t += exp( - a * a /
              ( 2.0 * ((double)(SIGMA)) * ((double)(SIGMA)) ) );
  }
  t *= 2;
  t += 1; // a = 0;
  a = ((double)(ALPHA));
  t += exp( - a * a / ( 2.0 * ((double)(SIGMA)) * ((double)(SIGMA)) ) );
  t /= ((double)(SIGMA)) * w2pi; // ausgeklammerter Vorfaktor
  t /= 2.0 * ((double)(INTEG));
  t *= ((double)(ALPHA));
  zufall[0] = EPSILON; // Startwert für Newton
  for( i=1; i<SCHRITT; i++ )
  {
    d = 1; // falls Startwert zu klein
    x = ((double)(i)) / ((double)(SCHRITT));
    do
    {
      // neuer Newton-Startwert aus letztem Ergebnis ableiten
      zufall[i] = zufall[i-1] * d;
      last = DBL_MAX;
      do
      {
        // Integral( q(t)dt, 0, zufall[i] )
        s = 0;
        for( j=1; j<INTEG; j++ )
        {
          a = ( zufall[i] * ((double)(j)) /
                ((double)(INTEG)) - ((double)(ALPHA)) );
          s += exp( - a * a /
                    ( 2.0 * ((double)(SIGMA)) * ((double)(SIGMA)) ) );
        }
        s *= 2;
        a = - ((double)(ALPHA));
        s += exp( - a * a /
                  ( 2.0 * ((double)(SIGMA)) * ((double)(SIGMA)) ) );
        a = zufall[i] - ((double)(ALPHA));
        s += exp( - a * a /
                  ( 2.0 * ((double)(SIGMA)) * ((double)(SIGMA)) ) );
        s /= ((double)(SIGMA)) * w2pi; // ausgeklammerter Vorfaktor
        s /= 2.0 * ((double)(INTEG));
        s *= zufall[i];
        // ERGEBNIS
        s -= x * ( 0.5 + t );
        a = zufall[i] - ((double)(ALPHA));
        a = exp( - a * a /
                 ( 2.0 * ((double)(SIGMA)) * ((double)(SIGMA)) ) );
        if( a == 0 ) // Division durch Null vermeiden
        {
          d *= 1.1; // anderer Startwert
          a = -1;
          break;
        }
        a /= ((double)(SIGMA)) * w2pi;
        s /= a;
        a = ( s > 0 ) ? s : -s;
        if( a > last ) // keine Konvergenz
        {
          d *= 1.1; // anderer Startwert
          a = -1;
          break;
        }
        last = a;
        zufall[i] -= s;
      } while( a > EPSILON );
    } while ( a < 0 );
  }
}
#endif


#if defined( ENDERGEBNIS ) || defined( SERIELL )
#ifdef SEQUENTIELL

// Berechnet die den Gütefaktor GD bei der seriellen Vereinigung der
// Module A und C mit den Gütefaktoren GA und GC zu Modul D bei der
// Rückführung auf sequentielle Laufzeiten
double vereinige_seriell( double ga, double gc, double* zufall )
{
  double summe;
  int i, j;
  summe = 0.0;
  for( i=1; i<SCHRITT; i++ )
  {
    for( j=1; j<SCHRITT; j++ )
    {
      if( zufall[j] > zufall[i] )
      {
        summe += ( ( zufall[j] * gc + zufall[i] ) /
                   ( zufall[j] + zufall[i] * gc ) );
      }
      else
      {
        summe += ( ( zufall[i] * ga + zufall[j] ) /
                   ( zufall[i] + zufall[j] * ga ) );
      }
    }
  }
  // i = SCHRITT exoder j = SCHRITT
  summe += ( SCHRITT - 1 ) * ( ga + gc );
  // i = SCHRITT und j = SCHRITT
  summe++;
  // normieren
  summe /= ((double)(SCHRITT)) * ((double)(SCHRITT));
  return summe;
}

#else // SEQUENTIELL nicht gesetzt

// Berechnet die den Gütefaktor GD bei der seriellen Vereinigung der
// Module A und C mit den Gütefaktoren GA und GC zu Modul D bei der
// Rückführung auf parallele Laufzeiten
double vereinige_seriell( double ga, double gc, double* zufall )
{
  double summe;
  int i, j;
  summe = 0.0;
  for( i=1; i<SCHRITT; i++ )
  {
    for( j=1; j<SCHRITT; j++ )
    {
      summe += ( ( ( zufall[i] + zufall[j] ) * ga * gc ) /
                 ( gc * zufall[i] + ga * zufall[j] ) );
    }
  }
  // i = SCHRITT exoder j = SCHRITT
  summe += ( SCHRITT - 1 ) * ( ga + gc );
  // i = SCHRITT und j = SCHRITT
  summe += 2 * ga * gc / ( ga + gc );
  // normieren
  summe /= ((double)(SCHRITT)) * ((double)(SCHRITT));
  return summe;
}
#endif // SEQUENTIELL gesetzt oder nicht


#ifndef ENDERGEBNIS
#define vereinige( ga, gc, zufall )                                    \
  vereinige_seriell( ga, gc, zufall )
#endif

#endif // SERIELL oder ENDERGEBNIS gesetzt


#if defined( ENDERGEBNIS ) || !defined( SERIELL )
#ifdef SEQUENTIELL

// Berechnet die den Gütefaktor GD bei der parallelen Vereinigung der
// Module A und C mit den Gütefaktoren GA und GC zu Modul D bei der
// Rückführung auf sequentielle Laufzeiten
double vereinige_parallel( double ga, double gc, double* zufall )
{
  double summe, min1, min2, da, dc, s;
  int i, j;
  summe = 0.0;
  for( i=1; i<SCHRITT; i++ )
  {
    for( j=1; j<SCHRITT; j++ )
    {
      s = zufall[i] + zufall[j];
      da = zufall[i] / s;
      min1 = da;
      if( ga < min1 )
      {
        min1 = ga;
      }
      dc = zufall[j] / s;
      min2 = dc;
      if( gc < min2 )
      {
        min2 = gc;
      }
      min1 += min2; // erster Term
      if( da > dc ) // Maximum suchen (2. und 3. Term)
      {
        if( da > min1 )
        {
          summe += min1;
        }
        else
        {
          summe += da;
        }
      }
      else
      {
        if( dc > min1 )
        {
          summe += min1;
        }
        else
        {
          summe += dc;
        }
      }
    }
  }
  // i = SCHRITT exoder j = SCHRITT
  summe += ( SCHRITT - 1 ) * ( ga + gc );
  // i = SCHRITT und j = SCHRITT
  if( ga + gc > 0.5 )
  {
    summe += 0.5;
  }
  else
  {
    summe += ga + gc;
  }
  // normieren
  summe /= ((double)(SCHRITT)) * ((double)(SCHRITT));
  return summe;
}

#else // SEQUENTIELL nicht gesetzt

// Berechnet die den Gütefaktor GD bei der parallelen Vereinigung der
// Module A und C mit den Gütefaktoren GA und GC zu Modul D bei der
// Rückführung auf parallele Laufzeiten
double vereinige_parallel( double ga, double gc, double* zufall )
{
  double summe, pa, pc, k, s, z, min1, min2, min3;
  int i, j;
  summe = 0.0;
  for( i=1; i<SCHRITT; i++ )
  {
    for( j=1; j<SCHRITT; j++ )
    {
      s = zufall[i] + zufall[j];
      k = ga * gc;
      pa = ga * zufall[j];
      pc = gc * zufall[i];
      min1 = zufall[i] / s;
      if( ga < min1 )
      {
        min1 = ga;
      }
      z = pc / zufall[j];
      if( z < min1 )
      {
        min1 = z;
      }
      min2 = zufall[j] / s;
      if( gc < min2 )
      {
        min2 = gc;
      }
      z = pa / zufall[i];
      if( z < min2 )
      {
        min2 = z;
      }
      min1 += min2;
      z = zufall[i] / ( zufall[i] + pa );
      min2 = pc / pa;
      if( min2 > z )
      {
        min2 = z;
      }
      z = zufall[j] / ( zufall[j] + pc );
      min3 = pa / pc;
      if( min3 > z )
      {
        min3 = z;
      }
      if( min2 < min3 ) // Maximum suchen
      {
        if( min1 < min3 )
        {
          summe += min1;
        }
        else
        {
          summe += min3;
        }
      }
      else
      {
        if( min1 < min2 )
        {
          summe += min1;
        }
        else
        {
          summe += min2;
        }
      }
    }
  }
  // i = SCHRITT exoder j = SCHRITT
  summe += ( SCHRITT - 1 ) * ( ga + gc );
  // i = SCHRITT und j = SCHRITT
  min1 = gc / ga;
  z = 1 / ( 1 + ga );
  if( z < min1 )
  {
    min1 = z;
  }
  min2 = ga / gc;
  z = 1 / ( 1 + gc );
  if( z < min2 )
  {
    min2 = z;
  }
  if( min2 > min1 )
  {
    min1 = min2;
  }
  if( gc < ga )
  {
    if( 2 * gc < min1 )
    {
      summe += 2 * gc;
    }
    else
    {
      summe += min1;
    }
  }
  else
  {
    if( 2 * ga < min1 )
    {
      summe += 2 * ga;
    }
    else
    {
      summe += min1;
    }
  }
  // normieren
  summe /= ((double)(SCHRITT)) * ((double)(SCHRITT));
  return summe;
}
#endif // SEQUENTIELL gesetzt oder nicht


#ifndef ENDERGEBNIS
#define vereinige( ga, gc, zufall )                                    \
  vereinige_parallel( ga, gc, zufall )
#endif

#endif // SERIELL nicht gesetzt oder ENDERGEBNIS gesetzt


#ifdef TABELLE

// Hauptroutine für (1)
int main()
{
  FILE* fi;
  int i, j;
  double a, c;
  double zufall[ SCHRITT ];
  berechne_verteilung( zufall );
  fi = fopen( FILENAME, "w" );
  if( fi == NULL )
  {
    return 1;
  }
  for( i=1; i<=GROESSE_A; i++ )
  {
    a = ( ((double)(AB_A)) +
          ((double)(i)) * ((double)(LAENGE_A)) /
          ((double)(GROESSE_A)) );
    for( j=1; j<=GROESSE_C; j++ )
    {
      c = ( ((double)(AB_C)) +
            ((double)(j)) * ((double)(LAENGE_C)) /
            ((double)(GROESSE_C)) );
      if( save_double_s( &fi, vereinige( a, c, zufall ) ) != 0 )
      {
        return 1;
      }
    }
    if( save_ret( &fi ) != 0 )
    {
      return 1;
    }
  }
  fclose( fi );
  return 0;
}
#endif


#ifdef FIXPUNKT

// Hauptroutine für (2)
int main()
{
  double a, c;
  double zufall[ SCHRITT ];
  FILE* fi;
  c = 0.5;
  berechne_verteilung( zufall );
  do
  {
    a = c;
    c = vereinige( a, a, zufall );
  } while( fabs( c - a ) > EPSILON );
  fi = fopen( FILENAME, "w" );
  if( fi == NULL )
  {
    return 1;
  }
  if( save_double_r( &fi, a ) != 0 )
  {
    return 1;
  }
  fclose( fi );
  return 0;
}
#endif


#ifdef ENDERGEBNIS

// ein Plausi
#if ( M_MIN != M_MAX ) && ( VERHAELTNIS_GROESSE != 1 )
#error \
  "Es ist nur möglich die Modulanzahl oder das Verhältnis zu variieren"
#endif


// Berechnet die Anzahl der möglichen Bäume mit bis zu m Blättern
void berechne_baum( double* baum, int m )
{
  int i, j;
  baum[0] = 1.0;
  for( i=1; i<m; i++ )
  {
    baum[i] = 0.0;
    for( j=i-1; j>=i/2; j-- )
    {
      baum[i] += baum[j] * baum[ i - 1 - j ];
    }
  }
}


// Berechnet die Ergebnisgüte von 2M1 bei m Blättern und einer
// Wahrscheinlichkeit von v für eine serielle Vereinigung
double guete_baum( double* baum, double* zufall, int m, double v )
{
  double c, d;
  int i;
  if( m == 0 )
  {
    return 1.0 / ((double)(N)); // bei einem Modul wird Güte 1 erzeugt
  }
  // Anzahl der Module in den Unterbäumen (rechts / links) bestimmen
  c = gleich_double( 0, 1 );
  c *= baum[ m ];
  i = m / 2;
  d = baum[i] * baum[ m - 1 - i ];
  while( d < c )
  {
    i++;
    d += baum[i] * baum[ m - 1 - i ];
  }
  c = guete_baum( baum, zufall, i, v );
  d = guete_baum( baum, zufall, m - 1 - i, v );
  if( gleich_double( 0, 1 ) < v )
  {
    c = vereinige_seriell( c, d, zufall );
  }
  else
  {
    c = vereinige_parallel( c, d, zufall );
  }
  return c;
}


// Hauptroutine für (3)
int main()
{
  double c;
  double zufall[ SCHRITT ];
  double baum[ M_MAX + 1 ];
  int i, m;
#if M_MIN != M_MAX
  int m_alt;
#endif
  int v_cnt;
  double v;
  FILE* fi;
  berechne_verteilung( zufall );
  fi = fopen( FILENAME, "w" );
  if( fi == NULL )
  {
    return 1;
  }
  m = M_MIN;
  v = VERHAELTNIS_AB;
  v_cnt = 0;
  while( m <= M_MAX && v_cnt < VERHAELTNIS_GROESSE )
  {
    berechne_baum( baum, m );
    c = 0.0;
    for( i=0; i<DURCHLAEUFE; i++ )
    {
      c += guete_baum( baum, zufall, m - 1, v );
    }
    c /= (double)(DURCHLAEUFE);
    // Speichere Ergebniszeile
    // -----------------------
#if M_MIN != M_MAX // Iteration über M
    if( save_int_s( &fi, m ) != 0 )
    {
      return 1;
    }
#endif
#if VERHAELTNIS_GROESSE != 1 // Iteration über das Verhältnis
    if( save_double_s( &fi, v ) != 0 )
    {
      return 1;
    }
#endif
    if( save_double_r( &fi, c ) != 0 )
    {
      return 1;
    }
    // Bereite nächsten Schleifendurchlauf vor
    // ---------------------------------------
#if M_MIN != M_MAX
    // Berechne nächstes M gemäß der Reihe 1,2,5,10,20,50,...
    m_alt = m;
    m = 1;
    while( 0 == m_alt % 10 )
    {
      m_alt /= 10;
      m *= 10;
    }
    switch( m_alt )
    {
      case 1:
        m *= 2;
        break;
      case 2:
        m *= 5;
        break;
      case 5:
        m *= 10;
        break;
      default:
        return 2;
    }
#endif
#if VERHAELTNIS_GROESSE != 1
    // Iteration über das Verhältnis von parallelen zu seriellen
    // Vereinigungen
    v += VERHAELTNIS_LAENGE / ((double)(VERHAELTNIS_GROESSE - 1));
    v_cnt ++;
#endif
  }
  fclose( fi );
  return 0;
}
#endif
