// Routine zum Erzeugen von Beispielen:
//   kein Abhängigkeitsgraph, moldable
//
// Aufruf:
// baue_bsp_datei(
//   char* Filename_für_Ergebnis,
//   char* Filename_Eingabedatei );
//
// oder:
// baue_bsp_array(
//   double** Feld_der_Größe_m*n (Speicher muß schon reserviert sein),
//   int*     m (wird belegt),
//   int*     n (wird belegt),
//   char*    Filename_Eingabedatei );
//
// Die Standardbeispiele werden mit
// baue_standard_bsp(
//   double** Feld_der_Größe_m*n (Speicher muß schon reserviert sein),
//   int      Prozessorenanzahl,
//   char*    Filename_Eingabedatei );
// erzeugt.
//
// Fehler: 1->zu wenig Speicher
//         2->konnte Ergebnisdatei nicht anlegen
//         3->nicht genug Einträge im Listenfeld / Fehler in Datei
//         4->konnte Eingabedatei nicht finden


#include <limits.h>
#include <malloc.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "../prgx/graphen.h"
#include "../prgx/heap.h"
#include "../prgx/load.h"
#include "../prgx/save.h"
#include "../prgx/suchsort.h"
#include "../prgx/zufall.h"


int baue_bsp_datei(
  char* fn,
  char* fin )
{
  int anz, h;
  double* ant;
  int i, n, arten;
  double lz, az, d;
  int pos;
  verteilung_int anzahl;
  verteilung_double vert;
  FILE* fi;
  FILE* fiin;
  fiin = fopen( fin, "r" );
  if( fiin == NULL )
  {
    return 4;
  }
  // Prozessorenanzahl einlesen
  if( lese_next_int_r( &fiin, &n ) != 0 )
  {
    return 3;
  }
  // zu erzeugende Modulanzahlverteilung einlesen
  if( lese_next_verteilung_int_r( &fiin, &anzahl ) != 0 )
  {
    return 3;
  }
  anz = zufall_int( &anzahl ); // zu erzeugende Modulanzahl
  // Verteilungsarten zählen
  pos = ftell( fiin );
  if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
  { // eine muß da sein!
    return 3;
  }
  arten = 1;
  while( 1 )
  {
    if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
    {
      break;
    }
    arten++;
  }
  // Speicher reservieren
  ant = (double*) malloc( arten * sizeof( double ) );
  if( ant == NULL )
  {
    return 1;
  }
  // Daten einlesen
  fseek( fiin, pos, SEEK_SET );
  if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
  {
    return 3;
  }
  ant[0] = zufall_double( &vert );
  for( i=1; i<arten; i++ )
  {
    if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    // gleich aufsummieren
    ant[i] = ant[ i - 1 ] + zufall_double( &vert );
  }
  fi = fopen( fn, "w" );
  if( fi == NULL )
  {
    return 2;
  }
  while( anz > 0 )
  {
    // Verteilung auswählen
    az = gleich_double( 0.0, ant[ arten - 1 ] );
    // dazugehörige Nummer suchen; da auch in Datei gesucht
    // werden muß ist lineare Laufzeit OK
    i = 0;
    fseek( fiin, pos, SEEK_SET );
    while( az >= ant[i] )
    {
      i++;
      if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
      {
        return 3;
      }
    }
    // Laufzeit mit einem Prozessor bestimmen (lz)
    // Wahrscheinlichkeit überlesen
    if( lese_next_verteilung_double( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    az = lz = zufall_double( &vert );
    if( save_double( &fi, lz ) != 0 )
    {
      return 2;
    }
    // Laufzeit mit i+1 Prozessoren bestimmen;
    // az enthält letzte Laufzeit
    if( lese_next_int_in_line( &fiin, &h ) != 0 )
    {
      return 3;
    }
    if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    for( i=1; i<n; i++ )
    {
      if( i == h )
      {
        if( lese_next_int_in_line( &fiin, &h ) != 0 )
        {
          return 3;
        }
        if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
        {
          return 3;
        }
      }
      d = lz / ((double)(i + 1)) - az;
      az += d * zufall_double( &vert );
      if( save_text( &fi, " " ) != 0 )
      {
        return 2;
      }
      if( save_double( &fi, az ) != 0 )
      {
        return 2;
      }
    }
    if( save_text( &fi, "\n" ) != 0 )
    {
      return 2;
    }
    anz--;
  }
  fclose( fi );
  fclose( fiin );
  free( ant );
  return 0;
}

int baue_bsp_array(
  double** laufz,
  int*     m,
  int*     n,
  char*    fin )
{
  int anz, h;
  double* ant;
  int i, k, arten;
  double d;
  int pos;
  verteilung_int anzahl;
  verteilung_double vert;
  FILE* fiin;
  fiin = fopen( fin, "r" );
  if( fiin == NULL )
  {
    return 4;
  }
  // Prozessorenanzahl einlesen
  if( lese_next_int_r( &fiin, n ) != 0 )
  {
    return 3;
  }
  // zu erzeugende Modulanzahlverteilung einlesen
  if( lese_next_verteilung_int_r( &fiin, &anzahl ) != 0 )
  {
    return 3;
  }
  *m = anz = zufall_int( &anzahl ); // zu erzeugende Modulanzahl
  // Verteilungsarten zählen
  pos = ftell( fiin );
  if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
  { // eine muß da sein!
    return 3;
  }
  arten = 1;
  while( 1 )
  {
    if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
    {
      break;
    }
    arten++;
  }
  // Speicher reservieren
  ant = (double*) malloc( arten * sizeof( double ) );
  if( ant == NULL )
  {
    return 1;
  }
  // Daten einlesen
  fseek( fiin, pos, SEEK_SET );
  if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
  {
    return 3;
  }
  ant[0] = zufall_double( &vert );
  for( i=1; i<arten; i++ )
  {
    if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    // gleich aufsummieren
    ant[i] = ant[ i - 1 ] + zufall_double( &vert );
  }
  for( k=0; k<anz; k++ )
  {
    // Verteilung auswählen
    d = gleich_double( 0.0, ant[ arten - 1 ] );
    // dazugehörige Nummer suchen; 
    // da auch in Datei gesucht werden muß ist lineare Laufzeit OK
    i = 0;
    fseek( fiin, pos, SEEK_SET );
    while( d >= ant[i] )
    {
      i++;
      if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
      {
        return 3;
      }
    }
    // Wahrscheinlichkeit für Modulauswahl überlesen
    if( lese_next_verteilung_double( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    // Laufzeit mit einem Prozessor bestimmen
    if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    laufz[k][0] = zufall_double( &vert );
    // Laufzeit mit i+1 Prozessoren bestimmen
    if( lese_next_int_in_line( &fiin, &h ) != 0 )
    {
      return 3;
    }
    if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    for( i=1; i<(*n); i++ )
    {
      if( i == h )
      {
        if( lese_next_int_in_line( &fiin, &h ) != 0 )
        {
          return 3;
        }
        if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
        {
          return 3;
        }
      }
      d = laufz[k][0] / ((double)(i + 1)) - laufz[k][ i-1 ];
      laufz[k][i] = laufz[k][ i-1 ] + d * zufall_double( &vert );
    }
  }
  fclose( fiin );
  free( ant );
  return 0;
}

// reserviert den maximal benötigten Speicher für eine Laufzeittabelle
// Fehler: 1->zu wenig Speicher
//         2->Configdatei nicht gefunden
//         3->Configdatei fehlerhaft
int init_laufzeit(
  double*** laufz,
  char*     fn )
{
  FILE* fi;
  int i, m, n;
  verteilung_int vert;
  fi = fopen( fn, "r" );
  if( fi == NULL )
  {
    return 2;
  }
  // Prozessorenanzahl einlesen
  if( lese_next_int_r( &fi, &n ) != 0 )
  {
    return 3;
  }
  // Modulanzahlverteilung einlesen
  if( lese_next_verteilung_int_r( &fi, &vert ) != 0 )
  {
    return 3;
  }
  // maximale Modulanzahl ermitteln
  m = zufall_max_int( &vert );
  fclose( fi );
  *laufz = (double**) malloc( m * sizeof( double* ) );
  if( *laufz == NULL )
  {
    return 1;
  }
  for( i=0; i<m; i++ )
  {
    (*laufz)[i] = (double*) malloc( n * sizeof( double ) );
    if( (*laufz)[i] == NULL )
    {
      return 1;
    }
  }
  return 0;
}

// gibt Speicher der Laufzeittabelle frei
// Fehler: 1->Configdatei nicht gefunden
//         2->Configdatei fehlerhaft
int frei_laufzeit(
  double** laufz,
  char*    fn )
{
  FILE* fi;
  int m;
  verteilung_int vert;
  fi = fopen( fn, "r" );
  if( fi == NULL )
  {
    return 1;
  }
  // Prozessorenanzahl überlesen
  if( lese_next_int_r( &fi, &m ) != 0 )
  {
    return 2;
  }
  // Modulanzahlverteilung einlesen
  if( lese_next_verteilung_int_r( &fi, &vert ) != 0 )
  {
    return 2;
  }
  // maximale Modulanzahl ermitteln
  m = zufall_max_int( &vert );
  fclose( fi );
  while( m > 0 )
  {
    m--;
    free( laufz[m] );
  }
  free( laufz );
  return 0;
}
