// Optimierungsroutinen moldable, kein Modulabhängigkeitsgraph;
// Verwendung vom "LMZ" - Einfügealgorithmus
// optimiert Module deren Laufzeiten in daten stehen
// (i-te Zeile j-te Spalte enthält Laufzeit vom i-ten Modul mit
//  j Prozessoren)
// oder mit Hilfe der übergebenen Funktion laufzeit
// (modulnummer>=1, prozessorenanzahl>=1) ermittelt werden können
//
// Aufruf: optimiere_datei( char* ergebnisdateiname,
//                          char* datendateiname )
//   oder: optimiere_formel( char*  ergebnisdateiname,
//                           int    gesamtmodulanzahl,
//                           int    gesamtprozessorenanzahl,
//                           double ((*laufzeit)( modulnummer,
//                                                prozessorenanzahl )) )
//
// Fehler: 1->Eingabedatei fehlerhaft
//         2->zu wenig Speicher
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
//         5->Fehler bei Arbeit mit Hilfsdatei


#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include "../prgx/baum.h"
#include "../prgx/load.h"
#include "../prgx/save.h"
#include "../prgx/suchsort.h"


// ------------------------ DATEI - ROUTINE MIT ABBRUCH ----------------
#if defined( ABBRUCH ) && !defined( FORMEL )

// optimiert Module deren Laufzeiten in daten stehen
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[n+1], beste[n+1], nr[m], nrh[m], rf[m],
//                       zeit[m], p_ende[n], lsg[2][m][2], stelle[2][m],
//                       reihenf[2][m], heap der Größe n und Puffer n
// Fehler: 1->Eingabedatei fehlerhaft
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
int opti_datei_abbr(
  char*     ergebnis,
  char*     daten,
  int       n,
  int       m,
  int*      breite,
  int*      beste,
  int*      nr,
  int*      nrh,
  int*      rf,
  double*   zeit,
  double*   p_ende,
  double*** lsg,
  int**     stelle,
  int**     reihenf,
  baum*     heap )
{
  FILE* fi;
  int pos;
  double h, best_h, a, b;
  int i, j, k;
  int hohe, akt_lsg;
  // Module nach sequentieller Laufzeit sortieren
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  pos = ftell( fi );
  h = 0; // Minimale Fläche bestimmen
  for( i=0; i<m; i++ )
  {
    nr[i] = i;
    if( lese_next_double_r( &fi, &(zeit[i]) ) != 0 )
    {
      return 1;
    }
    h += zeit[i];
  }
  merge_modul_zeit( zeit, nr, nrh, 0, m );
  hohe = 0; // Modulanzahl größer als min_fläche/n
  for( i=0; i<n; i++ )
  {
    if( i == m )
    {
      break;
    }
    if( zeit[ nr[i] ] * ((double)(n)) > h )
    {
      hohe++;
    }
  }
  such_heap_min_init( heap, p_ende, n );
  // mit einem Proz. einpacken
  for( i=0; i<m; i++ )
  {
    j = such_heap_min( heap );
    lsg[0][ nr[i] ][0] = p_ende[j]; // Ergebnis merken - Anfangszeit
    set_heap_min( heap, p_ende, p_ende[j] + zeit[ nr[i] ] );
    lsg[0][ nr[i] ][1] = p_ende[j]; // Ergebnis merken - Endzeit
    stelle[0][ nr[i] ] = j; // welches Modul steht wo
  }
  best_h = p_ende[ such_max( p_ende, n ) ];
  if( ( best_h > zeit[ nr[0] ] ) || ( hohe == 0 ) )
  { // gut genug gepackt ?
    fclose( fi );
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, nr[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, lsg[0][ nr[j] ][0] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, lsg[0][ nr[j] ][1] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, stelle[0][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi, stelle[0][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  else // es muß noch besser gepackt werden
  {
    akt_lsg = 1;
    beste[0] = 0;
    for( i=1; i<=hohe; i++ ) // eins mehr wegen Abschluß-Null
    {
      breite[i] = 0;
      beste[i] = 0;
    }
    // höchstes Modul verbreitern
    breite[0] = 1;
    fseek( fi, pos, SEEK_SET );
    if( lese_pos_double( &fi, nr[0], 1, &(zeit[ nr[0] ]) ) != 0 )
    {
      return 1;
    }
    while( 1 )
    {
      // einpacken
      for( i=0; i<n; i++ )
      {
        p_ende[i] = 0;
      }
      i = 0;
      while( breite[i] != 0 )
      {
        j = such_min_breit( p_ende, n, breite[i], &a );
        lsg[akt_lsg][ nr[i] ][0] = a; // Ergebnis merken - Anfangszeit
        a += zeit[ nr[i] ];
        lsg[akt_lsg][ nr[i] ][1] = a; // Ergebnis merken - Endzeit
        for( k=j; k<=j+breite[i]; k++ )
        {
          p_ende[ k ] = a;
        }
        stelle[akt_lsg][ nr[i] ] = j; // Position merken
        i++;
      }
      such_heap_min_init_werte( heap, p_ende, n );
      for( ; i<m; i++ )
      {
        j = such_heap_min( heap );
        // Ergebnis merken - Anfangszeit
        lsg[akt_lsg][ nr[i] ][0] = p_ende[j];
        set_heap_min( heap, p_ende, p_ende[j] + zeit[ nr[i] ] );
        // Ergebnis merken - Endzeit
        lsg[akt_lsg][ nr[i] ][1] = p_ende[j];
        // Position merken
        stelle[akt_lsg][ nr[i] ] = j;
      }
      h = p_ende[ such_max( p_ende, n ) ];
      if( h < best_h ) // immer jeweils bestes Ergebnis merken!!
      {
        best_h = h;
        i = 0;
        while( breite[i] != 0 )
        {
          beste[i] = breite[i];
          i++;
        }
        // wieder in anderer Ebene einpacken
        akt_lsg = ( akt_lsg + 1 ) & 1;
      }
      i = such_max_indirekt( zeit, nr, hohe );
      if( zeit[ nr[i] ] != h )
      { // wenn das höchste nicht mehr die Gesamthöhe bestimmt 
        // -> nächstes Verfahren
        break;
      }
      if( breite[i] == n-1 ) // kann noch verbreitert werden ?
      {
        break;
      }
      breite[i]++;
      fseek( fi, pos, SEEK_SET );
      if( lese_pos_double( &fi, nr[i], breite[i],
                           &(zeit[ nr[i] ]) ) != 0 )
      {
        return 1;
      }
    }
    // Flächenminimum suchen
    i = 0;
    while( breite[i] != 0 )
    {
      j = breite[i] + 1;
      h = zeit[ nr[i] ] * ((double)(j));
      fseek( fi, pos, SEEK_SET );
      if( lese_pos_double( &fi, nr[i], breite[i], &a ) != 0 )
      {
        return 1;
      }
      while( j < n )
      {
        j++;
        if( lese_next_double_in_line( &fi, &a ) != 0 )
        {
          return 1;
        }
        if( a * ((double)(j)) < h )
        {
          h = a * ((double)(j));
          breite[i] = j - 1;
          zeit[ nr[i] ] = a;
        }
      }
      i++;
    }
    j = ( akt_lsg + 1 ) & 1;
    for( i=0; i<m; i++ )
    {
      // Originalreihenfolge zwecks breite-Zuordnung invertiert merken
      rf[ nr[i] ] = i;
      reihenf[j][i] = nr[i]; // Einpackreihenfolge merken
    }
    // nach Höhe sortieren
    // mehr ist nicht durcheinander gekommen
    merge_modul_zeit( zeit, nr, nrh, 0, hohe );
    // die Liste mit den hohen und die andere zusammenmischen
    misch_modul_zeit( zeit, nr, nrh, 0, hohe, m );
    while( 1 )
    {
      // einpacken
      h = 0.0;
      i = n;
      b = 0.0; // aktuelle Fläche berechnen
      for( j=0; j<m; j++ )
      {
        stelle[akt_lsg][ nr[j] ] = i;
        i++;
        if( rf[ nr[j] ] < hohe )
        {
          i += breite[ rf[ nr[j] ] ];
          b += ( breite[ rf[ nr[j] ] ] + 1 ) * zeit[ nr[j] ];
          if( i > n )
          {
            i = breite[ rf[ nr[j] ] ] + 1;
            a = h;
            h += zeit[ nr[j] ];
            stelle[akt_lsg][ nr[j] ] = 0;
          }
        }
        else
        {
          b += zeit[ nr[j] ];
          if( i > n )
          {
            i = 1;
            a = h;
            h += zeit[ nr[j] ];
            stelle[akt_lsg][ nr[j] ] = 0;
          }
        }
        lsg[akt_lsg][ nr[j] ][0] = a; // Anfangszeit
        lsg[akt_lsg][ nr[j] ][1] = a + zeit[ nr[j] ]; // Endzeit
        reihenf[akt_lsg][j] = nr[j];
      }
      if( h < best_h )
      {
        i = 0;
        while( breite[i] != 0 )
        {
          beste[i] = breite[i];
          i++;
        }
        best_h = h;
        akt_lsg = ( akt_lsg + 1 ) & 1;
      }
      if( 3 * b > ((double)(n)) * best_h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        break;
      }
      // das höchste Modul flächenminimal verbreitern
      j = breite[ rf[ nr[0] ] ] + 2;
      if( j > n ) // ist noch eine Verbreiterung möglich
      {
        break;
      }
      breite[ rf[ nr[0] ] ]++;
      fseek( fi, pos, SEEK_SET );
      if( lese_pos_double( &fi, nr[0], breite[ rf[ nr[0] ] ],
                           &(zeit[ nr[0] ]) ) != 0 )
      {
        return 1;
      }
      h = zeit[ nr[0] ] * ((double)(j));
      while( j < n )
      {
        j++;
        if( lese_next_double_in_line( &fi, &a ) != 0 )
        {
          return 1;
        }
        if( a * ((double)(j)) < h )
        {
          h = a * ((double)(j));
          breite[ rf[ nr[0] ] ] = j - 1;
          zeit[ nr[0] ] = a;
        }
      }
      // das ehemals höchste einsortieren
      if( m != 1 )
      {
        i = nr[0];
        j = 0;
        while( zeit[ nr[ j+1 ] ] > zeit[i] )
        {
          nr[j] = nr[ j+1 ];
          j++;
          if( j == m - 1 )
          {
            break;
          }
        }
        if( j < m - 1 )
        {
          while( zeit[ nr[ j+1 ] ] == zeit[i] )
          {
            if( rf[ nr[ j+1 ] ] >= hohe )
            {
              break;
            }
            if( breite[ rf[ nr[ j+1 ] ] ] <= breite[ rf[i] ] )
            {
              break;
            }
            nr[j] = nr[ j+1 ];
            j++;
            if( j == m - 1 )
            {
              break;
            }
          }
        }
        nr[j] = i;
      }
    }
    fclose( fi );
    // Ergebnis erstellen
    akt_lsg = ( akt_lsg + 1 ) & 1;
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, reihenf[akt_lsg][j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi,
                       lsg[akt_lsg][ reihenf[akt_lsg][j] ][0] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi,
                       lsg[akt_lsg][ reihenf[akt_lsg][j] ][1] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int(
            &fi, stelle[akt_lsg][ reihenf[akt_lsg][j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( rf[ reihenf[akt_lsg][j] ] < hohe )
      {
        if( save_int_r(
              &fi, 
              ( stelle[akt_lsg][ reihenf[akt_lsg][j] ] +
                beste[ rf[ reihenf[akt_lsg][j] ] ] + 1 ) ) != 0 )
        {
          return 3;
        }
      }
      else
      {
        if( save_int_r(
              &fi, stelle[akt_lsg][ reihenf[akt_lsg][j] ] + 1 ) != 0 )
        {
          return 3;
        }
      }
    }
    fclose( fi );
  }
  return 0;
}
#endif


// ---------------------- DATEI - ROUTINE OHNE ABBRUCH -----------------
#if !defined( ABBRUCH ) && !defined( FORMEL )

// optimiert Module deren Laufzeiten in daten stehen
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[n+1], beste[n+1], nr[m], nrh[m], rf[m],
//                       zeit[m], p_ende[n], lsg[2][m][2], stelle[2][m],
//                       reihenf[2][m], heap der Größe n und Puffer n
// Fehler: 1->Eingabedatei fehlerhaft
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
int opti_datei_weiter(
  char*           ergebnis,
  char*           daten,
  int             n,
  int             m,
  int*            breite,
  int*            beste,
  int*            nr,
  int*            nrh,
  int*            rf,
  double*         zeit,
  double*         p_ende,
  double***       lsg,
  int**           stelle,
  int**           reihenf,
  baum*           heap )
{
  FILE* fi;
  int pos;
  double h, best_h, a;
  int i, j, k;
  int hohe, akt_lsg;
  // Module nach sequentieller Laufzeit sortieren
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  pos = ftell( fi );
  h = 0; // Minimale Fläche bestimmen
  for( i=0; i<m; i++ )
  {
    nr[i] = i;
    if( lese_next_double_r( &fi, &(zeit[i]) ) != 0 )
    {
      return 1;
    }
    h += zeit[i];
  }
  merge_modul_zeit( zeit, nr, nrh, 0, m );
  hohe = 0; // Modulanzahl größer als min_fläche/n
  for( i=0; i<n; i++ )
  {
    if( i == m )
    {
      break;
    }
    if( zeit[ nr[i] ] * ((double)(n)) > h )
    {
      hohe++;
    }
  }
  // mit einem Proz. einpacken
  such_heap_min_init( heap, p_ende, n );
  for( i=0; i<m; i++ )
  {
    j = such_heap_min( heap );
    lsg[0][ nr[i] ][0] = p_ende[j]; // Ergebnis merken - Anfangszeit
    set_heap_min( heap, p_ende, p_ende[j] + zeit[ nr[i] ] );
    lsg[0][ nr[i] ][1] = p_ende[j]; // Ergebnis merken - Endzeit
    stelle[0][ nr[i] ] = j; // welches Modul steht wo
  }
  best_h = p_ende[ such_max( p_ende, n ) ];
  if( ( best_h > zeit[ nr[0] ] ) || ( hohe == 0 ) )
  { // gut genug gepackt ?
    fclose( fi );
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, nr[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, lsg[0][ nr[j] ][0] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, lsg[0][ nr[j] ][1] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, stelle[0][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi, stelle[0][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  else // es muß noch besser gepackt werden
  {
    akt_lsg = 1;
    beste[0] = 0;
    for( i=1; i<=hohe; i++ ) // eins mehr wegen Abschluß-Null
    {
      breite[i] = 0;
      beste[i] = 0;
    }
    // höchstes Modul verbreitern
    breite[0] = 1;
    fseek( fi, pos, SEEK_SET );
    if( lese_pos_double( &fi, nr[0], 1, &(zeit[ nr[0] ]) ) != 0 )
    {
      return 1;
    }
    while( 1 )
    {
      // einpacken
      for( i=0; i<n; i++ )
      {
        p_ende[i] = 0;
      }
      i = 0;
      while( breite[i] != 0 )
      {
        j = such_min_breit( p_ende, n, breite[i], &a );
        lsg[akt_lsg][ nr[i] ][0] = a; // Ergebnis merken - Anfangszeit
        a += zeit[ nr[i] ];
        lsg[akt_lsg][ nr[i] ][1] = a; // Ergebnis merken - Endzeit
        for( k=j; k<=j+breite[i]; k++ )
        {
          p_ende[ k ] = a;
        }
        stelle[akt_lsg][ nr[i] ] = j; // Position merken
        i++;
      }
      such_heap_min_init_werte( heap, p_ende, n );
      for( ; i<m; i++ )
      {
        j = such_heap_min( heap );
        // Ergebnis merken - Anfangszeit
        lsg[akt_lsg][ nr[i] ][0] = p_ende[j];
        set_heap_min( heap, p_ende, p_ende[j] + zeit[ nr[i] ] );
        // Ergebnis merken - Endzeit
        lsg[akt_lsg][ nr[i] ][1] = p_ende[j];
        // Position merken
        stelle[akt_lsg][ nr[i] ] = j;
      }
      h = p_ende[ such_max( p_ende, n ) ];
      if( h < best_h ) // immer jeweils bestes Ergebnis merken!!
      {
        best_h = h;
        i = 0;
        while( breite[i] != 0 )
        {
          beste[i] = breite[i];
          i++;
        }
        // wieder in anderer Ebene einpacken
        akt_lsg = ( akt_lsg + 1 ) & 1;
      }
      i = such_max_indirekt( zeit, nr, hohe );
      if( zeit[ nr[i] ] != h )
      { // wenn das höchste nicht mehr die Gesamthöhe bestimmt
        // -> nächstes Verfahren
        break;
      }
      if( breite[i] == n-1 ) // kann noch verbreitert werden ?
      {
        break;
      }
      breite[i]++;
      fseek( fi, pos, SEEK_SET );
      if( lese_pos_double( &fi,
                           nr[i], breite[i], &(zeit[ nr[i] ]) ) != 0 )
      {
        return 1;
      }
    }
    // Flächenminimum suchen
    i = 0;
    while( breite[i] != 0 )
    {
      j = breite[i] + 1;
      h = zeit[ nr[i] ] * ((double)(j));
      fseek( fi, pos, SEEK_SET );
      if( lese_pos_double( &fi, nr[i], breite[i], &a ) != 0 )
      {
        return 1;
      }
      while( j < n )
      {
        j++;
        if( lese_next_double_in_line( &fi, &a ) != 0 )
        {
          return 1;
        }
        if( a * ((double)(j)) < h )
        {
          h = a * ((double)(j));
          breite[i] = j - 1;
          zeit[ nr[i] ] = a;
        }
      }
      i++;
    }
    j = ( akt_lsg + 1 ) & 1;
    for( i=0; i<m; i++ )
    {
      // Originalreihenfolge zwecks breite-Zuordnung invertiert merken
      rf[ nr[i] ] = i;
      reihenf[j][i] = nr[i]; // Einpackreihenfolge merken
    }
    // nach Höhe sortieren
    // mehr ist nicht durcheinander gekommen
    merge_modul_zeit( zeit, nr, nrh, 0, hohe );
    // die Liste mit den hohen und die andere zusammenmischen
    misch_modul_zeit( zeit, nr, nrh, 0, hohe, m );
    while( 1 )
    {
      // einpacken
      h = 0.0;
      i = n;
      for( j=0; j<m; j++ )
      {
        stelle[akt_lsg][ nr[j] ] = i;
        i++;
        if( rf[ nr[j] ] < hohe )
        {
          i += breite[ rf[ nr[j] ] ];
          if( i > n )
          {
            i = breite[ rf[ nr[j] ] ] + 1;
            a = h;
            h += zeit[ nr[j] ];
            stelle[akt_lsg][ nr[j] ] = 0;
          }
        }
        else
        {
          if( i > n )
          {
            i = 1;
            a = h;
            h += zeit[ nr[j] ];
            stelle[akt_lsg][ nr[j] ] = 0;
          }
        }
        lsg[akt_lsg][ nr[j] ][0] = a; // Anfangszeit
        lsg[akt_lsg][ nr[j] ][1] = a + zeit[ nr[j] ]; // Endzeit
        reihenf[akt_lsg][j] = nr[j];
      }
      if( h < best_h )
      {
        i = 0;
        while( breite[i] != 0 )
        {
          beste[i] = breite[i];
          i++;
        }
        best_h = h;
        akt_lsg = ( akt_lsg + 1 ) & 1;
      }
      // das höchste Modul flächenminimal verbreitern
      k = 0;
      i = 0;
      while( i < hohe )
      { // wenn alle hohen durch sind, kann abgebrochen werden
        if( rf[ nr[k] ] < hohe )
        {
          j = breite[ rf[ nr[k] ] ] + 2;
          if( j <= n ) // ist noch eine Verbreiterung möglich
          {
            break;
          }
          i++;
        }
        k++;
      }
      if( i == hohe )
      {
        break; // es kann kein hohes Modul mehr verbreitert werden
      }
      breite[ rf[ nr[k] ] ]++;
      fseek( fi, pos, SEEK_SET );
      if( lese_pos_double( &fi, nr[k], breite[ rf[ nr[k] ] ],
                           &(zeit[ nr[k] ]) ) != 0 )
      {
        return 1;
      }
      h = zeit[ nr[k] ] * ((double)(j));
      while( j < n )
      {
        j++;
        if( lese_next_double_in_line( &fi, &a ) != 0 )
        {
          return 1;
        }
        if( a * ((double)(j)) < h )
        {
          h = a * ((double)(j));
          breite[ rf[ nr[k] ] ] = j - 1;
          zeit[ nr[k] ] = a;
        }
      }
      // das verbreiterte einsortieren
      misch_modul_zeit( zeit, nr, nrh, k, k+1, m );
    }
    fclose( fi );
    // Ergebnis erstellen
    akt_lsg = ( akt_lsg + 1 ) & 1;
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, reihenf[akt_lsg][j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double(
            &fi, lsg[akt_lsg][ reihenf[akt_lsg][j] ][0] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double(
            &fi, lsg[akt_lsg][ reihenf[akt_lsg][j] ][1] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int(
            &fi, stelle[akt_lsg][ reihenf[akt_lsg][j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( rf[ reihenf[akt_lsg][j] ] < hohe )
      {
        if( save_int_r(
              &fi,
              ( stelle[akt_lsg][ reihenf[akt_lsg][j] ] +
                beste[ rf[ reihenf[akt_lsg][j] ] ] + 1 ) ) != 0 )
        {
          return 3;
        }
      }
      else
      {
        if( save_int_r(
              &fi, stelle[akt_lsg][ reihenf[akt_lsg][j] ] + 1 ) != 0 )
        {
          return 3;
        }
      }
    }
    fclose( fi );
  }
  return 0;
}
#endif


#ifndef FORMEL

// Hauptroutine für Datei; siehe oben
int optimiere_datei(
  char* ergebnis,
  char* daten )
{
  FILE* fi;
  int m;
  int n;
  double h;
  double* p_ende;
  int* breite;
  int* beste;
  int* nr;
  int* nrh;
  int* rf;
  double* zeit;
  double*** lsg;
  int** stelle;
  int** reihenf;
  int i;
  baum heap;

  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  // Prozessorenanzahl bestimmen
  if( lese_next_double( &fi, &h ) != 0 )
  {
    return 1;
  }
  n = 1;
  while( lese_next_double_in_line( &fi, &h ) == 0 )
  {
    n++;
  }
  // Modulanzahl bestimmen
  if( lese_next_ret( &fi ) != 0 )
  {
    return 1;
  }
  m = 1;
  while( 1 )
  {
    if( lese_next_double_r( &fi, &h ) != 0 )
    {
      break;
    }
    m++;
  }
  fclose( fi );

  // Speicherplatz reservieren
  p_ende = (double*) malloc( n * sizeof( double ) );
  if( p_ende == NULL )
  {
    return 2;
  }
  // eins mehr wegen Abschluß-Null
  breite = (int*) malloc( ( n + 1 ) * sizeof( int ) );
  if( breite == NULL )
  {
    return 2;
  }
  // eins mehr wegen Abschluß-Null
  beste = (int*) malloc( ( n + 1 ) * sizeof( int ) );
  if( beste == NULL )
  {
    return 2;
  }
  nr = (int*) malloc( m * sizeof( int ) );
  if( nr == NULL )
  {
    return 2;
  }
  nrh = (int*) malloc( m * sizeof( int ) );
  if( nrh == NULL )
  {
    return 2;
  }
  rf = (int*) malloc( m * sizeof( int ) );
  if( rf == NULL )
  {
    return 2;
  }
  zeit = (double*) malloc( m * sizeof( double ) );
  if( zeit == NULL )
  {
    return 2;
  }
  lsg = (double***) malloc( 2 * sizeof( double** ) );
  if( lsg == NULL )
  {
    return 2;
  }
  lsg[0] = (double**) malloc( m * sizeof( double* ) );
  if( lsg[0] == NULL )
  {
    return 2;
  }
  lsg[1] = (double**) malloc( m * sizeof( double* ) );
  if( lsg[1] == NULL )
  {
    return 2;
  }
  for( i=0; i<m; i++ )
  {
    lsg[0][i] = (double*) malloc( 2 * sizeof( double ) );
    if( lsg[0][i] == NULL )
    {
      return 2;
    }
    lsg[1][i] = (double*) malloc( 2 * sizeof( double ) );
    if( lsg[1][i] == NULL )
    {
      return 2;
    }
  }
  stelle = (int**) malloc( 2 * sizeof( int* ) );
  if( stelle == NULL )
  {
    return 2;
  }
  stelle[0] = (int*) malloc( m * sizeof( int ) );
  if( stelle[0] == NULL )
  {
    return 2;
  }
  stelle[1] = (int*) malloc( m * sizeof( int ) );
  if( stelle[1] == NULL )
  {
    return 2;
  }
  reihenf = (int**) malloc( 2 * sizeof( int* ) );
  if( reihenf == NULL )
  {
    return 2;
  }
  reihenf[0] = (int*) malloc( m * sizeof( int ) );
  if( reihenf[0] == NULL )
  {
    return 2;
  }
  reihenf[1] = (int*) malloc( m * sizeof( int ) );
  if( reihenf[1] == NULL )
  {
    return 2;
  }
  init_size_baum( &heap, n );
  for( i=0; (int)(i)<n; i++ ) // gleich n Instanzen anlegen
  {
    if( insert_ll_baum( &heap, NULL ) == NULL )
    {
      return 2;
    }
  }

#ifdef ABBRUCH
  n = opti_datei_abbr( ergebnis, daten, n, m, breite, beste, nr,
                       nrh, rf, zeit, p_ende, lsg, stelle,
                       reihenf, &heap );
#else
  n = opti_datei_weiter( ergebnis, daten, n, m, breite, beste, nr,
                         nrh, rf, zeit, p_ende, lsg, stelle,
                         reihenf, &heap );
#endif

  free_baum( &heap );
  free( reihenf[1] );
  free( reihenf[0] );
  free( reihenf );
  free( stelle[1] );
  free( stelle[0] );
  free( stelle );
  for( i=0; i<m; i++ )
  {
    free( lsg[0][i] );
    free( lsg[1][i] );
  }
  free( lsg[1] );
  free( lsg[0] );
  free( lsg );
  free( zeit );
  free( rf );
  free( nrh );
  free( nr );
  free( p_ende );
  free( breite );
  free( beste );
  return n;
}
#endif


// ------------------- FORMEL - ROUTINE MIT ABBRUCH --------------------
#if defined( ABBRUCH ) && defined( FORMEL )

// optimiert Module deren Laufzeiten über laufzeit() ermittelt
// werden können
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[n+1], beste[n+1], nr[m], nrh[m], rf[m],
//                       zeit[m], p_ende[n], lsg[2][m][2], stelle[2][m],
//                       reihenf[2][m], heap der Größe n mit Puffer n
// Fehler: 3->kann Ergebnisdatei nicht anlegen
int opti_formel_abbr(
  char*     ergebnis,
  double    (*laufzeit)( int modulnummer, int prozessorenanzahl ),
  int       n,
  int       m,
  int*      breite,
  int*      beste,
  int*      nr,
  int*      nrh,
  int*      rf,
  double*   zeit,
  double*   p_ende,
  double*** lsg,
  int**     stelle,
  int**     reihenf,
  baum*     heap )
{
  FILE* fi;
  double h, best_h, a, b;
  int i, j, k;
  int hohe, akt_lsg;
  // Module nach sequentieller Laufzeit sortieren
  zeit[0] = (*laufzeit)( 1, 1 );
  nr[0] = 0;
  h = zeit[0]; // Minimale Fläche bestimmen
  for( i=1; i<m; i++ )
  {
    nr[i] = i;
    zeit[i] = (*laufzeit)( i+1, 1 );
    h += zeit[i];
  }
  merge_modul_zeit( zeit, nr, nrh, 0, m );
  hohe = 0; // Modulanzahl größer als min_fläche/n
  for( i=0; i<n; i++ )
  {
    if( i == m )
    {
      break;
    }
    if( zeit[ nr[i] ] * ((double)(n)) > h )
    {
      hohe++;
    }
  }
  such_heap_min_init( heap, p_ende, n );
  // mit einem Proz. einpacken
  for( i=0; i<m; i++ )
  {
    j = such_heap_min( heap );
    lsg[0][ nr[i] ][0] = p_ende[j]; // Ergebnis merken - Anfangszeit
    set_heap_min( heap, p_ende, p_ende[j] + zeit[ nr[i] ] );
    lsg[0][ nr[i] ][1] = p_ende[j]; // Ergebnis merken - Endzeit
    stelle[0][ nr[i] ] = j; // welches Modul steht wo
  }
  best_h = p_ende[ such_max( p_ende, n ) ];
  if( ( best_h > zeit[ nr[0] ] ) || ( hohe == 0 ) )
  { // gut genug gepackt ?
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, nr[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, lsg[0][ nr[j] ][0] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, lsg[0][ nr[j] ][1] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, stelle[0][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi, stelle[0][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  else // es muß noch besser gepackt werden
  {
    akt_lsg = 1;
    beste[0] = 0;
    for( i=1; i<=hohe; i++ ) // eins mehr wegen Abschluß-Null
    {
      breite[i] = 0;
      beste[i] = 0;
    }
    // höchstes Modul verbreitern
    breite[0] = 1;
    zeit[ nr[0] ] = (*laufzeit)( nr[0]+1, 2 );
    while( 1 )
    {
      // einpacken
      for( i=0; i<n; i++ )
      {
        p_ende[i] = 0;
      }
      i = 0;
      while( breite[i] != 0 )
      {
        j = such_min_breit( p_ende, n, breite[i], &a );
        lsg[akt_lsg][ nr[i] ][0] = a; // Ergebnis merken - Anfangszeit
        a += zeit[ nr[i] ];
        lsg[akt_lsg][ nr[i] ][1] = a; // Ergebnis merken - Endzeit
        for( k=j; k<=j+breite[i]; k++ )
        {
          p_ende[ k ] = a;
        }
        stelle[akt_lsg][ nr[i] ] = j; // Position merken
        i++;
      }
      such_heap_min_init_werte( heap, p_ende, n );
      for( ; i<m; i++ )
      {
        j = such_heap_min( heap );
        // Ergebnis merken - Anfangszeit
        lsg[akt_lsg][ nr[i] ][0] = p_ende[j];
        set_heap_min( heap, p_ende, p_ende[j] + zeit[ nr[i] ] );
        // Ergebnis merken - Endzeit
        lsg[akt_lsg][ nr[i] ][1] = p_ende[j];
        // Position merken
        stelle[akt_lsg][ nr[i] ] = j;
      }
      h = p_ende[ such_max( p_ende, n ) ];
      if( h < best_h ) // immer jeweils bestes Ergebnis merken!!
      {
        best_h = h;
        i = 0;
        while( breite[i] != 0 )
        {
          beste[i] = breite[i];
          i++;
        }
        // wieder in anderer Ebene einpacken
        akt_lsg = ( akt_lsg + 1 ) & 1;
      }
      i = such_max_indirekt( zeit, nr, hohe );
      if( zeit[ nr[i] ] != h )
      { // wenn das höchste nicht mehr die Gesamthöhe bestimmt
        // -> nächstes Verfahren
        break;
      }
      if( breite[i] == n-1 ) // kann noch verbreitert werden ?
      {
        break;
      }
      breite[i]++;
      zeit[ nr[i] ] = (*laufzeit)( nr[i]+1, breite[i]+1 );
    }
    // Flächenminimum suchen
    i = 0;
    while( breite[i] != 0 )
    {
      j = breite[i] + 1;
      h = zeit[ nr[i] ] * ((double)(j));
      while( j < n )
      {
        j++;
        a = (*laufzeit)( nr[i]+1, j );
        if( a * ((double)(j)) < h )
        {
          h = a * ((double)(j));
          breite[i] = j - 1;
          zeit[ nr[i] ] = a;
        }
      }
      i++;
    }
    j = ( akt_lsg + 1 ) & 1;
    for( i=0; i<m; i++ )
    {
      // Originalreihenfolge zwecks breite-Zuordnung invertiert merken
      rf[ nr[i] ] = i;
      reihenf[j][i] = nr[i]; // Einpackreihenfolge merken
    }
    // nach Höhe sortieren
    // mehr ist nicht durcheinander gekommen
    merge_modul_zeit( zeit, nr, nrh, 0, hohe );
    // die Liste mit den hohen und die andere zusammenmischen
    misch_modul_zeit( zeit, nr, nrh, 0, hohe, m );
    while( 1 )
    {
      // einpacken
      h = 0.0;
      i = n;
      b = 0.0; // aktuelle Fläche
      for( j=0; j<m; j++ )
      {
        stelle[akt_lsg][ nr[j] ] = i;
        i++;
        if( rf[ nr[j] ] < hohe )
        {
          i += breite[ rf[ nr[j] ] ];
          b += ( breite[ rf[ nr[j] ] ] + 1 ) * zeit[ nr[j] ];
          if( i > n )
          {
            i = breite[ rf[ nr[j] ] ] + 1;
            a = h;
            h += zeit[ nr[j] ];
            stelle[akt_lsg][ nr[j] ] = 0;
          }
        }
        else
        {
          b += zeit[ nr[j] ];
          if( i > n )
          {
            i = 1;
            a = h;
            h += zeit[ nr[j] ];
            stelle[akt_lsg][ nr[j] ] = 0;
          }
        }
        lsg[akt_lsg][ nr[j] ][0] = a; // Anfangszeit
        lsg[akt_lsg][ nr[j] ][1] = a + zeit[ nr[j] ]; // Endzeit
        reihenf[akt_lsg][j] = nr[j];
      }
      if( h < best_h )
      {
        i = 0;
        while( breite[i] != 0 )
        {
          beste[i] = breite[i];
          i++;
        }
        best_h = h;
        akt_lsg = ( akt_lsg + 1 ) & 1;
      }
      if( 3 * b > ((double)(n)) * best_h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        break;
      }
      // das höchste Modul flächenminimal verbreitern
      j = breite[ rf[ nr[0] ] ] + 2;
      if( j > n ) // ist noch eine Verbreiterung möglich
      {
        break;
      }
      breite[ rf[ nr[0] ] ]++;
      h = zeit[ nr[0] ] * ((double)(j));
      while( j < n )
      {
        j++;
        a = (*laufzeit)( nr[0]+1, j );
        if( a * ((double)(j)) < h )
        {
          h = a * ((double)(j));
          breite[ rf[ nr[0] ] ] = j - 1;
          zeit[ nr[0] ] = a;
        }
      }
      // das ehemals höchste einsortieren
      misch_modul_zeit( zeit, nr, nrh, 0, 1, m );
    }
    // Ergebnis erstellen
    akt_lsg = ( akt_lsg + 1 ) & 1;
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, reihenf[akt_lsg][j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi,
                       lsg[akt_lsg][ reihenf[akt_lsg][j] ][0] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi,
                       lsg[akt_lsg][ reihenf[akt_lsg][j] ][1] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int(
            &fi, stelle[akt_lsg][ reihenf[akt_lsg][j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( rf[ reihenf[akt_lsg][j] ] < hohe )
      {
        if( save_int_r(
              &fi,
              ( stelle[akt_lsg][ reihenf[akt_lsg][j] ] +
                beste[ rf[ reihenf[akt_lsg][j] ] ] + 1 ) ) != 0 )
        {
          return 3;
        }
      }
      else
      {
        if( save_int_r( 
              &fi, stelle[akt_lsg][ reihenf[akt_lsg][j] ] + 1 ) != 0 )
        {
          return 3;
        }
      }
    }
    fclose( fi );
  }
  return 0;
}
#endif

// -------------------- FORMEL - ROUTINE OHNE ABBRUCH ------------------
#if !defined( ABBRUCH ) && defined( FORMEL )

// optimiert Module deren Laufzeiten über laufzeit() ermittelt
// werden können
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[n+1], beste[n+1], nr[m], nrh[m], rf[m],
//                       zeit[m], p_ende[n], lsg[2][m][2], stelle[2][m],
//                       reihenf[2][m], heap der Größe n mit Puffer n
// Fehler: 3->kann Ergebnisdatei nicht anlegen
int opti_formel_weiter(
  char*     ergebnis,
  double    (*laufzeit)( int modulnummer, int prozessorenanzahl ),
  int       n,
  int       m,
  int*      breite,
  int*      beste,
  int*      nr,
  int*      nrh,
  int*      rf,
  double*   zeit,
  double*   p_ende,
  double*** lsg,
  int**     stelle,
  int**     reihenf,
  baum*     heap )
{
  FILE* fi;
  double h, best_h, a;
  int i, j, k;
  int hohe, akt_lsg;
  // Module nach sequentieller Laufzeit sortieren
  zeit[0] = (*laufzeit)( 1, 1 );
  nr[0] = 0;
  h = zeit[0]; // Minimale Fläche bestimmen
  for( i=1; i<m; i++ )
  {
    nr[i] = i;
    zeit[i] = (*laufzeit)( i+1, 1 );
    h += zeit[i];
  }
  merge_modul_zeit( zeit, nr, nrh, 0, m );
  hohe = 0; // Modulanzahl größer als min_fläche/n
  for( i=0; i<n; i++ )
  {
    if( i == m )
    {
      break;
    }
    if( zeit[ nr[i] ] * ((double)(n)) > h )
    {
      hohe++;
    }
  }
  // mit einem Proz. einpacken
  such_heap_min_init( heap, p_ende, n );
  for( i=0; i<m; i++ )
  {
    j = such_heap_min( heap );
    lsg[0][ nr[i] ][0] = p_ende[j]; // Ergebnis merken - Anfangszeit
    set_heap_min( heap, p_ende, p_ende[j] + zeit[ nr[i] ] );
    lsg[0][ nr[i] ][1] = p_ende[j]; // Ergebnis merken - Endzeit
    stelle[0][ nr[i] ] = j; // welches Modul steht wo
  }
  best_h = p_ende[ such_max( p_ende, n ) ];
  if( ( best_h > zeit[ nr[0] ] ) || ( hohe == 0 ) )
  { // gut genug gepackt ?
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, nr[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, lsg[0][ nr[j] ][0] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, lsg[0][ nr[j] ][1] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, stelle[0][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi, stelle[0][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  else // es muß noch besser gepackt werden
  {
    akt_lsg = 1;
    beste[0] = 0;
    for( i=1; i<=hohe; i++ ) // eins mehr wegen Abschluß-Null
    {
      breite[i] = 0;
      beste[i] = 0;
    }
    // höchstes Modul verbreitern
    breite[0] = 1;
    zeit[ nr[0] ] = (*laufzeit)( nr[0]+1, 2 );
    while( 1 )
    {
      // einpacken
      for( i=0; i<n; i++ )
      {
        p_ende[i] = 0;
      }
      i = 0;
      while( breite[i] != 0 )
      {
        j = such_min_breit( p_ende, n, breite[i], &a );
        lsg[akt_lsg][ nr[i] ][0] = a; // Ergebnis merken - Anfangszeit
        a += zeit[ nr[i] ];
        lsg[akt_lsg][ nr[i] ][1] = a; // Ergebnis merken - Endzeit
        for( k=j; k<=j+breite[i]; k++ )
        {
          p_ende[ k ] = a;
        }
        stelle[akt_lsg][ nr[i] ] = j; // Position merken
        i++;
      }
      such_heap_min_init_werte( heap, p_ende, n );
      for( ; i<m; i++ )
      {
        j = such_heap_min( heap );
        // Ergebnis merken - Anfangszeit
        lsg[akt_lsg][ nr[i] ][0] = p_ende[j];
        set_heap_min( heap, p_ende, p_ende[j] + zeit[ nr[i] ] );
        // Ergebnis merken - Endzeit
        lsg[akt_lsg][ nr[i] ][1] = p_ende[j];
        // Position merken
        stelle[akt_lsg][ nr[i] ] = j;
      }
      h = p_ende[ such_max( p_ende, n ) ];
      if( h < best_h ) // immer jeweils bestes Ergebnis merken!!
      {
        best_h = h;
        i = 0;
        while( breite[i] != 0 )
        {
          beste[i] = breite[i];
          i++;
        }
        // wieder in anderer Ebene einpacken
        akt_lsg = ( akt_lsg + 1 ) & 1;
      }
      i = such_max_indirekt( zeit, nr, hohe );
      if( zeit[ nr[i] ] != h )
      { // wenn das höchste nicht mehr die Gesamthöhe bestimmt
        // -> nächstes Verfahren
        break;
      }
      if( breite[i] == n-1 ) // kann noch verbreitert werden ?
      {
        break;
      }
      breite[i]++;
      zeit[ nr[i] ] = (*laufzeit)( nr[i]+1, breite[i]+1 );
    }
    // Flächenminimum suchen
    i = 0;
    while( breite[i] != 0 )
    {
      j = breite[i] + 1;
      h = zeit[ nr[i] ] * ((double)(j));
      while( j < n )
      {
        j++;
        a = (*laufzeit)( nr[i]+1, j );
        if( a * ((double)(j)) < h )
        {
          h = a * ((double)(j));
          breite[i] = j - 1;
          zeit[ nr[i] ] = a;
        }
      }
      i++;
    }
    j = ( akt_lsg + 1 ) & 1;
    for( i=0; i<m; i++ )
    {
      // Originalreihenfolge zwecks breite-Zuordnung invertiert merken
      rf[ nr[i] ] = i;
      reihenf[j][i] = nr[i]; // Einpackreihenfolge merken
    }
    // nach Höhe sortieren
    // mehr ist nicht durcheinander gekommen
    merge_modul_zeit( zeit, nr, nrh, 0, hohe );
    // die Liste mit den hohen und die andere zusammenmischen
    misch_modul_zeit( zeit, nr, nrh, 0, hohe, m );
    while( 1 )
    {
      // einpacken
      h = 0.0;
      i = n;
      for( j=0; j<m; j++ )
      {
        stelle[akt_lsg][ nr[j] ] = i;
        i++;
        if( rf[ nr[j] ] < hohe )
        {
          i += breite[ rf[ nr[j] ] ];
          if( i > n )
          {
            i = breite[ rf[ nr[j] ] ] + 1;
            a = h;
            h += zeit[ nr[j] ];
            stelle[akt_lsg][ nr[j] ] = 0;
          }
        }
        else
        {
          if( i > n )
          {
            i = 1;
            a = h;
            h += zeit[ nr[j] ];
            stelle[akt_lsg][ nr[j] ] = 0;
          }
        }
        lsg[akt_lsg][ nr[j] ][0] = a; // Anfangszeit
        lsg[akt_lsg][ nr[j] ][1] = a + zeit[ nr[j] ]; // Endzeit
        reihenf[akt_lsg][j] = nr[j];
      }
      if( h < best_h )
      {
        i = 0;
        while( breite[i] != 0 )
        {
          beste[i] = breite[i];
          i++;
        }
        best_h = h;
        akt_lsg = ( akt_lsg + 1 ) & 1;
      }
      // das höchste Modul flächenminimal verbreitern
      k = 0;
      i = 0;
      // wenn alle hohen durch sind, kann abgebrochen werden
      while( i < hohe )
      {
        if( rf[ nr[k] ] < hohe )
        {
          j = breite[ rf[ nr[k] ] ] + 2;
          if( j <= n ) // ist noch eine Verbreiterung möglich
          {
            break;
          }
          i++;
        }
        k++;
      }
      if( i == hohe )
      {
        break; // es kann kein hohes Modul mehr verbreitert werden
      }
      breite[ rf[ nr[k] ] ]++;
      zeit[ nr[k] ] = (*laufzeit)( nr[k]+1, j );
      h = zeit[ nr[k] ] * ((double)(j));
      while( j < n )
      {
        j++;
        a = (*laufzeit)( nr[k]+1, j );
        if( a * ((double)(j)) < h )
        {
          h = a * ((double)(j));
          breite[ rf[ nr[k] ] ] = j - 1;
          zeit[ nr[k] ] = a;
        }
      }
      // das verbreiterte einsortieren
      misch_modul_zeit( zeit, nr, nrh, k, k+1, m );
    }
    // Ergebnis erstellen
    akt_lsg = ( akt_lsg + 1 ) & 1;
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, reihenf[akt_lsg][j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double(
            &fi, lsg[akt_lsg][ reihenf[akt_lsg][j] ][0] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double(
            &fi, lsg[akt_lsg][ reihenf[akt_lsg][j] ][1] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int(
            &fi, stelle[akt_lsg][ reihenf[akt_lsg][j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( rf[ reihenf[akt_lsg][j] ] < hohe )
      {
        if( save_int_r(
              &fi,
              ( stelle[akt_lsg][ reihenf[akt_lsg][j] ] +
                beste[ rf[ reihenf[akt_lsg][j] ] ] + 1 ) ) != 0 )
        {
          return 3;
        }
      }
      else
      {
        if( save_int_r(
              &fi, stelle[akt_lsg][ reihenf[akt_lsg][j] ] + 1 ) != 0 )
        {
          return 3;
        }
      }
    }
    fclose( fi );
  }
  return 0;
}
#endif


#ifdef FORMEL

// Hauptroutine für Formel; siehe oben
int optimiere_formel(
  char*  ergebnis,
  int    m,
  int    n,
  double ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  double* p_ende;
  int* breite;
  int* beste;
  int* nr;
  int* nrh;
  int* rf;
  double* zeit;
  double*** lsg;
  int** stelle;
  int** reihenf;
  int i;
  baum heap;

  // Speicherplatz reservieren
  p_ende = (double*) malloc( n * sizeof( double ) );
  if( p_ende == NULL )
  {
    return 2;
  }
  // eins mehr wegen Abschluß-Null
  breite = (int*) malloc( ( n + 1 ) * sizeof( int ) );
  if( breite == NULL )
  {
    return 2;
  }
  // eins mehr wegen Abschluß-Null
  beste = (int*) malloc( ( n + 1 ) * sizeof( int ) );
  if( beste == NULL )
  {
    return 2;
  }
  nr = (int*) malloc( m * sizeof( int ) );
  if( nr == NULL )
  {
    return 2;
  }
  nrh = (int*) malloc( m * sizeof( int ) );
  if( nrh == NULL )
  {
    return 2;
  }
  rf = (int*) malloc( m * sizeof( int ) );
  if( rf == NULL )
  {
    return 2;
  }
  zeit = (double*) malloc( m * sizeof( double ) );
  if( zeit == NULL )
  {
    return 2;
  }
  lsg = (double***) malloc( 2 * sizeof( double** ) );
  if( lsg == NULL )
  {
    return 2;
  }
  lsg[0] = (double**) malloc( m * sizeof( double* ) );
  if( lsg[0] == NULL )
  {
    return 2;
  }
  lsg[1] = (double**) malloc( m * sizeof( double* ) );
  if( lsg[1] == NULL )
  {
    return 2;
  }
  for( i=0; i<m; i++ )
  {
    lsg[0][i] = (double*) malloc( 2 * sizeof( double ) );
    if( lsg[0][i] == NULL )
    {
      return 2;
    }
    lsg[1][i] = (double*) malloc( 2 * sizeof( double ) );
    if( lsg[1][i] == NULL )
    {
      return 2;
    }
  }
  stelle = (int**) malloc( 2 * sizeof( int* ) );
  if( stelle == NULL )
  {
    return 2;
  }
  stelle[0] = (int*) malloc( m * sizeof( int ) );
  if( stelle[0] == NULL )
  {
    return 2;
  }
  stelle[1] = (int*) malloc( m * sizeof( int ) );
  if( stelle[1] == NULL )
  {
    return 2;
  }
  reihenf = (int**) malloc( 2 * sizeof( int* ) );
  if( reihenf == NULL )
  {
    return 2;
  }
  reihenf[0] = (int*) malloc( m * sizeof( int ) );
  if( reihenf[0] == NULL )
  {
    return 2;
  }
  reihenf[1] = (int*) malloc( m * sizeof( int ) );
  if( reihenf[1] == NULL )
  {
    return 2;
  }
  init_size_baum( &heap, n );
  for( i=0; (int)(i)<n; i++ ) // gleich n Instanzen anlegen
  {
    if( insert_ll_baum( &heap, NULL ) == NULL )
    {
      return 2;
    }
  }

#ifdef ABBRUCH
  n = opti_formel_abbr( ergebnis, laufzeit, n, m, breite, beste, nr,
                        nrh, rf, zeit, p_ende, lsg, stelle, reihenf,
                        &heap );
#else
  n = opti_formel_weiter( ergebnis, laufzeit, n, m, breite, beste, nr,
                          nrh, rf, zeit, p_ende, lsg, stelle, reihenf,
                          &heap );
#endif

  free_baum( &heap );
  free( reihenf[1] );
  free( reihenf[0] );
  free( reihenf );
  free( stelle[1] );
  free( stelle[0] );
  free( stelle );
  for( i=0; i<m; i++ )
  {
    free( lsg[0][i] );
    free( lsg[1][i] );
  }
  free( lsg[1] );
  free( lsg[0] );
  free( lsg );
  free( zeit );
  free( rf );
  free( nrh );
  free( nr );
  free( p_ende );
  free( breite );
  free( beste );
  return n;
}
#endif
