// Testprogramm für die einzelnen Routinen
// Aufruf test();
// Fehler: 1->zu wenig Speicher zum Erstellen von Beispielen
//         2->Beispielgenerator konnte Ergebnisdatei nicht anlegen
//         3->Beispielgenerator hat nicht genug Einträge im Listenfeld /
//            Fehler in Datei
//         4->Beispielgenerator konnte Eingabedatei nicht finden
//         21->Eingabedatei für Optimierung fehlerhaft
//         22->zu wenig Speicher für Optimierung
//         23->Optimierungsalgorithmus kann Ergebnisdatei nicht anlegen
//         24->Optimierungsalgorithmus hat Eingabedatei nicht gefunden
//         25->Optimierungsalgorithmus: Fehler bei Arbeit mit Hilfsdatei
//         41->Analyseroutine hat Ergebnisdatei nicht gefunden
//         42->der Analyseroutine übergebene Ergebnisdatei ist
//             fehlerhaft
//         43->Analyseroutine hat Zeitendatei nicht gefunden
//         44->der Analyseroutine übergebene Zeitendatei ist fehlerhaft
//         45->Dateneintragung der Analyseroutine in LOGFILE hat nicht
//             geklappt
//         60->kann LOGDATEI nicht anlegen / Eintragung vornehmen
//         81->zu wenig Speicher für Laufzeittabelle
//         82->Configdatei nicht gefunden
//         83->Configdatei fehlerhaft

// Standardbeispiele werden immer im Modus Formel ausgeführt
#ifdef STANDARD
#define FORMEL
#include "../prgx/bsp_bau_std.h"
#else
#include "bsp_bau1.h"
#endif

#ifdef USE_BMZ
#include "opti1b.h"
#endif

#ifdef USE_HMZ
#include "opti1h.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "../prgx/load.h"
#include "../prgx/save.h"


// -------------- Formelbasiert, zufällige Beispiele -------------------
#if defined( FORMEL ) && !defined( STANDARD )

// Feld zur Ablage der Laufzeitdaten und Modulgrößen
double** laufz;
int      m;
int      n;

// Laufzeitroutine
double laufzeit(
  int modulnummer,
  int prozessorenanzahl )
{
  return laufz[ modulnummer-1 ][ prozessorenanzahl-1 ];
}

// fügt Eintrag bzgl. der Mindestergebnisgüte ins logfile hinzu
// Fehler: 1->Ergebnisdatei nicht gefunden
//         2->Ergebnisdatei fehlerhaft
//         5->Dateneintragung in LOGFILE hat nicht geklappt
int analyse(
  char* logfile,
  char* ergeb )
{
  FILE* fi;
  double h, a, min_flae, max_hoehe;
  char c;
  int i;
  // Gesamthöhe auslesen
  fi = fopen( ergeb, "r" );
  if( fi == NULL )
  {
    return 1;
  }
  h = 0;
  while( lese_next_char( &fi, "i", &c ) == 0 )
  {
    if( lese_next_double_in_line( &fi, &a ) != 0 )
    {
      return 2;
    }
    if( a > h )
    {
      h = a;
    }
    if( lese_next_ret( &fi ) != 0 )
    {
      break;
    }
  }
  if( h == 0 )
  {
    return 2;
  }
  fclose( fi );
  // min_flae, max_höhe aus Laufzeiten bestimmen
  min_flae = laufz[0][0];
  max_hoehe = laufz[0][ n-1 ];
  for( i=1; i<m; i++ )
  {
    min_flae += laufz[i][0];
    if( max_hoehe < laufz[i][ n-1 ] )
    {
      max_hoehe = laufz[i][ n-1 ];
    }
  }
  min_flae /= ((double)(n));
  if( max_hoehe > min_flae )
  {
    min_flae = max_hoehe;
  }
  h /= min_flae;
  fi = fopen( logfile, "a" );
  if( fi == NULL )
  {
    return 5;
  }
  if( save_double_r( &fi, h ) != 0 )
  {
    return 5;
  }
  fclose( fi );
  return 0;
}

int test()
{
  int i, k;
  int fehler;
  double guete, max_guete, a;
  time_t anf, end, rahmen, zeit;
  FILE* fi;
  char dummy_datei[] = "dummy.txt";
  fi = fopen( LOGFILE, "w" );
  if( fi == NULL )
  {
    return 60;
  }
  if( save_text( &fi, "Die einzelnen Gütefaktoren lauten:\n" ) != 0 )
  {
    return 60;
  }
  fclose( fi );
  fehler = init_laufzeit( &laufz, CONFIGDATEI );
  if( fehler != 0 )
  {
    return 80 + fehler;
  }
  time( &anf ); // Startzeit
  for( i=0; i<TESTANZAHL; i++ )
  {
    fehler = baue_bsp_array( laufz, &m, &n, CONFIGDATEI );
    if( fehler != 0 )
    {
      return fehler;
    }
    for( k=0; k<OPTIMIERUNGEN; k++ )
    {
      fehler = optimiere_formel( ERGEBNISDATEI, m, n, &(laufzeit) );
      if( fehler != 0 )
      {
        return fehler + 20;
      }
    }
    fehler = analyse( LOGFILE, ERGEBNISDATEI );
    if( fehler != 0 )
    {
      return fehler + 40;
    }
  }
  time( &end ); // Endzeit = Rahmenanfangszeit
  for( i=0; i<TESTANZAHL; i++ )
  {
    fehler = baue_bsp_array( laufz, &m, &n, CONFIGDATEI );
    if( fehler != 0 )
    {
      return fehler;
    }
    fehler = analyse( dummy_datei, ERGEBNISDATEI );
    if( fehler != 0 )
    {
      return fehler + 40;
    }
  }
  time( &rahmen );
  zeit = 2 * end - anf - rahmen;
  frei_laufzeit( laufz, CONFIGDATEI );
  remove( dummy_datei );
  remove( ERGEBNISDATEI );
  fi = fopen( LOGFILE, "r" );
  if( fi == NULL )
  {
    return 60;
  }
  max_guete = 0.0;
  guete = 0.0;
  while( 1 )
  {
    if( lese_next_double_r( &fi, &a ) != 0 )
    {
      break;
    }
    guete += a;
    if( a > max_guete )
    {
      max_guete = a;
    }
  }
  fclose( fi );
  fi = fopen( LOGFILE, "a" );
  if( fi == NULL )
  {
    return 60;
  }
  if( save_text( &fi, "\nDurchschnittliche Zeit: " ) != 0 )
  {
    return 60;
  }
  if( save_double( &fi,
                   ( ((double)(zeit)) /
                     ((double)(TESTANZAHL)) /
                     ((double)(OPTIMIERUNGEN)) ) ) != 0 )
  {
    return 60;
  }
  if( save_text(
        &fi,
        " Sekunden\nDurchschnittliche Mindestergebnisgüte: " ) != 0 )
  {
    return 60;
  }
  if( save_double_r( &fi, guete / ((double)(TESTANZAHL)) ) != 0 )
  {
    return 60;
  }
  if( save_text( &fi, "Schlechteste Mindestgüte: " ) != 0 )
  {
    return 60;
  }
  if( save_double_r( &fi, max_guete ) != 0 )
  {
    return 60;
  }
  fclose( fi );
  printf( "\nDurchschnittliche Zeit: %lf Sekunden\n",
          ( ((double)(zeit)) /
            ((double)(TESTANZAHL)) /
            ((double)(OPTIMIERUNGEN)) ) );
  printf( "Durchschnittliche Mindestergebnisgüte: %lf\n",
          guete / ((double)(TESTANZAHL)) );
  printf( "Schlechteste Mindestgüte: %lf\n", max_guete );
  return 0;
}
#endif


// -------------- Dateibasiert, zufällige Beispiele --------------------
#if !defined( FORMEL ) && !defined( STANDARD )

// fügt Eintrag bzgl. der Mindestergebnisgüte ins logfile hinzu
// Fehler: 1->Ergebnisdatei nicht gefunden
//         2->Ergebnisdatei fehlerhaft
//         3->Zeitendatei nicht gefunden
//         4->Zeitendatei fehlerhaft
//         5->Dateneintragung in LOGFILE hat nicht geklappt
int analyse(
  char* logfile,
  char* ergeb,
  char* daten )
{
  FILE* fi;
  double h, a, min_flae, max_hoehe;
  char c;
  int i, n;
  // Gesamthöhe auslesen
  fi = fopen( ergeb, "r" );
  if( fi == NULL )
  {
    return 1;
  }
  h = 0;
  while( lese_next_char( &fi, "i", &c ) == 0 )
  {
    if( lese_next_double_in_line( &fi, &a ) != 0 )
    {
      return 2;
    }
    if( a > h )
    {
      h = a;
    }
    if( lese_next_ret( &fi ) != 0 )
    {
      break;
    }
  }
  if( h == 0 )
  {
    return 2;
  }
  fclose( fi );
  // n, min_flae, max_hoehe aus Zeitendatei bestimmen
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 3;
  }
  if( lese_next_double( &fi, &min_flae ) != 0 )
  {
    return 4;
  }
  n = 1;
  max_hoehe = 0;
  while( lese_next_double_in_line( &fi, &a ) == 0 )
  {
    n++;
    max_hoehe = a;
  }
  while( 1 )
  {
    if( lese_next_double( &fi, &a ) != 0 )
    {
      break;
    }
    min_flae += a;
    for( i=1; i<n; i++ )
    {
      if( lese_next_double_in_line( &fi, &a ) != 0 )
      {
        return 4;
      }
    }
    if( a > max_hoehe )
    {
      max_hoehe = a;
    }
  }
  min_flae /= ((double)(n));
  if( max_hoehe > min_flae )
  {
    min_flae = max_hoehe;
  }
  h /= min_flae;
  fclose( fi );
  fi = fopen( logfile, "a" );
  if( fi == NULL )
  {
    return 5;
  }
  if( save_double_r( &fi, h ) != 0 )
  {
    return 5;
  }
  fclose( fi );
  return 0;
}

int test()
{
  int i, k;
  int fehler;
  double guete, max_guete, a;
  time_t anf, end, rahmen, zeit;
  FILE* fi;
  char dummy_datei[] = "dummy.txt";
  fi = fopen( LOGFILE, "w" );
  if( fi == NULL )
  {
    return 60;
  }
  if( save_text( &fi, "Die einzelnen Gütefaktoren lauten:\n" ) != 0 )
  {
    return 60;
  }
  fclose( fi );
  time( &anf ); // Startzeit
  for( i=0; i<TESTANZAHL; i++ )
  {
    fehler = baue_bsp_datei( ZEITENDATEI, CONFIGDATEI );
    if( fehler != 0 )
    {
      return fehler;
    }
    for( k=0; k<OPTIMIERUNGEN; k++ )
    {
      fehler = optimiere_datei( ERGEBNISDATEI, ZEITENDATEI );
      if( fehler != 0 )
      {
        return fehler + 20;
      }
    }
    fehler = analyse( LOGFILE, ERGEBNISDATEI, ZEITENDATEI );
    if( fehler != 0 )
    {
      return fehler + 40;
    }
  }
  time( &end ); // Endzeit = Rahmenanfangszeit
  for( i=0; i<TESTANZAHL; i++ )
  {
    fehler = baue_bsp_datei( ZEITENDATEI, CONFIGDATEI );
    if( fehler != 0 )
    {
      return fehler;
    }
    fehler = analyse( dummy_datei, ERGEBNISDATEI, ZEITENDATEI );
    if( fehler != 0 )
    {
      return fehler + 40;
    }
  }
  time( &rahmen );
  zeit = 2 * end - anf - rahmen;
  remove( dummy_datei );
  remove( ERGEBNISDATEI );
  remove( ZEITENDATEI );
  fi = fopen( LOGFILE, "r" );
  if( fi == NULL )
  {
    return 60;
  }
  max_guete = 0.0;
  guete = 0.0;
  while( 1 )
  {
    if( lese_next_double_r( &fi, &a ) != 0 )
    {
      break;
    }
    guete += a;
    if( a > max_guete )
    {
      max_guete = a;
    }
  }
  fclose( fi );
  fi = fopen( LOGFILE, "a" );
  if( fi == NULL )
  {
    return 60;
  }
  if( save_text( &fi, "\nDurchschnittliche Zeit: " ) != 0 )
  {
    return 60;
  }
  if( save_double( &fi,
                   ( ((double)(zeit)) /
                     ((double)(TESTANZAHL)) /
                     ((double)(OPTIMIERUNGEN)) ) ) != 0 )
  {
    return 60;
  }
  if( save_text(
        &fi,
        " Sekunden\nDurchschnittliche Mindestergebnisgüte: " ) != 0 )
  {
    return 60;
  }
  if( save_double_r( &fi, guete / ((double)(TESTANZAHL)) ) != 0 )
  {
    return 60;
  }
  if( save_text( &fi, "Schlechteste Mindestgüte: " ) != 0 )
  {
    return 60;
  }
  if( save_double_r( &fi, max_guete ) != 0 )
  {
    return 60;
  }
  fclose( fi );
  printf( "\nDurchschnittliche Zeit: %lf Sekunden\n",
          ( ((double)(zeit)) /
            ((double)(TESTANZAHL)) /
            ((double)(OPTIMIERUNGEN)) ) );
  printf( "Durchschnittliche Mindestergebnisgüte: %lf\n",
          guete / ((double)(TESTANZAHL)) );
  printf( "Schlechteste Mindestgüte: %lf\n", max_guete );
  return 0;
}

#endif


// ------------------ Standardbeispiele --------------------------------
#ifdef STANDARD

// Feld zur Ablage der Laufzeitdaten und Modulgrößen
double** laufz;
int      m;
int      n;

// Laufzeitroutine
double laufzeit(
  int modulnummer,
  int prozessorenanzahl )
{
  return laufz[ modulnummer-1 ][ prozessorenanzahl-1 ];
}

// fügt Eintrag bzgl. der Mindestergebnisgüte ins logfile hinzu
// Fehler: 1->Ergebnisdatei nicht gefunden
//         2->Ergebnisdatei fehlerhaft
//         5->Dateneintragung in LOGFILE hat nicht geklappt
int analyse(
  char* logfile,
  char* ergeb )
{
  FILE* fi;
  double h, h_opt, a, min_flae, max_hoehe, guet;
  char c;
  int i;
  // Gesamthöhe auslesen
  fi = fopen( ergeb, "r" );
  if( fi == NULL )
  {
    return 1;
  }
  h = 0;
  while( lese_next_char( &fi, "i", &c ) == 0 )
  {
    if( lese_next_double_in_line( &fi, &a ) != 0 )
    {
      return 2;
    }
    if( a > h )
    {
      h = a;
    }
    if( lese_next_ret( &fi ) != 0 )
    {
      break;
    }
  }
  if( h == 0 )
  {
    return 2;
  }
  fclose( fi );
  // min_flae, max_hoehe aus Laufzeiten bestimmen
  min_flae = laufz[0][0];
  max_hoehe = laufz[0][ n-1 ];
  for( i=1; i<m; i++ )
  {
    min_flae += laufz[i][0];
    if( max_hoehe < laufz[i][ n-1 ] )
    {
      max_hoehe = laufz[i][ n-1 ];
    }
  }
  min_flae /= ((double)(n));
  if( max_hoehe > min_flae )
  {
    h_opt = max_hoehe;
  }
  else
  {
    h_opt = min_flae;
  }
  fi = fopen( logfile, "a" );
  if( fi == NULL )
  {
    return 5;
  }
  if( save_int( &fi, n ) != 0 )
  {
    return 5;
  }
  if( save_text( &fi, "    " ) != 0 )
  {
    return 5;
  }
  if( save_double( &fi, h ) != 0 )
  {
    return 5;
  }
  guet = h / h_opt;
  if( save_text( &fi, "    " ) != 0 )
  {
    return 5;
  }
  if( save_double( &fi, guet ) != 0 )
  {
    return 5;
  }
  h /= min_flae;
  if( save_text( &fi, "    " ) != 0 )
  {
    return 5;
  }
  if( save_double_r( &fi, h ) != 0 )
  {
    return 5;
  }
  fclose( fi );
  return 0;
}

int test()
{
  int i;
  int fehler, na, ne;
  graph gr;
  FILE* fi;
  fehler = bsp_groesse( &m, &na, &ne, STANDARDDATEI );
  if( fehler != 0 )
  {
    return fehler + 81;
  }
  // Speicher für Laufzeiten reservieren
  laufz = (double**)malloc( m * sizeof( double* ) );
  if( laufz == NULL )
  {
    return 81;
  }
  for( i=0; i<m; i++ )
  {
    laufz[i] = (double*)malloc( ne * sizeof( double ) );
    if( laufz[i] == NULL )
    {
      return 81;
    }
  }
  fi = fopen( LOGFILE, "w" );
  if( fi == NULL )
  {
    return 60;
  }
  if( save_text( &fi,
                 "Prozessoren    erzielte Höhe             "
                 "Gütefaktor                "
                 "Mehraufwand wg. Parallelisierung\n" ) != 0 )
  {
    return 60;
  }
  fclose( fi );
  init_graph( &gr );
  fehler = baue_standard_bsp( laufz, ne, &gr, STANDARDDATEI );
  if( fehler != 0 )
  {
    return fehler;
  }
  free_graph( &gr );
  for( n=na; n<=ne; n++ )
  {
    fehler = optimiere_formel( ERGEBNISDATEI, m, n, &(laufzeit) );
    if( fehler != 0 )
    {
      return fehler + 20;
    }
    fehler = analyse( LOGFILE, ERGEBNISDATEI );
    if( fehler != 0 )
    {
      return fehler + 40;
    }
  }
  for( i=0; i<m; i++ )
  {
    free( laufz[i] );
  }
  free( laufz );
  remove( ERGEBNISDATEI );
  return 0;
}
#endif



int main()
{
  int f;
  // nur wegen Fehlerausgabe; sonst könnte auch test
  // die main-Routine sein
  f = test();
  printf( "\nRückgabewert: %d\n", f );
  return f;
}
