// Optimierungsroutinen non-moldable, kein Modulabhängigkeitsgraph;
// Verwendung vom "BMZ" - Einfügealgorithmus
// optimiert Module deren Laufzeiten in daten stehen
// (i-te Zeile j-te Spalte enthält Laufzeit vom i-ten Modul mit
//  j Prozessoren)
// oder mit Hilfe der übergebenen Funktion laufzeit
// (modulnummer>=1, prozessorenanzahl>=1) ermittelt werden können
//
// Aufruf: optimiere_datei( char* ergebnisdateiname,
//                          char* datendateiname )
//   oder: optimiere_formel( char*  ergebnisdateiname,
//                           int    gesamtmodulanzahl,
//                           int    gesamtprozessorenanzahl,
//                           double ((*laufzeit)( modulnummer,
//                                                prozessorenanzahl )) )
//
// Fehler: 1->Eingabedatei fehlerhaft
//         2->zu wenig Speicher
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
//         5->Fehler bei Arbeit mit Hilfsdatei


#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include "../prgx/baum.h"
#include "../prgx/load.h"
#include "../prgx/save.h"
#include "../prgx/suchsort.h"


// ----------------------- DATEI - ROUTINEN MIT ABBRUCH ----------------
#if defined( ABBRUCH ) && !defined( FORMEL ) && !defined( NICHTSORTIEREN )

// optimiert Module deren Laufzeiten in daten stehen;
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m], p_ende[n],
//                       startzeit[2][m], startpos[2][m]
// Fehler: 1->Eingabedatei fehlerhaft
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
int opti_datei_abbr_sort(
  char*    ergebnis,
  char*    daten,
  int      n,
  int      m,
  double*  p_ende,
  int*     breite,
  int*     beste,
  int*     nr,
  int*     nrh,
  double*  zeit,
  double*  best_zeit,
  int*     best_rf,
  double** startzeit,
  int**    startpos )
{
  FILE* fi;
  int pos;
  double h, best_h, a, flae;
  int i, j, k;
  int akt_start;
  // flächenminimale Modullaufzeiten einlesen
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  pos = ftell( fi );
  flae = 0.0; // aktuelle Fläche
  for( i=0; i<m; i++ )
  {
    if( lese_next_double( &fi, &(zeit[i]) ) != 0 )
    {
      return 1;
    }
    j = 0;
    while( zeit[i] == 0 )
    {
      if( lese_next_double_in_line( &fi, &(zeit[i]) ) != 0 )
      {
        return 1;
      }
      j++;
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    flae += zeit[i] * ( breite[i] + 1 );
    nr[i] = i;
  }
  // nach Breite sortieren
  merge_modul_breit_zeit( zeit, breite, nr, nrh, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  akt_start = 0;
  while( 1 )
  {
    // einpacken
    for( i=0; i<n; i++ )
    {
      p_ende[i] = 0.0;
    }
    for( j=0; j<m; j++ )
    {
      i = such_min_breit( p_ende, n, breite[ nr[j] ], &a );
      startzeit[akt_start][ nr[j] ] = a;
      startpos[akt_start][ nr[j] ] = i;
      a += zeit[ nr[j] ];
      for( k=i; k<=i+breite[ nr[j] ]; k++ )
      {
        p_ende[k] = a;
      }
    }
    h = p_ende[ such_max( p_ende, n ) ];
    if( h < best_h )
    {
      if( 3 * flae > ((double)(n)) * h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 0; // Werte in breite nutzen
        break;
      }
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      akt_start = ( akt_start + 1 ) & 1; // bestes Ergebnis retten
      best_h = h;
    }
    else
    {
      if( 3 * flae > ((double)(n)) * best_h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 1; // Werte in beste nutzen
        break;
      }
    }
    i = such_max_indirekt( zeit, nr, m );
    flae -= ( breite[ nr[i] ] + 1 ) * zeit[ nr[i] ];
    // das höchste Modul flächenminimal verbreitern
    j = breite[ nr[i] ] + 2;
    if( j > n ) // ist noch eine Verbreiterung möglich
    {
      break;
    }
    breite[ nr[i] ]++;
    fseek( fi, pos, SEEK_SET );
    if( lese_pos_double( &fi, nr[i], breite[ nr[i] ],
                         &(zeit[ nr[i] ]) ) != 0 )
    {
      return 1;
    }
    h = zeit[ nr[i] ] * ((double)(j));
    while( j < n )
    {
      j++;
      if( lese_next_double( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[i] ] = j - 1;
        zeit[ nr[i] ] = a;
      }
    }
    flae += zeit[ nr[i] ] * ( breite[ nr[i] ] + 1 );
    // das ehemals höchste einsortieren
    misch_modul_breit_zeit( zeit, breite, nr, nrh, 0, i, i+1 );
  }
  // bestes Ergebnis wieder herstellen
  if( i == 1 ) // aus beste übernehmen
  {
    fclose( fi );
    // Ergebnis erstellen
    akt_start = ( akt_start + 1 ) & 1;
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, best_rf[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, startzeit[akt_start][ best_rf[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi,
                       ( startzeit[akt_start][ best_rf[j] ] +
                         best_zeit[ best_rf[j] ] ) ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, startpos[akt_start][ best_rf[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi,
                      ( startpos[akt_start][ best_rf[j] ] +
                        beste[ best_rf[j] ] + 1 ) ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  else // Reihenfolge OK, Breiten in breite
  {
    fclose( fi );
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, nr[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, startzeit[akt_start][ nr[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double(
            &fi,
            startzeit[akt_start][ nr[j] ] + zeit[ nr[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, startpos[akt_start][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi,
                      ( startpos[akt_start][ nr[j] ] +
                        breite[ nr[j] ] + 1 ) ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  return 0;
}
#endif


#if defined( ABBRUCH ) && !defined( FORMEL ) && defined( NICHTSORTIEREN )

// optimiert Module deren Laufzeiten in daten stehen;
// keine Sortierung nach Laufzeiten innerhalb von Modulen
// gleicher Breite
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m], p_ende[n], nrhh[n],
//                       startzeit[2][m], startpos[2][m]
// Fehler: 1->Eingabedatei fehlerhaft
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
int opti_datei_abbr_nosort(
  char*    ergebnis,
  char*    daten,
  int      n,
  int      m,
  double*  p_ende,
  int*     breite,
  int*     beste,
  int*     nr,
  int*     nrh,
  int*     nrhh,
  double*  zeit,
  double*  best_zeit,
  int*     best_rf,
  double** startzeit,
  int**    startpos )
{
  FILE* fi;
  int pos;
  double h, best_h, a, flae;
  int i, j, k;
  int akt_start;
  // flächenminimale Modullaufzeiten einlesen
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  pos = ftell( fi );
  flae = 0.0; // aktuelle Fläche
  for( i=0; i<m; i++ )
  {
    if( lese_next_double( &fi, &(zeit[i]) ) != 0 )
    {
      return 1;
    }
    j = 0;
    while( zeit[i] == 0 )
    {
      if( lese_next_double_in_line( &fi, &(zeit[i]) ) != 0 )
      {
        return 1;
      }
      j++;
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    flae += zeit[i] * ( breite[i] + 1 );
    nr[i] = i;
  }
  // nach Breite sortieren
  bucket_modul_breit( breite, nr, nrh, nrhh, n, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  akt_start = 0;
  while( 1 )
  {
    // einpacken
    for( i=0; i<n; i++ )
    {
      p_ende[i] = 0.0;
    }
    for( j=0; j<m; j++ )
    {
      i = such_min_breit( p_ende, n, breite[ nr[j] ], &a );
      startzeit[akt_start][ nr[j] ] = a;
      startpos[akt_start][ nr[j] ] = i;
      a += zeit[ nr[j] ];
      for( k=i; k<=i+breite[ nr[j] ]; k++ )
      {
        p_ende[k] = a;
      }
    }
    h = p_ende[ such_max( p_ende, n ) ];
    if( h < best_h )
    {
      if( 3 * flae > ((double)(n)) * h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 0; // Werte in breite nutzen
        break;
      }
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      akt_start = ( akt_start + 1 ) & 1; // bestes Ergebnis retten
      best_h = h;
    }
    else
    { // beim Einpacken Flächenfaktor 3 erreicht
      if( 3 * flae > ((double)(n)) * best_h )
      {
        i = 1; // Werte in beste nutzen
        break;
      }
    }
    i = such_max_indirekt( zeit, nr, m );
    flae -= ( breite[ nr[i] ] + 1 ) * zeit[ nr[i] ];
    // das höchste Modul flächenminimal verbreitern
    j = breite[ nr[i] ] + 2;
    if( j > n ) // ist noch eine Verbreiterung möglich
    {
      break;
    }
    breite[ nr[i] ]++;
    fseek( fi, pos, SEEK_SET );
    if( lese_pos_double( &fi, nr[i], breite[ nr[i] ],
                         &(zeit[ nr[i] ]) ) != 0 )
    {
      return 1;
    }
    h = zeit[ nr[i] ] * ((double)(j));
    while( j < n )
    {
      j++;
      if( lese_next_double( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[i] ] = j - 1;
        zeit[ nr[i] ] = a;
      }
    }
    flae += zeit[ nr[i] ] * ( breite[ nr[i] ] + 1 );
    misch_modul_breit( breite, nr, nrh, 0, i, i+1 );
  }
  // bestes Ergebnis wieder herstellen
  if( i == 1 ) // aus beste übernehmen
  {
    fclose( fi );
    // Ergebnis erstellen
    akt_start = ( akt_start + 1 ) & 1;
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, best_rf[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, startzeit[akt_start][ best_rf[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, 
                       ( startzeit[akt_start][ best_rf[j] ] +
                         best_zeit[ best_rf[j] ] ) ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, startpos[akt_start][ best_rf[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi,
                      ( startpos[akt_start][ best_rf[j] ] +
                        beste[ best_rf[j] ] + 1 ) ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  else // Reihenfolge OK, Breiten in breite
  {
    fclose( fi );
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, nr[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, startzeit[akt_start][ nr[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi,
                       ( startzeit[akt_start][ nr[j] ] +
                         zeit[ nr[j] ] ) ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, startpos[akt_start][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi, 
                      ( startpos[akt_start][ nr[j] ] +
                        breite[ nr[j] ] + 1 ) ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  return 0;
}
#endif


// ----------------------- DATEI - ROUTINEN OHNE ABBRUCH ---------------
#if !defined( ABBRUCH ) && !defined( FORMEL ) && !defined( NICHTSORTIEREN )

// optimiert Module deren Laufzeiten in daten stehen
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m], p_ende[n],
//                       startzeit[2][m], startpos[2][m]
// Fehler: 1->Eingabedatei fehlerhaft
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
int opti_datei_weiter_sort(
  char*    ergebnis,
  char*    daten,
  int      n,
  int      m,
  double*  p_ende,
  int*     breite,
  int*     beste,
  int*     nr,
  int*     nrh,
  double*  zeit,
  double*  best_zeit,
  int*     best_rf,
  double** startzeit,
  int**    startpos )
{
  FILE* fi;
  int pos;
  double h, best_h, a;
  int i, j, k;
  int akt_start;
  // flächenminimale Modullaufzeiten einlesen
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  pos = ftell( fi );
  for( i=0; i<m; i++ )
  {
    if( lese_next_double( &fi, &(zeit[i]) ) != 0 )
    {
      return 1;
    }
    j = 0;
    while( zeit[i] == 0 )
    {
      if( lese_next_double_in_line( &fi, &(zeit[i]) ) != 0 )
      {
        return 1;
      }
      j++;
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    nr[i] = i;
  }
  // nach Breite sortieren
  merge_modul_breit_zeit( zeit, breite, nr, nrh, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  akt_start = 0;
  while( 1 )
  {
    // einpacken
    for( i=0; i<n; i++ )
    {
      p_ende[i] = 0.0;
    }
    for( j=0; j<m; j++ )
    {
      i = such_min_breit( p_ende, n, breite[ nr[j] ], &a );
      startzeit[akt_start][ nr[j] ] = a;
      startpos[akt_start][ nr[j] ] = i;
      a += zeit[ nr[j] ];
      for( k=i; k<=i+breite[ nr[j] ]; k++ )
      {
        p_ende[k] = a;
      }
    }
    h = p_ende[ such_max( p_ende, n ) ];
    if( h < best_h )
    {
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      best_h = h;
      akt_start = ( akt_start + 1 ) & 1;
    }
    i = 0;
    while( breite[ nr[i] ] == n - 1 )
    {
      i++;
      if( i == m )
      {
        break;
      }
    }
    if( i == m )
    {
      break;
    }
    for( j=i+1; j<m; j++ ) // höchstes verbreiterbares Modul suchen
    {
      if( breite[ nr[j] ] < n - 1 && zeit[ nr[j] ] > zeit[ nr[i] ] )
      {
        i = j;
      }
    }
    // das höchste mögliche Modul flächenminimal verbreitern
    j = breite[ nr[i] ] + 2;
    breite[ nr[i] ]++;
    fseek( fi, pos, SEEK_SET );
    if( lese_pos_double( &fi, nr[i], breite[ nr[i] ],
                         &(zeit[ nr[i] ]) ) != 0 )
    {
      return 1;
    }
    h = zeit[ nr[i] ] * ((double)(j));
    while( j < n )
    {
      j++;
      if( lese_next_double( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[i] ] = j - 1;
        zeit[ nr[i] ] = a;
      }
    }
    // das verbreiterte einsortieren
    misch_modul_breit_zeit( zeit, breite, nr, nrh, 0, i, i+1 );
  }
  // bestes Ergebnis wieder herstellen
  fclose( fi );
  // Ergebnis erstellen
  akt_start = ( akt_start + 1 ) & 1;
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 3;
  }
  for( j=0; j<m; j++ )
  {
    if( save_text( &fi, "Modul " ) != 0 )
    {
      return 3;
    }
    if( save_int( &fi, best_rf[j] + 1 ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, ": von " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi, startzeit[akt_start][ best_rf[j] ] ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi, 
                     ( startzeit[akt_start][ best_rf[j] ] +
                       best_zeit[ best_rf[j] ] ) ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " auf Prozessoren " ) != 0 )
    {
      return 3;
    }
    if( save_int( &fi, startpos[akt_start][ best_rf[j] ] + 1 ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    if( save_int_r( &fi,
                    ( startpos[akt_start][ best_rf[j] ] +
                      beste[ best_rf[j] ] + 1 ) ) != 0 )
    {
      return 3;
    }
  }
  fclose( fi );
  return 0;
}
#endif


#if !defined( ABBRUCH ) && !defined( FORMEL ) && defined( NICHTSORTIEREN )

// optimiert Module deren Laufzeiten in daten stehen;
// keine Sortierung nach Laufzeiten innerhalb von Modulen mit gleicher
// Breite
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m], p_ende[n], nrhh[n],
//                       startzeit[2][m], startpos[2][m]
// Fehler: 1->Eingabedatei fehlerhaft
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
int opti_datei_weiter_nosort(
  char*    ergebnis,
  char*    daten,
  int      n,
  int      m,
  double*  p_ende,
  int*     breite,
  int*     beste,
  int*     nr,
  int*     nrh,
  int*     nrhh,
  double*  zeit,
  double*  best_zeit,
  int*     best_rf,
  double** startzeit,
  int**    startpos )
{
  FILE* fi;
  int pos;
  double h, best_h, a;
  int i, j, k;
  int akt_start;
  // flächenminimale Modullaufzeiten einlesen
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  pos = ftell( fi );
  for( i=0; i<m; i++ )
  {
    if( lese_next_double( &fi, &(zeit[i]) ) != 0 )
    {
      return 1;
    }
    j = 0;
    while( zeit[i] == 0 )
    {
      if( lese_next_double_in_line( &fi, &(zeit[i]) ) != 0 )
      {
        return 1;
      }
      j++;
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    nr[i] = i;
  }
  // nach Breite sortieren
  bucket_modul_breit( breite, nr, nrh, nrhh, n, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  akt_start = 0;
  while( 1 )
  {
    // einpacken
    for( i=0; i<n; i++ )
    {
      p_ende[i] = 0.0;
    }
    for( j=0; j<m; j++ )
    {
      i = such_min_breit( p_ende, n, breite[ nr[j] ], &a );
      startzeit[akt_start][ nr[j] ] = a;
      startpos[akt_start][ nr[j] ] = i;
      a += zeit[ nr[j] ];
      for( k=i; k<=i+breite[ nr[j] ]; k++ )
      {
        p_ende[k] = a;
      }
    }
    h = p_ende[ such_max( p_ende, n ) ];
    if( h < best_h )
    {
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      best_h = h;
      akt_start = ( akt_start + 1 ) & 1;
    }
    i = 0;
    while( breite[ nr[i] ] == n - 1 )
    {
      i++;
      if( i == m )
      {
        break;
      }
    }
    if( i == m )
    {
      break;
    }
    for( j=i+1; j<m; j++ ) // höchstes verbreiterbares Modul suchen
    {
      if( breite[ nr[j] ] < n - 1 && zeit[ nr[j] ] > zeit[ nr[i] ] )
      {
        i = j;
      }
    }
    // das höchste mögliche Modul flächenminimal verbreitern
    j = breite[ nr[i] ] + 2;
    breite[ nr[i] ]++;
    fseek( fi, pos, SEEK_SET );
    if( lese_pos_double( &fi, nr[i], breite[ nr[i] ],
                         &(zeit[ nr[i] ]) ) != 0 )
    {
      return 1;
    }
    h = zeit[ nr[i] ] * ((double)(j));
    while( j < n )
    {
      j++;
      if( lese_next_double( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[i] ] = j - 1;
        zeit[ nr[i] ] = a;
      }
    }
    misch_modul_breit( breite, nr, nrh, 0, i, i+1 );
  }
  // bestes Ergebnis wieder herstellen
  fclose( fi );
  // Ergebnis erstellen
  akt_start = ( akt_start + 1 ) & 1;
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 3;
  }
  for( j=0; j<m; j++ )
  {
    if( save_text( &fi, "Modul " ) != 0 )
    {
      return 3;
    }
    if( save_int( &fi, best_rf[j] + 1 ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, ": von " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi, startzeit[akt_start][ best_rf[j] ] ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi,
                     ( startzeit[akt_start][ best_rf[j] ] +
                       best_zeit[ best_rf[j] ] ) ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " auf Prozessoren " ) != 0 )
    {
      return 3;
    }
    if( save_int( &fi, startpos[akt_start][ best_rf[j] ] + 1 ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    if( save_int_r( &fi,
                    ( startpos[akt_start][ best_rf[j] ] +
                      beste[ best_rf[j] ] + 1 ) ) != 0 )
    {
      return 3;
    }
  }
  fclose( fi );
  return 0;
}
#endif


#ifndef FORMEL

// Hauptroutine für Datei; siehe oben
int optimiere_datei(
  char* ergebnis,
  char* daten )
{
  FILE* fi;
  int m;
  int n;
  double h;
  double* p_ende;
  int* nr;
  int* nrh;
  double* zeit;
  int* beste;
  int* breite;
  double* best_zeit;
  int* best_rf;
  int** startpos;
  double** startzeit;
  // für Bucketsort benötigte Variable
#ifdef NICHTSORTIEREN
  int* nrhh;
#endif
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  // Prozessorenanzahl bestimmen
  if( lese_next_double( &fi, &h ) != 0 )
  {
    return 1;
  }
  n = 1;
  while( lese_next_double_in_line( &fi, &h ) == 0 )
  {
    n++;
  }
  // Modulanzahl bestimmen
  if( lese_next_ret( &fi ) != 0 )
  {
    return 1;
  }
  m = 1;
  while( 1 )
  {
    if( lese_next_double_r( &fi, &h ) != 0 )
    {
      break;
    }
    m++;
  }
  fclose( fi );
  p_ende = (double*) malloc( n * sizeof( double ) );
  if( p_ende == NULL )
  {
    return 2;
  }
  nr = (int*) malloc( m * sizeof( int ) );
  if( nr == NULL )
  {
    return 2;
  }
  nrh = (int*) malloc( m * sizeof( int ) );
  if( nrh == NULL )
  {
    return 2;
  }
  zeit = (double*) malloc( m * sizeof( double ) );
  if( zeit == NULL )
  {
    return 2;
  }
  breite = (int*) malloc( m * sizeof( int ) );
  if( breite == NULL )
  {
    return 2;
  }
  beste = (int*) malloc( m * sizeof( int ) );
  if( beste == NULL )
  {
    return 2;
  }
  startzeit = (double**) malloc( 2 * sizeof( double* ) );
  if( startzeit == NULL )
  {
    return 2;
  }
  startzeit[0] = (double*) malloc( m * sizeof( double ) );
  if( startzeit[0] == NULL )
  {
    return 2;
  }
  startzeit[1] = (double*) malloc( m * sizeof( double ) );
  if( startzeit[1] == NULL )
  {
    return 2;
  }
  best_zeit = (double*) malloc( m * sizeof( double ) );
  if( best_zeit == NULL )
  {
    return 2;
  }
  best_rf = (int*) malloc( m * sizeof( int ) );
  if( best_rf == NULL )
  {
    return 2;
  }
  startpos = (int**) malloc( 2 * sizeof( int* ) );
  if( startpos == NULL )
  {
    return 2;
  }
  startpos[0] = (int*) malloc( m * sizeof( int ) );
  if( startpos[0] == NULL )
  {
    return 2;
  }
  startpos[1] = (int*) malloc( m * sizeof( int ) );
  if( startpos[1] == NULL )
  {
    return 2;
  }
#ifdef NICHTSORTIEREN
  nrhh = (int*) malloc( n * sizeof( int ) );
  if( nrhh == NULL )
  {
    return 2;
  }
#ifdef ABBRUCH
  n = opti_datei_abbr_nosort( ergebnis, daten, n, m, p_ende, breite,
                              beste, nr, nrh, nrhh, zeit, best_zeit,
                              best_rf, startzeit, startpos );
#else
  n = opti_datei_weiter_nosort( ergebnis, daten, n, m, p_ende, breite,
                                beste, nr, nrh, nrhh, zeit, best_zeit,
                                best_rf, startzeit, startpos );
#endif
  free( nrhh );
#else // NICHTSORTIEREN
#ifdef ABBRUCH
  n = opti_datei_abbr_sort( ergebnis, daten, n, m, p_ende, breite,
                            beste, nr, nrh, zeit, best_zeit, best_rf,
                            startzeit, startpos );
#else
  n = opti_datei_weiter_sort( ergebnis, daten, n, m, p_ende, breite,
                              beste, nr, nrh, zeit, best_zeit, best_rf,
                              startzeit, startpos );
#endif
#endif // NICHTSORTIEREN
  free( zeit );
  free( nrh );
  free( nr );
  free( breite );
  free( beste );
  free( best_zeit );
  free( best_rf );
  free( startzeit[1] );
  free( startzeit[0] );
  free( startzeit );
  free( startpos[1] );
  free( startpos[0] );
  free( startpos );
  free( p_ende );
  return n;
}

#endif


// --------------------- FORMEL - ROUTINEN MIT ABBRUCH -----------------
#if defined( ABBRUCH ) && defined( FORMEL ) && !defined( NICHTSORTIEREN )

// optimiert Module deren Laufzeiten in daten stehen
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m], p_ende[n],
//                       startzeit[2][m], startpos[2][m]
// Fehler: 3->kann Ergebnisdatei nicht anlegen
int opti_formel_abbr_sort(
  char*    ergebnis,
  double   ((*laufzeit)(int modulnummer, int prozessorenanzahl)),
  int      n,
  int      m,
  double*  p_ende,
  int*     breite,
  int*     beste,
  int*     nr,
  int*     nrh,
  double*  zeit,
  double*  best_zeit,
  int*     best_rf,
  double** startzeit,
  int**    startpos )
{
  FILE* fi;
  double h, best_h, a, flae;
  int i, j, k;
  int akt_start;
  // flächenminimale Modullaufzeiten einlesen
  flae = 0.0; // aktuelle Fläche
  for( i=0; i<m; i++ )
  {
    zeit[i] = (*laufzeit)( i+1, 1 );
    j = 0;
    while( zeit[i] == 0 )
    {
      j++;
      zeit[i] = (*laufzeit)( i+1, j+1 );
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      a = (*laufzeit)( i+1, j+1 );
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    flae += zeit[i] * ( breite[i] + 1 );
    nr[i] = i;
  }
  // nach Breite sortieren
  merge_modul_breit_zeit( zeit, breite, nr, nrh, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  akt_start = 0;
  while( 1 )
  {
    // einpacken
    for( i=0; i<n; i++ )
    {
      p_ende[i] = 0.0;
    }
    for( j=0; j<m; j++ )
    {
      i = such_min_breit( p_ende, n, breite[ nr[j] ], &a );
      startzeit[akt_start][ nr[j] ] = a;
      startpos[akt_start][ nr[j] ] = i;
      a += zeit[ nr[j] ];
      for( k=i; k<=i+breite[ nr[j] ]; k++ )
      {
        p_ende[k] = a;
      }
    }
    h = p_ende[ such_max( p_ende, n ) ];
    if( h < best_h )
    {
      if( 3 * flae > ((double)(n)) * h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 0; // Werte in breite nutzen
        break;
      }
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      best_h = h;
      akt_start = ( akt_start + 1 ) & 1;
    }
    else
    {
      if( 3 * flae > ((double)(n)) * best_h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 1; // Werte in beste nutzen
        break;
      }
    }
    i = such_max_indirekt( zeit, nr, m );
    flae -= ( breite[ nr[i] ] + 1 ) * zeit[ nr[i] ];
    // das höchste Modul flächenminimal verbreitern
    j = breite[ nr[i] ] + 2;
    if( j > n ) // ist noch eine Verbreiterung möglich
    {
      break;
    }
    breite[ nr[i] ]++;
    zeit[ nr[i] ] = (*laufzeit)( nr[i] + 1, breite[ nr[i] ] + 1 );
    h = zeit[ nr[i] ] * ((double)(j));
    while( j < n )
    {
      j++;
      a = (*laufzeit)( nr[i] + 1, j );
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[i] ] = j - 1;
        zeit[ nr[i] ] = a;
      }
    }
    flae += zeit[ nr[i] ] * ( breite[ nr[i] ] + 1 );
    // das ehemals höchste einsortieren
    misch_modul_breit_zeit( zeit, breite, nr, nrh, 0, i, i+1 );
  }
  // bestes Ergebnis wieder herstellen
  if( i == 1 ) // aus beste übernehmen
  {
    // Ergebnis erstellen
    akt_start = ( akt_start + 1 ) & 1;
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, best_rf[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, startzeit[akt_start][ best_rf[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi,
                       ( startzeit[akt_start][ best_rf[j] ] +
                         best_zeit[ best_rf[j] ] ) ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, startpos[akt_start][ best_rf[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi,
                      ( startpos[akt_start][ best_rf[j] ] +
                        beste[ best_rf[j] ] + 1 ) ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  else // Reihenfolge OK, Breiten in breite
  {
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, nr[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, startzeit[akt_start][ nr[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi,
                       ( startzeit[akt_start][ nr[j] ] +
                         zeit[ nr[j] ] ) ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, startpos[akt_start][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi,
                      ( startpos[akt_start][ nr[j] ] +
                        breite[ nr[j] ] + 1 ) ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  return 0;
}
#endif


#if defined( ABBRUCH ) && defined( FORMEL ) && defined( NICHTSORTIEREN )

// optimiert Module deren Laufzeiten in daten stehen;
// keine Sortierung nach Laufzeiten innerhalb von Modulen mit gleicher
// Breite
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m], p_ende[n], nrhh[n],
//                       startzeit[2][m], startpos[2][m]
// Fehler: 3->kann Ergebnisdatei nicht anlegen
int opti_formel_abbr_nosort(
  char*    ergebnis,
  double   ((*laufzeit)(int modulnummer, int prozessorenanzahl)),
  int      n,
  int      m,
  double*  p_ende,
  int*     breite,
  int*     beste,
  int*     nr,
  int*     nrh,
  int*     nrhh,
  double*  zeit,
  double*  best_zeit,
  int*     best_rf,
  double** startzeit,
  int**    startpos )
{
  FILE* fi;
  double h, best_h, a, flae;
  int i, j, k;
  int akt_start;
  // flächenminimale Modullaufzeiten einlesen
  flae = 0.0; // aktuelle Fläche
  for( i=0; i<m; i++ )
  {
    zeit[i] = (*laufzeit)( i+1, 1 );
    j = 0;
    while( zeit[i] == 0 )
    {
      j++;
      zeit[i] = (*laufzeit)( i+1, j+1 );
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      a = (*laufzeit)( i+1, j+1 );
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    flae += zeit[i] * ( breite[i] + 1 );
    nr[i] = i;
  }
  // nach Breite sortieren
  bucket_modul_breit( breite, nr, nrh, nrhh, n, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  akt_start = 0;
  while( 1 )
  {
    // einpacken
    for( i=0; i<n; i++ )
    {
      p_ende[i] = 0.0;
    }
    for( j=0; j<m; j++ )
    {
      i = such_min_breit( p_ende, n, breite[ nr[j] ], &a );
      startzeit[akt_start][ nr[j] ] = a;
      startpos[akt_start][ nr[j] ] = i;
      a += zeit[ nr[j] ];
      for( k=i; k<=i+breite[ nr[j] ]; k++ )
      {
        p_ende[k] = a;
      }
    }
    h = p_ende[ such_max( p_ende, n ) ];
    if( h < best_h )
    {
      if( 3 * flae > ((double)(n)) * h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 0; // Werte in breite nutzen
        break;
      }
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      best_h = h;
      akt_start = ( akt_start + 1 ) & 1;
    }
    else
    {
      if( 3 * flae > ((double)(n)) * best_h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 1; // Werte in beste nutzen
        break;
      }
    }
    i = such_max_indirekt( zeit, nr, m );
    flae -= ( breite[ nr[i] ] + 1 ) * zeit[ nr[i] ];
    // das höchste Modul flächenminimal verbreitern
    j = breite[ nr[i] ] + 2;
    if( j > n ) // ist noch eine Verbreiterung möglich
    {
      break;
    }
    breite[ nr[i] ]++;
    zeit[ nr[i] ] = (*laufzeit)( nr[i] + 1, breite[ nr[i] ] + 1 );
    h = zeit[ nr[i] ] * ((double)(j));
    while( j < n )
    {
      j++;
      a = (*laufzeit)( nr[i] + 1, j );
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[i] ] = j - 1;
        zeit[ nr[i] ] = a;
      }
    }
    flae += zeit[ nr[i] ] * ( breite[ nr[i] ] + 1 );
    // das ehemals höchste einsortieren
    misch_modul_breit( breite, nr, nrh, 0, i, i+1 );
  }
  // bestes Ergebnis wieder herstellen
  if( i == 1 ) // aus beste übernehmen
  {
    // Ergebnis erstellen
    akt_start = ( akt_start + 1 ) & 1;
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, best_rf[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, startzeit[akt_start][ best_rf[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi,
                       ( startzeit[akt_start][ best_rf[j] ] +
                         best_zeit[ best_rf[j] ] ) ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, startpos[akt_start][ best_rf[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi,
                      ( startpos[akt_start][ best_rf[j] ] +
                        beste[ best_rf[j] ] + 1 ) ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  else // Reihenfolge OK, Breiten in breite
  {
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    for( j=0; j<m; j++ )
    {
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, nr[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, startzeit[akt_start][ nr[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi,
                       ( startzeit[akt_start][ nr[j] ] +
                         zeit[ nr[j] ] ) ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, startpos[akt_start][ nr[j] ] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_int_r( &fi,
                      ( startpos[akt_start][ nr[j] ] +
                        breite[ nr[j] ] + 1 ) ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  return 0;
}
#endif


// ------------------- FORMEL - ROUTINEN OHNE ABBRUCH ------------------
#if !defined( ABBRUCH ) && defined( FORMEL ) && !defined( NICHTSORTIEREN )

// optimiert Module deren Laufzeiten in daten stehen
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m], p_ende[n],
//                       startzeit[2][m], startpos[2][m]
// Fehler: 3->kann Ergebnisdatei nicht anlegen
int opti_formel_weiter_sort(
  char*    ergebnis,
  double   ((*laufzeit)(int modulnummer, int prozessorenanzahl)),
  int      n,
  int      m,
  double*  p_ende,
  int*     breite,
  int*     beste,
  int*     nr,
  int*     nrh,
  double*  zeit,
  double*  best_zeit,
  int*     best_rf,
  double** startzeit,
  int**    startpos )
{
  FILE* fi;
  double h, best_h, a;
  int i, j, k;
  int akt_start;
  // flächenminimale Modullaufzeiten einlesen
  for( i=0; i<m; i++ )
  {
    zeit[i] = (*laufzeit)( i+1, 1 );
    j = 0;
    while( zeit[i] == 0 )
    {
      j++;
      zeit[i] = (*laufzeit)( i+1, j+1 );
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      a = (*laufzeit)( i+1, j+1 );
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    nr[i] = i;
  }
  // nach Breite sortieren
  merge_modul_breit_zeit( zeit, breite, nr, nrh, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  akt_start = 0;
  while( 1 )
  {
    // einpacken
    for( i=0; i<n; i++ )
    {
      p_ende[i] = 0.0;
    }
    for( j=0; j<m; j++ )
    {
      i = such_min_breit( p_ende, n, breite[ nr[j] ], &a );
      startzeit[akt_start][ nr[j] ] = a;
      startpos[akt_start][ nr[j] ] = i;
      a += zeit[ nr[j] ];
      for( k=i; k<=i+breite[ nr[j] ]; k++ )
      {
        p_ende[k] = a;
      }
    }
    h = p_ende[ such_max( p_ende, n ) ];
    if( h < best_h )
    {
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      best_h = h;
      akt_start = ( akt_start + 1 ) & 1;
    }
    i = 0;
    while( breite[ nr[i] ] == n - 1 )
    {
      i++;
      if( i == m )
      {
        break;
      }
    }
    if( i == m )
    {
      break;
    }
    for( j=i+1; j<m; j++ ) // höchstes verbreiterbares Modul suchen
    {
      if( breite[ nr[j] ] < n - 1 && zeit[ nr[j] ] > zeit[ nr[i] ] )
      {
        i = j;
      }
    }
    // das höchste mögliche Modul flächenminimal verbreitern
    j = breite[ nr[i] ] + 2;
    breite[ nr[i] ]++;
    zeit[ nr[i] ] = (*laufzeit)( nr[i] + 1, j );
    h = zeit[ nr[i] ] * ((double)(j));
    while( j < n )
    {
      j++;
      a = (*laufzeit)( nr[i] + 1, j );
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[i] ] = j - 1;
        zeit[ nr[i] ] = a;
      }
    }
    // das verbreiterte einsortieren
    misch_modul_breit_zeit( zeit, breite, nr, nrh, 0, i, i+1 );
  }
  // bestes Ergebnis wieder herstellen
  // Ergebnis erstellen
  akt_start = ( akt_start + 1 ) & 1;
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 3;
  }
  for( j=0; j<m; j++ )
  {
    if( save_text( &fi, "Modul " ) != 0 )
    {
      return 3;
    }
    if( save_int( &fi, best_rf[j] + 1 ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, ": von " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi, startzeit[akt_start][ best_rf[j] ] ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi,
                     ( startzeit[akt_start][ best_rf[j] ] +
                       best_zeit[ best_rf[j] ] ) ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " auf Prozessoren " ) != 0 )
    {
      return 3;
    }
    if( save_int( &fi, startpos[akt_start][ best_rf[j] ] + 1 ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    if( save_int_r( &fi,
                    ( startpos[akt_start][ best_rf[j] ] +
                      beste[ best_rf[j] ] + 1 ) ) != 0 )
    {
      return 3;
    }
  }
  fclose( fi );
  return 0;
}
#endif


#if !defined( ABBRUCH ) && defined( FORMEL ) && defined( NICHTSORTIEREN )
// optimiert Module deren Laufzeiten in daten stehen; 
// keine Sortierung nach Laufzeiten innerhalb von Modulen mit gleicher
// Breite
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m], nrhh[n], p_ende[n],
//                       startzeit[2][m], startpos[2][m]
// Fehler: 3->kann Ergebnisdatei nicht anlegen
int opti_formel_weiter_nosort(
  char*    ergebnis,
  double   ((*laufzeit)(int modulnummer, int prozessorenanzahl)),
  int      n,
  int      m,
  double*  p_ende,
  int*     breite,
  int*     beste,
  int*     nr,
  int*     nrh,
  int*     nrhh,
  double*  zeit,
  double*  best_zeit,
  int*     best_rf,
  double** startzeit,
  int**    startpos )
{
  FILE* fi;
  double h, best_h, a;
  int i, j, k;
  int akt_start;
  // flächenminimale Modullaufzeiten einlesen
  for( i=0; i<m; i++ )
  {
    zeit[i] = (*laufzeit)( i+1, 1 );
    j = 0;
    while( zeit[i] == 0 )
    {
      j++;
      zeit[i] = (*laufzeit)( i+1, j+1 );
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      a = (*laufzeit)( i+1, j+1 );
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    nr[i] = i;
  }
  // nach Breite sortieren
  bucket_modul_breit( breite, nr, nrh, nrhh, n, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  akt_start = 0;
  while( 1 )
  {
    // einpacken
    for( i=0; i<n; i++ )
    {
      p_ende[i] = 0.0;
    }
    for( j=0; j<m; j++ )
    {
      i = such_min_breit( p_ende, n, breite[ nr[j] ], &a );
      startzeit[akt_start][ nr[j] ] = a;
      startpos[akt_start][ nr[j] ] = i;
      a += zeit[ nr[j] ];
      for( k=i; k<=i+breite[ nr[j] ]; k++ )
      {
        p_ende[k] = a;
      }
    }
    h = p_ende[ such_max( p_ende, n ) ];
    if( h < best_h )
    {
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      best_h = h;
      akt_start = ( akt_start + 1 ) & 1;
    }
    i = 0;
    while( breite[ nr[i] ] == n - 1 )
    {
      i++;
      if( i == m )
      {
        break;
      }
    }
    if( i == m )
    {
      break;
    }
    for( j=i+1; j<m; j++ ) // höchstes verbreiterbares Modul suchen
    {
      if( breite[ nr[j] ] < n - 1 && zeit[ nr[j] ] > zeit[ nr[i] ] )
      {
        i = j;
      }
    }
    // das höchste mögliche Modul flächenminimal verbreitern
    j = breite[ nr[i] ] + 2;
    breite[ nr[i] ]++;
    zeit[ nr[i] ] = (*laufzeit)( nr[i] + 1, j );
    h = zeit[ nr[i] ] * ((double)(j));
    while( j < n )
    {
      j++;
      a = (*laufzeit)( nr[i] + 1, j );
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[i] ] = j - 1;
        zeit[ nr[i] ] = a;
      }
    }
    // das verbreiterte einsortieren
    misch_modul_breit( breite, nr, nrh, 0, i, i+1 );
  }
  // bestes Ergebnis wieder herstellen
  // Ergebnis erstellen
  akt_start = ( akt_start + 1 ) & 1;
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 3;
  }
  for( j=0; j<m; j++ )
  {
    if( save_text( &fi, "Modul " ) != 0 )
    {
      return 3;
    }
    if( save_int( &fi, best_rf[j] + 1 ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, ": von " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi, startzeit[akt_start][ best_rf[j] ] ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi,
                     ( startzeit[akt_start][ best_rf[j] ] +
                       best_zeit[ best_rf[j] ] ) ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " auf Prozessoren " ) != 0 )
    {
      return 3;
    }
    if( save_int( &fi, startpos[akt_start][ best_rf[j] ] + 1 ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    if( save_int_r( &fi,
                    ( startpos[akt_start][ best_rf[j] ] +
                      beste[ best_rf[j] ] + 1 ) ) != 0 )
    {
      return 3;
    }
  }
  fclose( fi );
  return 0;
}
#endif


#ifdef FORMEL

// Hauptroutine; siehe oben
int optimiere_formel(
  char*  ergebnis,
  int    m,
  int    n,
  double ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  double* p_ende;
  int* nr;
  int* nrh;
  double* zeit;
  int* beste;
  int* breite;
  double** startzeit;
  double* best_zeit;
  int* best_rf;
  int** startpos;
#ifdef NICHTSORTIEREN // Variable für Bucket-Sort
  int* nrhh;
#endif
  p_ende = (double*) malloc( n * sizeof( double ) );
  if( p_ende == NULL )
  {
    return 2;
  }
  nr = (int*) malloc( m * sizeof( int ) );
  if( nr == NULL )
  {
    return 2;
  }
  nrh = (int*) malloc( m * sizeof( int ) );
  if( nrh == NULL )
  {
    return 2;
  }
  zeit = (double*) malloc( m * sizeof( double ) );
  if( zeit == NULL )
  {
    return 2;
  }
  breite = (int*) malloc( m * sizeof( int ) );
  if( breite == NULL )
  {
    return 2;
  }
  beste = (int*) malloc( m * sizeof( int ) );
  if( beste == NULL )
  {
    return 2;
  }
  startzeit = (double**) malloc( 2 * sizeof( double* ) );
  if( startzeit == NULL )
  {
    return 2;
  }
  startzeit[0] = (double*) malloc( m * sizeof( double ) );
  if( startzeit[0] == NULL )
  {
    return 2;
  }
  startzeit[1] = (double*) malloc( m * sizeof( double ) );
  if( startzeit[1] == NULL )
  {
    return 2;
  }
  best_zeit = (double*) malloc( m * sizeof( double ) );
  if( best_zeit == NULL )
  {
    return 2;
  }
  best_rf = (int*) malloc( m * sizeof( int ) );
  if( best_rf == NULL )
  {
    return 2;
  }
  startpos = (int**) malloc( 2 * sizeof( int* ) );
  if( startpos == NULL )
  {
    return 2;
  }
  startpos[0] = (int*) malloc( m * sizeof( int ) );
  if( startpos[0] == NULL )
  {
    return 2;
  }
  startpos[1] = (int*) malloc( m * sizeof( int ) );
  if( startpos[1] == NULL )
  {
    return 2;
  }
#ifdef NICHTSORTIEREN
  nrhh = (int*) malloc( n * sizeof( int ) );
  if( nrhh == NULL )
  {
    return 2;
  }
#ifdef ABBRUCH
  n = opti_formel_abbr_nosort( ergebnis, laufzeit, n, m, p_ende, breite,
                               beste, nr, nrh, nrhh, zeit, best_zeit,
                               best_rf, startzeit, startpos );
#else
  n = opti_formel_weiter_nosort( ergebnis, laufzeit, n, m, p_ende,
                                 breite, beste, nr, nrh, nrhh, zeit,
                                 best_zeit, best_rf, startzeit,
                                 startpos );
#endif
  free( nrhh );
#else // NICHTSORTIEREN
#ifdef ABBRUCH
  n = opti_formel_abbr_sort( ergebnis, laufzeit, n, m, p_ende, breite,
                             beste, nr, nrh, zeit, best_zeit, best_rf,
                             startzeit, startpos );
#else
  n = opti_formel_weiter_sort( ergebnis, laufzeit, n, m, p_ende, breite,
                               beste, nr, nrh, zeit, best_zeit, best_rf,
                               startzeit, startpos );
#endif
#endif // NICHTSORTIEREN
  free( zeit );
  free( nrh );
  free( nr );
  free( breite );
  free( beste );
  free( best_zeit );
  free( best_rf );
  free( startzeit[1] );
  free( startzeit[0] );
  free( startzeit );
  free( startpos[1] );
  free( startpos[0] );
  free( startpos );
  free( p_ende );
  return n;
}
#endif
