// Optimierungsroutinen non-moldable, kein Modulabhängigkeitsgraph;
// Verwendung vom "HMZ" - Einfügealgorithmus
// optimiert Module deren Laufzeiten in daten stehen
// (i-te Zeile j-te Spalte enthält Laufzeit vom i-ten Modul mit
//  j Prozessoren)
// oder mit Hilfe der übergebenen Funktion laufzeit
// (modulnummer>=1, prozessorenanzahl>=1) ermittelt werden können
//
// Aufruf: optimiere_datei( char* ergebnisdateiname,
//                          char* datendateiname )
//   oder: optimiere_formel( char*  ergebnisdateiname,
//                           int    gesamtmodulanzahl,
//                           int    gesamtprozessorenanzahl,
//                           double ((*laufzeit)( modulnummer,
//                                                prozessorenanzahl )) )
//
// Fehler: 1->Eingabedatei fehlerhaft
//         2->zu wenig Speicher
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
//         5->Fehler bei Arbeit mit Hilfsdatei


#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include "../prgx/baum.h"
#include "../prgx/load.h"
#include "../prgx/save.h"
#include "../prgx/suchsort.h"


// --------------------- DATEI - ROUTINE MIT ABBRUCH -------------------
#if defined( ABBRUCH ) && !defined( FORMEL )

// optimiert Module deren Laufzeiten in daten stehen
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m]
// Fehler: 1->Eingabedatei fehlerhaft
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
int opti_datei_abbr(
  char*   ergebnis,
  char*   daten,
  int     n,
  int     m,
  int*    breite,
  int*    beste,
  int*    nr,
  int*    nrh,
  double* zeit,
  double* best_zeit,
  int*    best_rf )
{
  FILE* fi;
  int pos;
  double h, best_h, a, flae;
  int i, j;
  // flächenminimale Modullaufzeiten einlesen
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  pos = ftell( fi );
  flae = 0.0; // aktuelle Fläche
  for( i=0; i<m; i++ )
  {
    if( lese_next_double( &fi, &(zeit[i]) ) != 0 )
    {
      return 1;
    }
    j = 0;
    while( zeit[i] == 0 )
    {
      if( lese_next_double_in_line( &fi, &(zeit[i]) ) != 0 )
      {
        return 1;
      }
      j++;
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    flae += zeit[i] * ( breite[i] + 1 );
    nr[i] = i;
    lese_next_ret( &fi );
  }
  // nach Höhe sortieren
  merge_modul_zeit( zeit, nr, nrh, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  while( 1 )
  {
    // einpacken
    h = 0.0;
    i = n;
    for( j=0; j<m; j++ )
    {
      i += breite[ nr[j] ] + 1;
      if( i > n )
      {
        i = breite[ nr[j] ] + 1;
        h += zeit[ nr[j] ];
      }
    }
    if( h < best_h )
    {
      if( 3 * flae > ((double)(n)) * h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 0; // Werte in breite nutzen
        break;
      }
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      best_h = h;
    }
    else
    {
      if( 3 * flae > ((double)(n)) * best_h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 1; // Werte in beste nutzen
        break;
      }
    }
    flae -= ( breite[ nr[0] ] + 1 ) * zeit[ nr[0] ];
    // das höchste Modul flächenminimal verbreitern
    j = breite[ nr[0] ] + 2;
    if( j > n ) // ist noch eine Verbreiterung möglich
    {
      break;
    }
    breite[ nr[0] ]++;
    fseek( fi, pos, SEEK_SET );
    if( lese_pos_double( &fi, nr[0], breite[ nr[0] ],
                         &(zeit[ nr[0] ]) ) != 0 )
    {
      return 1;
    }
    h = zeit[ nr[0] ] * ((double)(j));
    while( j < n )
    {
      j++;
      if( lese_next_double( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[0] ] = j - 1;
        zeit[ nr[0] ] = a;
      }
    }
    flae += zeit[ nr[0] ] * ( breite[ nr[0] ] + 1 );
    // das ehemals höchste einsortieren
    misch_modul_zeit( zeit, nr, nrh, 0, 1, m );
  }
  // bestes Ergebnis wieder herstellen
  if( i == 1 ) // aus beste übernehmen
  {
    fclose( fi );
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    a = 0.0;
    i = n;
    for( j=0; j<m; j++ )
    {
      if( i + beste[ best_rf[j] ] + 1 > n )
      {
        i = 0;
        h = a;
        a += best_zeit[ best_rf[j] ];
      }
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, best_rf[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, h ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, h + best_zeit[ best_rf[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      i++;
      if( save_int( &fi, i ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      i += beste[ best_rf[j] ];
      if( save_int_r( &fi, i ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  else // Reihenfolge OK, Breiten in breite
  {
    fclose( fi );
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    a = 0.0;
    i = n;
    for( j=0; j<m; j++ )
    {
      if( i + breite[ nr[j] ] + 1 > n )
      {
        i = 0;
        h = a;
        a += zeit[ nr[j] ];
      }
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, nr[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, h ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, h + zeit[ nr[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      i++;
      if( save_int( &fi, i ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      i += breite[ nr[j] ];
      if( save_int_r( &fi, i ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  return 0;
}
#endif

// -------------------- DATEI - ROUTINE OHNE ABBRUCH -------------------
#if !defined( ABBRUCH ) && !defined( FORMEL )

// optimiert Module deren Laufzeiten in daten stehen
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m]
// Fehler: 1->Eingabedatei fehlerhaft
//         3->kann Ergebnisdatei nicht anlegen
//         4->Eingabedatei nicht gefunden
int opti_datei_weiter(
  char*   ergebnis,
  char*   daten,
  int     n,
  int     m,
  int*    breite,
  int*    beste,
  int*    nr,
  int*    nrh,
  double* zeit,
  double* best_zeit,
  int*    best_rf )
{
  FILE* fi;
  int pos;
  double h, best_h, a;
  int i, j;
  // flächenminimale Modullaufzeiten einlesen
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  pos = ftell( fi );
  for( i=0; i<m; i++ )
  {
    if( lese_next_double( &fi, &(zeit[i]) ) != 0 )
    {
      return 1;
    }
    j = 0;
    while( zeit[i] == 0 )
    {
      if( lese_next_double_in_line( &fi, &(zeit[i]) ) != 0 )
      {
        return 1;
      }
      j++;
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    nr[i] = i;
    lese_next_ret( &fi );
  }
  // nach Höhe sortieren
  merge_modul_zeit( zeit, nr, nrh, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  while( 1 )
  {
    // einpacken
    h = 0.0;
    i = n;
    for( j=0; j<m; j++ )
    {
      i += breite[ nr[j] ] + 1;
      if( i > n )
      {
        i = breite[ nr[j] ] + 1;
        h += zeit[ nr[j] ];
      }
    }

    if( h < best_h )
    {
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      best_h = h;
    }
    i = 0;
    while( breite[ nr[i] ] == n - 1 )
    {
      i++;
      if( i == m )
      {
        break;
      }
    }
    if( i == m )
    {
      break;
    }
    // das höchste mögliche Modul flächenminimal verbreitern
    j = breite[ nr[i] ] + 2;
    breite[ nr[i] ]++;
    fseek( fi, pos, SEEK_SET );
    if( lese_pos_double( &fi, nr[i], breite[ nr[i] ],
                         &(zeit[ nr[i] ]) ) != 0 )
    {
      return 1;
    }
    h = zeit[ nr[i] ] * ((double)(j));
    while( j < n )
    {
      j++;
      if( lese_next_double( &fi, &a ) != 0 )
      {
        return 1;
      }
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[i] ] = j - 1;
        zeit[ nr[i] ] = a;
      }
    }
    // das verbreiterte einsortieren
    misch_modul_zeit( zeit, nr, nrh, i, i+1, m );
  }
  // bestes Ergebnis wieder herstellen
  fclose( fi );
  // Ergebnis erstellen
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 3;
  }
  a = 0.0;
  i = n;
  for( j=0; j<m; j++ )
  {
    if( i + beste[ best_rf[j] ] + 1 > n )
    {
      i = 0;
      h = a;
      a += best_zeit[ best_rf[j] ];
    }
    if( save_text( &fi, "Modul " ) != 0 )
    {
      return 3;
    }
    if( save_int( &fi, best_rf[j] + 1 ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, ": von " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi, h ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi, h + best_zeit[ best_rf[j] ] ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " auf Prozessoren " ) != 0 )
    {
      return 3;
    }
    i++;
    if( save_int( &fi, i ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    i += beste[ best_rf[j] ];
    if( save_int_r( &fi, i ) != 0 )
    {
      return 3;
    }
  }
  fclose( fi );
  return 0;
}
#endif


#ifndef FORMEL

// Hauptroutine für Datei; siehe oben
int optimiere_datei(
  char* ergebnis,
  char* daten )
{
  FILE* fi;
  int m;
  int n;
  double h;
  int* nr;
  int* nrh;
  double* zeit;
  int* beste;
  int* breite;
  double* best_zeit;
  int* best_rf;
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  // Prozessorenanzahl bestimmen
  if( lese_next_double( &fi, &h ) != 0 )
  {
    return 1;
  }
  n = 1;
  while( lese_next_double_in_line( &fi, &h ) == 0 )
  {
    n++;
  }
  // Modulanzahl bestimmen
  m = 1;
  while( 1 )
  {
    if( lese_next_double_r( &fi, &h ) != 0 )
    {
      break;
    }
    m++;
  }
  fclose( fi );
  nr = (int*) malloc( m * sizeof( int ) );
  if( nr == NULL )
  {
    return 2;
  }
  nrh = (int*) malloc( m * sizeof( int ) );
  if( nrh == NULL )
  {
    return 2;
  }
  zeit = (double*) malloc( m * sizeof( double ) );
  if( zeit == NULL )
  {
    return 2;
  }
  breite = (int*) malloc( m * sizeof( int ) );
  if( breite == NULL )
  {
    return 2;
  }
  beste = (int*) malloc( m * sizeof( int ) );
  if( beste == NULL )
  {
    return 2;
  }
  best_zeit = (double*) malloc( m * sizeof( double ) );
  if( best_zeit == NULL )
  {
    return 2;
  }
  best_rf = (int*) malloc( m * sizeof( int ) );
  if( best_rf == NULL )
  {
    return 2;
  }
#ifdef ABBRUCH
  n = opti_datei_abbr( ergebnis, daten, n, m, breite, beste,
                       nr, nrh, zeit, best_zeit, best_rf );
#else
  n = opti_datei_weiter( ergebnis, daten, n, m, breite, beste,
                         nr, nrh, zeit, best_zeit, best_rf );
#endif
  free( zeit );
  free( nrh );
  free( nr );
  free( breite );
  free( beste );
  free( best_zeit );
  free( best_rf );
  return n;
}
#endif


// -------------------- FORMEL - ROUTINE MIT ABBRUCH -------------------
#if defined( ABBRUCH ) && defined( FORMEL )

// optimiert Module deren Laufzeiten in daten stehen
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m]
// Fehler: 3->kann Ergebnisdatei nicht anlegen
int opti_formel_abbr(
  char*   ergebnis,
  double  ((*laufzeit)(int modulnummer, int prozessorenanzahl)),
  int     n,
  int     m,
  int*    breite,
  int*    beste,
  int*    nr,
  int*    nrh,
  double* zeit,
  double* best_zeit,
  int*    best_rf )
{
  FILE* fi;
  double h, best_h, a, flae;
  int i, j;
  // flächenminimale Modullaufzeiten einlesen
  flae = 0.0; // aktuelle Fläche
  for( i=0; i<m; i++ )
  {
    zeit[i] = (*laufzeit)( i+1, 1 );
    j = 0;
    while( zeit[i] == 0 )
    {
      j++;
      zeit[i] = (*laufzeit)( i+1, j+1 );
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      a = (*laufzeit)( i+1, j+1 );
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    flae += zeit[i] * ( breite[i] + 1 );
    nr[i] = i;
  }
  // nach Höhe sortieren
  merge_modul_zeit( zeit, nr, nrh, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  while( 1 )
  {
    // einpacken
    h = 0.0;
    i = n;
    for( j=0; j<m; j++ )
    {
      i += breite[ nr[j] ] + 1;
      if( i > n )
      {
        i = breite[ nr[j] ] + 1;
        h += zeit[ nr[j] ];
      }
    }
    if( h < best_h )
    {
      if( 3 * flae > ((double)(n)) * h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 0; // Werte in breite nutzen
        break;
      }
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      best_h = h;
    }
    else
    {
      if( 3 * flae > ((double)(n)) * best_h )
      { // beim Einpacken Flächenfaktor 3 erreicht
        i = 1; // Werte in beste nutzen
        break;
      }
    }
    flae -= ( breite[ nr[0] ] + 1 ) * zeit[ nr[0] ];
    // das höchste Modul flächenminimal verbreitern
    j = breite[ nr[0] ] + 2;
    if( j > n ) // ist noch eine Verbreiterung möglich
    {
      break;
    }
    breite[ nr[0] ]++;
    zeit[ nr[0] ] = (*laufzeit)( nr[0] + 1, breite[ nr[0] ] + 1 );
    h = zeit[ nr[0] ] * ((double)(j));
    while( j < n )
    {
      j++;
      a = (*laufzeit)( nr[0] + 1, j );
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[0] ] = j - 1;
        zeit[ nr[0] ] = a;
      }
    }
    flae += zeit[ nr[0] ] * ( breite[ nr[0] ] + 1 );
    // das ehemals höchste einsortieren
    misch_modul_zeit( zeit, nr, nrh, 0, 1, m );
  }
  // bestes Ergebnis wieder herstellen
  if( i == 1 ) // aus beste übernehmen
  {
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    a = 0.0;
    i = n;
    for( j=0; j<m; j++ )
    {
      if( i + beste[ best_rf[j] ] + 1 > n )
      {
        i = 0;
        h = a;
        a += best_zeit[ best_rf[j] ];
      }
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, best_rf[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, h ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, h + best_zeit[ best_rf[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      i++;
      if( save_int( &fi, i ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      i += beste[ best_rf[j] ];
      if( save_int_r( &fi, i ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  else // Reihenfolge OK, Breiten in breite
  {
    // Ergebnis erstellen
    fi = fopen( ergebnis, "w" );
    if( fi == NULL )
    {
      return 3;
    }
    a = 0.0;
    i = n;
    for( j=0; j<m; j++ )
    {
      if( i + breite[ nr[j] ] + 1 > n )
      {
        i = 0;
        h = a;
        a += zeit[ nr[j] ];
      }
      if( save_text( &fi, "Modul " ) != 0 )
      {
        return 3;
      }
      if( save_int( &fi, nr[j] + 1 ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, ": von " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, h ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      if( save_double( &fi, h + zeit[ nr[j] ] ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " auf Prozessoren " ) != 0 )
      {
        return 3;
      }
      i++;
      if( save_int( &fi, i ) != 0 )
      {
        return 3;
      }
      if( save_text( &fi, " bis " ) != 0 )
      {
        return 3;
      }
      i += breite[ nr[j] ];
      if( save_int_r( &fi, i ) != 0 )
      {
        return 3;
      }
    }
    fclose( fi );
  }
  return 0;
}
#endif

// -------------------- FORMEL - ROUTINE OHNE ABBRUCH ------------------
#if !defined( ABBRUCH ) && defined( FORMEL )

// optimiert Module deren Laufzeiten in daten stehen
// n..Prozessorenanzahl; m..Modulanzahl
// benötigte Feldgrößen: breite[m], beste[m], nr[m], nrh[m], zeit[m],
//                       best_zeit[m], best_rf[m]
// Fehler: 3->kann Ergebnisdatei nicht anlegen
int opti_formel_weiter(
  char*   ergebnis,
  double  ((*laufzeit)(int modulnummer, int prozessorenanzahl)),
  int     n,
  int     m,
  int*    breite,
  int*    beste,
  int*    nr,
  int*    nrh,
  double* zeit,
  double* best_zeit,
  int*    best_rf )
{
  FILE* fi;
  double h, best_h, a;
  int i, j;
  // flächenminimale Modullaufzeiten einlesen
  for( i=0; i<m; i++ )
  {
    zeit[i] = (*laufzeit)( i+1, 1 );
    j = 0;
    while( zeit[i] == 0 )
    {
      j++;
      zeit[i] = (*laufzeit)( i+1, j+1 );
    }
    breite[i] = j;
    for( j++; j<n; j++ )
    {
      a = (*laufzeit)( i+1, j+1 );
      if( zeit[i] * ( breite[i] + 1 ) > a * ( j + 1 ) )
      {
        zeit[i] = a;
        breite[i] = j;
      }
    }
    nr[i] = i;
  }
  // nach Höhe sortieren
  merge_modul_zeit( zeit, nr, nrh, 0, m );
  best_h = DBL_MAX; // so schlimm sollte es nicht werden
  while( 1 )
  {
    // einpacken
    h = 0.0;
    i = n;
    for( j=0; j<m; j++ )
    {
      i += breite[ nr[j] ] + 1;
      if( i > n )
      {
        i = breite[ nr[j] ] + 1;
        h += zeit[ nr[j] ];
      }
    }

    if( h < best_h )
    {
      for( i=0; i<m; i++ )
      {
        beste[i] = breite[i];
        best_rf[i] = nr[i]; // Reihenfolge retten
        best_zeit[i] = zeit[i]; // Zeiten retten
      }
      best_h = h;
    }
    i = 0;
    while( breite[ nr[i] ] == n - 1 )
    {
      i++;
      if( i == m )
      {
        break;
      }
    }
    if( i == m )
    {
      break;
    }
    // das höchste mögliche Modul flächenminimal verbreitern
    j = breite[ nr[i] ] + 2;
    breite[ nr[i] ]++;
    zeit[ nr[i] ] = (*laufzeit)( nr[i] + 1, j );
    h = zeit[ nr[i] ] * ((double)(j));
    while( j < n )
    {
      j++;
      a = (*laufzeit)( nr[i] + 1, j );
      if( a * ((double)(j)) < h )
      {
        h = a * ((double)(j));
        breite[ nr[i] ] = j - 1;
        zeit[ nr[i] ] = a;
      }
    }
    // das verbreiterte einsortieren
    misch_modul_zeit( zeit, nr, nrh, i, i+1, m );
  }
  // bestes Ergebnis wieder herstellen
  // Ergebnis erstellen
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 3;
  }
  a = 0.0;
  i = n;
  for( j=0; j<m; j++ )
  {
    if( i + beste[ best_rf[j] ] + 1 > n )
    {
      i = 0;
      h = a;
      a += best_zeit[ best_rf[j] ];
    }
    if( save_text( &fi, "Modul " ) != 0 )
    {
      return 3;
    }
    if( save_int( &fi, best_rf[j] + 1 ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, ": von " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi, h ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    if( save_double( &fi, h + best_zeit[ best_rf[j] ] ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " auf Prozessoren " ) != 0 )
    {
      return 3;
    }
    i++;
    if( save_int( &fi, i ) != 0 )
    {
      return 3;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 3;
    }
    i += beste[ best_rf[j] ];
    if( save_int_r( &fi, i ) != 0 )
    {
      return 3;
    }
  }
  fclose( fi );
  return 0;
}
#endif


#ifdef FORMEL

// Hauptroutine; siehe oben
int optimiere_formel(
  char*  ergebnis,
  int    m,
  int    n,
  double ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  int* nr;
  int* nrh;
  double* zeit;
  int* beste;
  int* breite;
  double* best_zeit;
  int* best_rf;
  nr = (int*) malloc( m * sizeof( int ) );
  if( nr == NULL )
  {
    return 2;
  }
  nrh = (int*) malloc( m * sizeof( int ) );
  if( nrh == NULL )
  {
    return 2;
  }
  zeit = (double*) malloc( m * sizeof( double ) );
  if( zeit == NULL )
  {
    return 2;
  }
  breite = (int*) malloc( m * sizeof( int ) );
  if( breite == NULL )
  {
    return 2;
  }
  beste = (int*) malloc( m * sizeof( int ) );
  if( beste == NULL )
  {
    return 2;
  }
  best_zeit = (double*) malloc( m * sizeof( double ) );
  if( best_zeit == NULL )
  {
    return 2;
  }
  best_rf = (int*) malloc( m * sizeof( int ) );
  if( best_rf == NULL )
  {
    return 2;
  }
#ifdef ABBRUCH
  n = opti_formel_abbr( ergebnis, laufzeit, n, m, breite, beste,
                        nr, nrh, zeit, best_zeit, best_rf );
#else
  n = opti_formel_weiter( ergebnis, laufzeit, n, m, breite, beste,
                          nr, nrh, zeit, best_zeit, best_rf );
#endif
  free( zeit );
  free( nrh );
  free( nr );
  free( breite );
  free( beste );
  free( best_zeit );
  free( best_rf );
  return n;
}
#endif
