// Routine zum Erzeugen von Beispielen:
//   serien - paralleler Abhängigkeitsgraph, moldable
//
// Aufruf für zufällige Beispiele:
// baue_bsp_array(
//   double**                 Feld der Größe m*n (Speicher muß schon
//                                                reserviert sein),
//   graph*                   Modulabhängigkeitsgraph,
//   int                      Prozessorenanzahl,
//   verteilung_int*          Gesamtanzahl der Module,
//   verteilung_double*       Verhältnis von paralleler und serieller
//                            Vereinigung (0..nur parallel
//                                         1..nur seriell),
//   int                      Anzahl der verschiedenen Verteilungen
//                            zum Erzeugen von Modulen,
//   verteilung_double_liste* Listenfeld mit
//                            1. Eintrag: Anteile der Verteilungsarten
//                                        (die Wahrscheinlichkeit, daß
//                                         Verteilung i genommen wird
//                                         ist ant[i]/summe(ant[?]),
//                                         wobei ant[i] =
//                                         zufall(erster Eintrag[i]) )
//                             2. Eintrag: Laufzeit des Moduls mit
//                                         einem Prozessor
//                             ab 3. Eintrag: Parallelisierungsdaten
//                                            (bis bestimmt dabei wie
//                                             lange der Eintrag
//                                             gültig ist) );
//
// oder:
// baue_bsp_datei(
//   char* Filename für Ergebnis,
//   char* Filename für Modulabhängigkeitsgraphen,
//   char* Filename Eingabedatei );
//
// Aufruf für Standardbeispiele
// baue_standard_bsp(
//   double** Feld der Größe m*n (Speicher muß schon reserviert sein),
//   int      Prozessorenanzahl,
//   graph*   Modulabhängigkeitsgraph (muß initialisiert sein)
//   char*    Filename Eingabedatei );
//
// Fehler: 1->zu wenig Speicher
//         2->konnte Ergebnisdatei nicht anlegen
//         3->nicht genug Einträge im Listenfeld / Fehler in Datei
//         4->konnte Eingabedatei nicht finden

#include "../prgx/graphen.h"
#include "../prgx/zufall.h"

int baue_bsp_array(
  double**                 laufz,
  graph*                   abhgraph,
  int                      n,
  verteilung_int*          vert_m,
  verteilung_double*       vert_verh,
  int                      anz_vert,
  verteilung_double_liste* liste_vert );

int baue_bsp_datei(
  char* fn,
  char* fg,
  char* fin );

// --------------------- weitere Routinen ------------------------------

// reserviert den maximal benötigten Speicher für eine Laufzeittabelle
// Parameter 2 und 3 entsprechen den Parametern 2 und 3 von
// load_configdatei
// Fehler: 1->zu wenig Speicher
int init_laufzeit_verteilung(
  double***       laufz,
  int             n,
  verteilung_int* m_vert );

// gibt Speicher der Laufzeittabelle frei
// Parameter 2 entspricht den Parameter 3 von load_configdatei
void frei_laufzeit_verteilung(
  double**        laufz,
  verteilung_int* vert_m );

// lädt eine Beispielgenerierungsdatei
// der Speicher für den letzten Parameter wird reserviert, der Rest
// sind einfache Variablen von denen der Pointer übergeben wird
// Fehler: 1->zu wenig Speicher
//         2->Configdatei fehlerhaft
//         3->Configdatei nicht gefunden
int load_configdatei(
  char*                     configdatei,
  int*                      n,
  verteilung_int*           vert_m,
  verteilung_double*        vert_verh,
  int*                      anz_vert,
  verteilung_double_liste** liste_vert );

// gibt beim Laden einer Beispielgenerierungsdatei reservierten
// Speicher wieder frei
// es werden die letzten beiden Parameter von load_configdatei benötigt
void frei_configdatei(
  int                      anz_vert,
  verteilung_double_liste* liste_vert );
