// Optimierungsroutinen moldable,
// Modulabhängigkeitsgraph ist ein serien-paralleler Graph
// optimiert Module deren Laufzeiten in der Datei daten stehen
// (i-te Zeile j-te Spalte enthält Laufzeit vom i-ten Modul
//  mit j Prozessoren)
// oder mit Hilfe der übergebenen Funktion
// laufzeit (modulnummer>=1, prozessorenanzahl>=1) ermittelt
// werden können
//
// Aufruf:
//   optimiere_datei(
//     char* ergebnis_datei,
//     char* laufzeit_daten_datei,
//     char* modulabhängigkeitsgraph_datei )
// oder:
//   optimiere_formel(
//     double** startzeit_der_module,
//     int**    prozessoranzahl_der_module,
//     int**    kleinste_prozessornummer_der_module,
//     graph*   modulabhängigkeitsgraph,
//     int      maximalprozessoranzahl,
//     double   ((*laufzeit)( modulnummer, prozessorenanzahl )) )
//
//   Speicher für Rückgabewerte startzeit_der_module,
//   prozessoranzahl_der_module und kleinste_prozessornummer_der_module
//   wird reserviert
//
// Fehler: 1->zu wenig Speicher
//         2->Eingabedatei mit Laufzeiten nicht gefunden / fehlerhaft
//         3->Eingabedatei mit Graphen nicht gefunden / fehlerhaft
//         4->kann Ergebnisdatei nicht anlegen / Schreibfehler

#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include "../prgx/graphen.h"
#include "../prgx/load.h"
#include "../prgx/save.h"
#include "../prgx/sp_graph.h"


// ----------------------- Dateibasierte Routinen ----------------------
#ifndef FORMEL

// berechnet die Zeiten za vom fiktiven Modul w;
// b ist der Konstruktionsbaum mit der Beschriftung beschrift;
// in typ wird die durchgeführte Vereinigung gespeichert
// Fehler: 1->zu wenig Speicher
int verein_modul_array(
  graph*  b,
  int     w,
  int*    beschrift,
  int     o,
  int*    typ,
  int     n,
  double* za,
  double* t )
{
  int i, j;
  double* zb;
  int x, y;
  if( w < o ) // Blatt erreicht; Modullaufzeiten zurückgeben
  {
    for( i=0; i<n; i++ )
    {
      za[i] = t[ (w-1) * n + i ];
    }
    return 0;
  }
  reset_kind( b, w );
  x = next_kind( b, w );
  y = next_kind( b, w );
  zb = (double*) malloc( sizeof( double ) * n );
  if( zb == NULL )
  {
    return 1;
  }
  if( verein_modul_array( b, x, beschrift, o, typ, n, za, t ) != 0 )
  {
    return 1;
  }
  if( verein_modul_array( b, y, beschrift, o, typ, n, zb, t ) != 0 )
  {
    return 1;
  }
  if( beschrift[ w-o ] == 2 ) // sequentielle Vereinigung
  {
    for( i=0; i<n; i++ )
    {
      za[i] += zb[i];
    }
  }
  else // parallele Vereinigung
  {
    // Dimension n wegen Bestimmung vom Wechsel extra betrachten
    za[n-1] += zb[n-1];
    typ[ (w-o) * n + n - 1 ] = 0;
    x = 0; // Stelle mit Wechsel der Größen von za und zb suchen
    y = n;
    while( x + 1 < y )
    {
      j = ( x + y ) / 2;
      if( za[j-1] <= zb[n-j-1] ) // hier ist j immer >=1 und <=n-1
      {
        y = j;
      }
      else
      {
        x = j;
      }
    }
    // wenn 1 <= x <= n-2 gilt za[x] <= zb[n-x-2] und
    //                         za[x-1] > zb[n-x-1]
    // wenn x == 0 gilt für alle i: za[i] <= za[0] <= zb[n-2]
    // wenn x == n-1 gilt für alle i: za[n-2] > zb[0] >= zb[i]
    if( x == 0 ) // za[x-1] existiert nicht
    {
      if( za[n-1] >= zb[n-2] ) // parallele Ausführung besser ?
      {
        za[n-1] = zb[n-2];
        typ[ (w-o) * n + n - 1 ] = 1;
      }
    }
    else // x >= 1
    {
      if( x == n-1 )
      {
        if( za[n-1] >= za[n-2] )
        {
          za[n-1] = za[n-2];
          typ[ (w-o) * n + n - 1 ] = n-1;
        }
        x--; // nächster Wert
      }
      else // 1 <= x <= n-2
      {
        // Minimum aus beiden Möglichkeiten suchen
        if( za[x-1] <= zb[n-x-2] )
        {
          if( za[n-1] >= za[x-1] )
          {
            za[n-1] = za[x-1];
            typ[ (w-o) * n + n - 1 ] = x;
          }
          x--; // nächster Wert
        }
        else
        {
          if( za[n-1] >= zb[n-x-2] )
          {
            za[n-1] = zb[n-x-2];
            typ[ (w-o) * n + n - 1 ] = x + 1;
          }
        }
      }
    }
    for( i=n-2; i>=1; i-- )
    {
      za[i] += zb[i];
      typ[ (w-o) * n + i ] = 0;
      if( x == 0 ) // za[x-1] existiert nicht
      {
        if( za[i] >= zb[i-1] ) // parallele Ausführung besser ?
        {
          za[i] = zb[i-1];
          typ[ (w-o) * n + i ] = 1;
        }
      }
      else // x >= 1
      {
        if( x == i )
        {
          if( za[i] >= za[i-1] )
          {
            za[i] = za[i-1];
            typ[ (w-o) * n + i ] = i;
          }
          x--; // nächster Wert
        }
        else // 1 <= x <= i-1
        {
          // Minimum aus beiden Möglichkeiten suchen
          if( za[x-1] <= zb[i-x-1] )
          {
            if( za[i] >= za[x-1] )
            {
              za[i] = za[x-1];
              typ[ (w-o) * n + i ] = x;
            }
            x--; // nächster Wert
          }
          else
          {
            if( za[i] >= zb[i-x-1] )
            {
              za[i] = zb[i-x-1];
              typ[ (w-o) * n + i ] = x + 1;
            }
          }
        }
      }
    }
    za[0] += zb[0]; // bei einem Prozessor gibt es nur eine Möglichkeit
    typ[ (w-o) * n ] = 0;
  }
  free( zb );
  return 0;
}

// speichert das Ergebnis indem der rekursive Abstieg, wie in
// verein_modul berechnet, durchgeführt wird
// Rückgabe: >0 Endzeit des fiktiven Knotens w bei Startzeit st und
//              Breite re-li und linken Prozessor li
//           -1->Schreibfehler
double setze_zeiten_array(
  graph*  b,
  int     w,
  int*    beschrift,
  int     o,
  int*    typ,
  int     n,
  double* t,
  FILE**  fi,
  int     li,
  int     re,
  double  st )
{
  int x, y;
  double sx;
  if( w < o ) // Blatt erreicht
  {
    if( save_text( fi, "Modul " ) != 0 )
    {
      return -1;
    }
    if( save_int( fi, w ) != 0 )
    {
      return -1;
    }
    if( save_text( fi, ": von " ) != 0 )
    {
      return -1;
    }
    if( save_double( fi, st ) != 0 )
    {
      return -1;
    }
    st += t[ (w-1) * n + re - li - 1 ];
    if( save_text( fi, " bis " ) != 0 )
    {
      return -1;
    }
    if( save_double( fi, st ) != 0 )
    {
      return -1;
    }
    if( save_text( fi, " auf Prozessoren " ) != 0 )
    {
      return -1;
    }
    if( save_int( fi, li+1 ) != 0 )
    {
      return -1;
    }
    if( save_text( fi, " bis " ) != 0 )
    {
      return -1;
    }
    if( save_int_r( fi, re ) != 0 )
    {
      return -1;
    }
    return st;
  }
  reset_kind( b, w );
  x = next_kind( b, w );
  y = next_kind( b, w );
  if( beschrift[ w-o ] == 2 ) // sequentielle Vereinigung
  {
    st = setze_zeiten_array( b, x, beschrift, o, typ, n, t, fi,
                             li, re, st );
    if( st < 0 )
    {
      return -1;
    }
    return setze_zeiten_array( b, y, beschrift, o, typ, n, t, fi,
                               li, re, st );
  }
  // parallele Vereinigung
  // es wurde aber trotzdem die sequentielle Vereinigung angewendet
  if( typ[ (w-o) * n + re - li - 1 ] == 0 )
  {
    st = setze_zeiten_array( b, x, beschrift, o, typ, n, t, fi,
                             li, re, st );
    if( st < 0 )
    {
      return -1;
    }
    return setze_zeiten_array( b, y, beschrift, o, typ, n, t, fi,
                               li, re, st );
  }
  sx = setze_zeiten_array( b, x, beschrift, o, typ, n, t, fi, li,
                           li+typ[(w-o)*n+re-li-1], st );
  if( sx < 0 )
  {
    return -1;
  }
  st = setze_zeiten_array( b, y, beschrift, o, typ, n, t, fi,
                           li+typ[(w-o)*n+re-li-1], re, st );
  if( st < 0 )
  {
    return -1;
  }
  if( st < sx )
  {
    return sx;
  }
  return st;
}

// Optimierungsroutine siehe oben
int optimiere_datei(
  char* ergebnis,
  char* daten,
  char* graph_datei )
{
  graph g, b;
  int* beschrift;
  int w, m, o, i, j, n;
  int* typ;
  int pos;
  double d;
  double* za;
  double* t;
  FILE* fi;
  init_graph( &g );
  init_graph( &b );
  // Graph laden
  if( load_graph_rueck_mehr( graph_datei, &g, 2 ) != 0 )
  {
    return 3;
  }
  // Maximale Knotenanzahl ermitteln
  m = max_knoten( &g );
  // Konstruktionsbaum berechnen
#ifdef BALANCIERT
  w = baue_sp_baum_binaer( &g, &b, &beschrift, &o, 0, 0 );
  if( w < 0 )
  {
    return 1;
  }
#else
  w = baue_sp_baum_linear( &g, &b, &beschrift, &o, 0, 0 );
  if( w < 0 )
  {
    return 1;
  }
#endif
  free_graph( &g );
  // Berechnen der Lösung
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 2;
  }
  // Bestimmen der Prozessorenanzahl
  pos = ftell( fi );
  n = 1;
  if( lese_next_double( &fi, &d ) != 0 )
  {
    return 2;
  }
  while( lese_next_double_in_line( &fi, &d ) == 0 )
  {
    n++;
  }
  fseek( fi, pos, SEEK_SET );
  // Hilfsvariablen
  typ = (int*) malloc( sizeof( int ) * m * n );
  if( typ == NULL )
  {
    return 1;
  }
  za = (double*) malloc( sizeof( double ) * n );
  if( za == NULL )
  {
    return 1;
  }
  t = (double*) malloc( sizeof( double ) * m * n );
  if( t == NULL )
  {
    return 1;
  }
  // Laufzeiten laden
  for( i=0; i<m; i++ )
  {
    if( lese_next_double( &fi, &(t[i*n]) ) != 0 )
    {
      return 2;
    }
    for( j=1; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &(t[i*n+j]) ) != 0 )
      {
        return 2;
      }
    }
  }
  fclose( fi );
  if( verein_modul_array( &b, w, beschrift, o, typ, n, za, t ) != 0 )
  {
    return 1;
  }
  free( za );
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 4;
  }
  if( setze_zeiten_array( &b, w, beschrift, o, typ, n, t, &fi,
                          0, n, 0.0 ) < 0 )
  {
    return 4;
  }
  fclose( fi );
  free( typ );
  free( t );
  free_graph( &b );
  free( beschrift );
  return 0;
}

#endif


// ----------------- Formelorientierte Routinen ------------------------
#ifdef FORMEL

// berechnet die Zeiten za vom fiktiven Modul w;
// b ist der Konstruktionsbaum mit der Beschriftung beschrift;
// in typ wird die durchgeführte Vereinigung gespeichert
// Fehler: 1->zu wenig Speicher
int verein_modul_formel(
  graph*  b,
  int     w,
  int*    beschrift,
  int     o,
  int*    typ,
  int     n,
  double* za,
  double  ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  int i, j;
  double* zb;
  int x, y;
  if( w < o ) // Blatt erreicht; Modullaufzeiten zurückgeben
  {
    for( i=0; i<n; i++ )
    {
      za[i] = laufzeit( w, i+1 );
    }
    return 0;
  }
  reset_kind( b, w );
  x = next_kind( b, w );
  y = next_kind( b, w );
  zb = (double*) malloc( sizeof( double ) * n );
  if( zb == NULL )
  {
    return 1;
  }
  if( verein_modul_formel( b, x, beschrift, o, typ,
                           n, za, laufzeit ) != 0 )
  {
    return 1;
  }
  if( verein_modul_formel( b, y, beschrift, o, typ,
                           n, zb, laufzeit ) != 0 )
  {
    return 1;
  }
  if( beschrift[ w-o ] == 2 ) // sequentielle Vereinigung
  {
    for( i=0; i<n; i++ )
    {
      za[i] += zb[i];
    }
  }
  else // parallele Vereinigung
  {
    // Dimension n wegen Bestimmung vom Wechsel extra betrachten
    za[n-1] += zb[n-1];
    typ[ (w-o) * n + n - 1 ] = 0;
    x = 0; // Stelle mit Wechsel der Größen von za und zb suchen
    y = n;
    while( x + 1 < y )
    {
      j = ( x + y ) / 2;
      if( za[j-1] <= zb[n-j-1] ) // hier ist j immer >=1 und <=n-1
      {
        y = j;
      }
      else
      {
        x = j;
      }
    }
    // wenn 1 <= x <= n-2 gilt za[x] <= zb[n-x-2] und
    //                         za[x-1] > zb[n-x-1]
    // wenn x == 0 gilt für alle i: za[i] <= za[0] <= zb[n-2]
    // wenn x == n-1 gilt für alle i: za[n-2] > zb[0] >= zb[i]
    if( x == 0 ) // za[x-1] existiert nicht
    {
      if( za[n-1] >= zb[n-2] ) // parallele Ausführung besser ?
      {
        za[n-1] = zb[n-2];
        typ[ (w-o) * n + n - 1 ] = 1;
      }
    }
    else // x >= 1
    {
      if( x == n-1 )
      {
        if( za[n-1] >= za[n-2] )
        {
          za[n-1] = za[n-2];
          typ[ (w-o) * n + n - 1 ] = n-1;
        }
        x--; // nächster Wert
      }
      else // 1 <= x <= n-2
      {
        // Minimum aus beiden Möglichkeiten suchen
        if( za[x-1] <= zb[n-x-2] )
        {
          if( za[n-1] >= za[x-1] )
          {
            za[n-1] = za[x-1];
            typ[ (w-o) * n + n - 1 ] = x;
          }
          x--; // nächster Wert
        }
        else
        {
          if( za[n-1] >= zb[n-x-2] )
          {
            za[n-1] = zb[n-x-2];
            typ[ (w-o) * n + n - 1 ] = x + 1;
          }
        }
      }
    }
    for( i=n-2; i>=1; i-- )
    {
      za[i] += zb[i];
      typ[ (w-o) * n + i ] = 0;
      if( x == 0 ) // za[x-1] existiert nicht
      {
        if( za[i] >= zb[i-1] ) // parallele Ausführung besser ?
        {
          za[i] = zb[i-1];
          typ[ (w-o) * n + i ] = 1;
        }
      }
      else // x >= 1
      {
        if( x == i )
        {
          if( za[i] >= za[i-1] )
          {
            za[i] = za[i-1];
            typ[ (w-o) * n + i ] = i;
          }
          x--; // nächster Wert
        }
        else // 1 <= x <= i-1
        {
          // Minimum aus beiden Möglichkeiten suchen
          if( za[x-1] <= zb[i-x-1] )
          {
            if( za[i] >= za[x-1] )
            {
              za[i] = za[x-1];
              typ[ (w-o) * n + i ] = x;
            }
            x--; // nächster Wert
          }
          else
          {
            if( za[i] >= zb[i-x-1] )
            {
              za[i] = zb[i-x-1];
              typ[ (w-o) * n + i ] = x + 1;
            }
          }
        }
      }
    }
    za[0] += zb[0]; // bei einem Prozessor gibt es nur eine Möglichkeit
    typ[ (w-o) * n ] = 0;
  }
  free( zb );
  return 0;
}

// speichert das Ergebnis indem der rekursive Abstieg, wie in
// verein_modul berechnet, durchgeführt wird
// Rückgabe: Endzeit des fiktiven Knotens w bei Startzeit st
//           und Breite re-li und linken Prozessor li
double setze_zeiten_formel(
  graph*  b,
  int     w,
  int*    beschrift,
  int     o,
  int*    typ,
  int     n,
  double  ((*laufzeit)( int modulnummer, int prozessorenanzahl )),
  double* startzeit,
  int*    breite,
  int*    linker,
  int     li,
  int     re,
  double  st )
{
  int x, y;
  double sx;
  if( w < o ) // Blatt erreicht
  {
    startzeit[w-1] = st;
    st += laufzeit( w, re - li );
    breite[w-1] = re - li;
    linker[w-1] = li + 1;
    return st;
  }
  reset_kind( b, w );
  x = next_kind( b, w );
  y = next_kind( b, w );
  if( beschrift[ w-o ] == 2 ) // sequentielle Vereinigung
  {
    st = setze_zeiten_formel( b, x, beschrift, o, typ, n, laufzeit,
                              startzeit, breite, linker, li, re, st );
    return setze_zeiten_formel( b, y, beschrift, o, typ, n, laufzeit,
                                startzeit, breite, linker, li, re, st );
  }
  // parallele Vereinigung
  // es wurde aber trotzdem die sequentielle Vereinigung angewendet
  if( typ[ (w-o) * n + re - li - 1 ] == 0 )
  {
    st = setze_zeiten_formel( b, x, beschrift, o, typ, n, laufzeit,
                              startzeit, breite, linker, li, re, st );
    return setze_zeiten_formel( b, y, beschrift, o, typ, n, laufzeit,
                                startzeit, breite, linker, li, re, st );
  }
  sx = setze_zeiten_formel( b, x, beschrift, o, typ, n, laufzeit,
                            startzeit, breite, linker, li,
                            li+typ[(w-o)*n+re-li-1], st );
  st = setze_zeiten_formel( b, y, beschrift, o, typ, n, laufzeit,
                            startzeit, breite, linker,
                            li+typ[(w-o)*n+re-li-1], re, st );
  if( st < sx )
  {
    return sx;
  }
  return st;
}

// Optimierungsroutine siehe oben
int optimiere_formel(
  double** startzeit,
  int**    breite,
  int**    linker,
  graph*   g,
  int      n,
  double   ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  graph b;
  int* beschrift;
  int w, m, o;
  int* typ;
  double* za;
  init_graph( &b );
  // Maximale Knotenanzahl ermitteln
  m = max_knoten( g );
  // Konstruktionsbaum berechnen
#ifdef BALANCIERT
  w = baue_sp_baum_binaer( g, &b, &beschrift, &o, 0, 0 );
  if( w < 0 )
  {
    return 1;
  }
#else
  w = baue_sp_baum_linear( g, &b, &beschrift, &o, 0, 0 );
  if( w < 0 )
  {
    return 1;
  }
#endif
  // Hilfsvariablen
  typ = (int*) malloc( sizeof( int ) * m * n );
  if( typ == NULL )
  {
    return 1;
  }
  za = (double*) malloc( sizeof( double ) * n );
  if( za == NULL )
  {
    return 1;
  }
  // Ergebnisvariablen
  *startzeit = (double*) malloc( sizeof( double ) * m );
  if( *startzeit == NULL )
  {
    return 1;
  }
  *breite = (int*) malloc( sizeof( int ) * m );
  if( *breite == NULL )
  {
    return 1;
  }
  *linker = (int*) malloc( sizeof( int ) * m );
  if( *linker == NULL )
  {
    return 1;
  }
  if( verein_modul_formel( &b, w, beschrift, o, typ, n,
                           za, laufzeit ) != 0 )
  {
    return 1;
  }
  free( za );
  if( setze_zeiten_formel( &b, w, beschrift, o, typ, n, laufzeit,
                           *startzeit, *breite, *linker, 0,
                           n, 0.0 ) < 0 )
  {
    return 4;
  }
  free( typ );
  free_graph( &b );
  free( beschrift );
  return 0;
}

#endif
