// Routine zum Erzeugen von Beispielen:
//   serien - paralleler Abhängigkeitsgraph, non-moldable
// Aufruf:
//   baue_bsp_array(
//     double**                  Feld der Größe m*n (Speicher muß schon
//                                                   reserviert sein),
//     graph*                    Modulabhängigkeitsgraph,
//     int                       Prozessorenanzahl,
//     verteilung_int*           Gesamtanzahl der Module,
//     verteilung_double*        Verhältnis von paralleler und serieller
//                               Vereinigung (0..nur parallel;
//                                            1..nur seriell),
//     int                       Anzahl der verschiedenen Verteilungen
//                               zum Erzeugen von Modulen,
//     verteilung_double_liste** Listenfeld mit
//                                1. Eintrag:
//                                 Anteile der Verteilungsarten (die
//                                 Wahrscheinlichkeit, daß Verteilung i
//                                 genommen wird ist
//                                 ant[i]/summe(ant[?]), wobei
//                                 ant[i] = zufall(erster Eintrag[i]) )
//                                2. Eintrag:
//                                 Mindestprozessorenanzahl
//                                3. Eintrag:
//                                 Höhe des Moduls mit
//                                 Mindestprozessorenanzahl
//                                ab 4. Eintrag:
//                                 Parallelisierungsdaten (bis bestimmt
//                                 dabei wie lange der Eintrag gültig
//                                 ist)
//   );
//
// oder:
//   baue_bsp_datei(
//     char* Filename für Ergebnis,
//     char* Filename für Modulabhängigkeitsgraphen,
//     char* Filename Eingabedatei );
//
// Fehler: 1->zu wenig Speicher
//         2->konnte Ergebnisdatei nicht anlegen
//         3->nicht genug Einträge im Listenfeld / Fehler in Datei
//         4->konnte Eingabedatei nicht finden

#include <float.h>
#include <limits.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "../prgx/graphen.h"
#include "../prgx/load.h"
#include "../prgx/save.h"
#include "../prgx/sp_graph.h"
#include "../prgx/zufall.h"

#define EPSILON DBL_MIN // minimale zulässige Laufzeit

int baue_bsp_array(
  double**                 laufz,
  graph*                   abhgraph,
  int                      n,
  verteilung_int*          vert_m,
  verteilung_double*       vert_verh,
  int                      anz_vert,
  verteilung_double_liste* liste_vert )
{
  int anz;
  double* ant;
  int i, j, k, nr;
  double az, d;
  verteilung_double_liste* vl;
  ant = (double*) malloc( anz_vert * sizeof( double ) );
  if( ant == NULL )
  {
    return 1;
  }
  anz = zufall_int( vert_m ); // zu erzeugende Modulanzahl
  // Abhängigkeitsgraphen erzeugen
  if( rnd_sp_graph( abhgraph, zufall_double( vert_verh ),
                    anz, anz+2 ) != 0 )
  {
    return 1;
  }
  if( &(liste_vert[0]) == NULL )
  {
    return 3;
  }
  ant[0] = zufall_double( &((liste_vert[0]).v) );
  for( i=1; i<anz_vert; i++ )
  {
    // gleich aufsummieren
    ant[i] = ant[ i - 1 ] + zufall_double( &((liste_vert[i]).v) );
  }
  for( k=0; k<anz; k++ )
  {
    // Verteilung auswählen
    az = gleich_double( 0.0, ant[ anz_vert - 1 ] );
    // dazugehörige Nummer suchen
    i = 0;
    nr = anz_vert - 1;
    while( i != nr )
    {
      j = ( i + nr ) / 2;
      if( az < ant[ j ] )
      {
        nr = j;
      }
      else
      {
        i = j + 1;
      }
    }
    // Mindestprozessorenanzahl bestimmen
    if( (liste_vert[nr]).next == NULL )
    {
      return 3;
    }
    i = zufall_int_from_double( &((liste_vert[nr]).next->v) );
    if( i == 1 )
    {
      // Laufzeit mit einem Prozessor bestimmen
      if( (liste_vert[nr]).next->next == NULL )
      {
        return 3;
      }
      laufz[k][0] =
        zufall_double( &((liste_vert[nr]).next->next->v) );
      // Laufzeit mit i+1 Prozessoren bestimmen
      vl = (liste_vert[nr]).next->next->next;
      for( i=1; i<n; i++ )
      {
        if( i == vl->bis )
        {
          if( vl->next == NULL )
          {
            return 3;
          }
          vl = vl->next;
        }
        d = ( laufz[k][0] / ((double)(i + 1)) - laufz[k][i-1] );
        laufz[k][i] = laufz[k][i-1] + d * zufall_double( &(vl->v) );
      }
    }
    else
    {
      for( j=0; j<i-1; j++ )
      {
        laufz[k][j] = 0.0;
      }
      // Laufzeit mit i Prozessoren bestimmen
      if( (liste_vert[nr]).next->next == NULL )
      {
        return 3;
      }
      laufz[k][i-1] =
        zufall_double( &((liste_vert[nr]).next->next->v) );
      // Laufzeit mit i+1 Prozessoren bestimmen
      vl = (liste_vert[nr]).next->next->next;
      while( vl->bis < i )
      {
        if( vl->next == NULL )
        {
          return 3;
        }
        vl = vl->next;
      }
      for( ; i<n; i++ )
      {
        if( i == vl->bis )
        {
          if( vl->next == NULL )
          {
            return 3;
          }
          vl = vl->next;
        }
        laufz[k][i] =
          laufz[k][i-1] * ( 1 - zufall_double( &(vl->v) ) );
        if( laufz[k][i] < EPSILON )
        {
          laufz[k][i] = EPSILON;
        }
      }
    }
  }
  free( ant );
  return 0;
}

int baue_bsp_datei( char* fn, char* fg, char* fin )
{
  int anz, h;
  double* ant;
  int i, j, n, arten;
  double lz, az, d;
  int pos;
  verteilung_int anzahl;
  verteilung_double vert;
  FILE* fi;
  FILE* fiin;
  graph g;
  fiin = fopen( fin, "r" );
  if( fiin == NULL )
  {
    return 4;
  }
  // Prozessorenanzahl einlesen
  if( lese_next_int_r( &fiin, &n ) != 0 )
  {
    return 3;
  }
  // zu erzeugende Modulanzahlverteilung einlesen
  if( lese_next_verteilung_int_r( &fiin, &anzahl ) != 0 )
  {
    return 3;
  }
  anz = zufall_int( &anzahl ); // zu erzeugende Modulanzahl
  // Verhältnis zwischen paralleler und sequentieller
  // Vereinigung einlesen
  if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
  {
    return 3;
  }
  // Modulabhängigkeitsgraph erzeugen und speichern
  init_graph( &g );
  if( rnd_sp_graph( &g, zufall_double( &vert ), anz, anz+2 ) != 0 )
  {
    return 1;
  }
  if( save_graph_rueck( &g, fg ) != 0 )
  {
    return 2;
  }
  free_graph( &g );
  // Verteilungsarten zählen
  pos = ftell( fiin );
  // eine muß da sein!
  if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
  {
    return 3;
  }
  arten = 1;
  while( 1 )
  {
    if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
    {
      break;
    }
    arten++;
  }
  // Speicher reservieren
  ant = (double*) malloc( arten * sizeof( double ) );
  if( ant == NULL )
  {
    return 1;
  }
  // Daten einlesen
  fseek( fiin, pos, SEEK_SET );
  if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
  {
    return 3;
  }
  ant[0] = zufall_double( &vert );
  for( i=1; i<arten; i++ )
  {
    if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    // gleich aufsummieren
    ant[i] = ant[ i - 1 ] + zufall_double( &vert );
  }
  fi = fopen( fn, "w" );
  if( fi == NULL )
  {
    return 2;
  }
  while( anz > 0 )
  {
    // Verteilung auswählen
    az = gleich_double( 0.0, ant[ arten - 1 ] );
    // dazugehörige Nummer suchen; da auch in Datei gesucht werden muß,
    // ist lineare Laufzeit OK
    i = 0;
    fseek( fiin, pos, SEEK_SET );
    while( az >= ant[i] )
    {
      i++;
      if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
      {
        return 3;
      }
    }
    // Mindestprozessorenanzahl bestimmen
    // Wahrscheinlichkeit überlesen
    if( lese_next_verteilung_double( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
    {
      return 3;
    }
    i = zufall_int_from_double( &vert );
    if( i == 1 )
    {
      // Laufzeit mit einem Prozessor bestimmen (lz)
      if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
      {
        return 3;
      }
      az = lz = zufall_double( &vert );
      if( save_double( &fi, lz ) != 0 )
      {
        return 2;
      }
      // Laufzeit mit i+1 Prozessoren bestimmen;
      // az enthält letzte Laufzeit
      if( lese_next_int_in_line( &fiin, &h ) != 0 )
      {
        return 3;
      }
      if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
      {
        return 3;
      }
      for( i=1; i<n; i++ )
      {
        if( i == h )
        {
          if( lese_next_int_in_line( &fiin, &h ) != 0 )
          {
            return 3;
          }
          if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
          {
            return 3;
          }
        }
        d = ( lz / ((double)(i + 1)) - az );
        az += d * zufall_double( &vert );
        if( save_text( &fi, " " ) != 0 )
        {
          return 2;
        }
        if( save_double( &fi, az ) != 0 )
        {
          return 2;
        }
      }
    }
    else // Modul ist nonmaldable
    {
      // Nullen als unendlich speichern
      for( j=1; j<i; j++ )
      {
        if( save_double_s( &fi, 0.0 ) != 0 )
        {
          return 2;
        }
      }
      // Laufzeit mit i Prozessoren bestimmen
      if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
      {
        return 3;
      }
      az = zufall_double( &vert );
      if( save_double( &fi, az ) != 0 )
      {
        return 2;
      }
      // Laufzeit mit i+1 Prozessoren bestimmen;
      // az enthält letzte Laufzeit
      do
      {
        if( lese_next_int_in_line( &fiin, &h ) != 0 )
        {
          return 3;
        }
        if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
        {
          return 3;
        }
      } while ( h < i );
      for( ; i<n; i++ )
      {
        if( i == h )
        {
          if( lese_next_int_in_line( &fiin, &h ) != 0 )
          {
            return 3;
          }
          if( lese_next_verteilung_double_in_line( &fiin, &vert ) != 0 )
          {
            return 3;
          }
        }
        az *= 1.0 - zufall_double( &vert );
        if( az < EPSILON )
        {
          az = EPSILON;
        }
        if( save_text( &fi, " " ) != 0 )
        {
          return 2;
        }
        if( save_double( &fi, az ) != 0 )
        {
          return 2;
        }
      }
    }
    if( save_text( &fi, "\n" ) != 0 )
    {
      return 2;
    }
    anz--;
  }
  fclose( fi );
  fclose( fiin );
  free( ant );
  return 0;
}

// reserviert den maximal benötigten Speicher für eine Laufzeittabelle
// Parameter 2 und 3 entsprechen den Parametern 2 und 3
// von load_configdatei
// Fehler: 1->zu wenig Speicher
int init_laufzeit_array(
  double***       laufz,
  int             n,
  verteilung_int* vert_m )
{
  int i, m;
  // maximale Modulanzahl ermitteln
  m = zufall_max_int( vert_m );
  *laufz = (double**) malloc( m * sizeof( double* ) );
  if( *laufz == NULL )
  {
    return 1;
  }
  for( i=0; i<m; i++ )
  {
    (*laufz)[i] = (double*) malloc( n * sizeof( double ) );
    if( (*laufz)[i] == NULL )
    {
      while( i > 0 )
      {
        i--;
        free( (*laufz)[i] );
      }
      free( *laufz );
      return 1;
    }
  }
  return 0;
}

// gibt Speicher der Laufzeittabelle frei
// Parameter 2 entspricht dem Parameter 3 von load_configdatei
int frei_laufzeit_array( double** laufz, verteilung_int* m_vert )
{
  int i, m;
  // maximale Modulanzahl ermitteln
  m = zufall_max_int( m_vert );
  for( i=0; i<m; i++ )
  {
    free( laufz[i] );
  }
  free( laufz );
  return 0;
}

// lädt eine Beispielgenerierungsdatei für nonmoldable Module
// der Speicher für den letzten Parameter wird reserviert,
// der Rest sind einfache Variablen von denen der Pointer übergeben wird
// Fehler: 1->zu wenig Speicher
//         2->Configdatei fehlerhaft
//         3->Configdatei nicht gefunden
int load_configdatei(
  char*                     configdatei,
  int*                      n,
  verteilung_int*           vert_m,
  verteilung_double*        vert_verh,
  int*                      anz_vert,
  verteilung_double_liste** liste_vert )
{
  int h;
  int i;
  int pos;
  verteilung_double vert;
  verteilung_double_liste* a;
  FILE* fiin;
  fiin = fopen( configdatei, "r" );
  if( fiin == NULL )
  {
    return 3;
  }
  // Prozessorenanzahl einlesen
  if( lese_next_int_r( &fiin, n ) != 0 )
  {
    return 2;
  }
  // zu erzeugende Modulanzahlverteilung einlesen
  if( lese_next_verteilung_int_r( &fiin, vert_m ) != 0 )
  {
    return 2;
  }
  // Verhältnis zwischen paralleler und sequentieller
  // Vereinigung einlesen
  if( lese_next_verteilung_double_r( &fiin, vert_verh ) != 0 )
  {
    return 2;
  }
  // Verteilungsarten zählen
  pos = ftell( fiin );
  // eine muß da sein!
  if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
  {
    return 2;
  }
  *anz_vert = 1;
  while( 1 )
  {
    if( lese_next_verteilung_double_r( &fiin, &vert ) != 0 )
    {
      break;
    }
    (*anz_vert)++;
  }
  // Speicher reservieren
  *liste_vert = (verteilung_double_liste*) malloc(
    (*anz_vert) * sizeof( verteilung_double_liste ) );
  if( *liste_vert == NULL )
  {
    return 1;
  }
  // Daten einlesen
  fseek( fiin, pos, SEEK_SET );
  for( i=0; i<(*anz_vert); i++ )
  {
    // Verteilungswahrscheinlichkeit
    if( lese_next_verteilung_double( &fiin,
                                     &(((*liste_vert)[i]).v) ) != 0 )
    {
      return 2;
    }
    a = ((*liste_vert)[i]).next = (verteilung_double_liste*) malloc(
      sizeof( verteilung_double_liste ) );
    if( a == NULL )
    {
      return 1;
    }
    // Mindestprozessorenanzahl
    if( lese_next_verteilung_double_in_line( &fiin, &(a->v) ) != 0 )
    {
      return 2;
    }
    a = ((*liste_vert)[i]).next->next =
      (verteilung_double_liste*) malloc(
        sizeof( verteilung_double_liste ) );
    if( a == NULL )
    {
      return 1;
    }
    // Laufzeit mit Mindestprozessorenanzahl
    if( lese_next_verteilung_double_in_line( &fiin, &(a->v) ) != 0 )
    {
      return 2;
    }
    // Parallelisierungsparameter
    do
    {
      a->next = (verteilung_double_liste*) malloc(
        sizeof( verteilung_double_liste ) );
      a = a->next;
      if( a == NULL )
      {
        return 1;
      }
      if( lese_next_int_in_line( &fiin, &h ) != 0 )
      {
        return 2;
      }
      a->bis = h;
      if( lese_next_verteilung_double_in_line( &fiin, &(a->v) ) != 0 )
      {
        return 2;
      }
    } while( a->bis < (*n) );
    a->next = NULL;
    lese_next_ret( &fiin );
  }
  fclose( fiin );
  return 0;
}

// gibt beim Laden einer Beispielgenerierungsdatei reservierten
// Speicher wieder frei
// es werden die letzten beiden Parameter von load_configdatei benötigt
int frei_configdatei(
  int                      anz_vert,
  verteilung_double_liste* liste_vert )
{
  int i;
  verteilung_double_liste* a;
  verteilung_double_liste* b;
  for( i=0; i<anz_vert; i++ )
  {
    a = (liste_vert[i]).next;
    while( a != NULL )
    {
      b = a->next;
      free( a );
      a = b;
    }
  }
  free( liste_vert );
  return 0;
}
