// Optimierungsroutinen moldable, beliebiger Modulabhängigkeitsgraph
// Graphüberführung in Schichten mit unabhängigen Knoten, Schichtenwahl
// erfolgt nach Tiefe im Modulabhängigkeitsgraphen
//
// optimiert Module deren Laufzeiten in der Datei daten stehen
// (i-te Zeile j-te Spalte enthält Laufzeit vom i-ten Modul
//  mit j Prozessoren)
// oder mit Hilfe der übergebenen Funktion laufzeit
// (modulnummer>=1, prozessorenanzahl>=1) ermittelt werden können
//
// Aufruf:
//   optimiere_datei(
//     char* Ergebnisdatei,
//     char* Laufzeitendatei,
//     char* Graphendatei )
// oder:
//   optimiere_formel(
//     double** Startzeit der Module,
//     int**    Modulbreiten,
//     int**    linker Prozessor jedes Moduls,
//     graph*   Modulabhängigkeitsgraph,
//     int      Maximalprozessorenanzahl,
//     double   ((*laufzeit)( modulnummer, prozessorenanzahl )) )
//   Der Speicher für die Startzeiten, Breiten und linker Prozessor
//   wird reserviert
//
// Fehler: 1->zu wenig Speicher
//         2->Eingabedatei mit Laufzeiten nicht gefunden / fehlerhaft
//         3->Eingabedatei mit Graphen nicht gefunden / fehlerhaft /
//            Graph enthält Zyklus
//         4->kann Ergebnisdatei nicht anlegen / Schreibfehler

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include "../prgx/baum.h"
#include "../prgx/graphen.h"
#include "../prgx/load.h"
#include "../prgx/save.h"
#include "../prgx/suchsort.h"


// -------------- Formelbasierte Optimierungsroutinen ------------------
#ifdef FORMEL

// Optimierungsroutine; siehe oben
int optimiere_formel(
  double** startzeit,
  int**    breite,
  int**    linker,
  graph*   g,
  int      n,
  double   ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  int m, i, j, k, q, w, p, akt;
  int* tief_num;
  double* lz;
  double* endzeit;
  double* startzei[2];
  int* breit[2];
  int* linke[2];
  int* sort;
  int* hilfsfeld;
  int* anz;
  int* sor[2];
  double st, sr, r, fl;
  baum heap;
  m = max_knoten( g );
  *startzeit = (double*)malloc( m * sizeof( double ) );
  if( *startzeit == NULL )
  {
    return 1;
  }
  *breite = (int*)malloc( m * sizeof( int ) );
  if( *breite == NULL )
  {
    return 1;
  }
  *linker = (int*)malloc( m * sizeof( int ) );
  if( *linker == NULL )
  {
    return 1;
  }
  startzei[0] = *startzeit;
  breit[0] = *breite;
  linke[0] = *linker;
  startzei[1] = (double*)malloc( m * sizeof( double ) );
  if( startzei[1] == NULL )
  {
    return 1;
  }
  breit[1] = (int*)malloc( m * sizeof( int ) );
  if( breit[1] == NULL )
  {
    return 1;
  }
  linke[1] = (int*)malloc( m * sizeof( int ) );
  if( linke[1] == NULL )
  {
    return 1;
  }
  tief_num = (int*)malloc( ( m + 1 ) * sizeof( int ) );
  if( tief_num == NULL )
  {
    return 1;
  }
  sort = (int*)malloc( m * sizeof( int ) );
  if( sort == NULL )
  {
    return 1;
  }
  lz = (double*)malloc( m * sizeof( double ) );
  if( lz == NULL )
  {
    return 1;
  }
  hilfsfeld = (int*)malloc( m * sizeof( int ) );
  if( hilfsfeld == NULL )
  {
    return 1;
  }
  sor[1] = (int*)malloc( m * sizeof( int ) );
  if( sor[1] == NULL )
  {
    return 1;
  }
  sor[0] = hilfsfeld;
  endzeit = (double*)malloc( n * sizeof( double ) );
  if( endzeit == NULL )
  {
    return 1;
  }
  // Tiefe bestimmen
  if( tiefe( g, tief_num ) != 0 )
  {
    return 3;
  }
  anz = (int*)malloc( ( tief_num[0] + 1 ) * sizeof( int ) );
  if( anz == NULL )
  {
    return 1;
  }
  for( i=0; i<=tief_num[0]; i++ )
  {
    anz[i] = 0; // Elemente in einer Schicht
  }
  for( i=1; i<=m; i++ )
  {
    lz[i-1] = laufzeit( i, 1 );
    (anz[ tief_num[i] ])++;
    sort[i-1] = i-1;
  }
  // Bucketsort der Module in die Schichten ohne Zerstörung
  // der Reihenfolge in Sort
  q = 0;
  for( i=1; i<=tief_num[0]; i++ )
  {
    w = q + anz[i];
    anz[i] = q;
    q = w;
  }
  for( i=0; i<m; i++ ) // übernehmen
  {
    hilfsfeld[ anz[ tief_num[ sort[i] + 1 ] ] ] = sort[i];
    (anz[ tief_num[ sort[i] + 1 ] ])++;
    (*breite)[i] = 1; // Ergebnis vorbereiten
  }
  st = 0.0; // bis dahin wurde bereits eingepackt
  init_size_baum( &heap, n );
  for( i=0; i<tief_num[0]; i++ ) // für alle Schichten
  {
    // Module ordnen
    merge_modul_zeit( lz, hilfsfeld, sort, anz[i], anz[i+1] );
    // einpacken mit einem Proz.
    if( such_heap_min_init_set_werte( &heap, endzeit, n, st ) != 0 )
    {
      return 1;
    }
    (*linker)[ hilfsfeld[ anz[i] ] ] = 1;
    (*startzeit)[ hilfsfeld[ anz[i] ] ] = st;
    set_heap_min( &heap, endzeit, st + lz[ hilfsfeld[ anz[i] ] ] );
    q = 0;
    for( k=anz[i]+1; k<anz[i+1]; k++ )
    {
      (*linker)[ hilfsfeld[k] ] = such_heap_min( &heap );
      (*startzeit)[ hilfsfeld[k] ] =
        endzeit[ (*linker)[ hilfsfeld[k] ] ];
      set_heap_min( 
        &heap, endzeit,
        endzeit[ (*linker)[ hilfsfeld[k] ] ] + lz[ hilfsfeld[k] ] );
      if( (*linker)[ hilfsfeld[k] ] == 0 )
      {
        q = 1; // auf dem höchsten steht ein anderer
      }
      ((*linker)[ hilfsfeld[k] ])++;
    }
    sr = endzeit[0];
    for( k=1; k<n; k++ )
    {
      if( endzeit[k] > sr )
      {
        sr = endzeit[k];
        // ein anderer Prozessor läuft länger als der mit dem höchsten
        q = 1;
      }
    }
    if( q == 1 )
    {
      // Gütefaktor 2 in dieser Schicht erreicht
      st = sr;
      continue; // nächste Schicht
    }
    akt = 1; // in 0 steht die aktuelle Lösung
    // kopieren der Breite-Daten und Sortierungsinformation
    for( j=anz[i]; j<anz[i+1]; j++ )
    {
      breit[1][ hilfsfeld[j] ] = breit[0][ hilfsfeld[j] ];
      sor[1][j] = sor[0][j];
    }
    // verbreitern notwendig
    j = 0; // maximal n Verbreiterungen
    while( 1 )
    {
      // höchstes Modul suchen; sor[0][], da Reihenfolge
      // nicht verändert wird
      q = such_max_indirekt_range( lz, sor[0], anz[i], anz[i+1] );
      if( breit[akt][ sor[0][q] ] == 1 )
      {  // erstmalig ausgewählt (eine Verbreiterung weniger möglich)
        j += 2;
      }
      else
      {
        j++;
      }
      if( j > n )
      {
        // kein Platz mehr ab Zeitpunkt st für alle verbreiterten Module
        break;
      }
      breit[akt][ sor[0][q] ]++;
      lz[ sor[0][q] ] = laufzeit( sor[0][q]+1,
                                  breit[akt][ sor[0][q] ] );
      k = 0;
      w = anz[i];
      // breite Module einpacken
      while( ( k < n ) && ( w < anz[i+1] ) )
      {
        for( p=k; p<k+breit[akt][ sor[0][w] ]; p++ )
        {
          startzei[akt][ sor[0][w] ] = st;
          linke[akt][ sor[0][w] ] = k + 1;
          endzeit[p] = st + lz[ sor[0][w] ];
        }
        k = p;
        w++;
      }
      // restliche Module einpacken
      such_heap_min_init_werte( &heap, endzeit, n );
      for( ; w<anz[i+1]; w++ )
      {
        p = such_heap_min( &heap );
        startzei[akt][ sor[0][w] ] = endzeit[p];
        linke[akt][ sor[0][w] ] = p + 1;
        set_heap_min( &heap, endzeit, endzeit[p] + lz[ sor[0][w] ] );
      }
      // Ergebnishöhe bestimmen
      p = such_max( endzeit, k );
      if( endzeit[p] > sr )
      {
        break; // letzte Verbreiterung war nicht mehr sinnvoll
      }
      akt = ( akt + 1 ) & 1;
      sr = endzeit[p];
      breit[akt][ sor[0][q] ]++; // anpassen der anderen Lösung
    }
    // Flächenminimum suchen
    j = anz[i];
    fl = 0.0; // Gesamtfläche der Module in dieser Schicht
    while( breit[akt][ sor[0][j] ] != 1 ) // alle verbreiterten Module
    {
      k = breit[akt][ sor[0][j] ] + 1;
      while( k <= n )
      {
        r = laufzeit( sor[0][j] + 1, k );
        if( r * k < breit[akt][ sor[0][j] ] * lz[ sor[0][j] ] )
        {
          breit[akt][ sor[0][j] ] = k;
          lz[ sor[0][j] ] = r;
        }
        k++;
      }
      // Fläche der aktuellen Schicht aktualisieren
      fl += breit[akt][ sor[0][j] ] * lz[ sor[0][j] ];
      j++;
      if( j == anz[i+1] )
      {
        break;
      }
    }
    for( k=j; k<anz[i+1]; k++ )
    {
      fl += lz[ sor[0][k] ]; // Fläche der restlichen Module hinzunehmen
    }
    // sortieren
    merge_modul_breit_zeit( lz, breit[akt], sor[akt], sort, anz[i], j );
    while( 1 )
    {
      // einpacken
      for( j=0; j<n; j++ )
      {
        endzeit[j] = st;
      }
      j = anz[i];
      // breite Module
      while( breit[akt][ sor[akt][j] ] != 1 )
      {
        k = such_min_breit( endzeit, n, breit[akt][ sor[akt][j] ] - 1,
                            &r );
        startzei[akt][ sor[akt][j] ] = r;
        linke[akt][ sor[akt][j] ] = k + 1;
        r += lz[ sor[akt][j] ];
        for( q=k; q<k+breit[akt][ sor[akt][j] ]; q++ )
        {
          endzeit[q] = r;
        }
        j++;
        if( j == anz[i+1] )
        {
          break;
        }
      }
      // Module mit Breite 1
      such_heap_min_init_werte( &heap, endzeit, n );
      while( j < anz[i+1] )
      {
        k = such_heap_min( &heap );
        startzei[akt][ sor[akt][j] ] = endzeit[k];
        linke[akt][ sor[akt][j] ] = k + 1;
        set_heap_min( &heap, endzeit, endzeit[k] + lz[ sor[akt][j] ] );
        j++;
      }
      j = such_max( endzeit, n );
      if( endzeit[j] < sr ) // besser geworden ?
      {
        sr = endzeit[j];
        q = akt;
        akt = ( akt + 1 ) & 1;
        // Breiten übernehmen
        for( k=anz[i]; k<min( anz[i+1], anz[i]+n ); k++ )
        {
          breit[akt][ sor[1][k] ] = breit[q][ sor[1][k] ];
        }
      }
      if( n * ( endzeit[j] - st ) <= 3 * fl )
      {
        break; // Ergebnisgüte 3 erreicht
      }
      // höchstes Modul verbreitern
      j = such_max_indirekt_range( lz, sor[akt], anz[i],
                                   min( anz[i+1], anz[i] + n ) );
      k = breit[akt][ sor[akt][j] ];
      if( k == n )
      {
        break; // wird bereits mit n Prozessoren ausgeführt
      }
      // Fläche der aktuellen Schicht aktualisieren
      fl -= breit[akt][ sor[akt][j] ] * lz[ sor[akt][j] ];
      k++;
      breit[akt][ sor[akt][j] ] = k; // mind. um 1 verbreitern
      lz[ sor[akt][j] ] = laufzeit( sor[akt][j] + 1, k );
      while( k <= n )
      {
        r = laufzeit( sor[akt][j] + 1, k );
        if( r * k < breit[akt][ sor[akt][j] ] * lz[ sor[akt][j] ] )
        {
          breit[akt][ sor[akt][j] ] = k;
          lz[ sor[akt][j] ] = r;
        }
        k++;
      }
      // Fläche der aktuellen Schicht aktualisieren
      fl += breit[akt][ sor[akt][j] ] * lz[ sor[akt][j] ];
      // verbreitertes Modul wieder einsortieren
      misch_modul_breit_zeit( lz, breit[akt], sor[akt], sort,
                              anz[i], j, j+1 );
    }
    if( akt == 0 )
    {
      // Lösung nach 0 kopieren
      for( j=anz[i]; j<anz[i+1]; j++ )
      {
        breit[0][ sor[1][j] ] = breit[1][ sor[1][j] ];
        startzei[0][ sor[1][j] ] = startzei[1][ sor[1][j] ];
        linke[0][ sor[1][j] ] = linke[1][ sor[1][j] ];
      }
    }
    st = sr;
  }
  free_baum( &heap );
  free( tief_num );
  free( sort );
  free( hilfsfeld );
  free( lz );
  free( endzeit );
  free( anz );
  free( linke[1] );
  free( breit[1] );
  free( startzei[1] );
  free( sor[1] );
  return 0;
}

#endif


// -------------- Formelbasierte Optimierungsroutinen ------------------
#ifndef FORMEL

// Optimierungsroutine siehe oben
int optimiere_datei(
  char* ergebnis,
  char* daten,
  char* graph_datei )
{
  graph g;
  double* t;
  int pos;
  int m, n, i, j, k, q, w, p, akt;
  int* tief_num;
  double* lz;
  double* endzeit;
  double* startzei[2];
  double* startzeit;
  int* breite;
  int* linker;
  int* breit[2];
  int* linke[2];
  int* sort;
  int* hilfsfeld;
  int* anz;
  int* sor[2];
  double st, sr, r, fl;
  FILE* fi;
  baum heap;
  init_graph( &g );
  // Graph laden
  if( load_graph_rueck_mehr( graph_datei, &g, 2 ) != 0 )
  {
    return 3;
  }
  // Maximale Knotenanzahl ermitteln
  m = max_knoten( &g );
  // Berechnen der Lösung
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 2;
  }
  // Bestimmen der Prozessorenanzahl
  pos = ftell( fi );
  n = 1;
  if( lese_next_double( &fi, &r ) != 0 )
  {
    return 2;
  }
  while( lese_next_double_in_line( &fi, &r ) == 0 )
  {
    n++;
  }
  fseek( fi, pos, SEEK_SET );
  // Hilfsvariablen
  t = (double*) malloc( sizeof( double ) * m * n );
  if( t == NULL )
  {
    return 1;
  }
  // Laufzeiten laden
  for( i=0; i<m; i++ )
  {
    if( lese_next_double( &fi, &(t[i*n]) ) != 0 )
    { // m ist nur eine obere Schranke
      break;
    }
    for( j=1; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &(t[i*n+j]) ) != 0 )
      {
        return 2;
      }
    }
    lese_next_ret( &fi );
  }
  fclose( fi );
  startzeit = (double*)malloc( m * sizeof( double ) );
  if( startzeit == NULL )
  {
    return 1;
  }
  breite = (int*)malloc( m * sizeof( int ) );
  if( breite == NULL )
  {
    return 1;
  }
  linker = (int*)malloc( m * sizeof( int ) );
  if( linker == NULL )
  {
    return 1;
  }
  startzei[0] = startzeit;
  breit[0] = breite;
  linke[0] = linker;
  startzei[1] = (double*)malloc( m * sizeof( double ) );
  if( startzei[1] == NULL )
  {
    return 1;
  }
  breit[1] = (int*)malloc( m * sizeof( int ) );
  if( breit[1] == NULL )
  {
    return 1;
  }
  linke[1] = (int*)malloc( m * sizeof( int ) );
  if( linke[1] == NULL )
  {
    return 1;
  }
  tief_num = (int*)malloc( ( m + 1 ) * sizeof( int ) );
  if( tief_num == NULL )
  {
    return 1;
  }
  sort = (int*)malloc( m * sizeof( int ) );
  if( sort == NULL )
  {
    return 1;
  }
  lz = (double*)malloc( m * sizeof( double ) );
  if( lz == NULL )
  {
    return 1;
  }
  hilfsfeld = (int*)malloc( m * sizeof( int ) );
  if( hilfsfeld == NULL )
  {
    return 1;
  }
  sor[1] = (int*)malloc( m * sizeof( int ) );
  if( sor[1] == NULL )
  {
    return 1;
  }
  sor[0] = hilfsfeld;
  endzeit = (double*)malloc( n * sizeof( double ) );
  if( endzeit == NULL )
  {
    return 1;
  }
  // Tiefe bestimmen
  if( tiefe( &g, tief_num ) != 0 )
  {
    return 3;
  }
  anz = (int*)malloc( ( tief_num[0] + 1 ) * sizeof( int ) );
  if( anz == NULL )
  {
    return 1;
  }
  for( i=0; i<=tief_num[0]; i++ )
  {
    anz[i] = 0; // Elemente in einer Schicht
  }
  for( i=1; i<=m; i++ )
  {
    lz[i-1] = t[ (i-1) * n ]; // für schnelleren Zugriff kopieren
    (anz[ tief_num[i] ])++;
    sort[i-1] = i-1;
  }
  // Bucketsort der Module in die Schichten ohne Zerstörung
  // der Reihenfolge in Sort
  q = 0;
  for( i=1; i<=tief_num[0]; i++ )
  {
    w = q + anz[i];
    anz[i] = q;
    q = w;
  }
  for( i=0; i<m; i++ ) // übernehmen
  {
    hilfsfeld[ anz[ tief_num[ sort[i] + 1 ] ] ] = sort[i];
    (anz[ tief_num[ sort[i] + 1 ] ])++;
    breite[i] = 1; // Ergebnis vorbereiten
  }
  st = 0.0; // bis dahin wurde bereits eingepackt
  init_size_baum( &heap, n );
  for( i=0; i<tief_num[0]; i++ ) // für alle Schichten
  {
    // Module ordnen
    merge_modul_zeit( lz, hilfsfeld, sort, anz[i], anz[i+1] );
    // einpacken mit einem Proz.
    if( such_heap_min_init_set_werte( &heap, endzeit, n, st ) != 0 )
    {
      return 1;
    }
    linker[ hilfsfeld[ anz[i] ] ] = 1;
    startzeit[ hilfsfeld[ anz[i] ] ] = st;
    set_heap_min( &heap, endzeit, st + lz[ hilfsfeld[ anz[i] ] ] );
    q = 0;
    for( k=anz[i]+1; k<anz[i+1]; k++ )
    {
      linker[ hilfsfeld[k] ] = such_heap_min( &heap );
      startzeit[ hilfsfeld[k] ] = endzeit[ linker[ hilfsfeld[k] ] ];
      set_heap_min(
        &heap, endzeit,
        endzeit[ linker[ hilfsfeld[k] ] ] + lz[ hilfsfeld[k] ] );
      if( linker[ hilfsfeld[k] ] == 0 )
      {
        q = 1; // auf dem höchsten steht ein anderer
      }
      (linker[ hilfsfeld[k] ])++;
    }
    sr = endzeit[0];
    for( k=1; k<n; k++ )
    {
      if( endzeit[k] > sr )
      {
        sr = endzeit[k];
        // ein anderer Prozessor läuft länger als der mit dem höchsten
        q = 1;
      }
    }
    if( q == 1 )
    {
      // Gütefaktor 2 in dieser Schicht erreicht
      st = sr;
      continue; // nächste Schicht
    }
    akt = 1; // in 0 steht die aktuelle Lösung
    // kopieren der Breite-Daten und Sortierungsinformation
    for( j=anz[i]; j<anz[i+1]; j++ )
    {
      breit[1][ hilfsfeld[j] ] = breit[0][ hilfsfeld[j] ];
      sor[1][j] = sor[0][j];
    }
    // verbreitern notwendig
    j = 0; // maximal n Verbreiterungen
    while( 1 )
    {
      // höchstes Modul suchen; sor[0][], da Reihenfolge
      // nicht verändert wird
      q = such_max_indirekt_range( lz, sor[0], anz[i], anz[i+1] );
      if( breit[akt][ sor[0][q] ] == 1 )
      { // erstmalig ausgewählt (eine Verbreiterung weniger möglich)
        j += 2;
      }
      else
      {
        j++;
      }
      if( j > n )
      { // kein Platz mehr ab Zeitpunkt st für alle verbreiterten Module
        break;
      }
      breit[akt][ sor[0][q] ]++;
      lz[ sor[0][q] ] =
        t[ ( sor[0][q] ) * n + breit[akt][ sor[0][q] ] - 1 ];
      k = 0;
      w = anz[i];
      // breite Module einpacken
      while( ( k < n ) && ( w < anz[i+1] ) )
      {
        for( p=k; p<k+breit[akt][ sor[0][w] ]; p++ )
        {
          startzei[akt][ sor[0][w] ] = st;
          linke[akt][ sor[0][w] ] = k + 1;
          endzeit[p] = st + lz[ sor[0][w] ];
        }
        k = p;
        w++;
      }
      // restliche Module einpacken
      such_heap_min_init_werte( &heap, endzeit, n );
      for( ; w<anz[i+1]; w++ )
      {
        p = such_heap_min( &heap );
        startzei[akt][ sor[0][w] ] = endzeit[p];
        linke[akt][ sor[0][w] ] = p + 1;
        set_heap_min( &heap, endzeit, endzeit[p] + lz[ sor[0][w] ] );
      }
      // Ergebnishöhe bestimmen
      p = such_max( endzeit, k );
      if( endzeit[p] > sr )
      {
        break; // letzte Verbreiterung war nicht mehr sinnvoll
      }
      akt = ( akt + 1 ) & 1;
      sr = endzeit[p];
      breit[akt][ sor[0][q] ]++; // anpassen der anderen Lösung
    }
    // Flächenminimum suchen
    j = anz[i];
    fl = 0.0; // Fläche der aktuellen Schicht
    while( breit[akt][ sor[0][j] ] != 1 ) // alle verbreiterten Module
    {
      k = breit[akt][ sor[0][j] ] + 1;
      while( k <= n )
      {
        r = t[ ( sor[0][j] ) * n + k-1 ];
        if( r * k < breit[akt][ sor[0][j] ] * lz[ sor[0][j] ] )
        {
          breit[akt][ sor[0][j] ] = k;
          lz[ sor[0][j] ] = r;
        }
        k++;
      }
      // Fläche der aktuellen Schicht berechnen
      fl += breit[akt][ sor[0][j] ] * lz[ sor[0][j] ];
      j++;
      if( j == anz[i+1] )
      {
        break;
      }
    }
    for( k=j; k<anz[i+1]; k++ )
    {
      fl += lz[ sor[0][k] ]; // Fläche der restlichen Module hinzunehmen
    }
    // sortieren
    merge_modul_breit_zeit( lz, breit[akt], sor[akt], sort, anz[i], j );
    while( 1 )
    {
      // einpacken
      for( j=0; j<n; j++ )
      {
        endzeit[j] = st;
      }
      j = anz[i];
      // breite Module
      while( breit[akt][ sor[akt][j] ] != 1 )
      {
        k = such_min_breit( endzeit, n, breit[akt][ sor[akt][j] ] - 1,
                            &r );
        startzei[akt][ sor[akt][j] ] = r;
        linke[akt][ sor[akt][j] ] = k + 1;
        r += lz[ sor[akt][j] ];
        for( q=k; q<k+breit[akt][ sor[akt][j] ]; q++ )
        {
          endzeit[q] = r;
        }
        j++;
        if( j == anz[i+1] )
        {
          break;
        }
      }
      // Module mit Breite 1
      such_heap_min_init_werte( &heap, endzeit, n );
      while( j < anz[i+1] )
      {
        k = such_heap_min( &heap );
        startzei[akt][ sor[akt][j] ] = endzeit[k];
        linke[akt][ sor[akt][j] ] = k + 1;
        set_heap_min( &heap, endzeit, endzeit[k] + lz[ sor[akt][j] ] );
        j++;
      }
      j = such_max( endzeit, n );
      if( endzeit[j] < sr ) // besser geworden ?
      {
        sr = endzeit[j];
        q = akt;
        akt = ( akt + 1 ) & 1;
        for( k=anz[i]; k<min( anz[i+1], anz[i]+n ); k++ )
        { // Breiten übernehmen
          breit[akt][ sor[1][k] ] = breit[q][ sor[1][k] ];
        }
      }
      if( n * ( endzeit[j] - st ) <= 3 * fl )
      {
        break; // Ergebnisgüte 3 erreicht
      }
      // höchstes Modul verbreitern
      j = such_max_indirekt_range( lz, sor[akt], anz[i],
                                   min( anz[i+1], anz[i] + n ) );
      k = breit[akt][ sor[akt][j] ];
      if( k == n )
      {
        break; // wird bereits mit n Prozessoren ausgeführt
      }
      // Fläche der aktuellen Schicht aktualisieren
      fl -= breit[akt][ sor[akt][j] ] * lz[ sor[akt][j] ];
      k++;
      breit[akt][ sor[akt][j] ] = k; // mind. um 1 verbreitern
      lz[ sor[akt][j] ] = t[ ( sor[akt][j] ) * n + k-1 ];
      while( k <= n )
      {
        r = t[ ( sor[akt][j] ) * n + k-1 ];
        if( r * k < breit[akt][ sor[akt][j] ] * lz[ sor[akt][j] ] )
        {
          breit[akt][ sor[akt][j] ] = k;
          lz[ sor[akt][j] ] = r;
        }
        k++;
      }
      // Fläche der aktuellen Schicht aktualisieren
      fl += breit[akt][ sor[akt][j] ] * lz[ sor[akt][j] ];
      // verbreitertes Modul wieder einsortieren
      misch_modul_breit_zeit( lz, breit[akt], sor[akt], sort,
                              anz[i], j, j+1 );
    }
    if( akt == 0 )
    {
      // Lösung nach 0 kopieren
      for( j=anz[i]; j<anz[i+1]; j++ )
      {
        breit[0][ sor[1][j] ] = breit[1][ sor[1][j] ];
        startzei[0][ sor[1][j] ] = startzei[1][ sor[1][j] ];
        linke[0][ sor[1][j] ] = linke[1][ sor[1][j] ];
      }
    }
    st = sr;
  }
  free_baum( &heap );
  free( tief_num );
  free( sort );
  free( hilfsfeld );
  free( lz );
  free( endzeit );
  free( anz );
  free( linke[1] );
  free( breit[1] );
  free( startzei[1] );
  free( sor[1] );
  free_graph( &g );
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 4;
  }
  for( i=0; i<m; i++ )
  {
    if( save_text( &fi, "Modul " ) != 0 )
    {
      return 4;
    }
    if( save_int( &fi, i+1 ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, ": von " ) != 0 )
    {
      return 4;
    }
    if( save_double( &fi, startzeit[i] ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 4;
    }
    if( save_double( &fi, startzeit[i] + t[ i*n + breite[i]-1 ] ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, " auf Prozessoren " ) != 0 )
    {
      return 4;
    }
    if( save_int( &fi, linker[i] ) != 0 )
    {
      return 4;
    }
    if( save_text( &fi, " bis " ) != 0 )
    {
      return 4;
    }
    if( save_int_r( &fi, linker[i] + breite[i] - 1 ) != 0 )
    {
      return 4;
    }
  }
  fclose( fi );
  free( t );
  free( startzeit );
  free( breite );
  free( linker );
  return 0;
}

#endif
