// Routine zum Erzeugen von Beispielen:
//  beliebiger Abhängigkeitsgraph, nonmoldable
// Aufruf:
//   baue_bsp_array(
//     double**                  Feld der Größe m*n (Speicher muß schon
//                                                   reserviert sein),
//     graph*                    Modulabhängigkeitsgraph,
//     int                       Prozessorenanzahl,
//     verteilung_int*           Gesamtanzahl der Module,
//     verteilung_text*          Anzahl der Kanten (als Variable ist m
//                               und M zulässig für Modulanzahl),
//     verteilung_text*          Knoten i darf maximal mit_Knoten ?+i
//                               verbunden werden (m und M für
//                               Modulanzahl, k und K für Kantenanzahl),
//     int                       Anzahl der verschiedenen Verteilungen
//                               zum Erzeugen von Modulen,
//     verteilung_double_liste** Listenfeld mit
//                                1. Eintrag:
//                                 Anteile der Verteilungsarten (die
//                                 Wahrscheinlichkeit, daß Verteilung i
//                                 genommen wird ist
//                                 ant[i]/summe(ant[?]), wobei
//                                 ant[i] = zufall(erster Eintrag[i]) )
//                                2. Eintrag:
//                                 Mindestprozessorenanzahl
//                                3. Eintrag:
//                                 Höhe des Moduls mit
//                                 Mindestprozessorenanzahl
//                                ab 4. Eintrag:
//                                 Parallelisierungsdaten (bis bestimmt
//                                 dabei wie lange der Eintrag gültig
//                                 ist)
//   );
//
// oder:
//   baue_bsp_datei(
//     char* Filename für Ergebnis,
//     char* Filename für Modulabhängigkeitsgraphen,
//     char* Filename Eingabedatei );
//
// Fehler: 1->zu wenig Speicher
//         2->konnte Ergebnisdatei nicht anlegen
//         3->nicht genug Einträge im Listenfeld / Fehler in Datei
//         4->konnte Eingabedatei nicht finden


#include "../prgx/graphen.h"
#include "../prgx/zufall.h"

int baue_bsp_array(
  double**                 laufz,
  graph*                   abhgraph,
  int                      n,
  verteilung_int*          vert_m,
  verteilung_text*         vert_kanten,
  verteilung_text*         vert_seq,
  int                      anz_vert,
  verteilung_double_liste* liste_vert );

int baue_bsp_datei(
  char* fn,
  char* fg,
  char* fin );

// --------------------------- weitere Routinen ------------------------

// reserviert den maximal benötigten Speicher für eine Laufzeittabelle
// Parameter 2 und 3 entsprechen den Parametern 2 und 3
// von load_configdatei
// Fehler: 1->zu wenig Speicher
int init_laufzeit( double** *laufz, int n, verteilung_int* m_vert );

// gibt Speicher der Laufzeittabelle frei
// Parameter 2 entspricht dem Parameter 3 von load_configdatei
void frei_laufzeit( double** laufz, verteilung_int* vert_m );

// lädt eine Beispielgenerierungsdatei für nonmoldable Module
// der Speicher für den letzten Parameter wird reserviert, der Rest
// sind einfache Variablen von denen der Pointer übergeben wird
// Fehler: 1->zu wenig Speicher
//         2->Configdatei fehlerhaft
//         3->Configdatei nicht gefunden
int load_configdatei_nonmoldable(
  char*                     configdatei,
  int*                      n,
  verteilung_int*           vert_m,
  verteilung_text*          vert_kanten,
  verteilung_text*          vert_seq,
  int*                      anz_vert,
  verteilung_double_liste** liste_vert );

// gibt beim Laden einer Beispielgenerierungsdatei reservierten Speicher
// wieder frei
// es werden die letzten beiden Parameter von load_configdatei benötigt
void frei_configdatei(
  int                      anz_vert,
  verteilung_double_liste* liste_vert );
