// Optimierungsroutinen nonmoldable, beliebiger Modulabhängigkeitsgraph
// Modulabhängigkeitsgraph wird in einen serien-parallelen Graphen
// überführt
//
// optimiert Module deren Laufzeiten in der Datei daten stehen
// (i-te Zeile j-te Spalte enthält Laufzeit vom i-ten Modul
//  mit j Prozessoren)
// oder mit Hilfe der übergebenen Funktion laufzeit
// (modulnummer>=1, prozessorenanzahl>=1) ermittelt werden können
//
// Aufruf:
//   optimiere_datei(
//     char* Ergebnisdatei,
//     char* Laufzeitendatei,
//     char* Graphendatei )
// oder:
//   optimiere_formel(
//     double** Startzeit der Module,
//     int**    Modulbreiten,
//     int**    linker Prozessor jedes Moduls,
//     graph*   Modulabhängigkeitsgraph,
//     int      Maximalprozessorenanzahl,
//     double   ((*laufzeit)( modulnummer, prozessorenanzahl )) )
//   Der Speicher für die Startzeiten, Breiten und linker Prozessor
//   wird reserviert
//
// Fehler: 1->zu wenig Speicher
//         2->Eingabedatei mit Laufzeiten nicht gefunden / fehlerhaft
//         3->Eingabedatei mit Graphen nicht gefunden / fehlerhaft /
//            Graph enthält Zyklus
//         4->kann Ergebnisdatei nicht anlegen / Schreibfehler

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include "../prgx/dgraphen.h"
#include "../prgx/save.h"
#include "../prgx/load.h"
#include "../prgx/listen.h"

// der übergebene Graph darf von der Optimierungsroutine nicht
// verändert werden (wichtig bei den Testläufen mit FORMEL,
// anderenfalls sinnlos)
#ifdef FORMEL
#define GRAPH_ERHALTEN
#endif

// fügt Knoten m in die entsprechende Liste ein
// Fehler: 1->zu wenig Speicher
int knoten_eintragen(
  graph*    h,
  int       m,
  intliste* inoutnull,
  intliste* einzelkante,
  intliste* outnullineins,
  intliste* innullouteins,
  intliste* inouteins,
  intliste* outeins,
  intliste* ineins,
  intliste* inoutgross,
  int*      aktinliste )
{
  int i;
  if( ziel_knoten( h, m ) == 0 )
  {
    if( quell_knoten( h, m ) == 0 )
    {
      aktinliste[m-1] = 0;
      return add_end_il( inoutnull, m );
    }
    if( vgl_eing_grad( h, m, 1 ) == 0 )
    {
      aktinliste[m-1] = 2;
      return add_end_il( outnullineins, m );
    }
    aktinliste[m-1] = 8;
    return 0;
  }
  if( vgl_ausg_grad( h, m, 1 ) == 0 )
  {
    reset_kind( h, m );
    i = next_kind( h, m );
    if( vgl_eing_grad( h, i, 1 ) == 0 )
    {
      aktinliste[m-1] = 1;
      if( add_end_il( einzelkante, m ) != 0 )
      {
        return 1;
      }
      aktinliste[i-1] = 1;
      return add_end_il( einzelkante, i );
    }
    if( quell_knoten( h, m ) == 0 )
    {
      aktinliste[m-1] = 3;
      return add_end_il( innullouteins, m );
    }
    if( vgl_eing_grad( h, m, 1 ) == 0 )
    {
      aktinliste[m-1] = 4;
      return add_end_il( inouteins, m );
    }
    aktinliste[m-1] = 5;
    return add_end_il( outeins, m );
  }
  if( vgl_eing_grad( h, m, 1 ) == 0 )
  {
    aktinliste[m-1] = 6;
    return add_end_il( ineins, m );
  }
  aktinliste[m-1] = 7;
  return add_end_il( inoutgross, m );
}

// fügt Knoten m in die entsprechende Liste ein, sofern er nicht
// schon in dieser ist
// ist er bereits als Zielknoten in einzelkante gespeichert,
// so wird er (auch wenn er es ist) nicht noch einmal als
// Quellknoten abgelegt
// Fehler: 1->zu wenig Speicher
int knoten_umsortieren(
  graph*    h,
  int       m,
  intliste* einzelkante,
  intliste* outnullineins,
  intliste* innullouteins,
  intliste* inouteins,
  intliste* outeins,
  intliste* ineins,
  int*      aktinliste )
{
  int i;
  switch( aktinliste[m-1] )
  {
    case 5: // bisher in outeins
    {
      if( vgl_eing_grad( h, m, 1 ) == 0 )
      {
        aktinliste[m-1] = 4;
        return add_end_il( inouteins, m );
      }
      return 0;
    }
    case 6: // bisher in ineins
    {
      if( vgl_ausg_grad( h, m, 1 ) == 0 )
      {
        reset_kind( h, m );
        i = next_kind( h, m );
        if( vgl_eing_grad( h, i, 1 ) == 0 )
        {
          aktinliste[m-1] = 1;
          aktinliste[i-1] = 1;
          if( add_end_il( einzelkante, m ) != 0 )
          {
            return 1;
          }
          return add_end_il( einzelkante, i );
        }
        aktinliste[m-1] = 4;
        return add_end_il( inouteins, m );
      }
      return 0;
    }
    case 7: // bisher in inoutgross
    {
      if( vgl_ausg_grad( h, m, 1 ) == 0 )
      {
        reset_kind( h, m );
        i = next_kind( h, m );
        if( vgl_eing_grad( h, i, 1 ) == 0 )
        {
          aktinliste[m-1] = 1;
          aktinliste[i-1] = 1;
          if( add_end_il( einzelkante, m ) != 0 )
          {
            return 1;
          }
          return add_end_il( einzelkante, i );
        }
        if( quell_knoten( h, m ) == 0 )
        {
          aktinliste[m-1] = 3;
          return add_end_il( innullouteins, m );
        }
        aktinliste[m-1] = 5;
        return add_end_il( outeins, m );
      }
      if( vgl_eing_grad( h, m, 1 ) == 0 )
      {
        aktinliste[m-1] = 6;
        return add_end_il( ineins, m );
      }
      return 0;
    }
    case 8: // bisher in keiner Liste (in>=2; out=0)
    {
      if( vgl_eing_grad( h, m, 1 ) == 0 )
      {
        aktinliste[m-1] = 2;
        return add_end_il( outnullineins, m );
      }
      return 0;
    }
    default:
    {
      return 0;
    }
  }
}

// führt zu m auch eine Einzelkante; wird nur eingetragen,
// wenn m kein Startpunkt einer Einzelkante ist
// Fehler: 1->zu wenig Speicher
int vorg_einzelkante(
  graph*    h,
  int       m,
  intliste* einzelkante,
  int*      aktinliste )
{
  int i, j;
  i = aktinliste[m-1];
  if( ( i != 2 ) && ( i != 4 ) && ( i != 6 ) )
  {
    return 0;
  }
  reset_vorg( h, m );
  i = next_vorg( h, m );
  if( i != 0 )
  {
    j = next_vorg( h, m );
    if( j == 0 ) // m hat Eingangsgrad 1 mit Vorgänger i
    {
      if( vgl_ausg_grad( h, i, 1 ) == 0 ) // hat i nur ein Kind
      {
        // (i,m) ist Einzelkante
        aktinliste[m-1] = 1;
        aktinliste[i-1] = 1;
        if( add_end_il( einzelkante, i ) != 0 )
        {
          return 1;
        }
        return add_end_il( einzelkante, m );
      }
    }
  }
  return 0;
}

int verein_baum(
  graph*    b,
  int*      m,
  intliste* inoutnull,
  int       i,
  int*      beschrift,
  int       plus,
  int       mehrb )
{
  int j;
  if( i == 1 )
  {
    return del_anf_il( inoutnull );
  }
  j = i / 2;
  i -= j;
  i = verein_baum( b, m, inoutnull, i, beschrift, plus, mehrb );
  if( i < 0 )
  {
    return -1;
  }
  j = verein_baum( b, m, inoutnull, j, beschrift, plus, mehrb );
  if( j < 0 )
  {
    return -1;
  }
  (*m)++;
  if( add_kante_mehr( b, *m, i, mehrb ) != 0 )
  {
    return -1;
  }
  if( add_kante_mehr( b, *m, j, mehrb ) != 0 )
  {
    return -1;
  }
  beschrift[ *m - plus ] = 1; // parallele Vereinigung
  return *m;
}

// berechnet den Pseudo-Konstruktionsbaum b zu einem beliebigen
// Graphen g
// für den aus b erstellbaren serien-parallelen Graphen h gilt:
//   existiert ein Pfad von Knoten x nach Knoten y in g, so auch in h
//   (die Umkehrung dieser Aussage gilt nicht!)
// Info für Wahl von "mehr" bei vorhergehenden Routinen:
//   baue_sp_baum_pseudo verdoppelt die Knotenanzahl von g (nur
//   wenn GRAPH_ERHALTEN nicht gesetzt ist)
// mehrg gibt die zu erwartende Knotenanzahl für g an, mehrb die für b
// Rückgabe: -1->zu wenig Speicher
//           Nummer der Wurzel von b
//           in beschrift (Speicher wird reserviert!) liegt unter i
//           die Beschriftung (1..parallel; 2..sequentiell) für
//           Knoten i+plus
//           die Knoten mit einer Nummer kleiner als plus sind Blätter
//           in b und entsprechen den Knoten von g
int baue_sp_baum_pseudo(
  graph* g,
  graph* b,
  int**  beschrift,
  int*   plus,
  int    mehrg,
  int    mehrb )
{
  // Knoten ohne Kanten
  intliste inoutnull;
  // einzige Kante zwischen zwei Knoten (liegen nacheinander in der
  // Liste)
  intliste einzelkante;
  // Quellknoten mit Ausgangsgrad Eins
  intliste innullouteins;
  // Zielknoten mit Eingangsgrad Eins
  intliste outnullineins;
  // Knoten mit Ein- und Ausgangsgrad Eins
  intliste inouteins;
  // Knoten mit Ausgangsgrad Eins und Eingangsgrad größer als Eins
  intliste outeins;
  // Knoten mit Eingangsgrad Eins und Ausgangsgrad größer als Eins
  intliste ineins;
  // Knoten mit Ausgangsgrad von mindestens Zwei und Eingangsgrad
  // ungleich Eins
  intliste inoutgross;
  // transitive Kanten
  kantenliste trans;
  int m, i, j, k, p, q;
  int* aktinliste;
  kante* v;
  graph* h;
#ifdef GRAPH_ERHALTEN
  graph h_h;
  init_graph( &h_h );
  h = &h_h;
  m = max_knoten( g );
  if( mehrg < 2 * m )
  {
    mehrg = 2 * m;
  }
  if( copy_graph_mehr( g, h, mehrg ) != 0 )
  {
    return -1;
  }
#else
  m = max_knoten( g );
  if( mehrg < 2 * m )
  {
    mehrg = 2 * m;
  }
  h = g;
#endif
  init_il( &inoutnull );
  init_il( &einzelkante );
  init_il( &outnullineins );
  init_il( &innullouteins );
  init_il( &inouteins );
  init_il( &outeins );
  init_il( &ineins );
  init_il( &inoutgross );
  init_kl( &trans );
  if( mehrb < 2 * m )
  {
    mehrb = 2 * m;
  }
  *beschrift = (int*)malloc( m * sizeof( int ) );
  if( *beschrift == NULL )
  {
    return -1;
  }
  aktinliste = (int*)malloc( 2 * m * sizeof( int ) );
  if( aktinliste == NULL )
  {
    return -1;
  }
  for( i=1; i<=m; i++ ) // transitive Kanten löschen
  {
    switch( transitiv_kante_aus_mehr( h, i, &trans, mehrg ) )
    {
      case 1: // zu wenig Speicher
        return -1;
      case 0: // Knoten existiert
        while( 1 ) // gibt es transitive Kanten ?
        {
          v = del_anf_kl( &trans );
          if( v == NULL ) // Liste ist leer
          {
            break;
          }
          del_kante_verweis( h, v );
        }
        break;
    }
  }
  // gegebene Knoten in entsprechende Listen sortieren
  for( i=1; i<=m; i++ )
  {
    if( test_knoten( h, i ) == 0 ) // Knoten existiert
    {
      if( knoten_eintragen( h, i, &inoutnull, &einzelkante,
                            &outnullineins, &innullouteins,
                            &inouteins, &outeins, &ineins,
                            &inoutgross, aktinliste ) != 0 )
      {
        return -1;
      }
    }
  }
  del_graph( b );
  *plus = m + 1;
  while( 1 ) // schrittweise den Graphen reduzieren
  {
  neu_einzelkante:
    i = del_anf_il( &einzelkante );
    if( i != 0 ) // es gibt eine Einzelkante
    {
      j = del_anf_il( &einzelkante ); // Zielknoten
      // existieren die Knoten noch?
      if( ( test_knoten( h, i ) == 0 ) && ( test_knoten( h, j ) == 0 ) )
      {
        m++;
        // Eingangskanten kopieren von i zu Eingangskanten von j machen
        reset_vorg( h, i );
        while( 1 )
        {
          k = next_vorg( h, i );
          if( k == 0 )
          {
            break;
          }
          if( add_kante_mehr( h, k, j, mehrg ) != 0 )
          {
            return -1;
          }
        }
        del_knoten( h, i );
        // j eine neue Nummer geben
        if( move_knoten_mehr( h, j, m, mehrg ) != 0 )
        {
          return -1;
        }
        // Änderung im Konstruktionsbaum speichern
        if( add_kante_mehr( b, m, j, mehrb ) != 0 )
        {
          return -1;
        }
        if( add_kante_mehr( b, m, i, mehrb ) != 0 )
        {
          return -1;
        }
        (*beschrift)[ m - *plus ] = 2; // sequentielle Vereinigung
        // neuen Knoten in entsprechende Liste schreiben
        if( knoten_eintragen( h, m, &inoutnull, &einzelkante,
                              &outnullineins, &innullouteins,
                              &inouteins, &outeins, &ineins,
                              &inoutgross, aktinliste ) != 0 )
        {
          return -1;
        }
        // führt zu m auch eine Einzelkante
        if( vorg_einzelkante( h, m, &einzelkante, aktinliste ) != 0 )
        {
          return -1;
        }
        continue;
      }
      goto neu_einzelkante;
    }
  neu_outnullineins:
    i = del_anf_il( &outnullineins );
    if( i != 0 )
    {
      if( test_knoten( h, i ) == 0 )
      {
        // da keine Einzelkante existiert, hat der Vorgänger j von i
        // mind. Ausgangsgrad 2
        // k ist weiteres Kind von j; vereinige i und k parallel
        reset_vorg( h, i );
        j = next_vorg( h, i );
        reset_kind( h, j );
        k = next_kind( h, j );
        if( k == i )
        {
          k = next_kind( h, j );
        }
        m++;
        if( move_knoten_mehr( h, k, m, mehrg ) != 0 )
        {
          return -1;
        }
        del_knoten( h, i );
        // Änderung im Konstruktionsbaum speichern
        if( add_kante_mehr( b, m, k, mehrb ) != 0 )
        {
          return -1;
        }
        if( add_kante_mehr( b, m, i, mehrb ) != 0 )
        {
          return -1;
        }
        (*beschrift)[ m - *plus ] = 1; // parallele Vereinigung
        // neuen Knoten in entsprechende Liste schreiben
        if( knoten_eintragen( h, m, &inoutnull, &einzelkante,
                              &outnullineins, &innullouteins,
                              &inouteins, &outeins, &ineins,
                              &inoutgross, aktinliste ) != 0 )
        {
          return -1;
        }
        // Knoten j umsortieren
        if( knoten_umsortieren( h, j, &einzelkante, &outnullineins,
                                &innullouteins, &inouteins, &outeins,
                                &ineins, aktinliste ) != 0 )
        {
          return -1;
        }
        continue;
      }
      goto neu_outnullineins;
    }
  neu_innullouteins:
    i = del_anf_il( &innullouteins );
    if( i != 0 )
    {
      if( test_knoten( h, i ) == 0 )
      {
        // da keine Einzelkante existiert, hat der Nachfolger j von i
        // mind. Eingangsgrad 2
        // k ist weiterer Vorgänger von j; vereinige i und k parallel
        reset_kind( h, i );
        j = next_kind( h, i );
        reset_vorg( h, j );
        k = next_vorg( h, j );
        if( k == i )
        {
          k = next_vorg( h, j );
        }
        m++;
        if( move_knoten_mehr( h, k, m, mehrg ) != 0 )
        {
          return -1;
        }
        del_knoten( h, i );
        // Änderung im Konstruktionsbaum speichern
        if( add_kante_mehr( b, m, k, mehrb ) != 0 )
        {
          return -1;
        }
        if( add_kante_mehr( b, m, i, mehrb ) != 0 )
        {
          return -1;
        }
        (*beschrift)[ m - *plus ] = 1; // parallele Vereinigung
        // neuen Knoten in entsprechende Liste schreiben
        if( knoten_eintragen( h, m, &inoutnull, &einzelkante,
                              &outnullineins, &innullouteins,
                              &inouteins, &outeins, &ineins,
                              &inoutgross, aktinliste ) != 0 )
        {
          return -1;
        }
        // Knoten j umsortieren
        if( knoten_umsortieren( h, j, &einzelkante, &outnullineins,
                                &innullouteins, &inouteins, &outeins,
                                &ineins, aktinliste ) != 0 )
        {
          return -1;
        }
        continue;
      }
      goto neu_innullouteins;
    }
  neu_inouteins:
    i = del_anf_il( &inouteins );
    if( i != 0 )
    {
      if( test_knoten( h, i ) == 0 )
      { // Knoten wurde noch nicht anders bearbeitet
        reset_vorg( h, i );
        j = next_vorg( h, i ); // j ist Vorgänger von i
        reset_kind( h, i );
        k = next_kind( h, i ); // k ist Kind von i
        del_knoten( h, i );
        p = get_mitte_pfad_mehr( h, j, k, mehrg );
        if( p != 0 ) // es gibt einen Pfad von j nach k über p
        {
          // p = j ist nicht möglich, da es sonst eine Vorwärtskante
          // wäre -> einfach i dem Knoten p parallel zuordnen
          del_knoten( h, i );
          m++;
          // der neue Knoten hat die gleichen Kanten wie p
          if( move_knoten_mehr( h, p, m, mehrg ) != 0 )
          {
            return -1;
          }
          // Änderung im Konstruktionsbaum speichern
          if( add_kante_mehr( b, m, p, mehrb ) != 0 )
          {
            return -1;
          }
          if( add_kante_mehr( b, m, i, mehrb ) != 0 )
          {
            return -1;
          }
          (*beschrift)[ m - *plus ] = 1; // parallele Vereinigung
          // neuen Knoten in entsprechende Liste schreiben
          if( knoten_eintragen( h, m, &inoutnull, &einzelkante,
                                &outnullineins, &innullouteins,
                                &inouteins, &outeins, &ineins,
                                &inoutgross, aktinliste ) != 0 )
          {
            return -1;
          }
          // ggf. Knoten j in eine andere Liste schreiben
          if( knoten_umsortieren( h, j, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          // ggf. Knoten k in eine andere Liste schreiben
          if( knoten_umsortieren( h, k, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          // führt zu k auch eine Einzelkante
          if( vorg_einzelkante( h, k, &einzelkante, aktinliste ) != 0 )
          {
            return -1;
          }
          continue;
        }
        // es gibt keinen parallelen Pfad -> mit Nachbar vereinigen
        m++;
        reset_kind( h, j ); // Nachbar von i = Kind von j suchen
        p = next_kind( h, j );
        if( p == i )
        {
          p = next_kind( h, j );
        }
        // p und i parallel vereinigen
        if( move_knoten_mehr( h, p, m, mehrg ) != 0 )
        {
          return -1;
        }
        if( add_kante_mehr( h, m, k, mehrg ) != 0 )
        {
          return -1;
        }
        // neue transitive Kanten löschen; 1) sie nutzen den Pfad über m
        if( transitiv_kante_ueber_mehr( h, m, &trans, mehrg ) != 0 )
        {
          return -1;
        }
        while ( 1 )
        {
          v = del_anf_kl( &trans );
          if( v == NULL )
          {
            break;
          }
          k = start_knoten( v );
          q = end_knoten( v );
          del_kante_verweis( h, v );
          // die beiden Knoten umsortieren
          if( knoten_umsortieren( h, k, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          if( knoten_umsortieren( h, q, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          // führt zum Zielknoten der Kante eine Einzelkante
          if( vorg_einzelkante( h, q, &einzelkante, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        // neue transitive Kanten löschen; 2) m ist Startpunkt der Kante
        if( transitiv_kante_aus_mehr( h, m, &trans, mehrg ) != 0 )
        {
          return -1;
        }
        while ( 1 )
        {
          v = del_anf_kl( &trans );
          if( v == NULL )
          {
            break;
          }
          k = end_knoten( v );
          del_kante_verweis( h, v );
          // den Endknoten k umsortieren
          if( knoten_umsortieren( h, k, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          // führt zu k auch eine Einzelkante
          if( vorg_einzelkante( h, k, &einzelkante, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        // Änderung im Konstruktionsbaum speichern
        if( add_kante_mehr( b, m, i, mehrb ) != 0 )
        {
          return -1;
        }
        if( add_kante_mehr( b, m, p, mehrb ) != 0 )
        {
          return -1;
        }
        (*beschrift)[ m - *plus ] = 1; // parallele Vereinigung
        // neuen Knoten in entsprechende Liste schreiben
        if( knoten_eintragen( h, m, &inoutnull, &einzelkante,
                              &outnullineins, &innullouteins,
                              &inouteins, &outeins, &ineins,
                              &inoutgross, aktinliste ) != 0 )
        {
          return -1;
        }
        // ggf. Knoten j in eine andere Liste schreiben
        if( knoten_umsortieren( h, j, &einzelkante, &outnullineins,
                                &innullouteins, &inouteins, &outeins,
                                &ineins, aktinliste ) != 0 )
        {
          return -1;
        }
        continue;
      }
      goto neu_inouteins;
    }
  neu_outeins:
    i = del_anf_il( &outeins );
    if( i != 0 )
    {
      if( test_knoten( h, i ) == 0 )
      {
        // das Kind von i sei j; j hat mind. Eingangsgrad 2;
        // k ist Nachbar von i
        // der neue Knoten ergibt sich aus i parallel k
        reset_kind( h, i );
        j = next_kind( h, i );
        reset_vorg( h, j );
        k = next_vorg( h, j );
        if( k == i )
        {
          k = next_vorg( h, j );
        }
        m++;
        // Eingangskanten von i zu Eingangskanten von k machen
        if( add_kanten_test_nach_init_mehr( h, k, mehrg ) != 0 )
        {
          return -1;
        }
        reset_vorg( h, i );
        while( 1 )
        {
          p = next_vorg( h, i );
          if( p == 0 )
          {
            break;
          }
          if( add_kanten_test_nach_mehr( h, p, k, mehrg ) != 0 )
          {
            return -1;
          }
        }
        if( move_knoten_mehr( h, k, m, mehrg ) != 0 )
        {
          return -1;
        }
        del_knoten( h, i );
        // neue transitive Kanten löschen; 1) sie nutzen den Pfad über m
        if( transitiv_kante_ueber_mehr( h, m, &trans, mehrg ) != 0 )
        {
          return -1;
        }
        while ( 1 )
        {
          v = del_anf_kl( &trans );
          if( v == NULL )
          {
            break;
          }
          p = start_knoten( v );
          q = end_knoten( v );
          del_kante_verweis( h, v );
          // die beiden Knoten umsortieren
          if( knoten_umsortieren( h, p, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          if( knoten_umsortieren( h, q, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          // führt zum Zielknoten der Kante eine Einzelkante
          if( vorg_einzelkante( h, q, &einzelkante, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        // neue transitive Kanten löschen; 2) m ist Endpunkt der Kante
        if( transitiv_kante_ein_mehr( h, m, &trans, mehrg ) != 0 )
        {
          return -1;
        }
        while ( 1 )
        {
          v = del_anf_kl( &trans );
          if( v == NULL )
          {
            break;
          }
          p = start_knoten( v );
          del_kante_verweis( h, v );
          // den Startknoten umsortieren
          if( knoten_umsortieren( h, p, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        // Änderung im Konstruktionsbaum speichern
        if( add_kante_mehr( b, m, i, mehrb ) != 0 )
        {
          return -1;
        }
        if( add_kante_mehr( b, m, k, mehrb ) != 0 )
        {
          return -1;
        }
        (*beschrift)[ m - *plus ] = 1; // parallele Vereinigung
        // neuen Knoten in entsprechende Liste schreiben
        if( knoten_eintragen( h, m, &inoutnull, &einzelkante,
                              &outnullineins, &innullouteins,
                              &inouteins, &outeins, &ineins,
                              &inoutgross, aktinliste ) != 0 )
        {
          return -1;
        }
        // Knoten j umsortieren
        if( knoten_umsortieren( h, j, &einzelkante, &outnullineins,
                                &innullouteins, &inouteins, &outeins,
                                &ineins, aktinliste ) != 0 )
        {
          return -1;
        }
        // alle Vorgänger des neuen Knoten umsortieren
        reset_vorg( h, m );
        while( 1 )
        {
          k = next_vorg( h, m );
          if( k == 0 )
          {
            break;
          }
          if( knoten_umsortieren( h, k, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        continue;
      }
      goto neu_outeins;
    }
  neu_ineins:
    i = del_anf_il( &ineins );
    if( i != 0 )
    {
      if( test_knoten( h, i ) == 0 )
      {
        // der Vorgänger von i sei j; j hat mind. Ausgangsgrad 2;
        // k ist Nachbar von i
        // der neue Knoten ergibt sich aus i parallel k
        reset_vorg( h, i );
        j = next_vorg( h, i );
        reset_kind( h, j );
        k = next_kind( h, j );
        if( k == i )
        {
          k = next_kind( h, j );
        }
        m++;
        // Ausgangskanten von i zu Ausgangskanten von k machen
        if( add_kanten_test_von_init_mehr( h, k, mehrg ) != 0 )
        {
          return -1;
        }
        reset_kind( h, i );
        while( 1 )
        {
          p = next_kind( h, i );
          if( p == 0 )
          {
            break;
          }
          if( add_kanten_test_von_mehr( h, k, p, mehrg ) != 0 )
          {
            return -1;
          }
        }
        if( move_knoten_mehr( h, k, m, mehrg ) != 0 )
        {
          return -1;
        }
        del_knoten( h, i );
        // neue transitive Kanten löschen; 1) sie nutzen den Pfad über m
        if( transitiv_kante_ueber_mehr( h, m, &trans, mehrg ) != 0 )
        {
          return -1;
        }
        while ( 1 )
        {
          v = del_anf_kl( &trans );
          if( v == NULL )
          {
            break;
          }
          p = start_knoten( v );
          q = end_knoten( v );
          del_kante_verweis( h, v );
          // die beiden Knoten umsortieren
          if( knoten_umsortieren( h, p, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          if( knoten_umsortieren( h, q, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          // führt zum Zielknoten der Kante eine Einzelkante
          if( vorg_einzelkante( h, q, &einzelkante, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        // neue transitive Kanten löschen; 2) m ist Startpunkt der Kante
        if( transitiv_kante_aus_mehr( h, m, &trans, mehrg ) != 0 )
        {
          return -1;
        }
        while ( 1 )
        {
          v = del_anf_kl( &trans );
          if( v == NULL )
          {
            break;
          }
          p = end_knoten( v );
          del_kante_verweis( h, v );
          // den Endknoten p umsortieren
          if( knoten_umsortieren( h, p, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          // führt zu p auch eine Einzelkante
          if( vorg_einzelkante( h, p, &einzelkante, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        // Änderung im Konstruktionsbaum speichern
        if( add_kante_mehr( b, m, i, mehrb ) != 0 )
        {
          return -1;
        }
        if( add_kante_mehr( b, m, k, mehrb ) != 0 )
        {
          return -1;
        }
        (*beschrift)[ m - *plus ] = 1; // parallele Vereinigung
        // neuen Knoten in entsprechende Liste schreiben
        if( knoten_eintragen( h, m, &inoutnull, &einzelkante,
                              &outnullineins, &innullouteins,
                              &inouteins, &outeins, &ineins,
                              &inoutgross, aktinliste ) != 0 )
        {
          return -1;
        }
        // Knoten j umsortieren
        if( knoten_umsortieren( h, j, &einzelkante, &outnullineins,
                                &innullouteins, &inouteins, &outeins,
                                &ineins, aktinliste ) != 0 )
        {
          return -1;
        }
        // alle Nachfolger des neuen Knoten umsortieren
        reset_kind( h, m );
        while( 1 )
        {
          k = next_kind( h, m );
          if( k == 0 )
          {
            break;
          }
          if( knoten_umsortieren( h, k, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        continue;
      }
      goto neu_ineins;
    }
  neu_inoutgross:
    i = del_anf_il( &inoutgross );
    if( i != 0 )
    {
      if( test_knoten( h, i ) == 0 )
      {
        // vereinige zwei Kinder von i
        reset_kind( h, i );
        j = next_kind( h, i );
        k = next_kind( h, i );
        m++;
        reset_vorg( h, k ); // Eingangskanten kopieren
        if( add_kanten_test_nach_init_mehr( h, j, mehrg ) != 0 )
        {
          return -1;
        }
        while( 1 )
        {
          p = next_vorg( h, k );
          if( p == 0 )
          {
            break;
          }
          if( add_kanten_test_nach_mehr( h, p, j, mehrg ) != 0 )
          {
            return -1;
          }
        }
        reset_kind( h, k ); // Ausgangskanten kopieren
        if( add_kanten_test_von_init_mehr( h, j, mehrg ) != 0 )
        {
          return -1;
        }
        while( 1 )
        {
          p = next_kind( h, k );
          if( p == 0 )
          {
            break;
          }
          if( add_kanten_test_von_mehr( h, j, p, mehrg ) != 0 )
          {
            return -1;
          }
        }
        del_knoten( h, k );
        if( move_knoten_mehr( h, j, m, mehrg ) != 0 )
        {
          return -1;
        }
        // neue transitive Kanten löschen; 1) sie nutzen den Pfad über m
        if( transitiv_kante_ueber_mehr( h, m, &trans, mehrg ) != 0 )
        {
          return -1;
        }
        while ( 1 )
        {
          v = del_anf_kl( &trans );
          if( v == NULL )
          {
            break;
          }
          p = start_knoten( v );
          q = end_knoten( v );
          del_kante_verweis( h, v );
          // die beiden Knoten umsortieren
          if( knoten_umsortieren( h, p, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          if( knoten_umsortieren( h, q, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          // führt zum Zielknoten der Kante eine Einzelkante
          if( vorg_einzelkante( h, q, &einzelkante, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        // neue transitive Kanten löschen; 2) m ist Startpunkt der Kante
        if( transitiv_kante_aus_mehr( h, m, &trans, mehrg ) != 0 )
        {
          return -1;
        }
        while ( 1 )
        {
          v = del_anf_kl( &trans );
          if( v == NULL )
          {
            break;
          }
          p = end_knoten( v );
          del_kante_verweis( h, v );
          // den Endknoten p umsortieren
          if( knoten_umsortieren( h, p, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
          // führt zu p auch eine Einzelkante
          if( vorg_einzelkante( h, p, &einzelkante, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        // neue transitive Kanten löschen; 3) m ist Endpunkt der Kante
        if( transitiv_kante_ein_mehr( h, m, &trans, mehrg ) != 0 )
        {
          return -1;
        }
        while ( 1 )
        {
          v = del_anf_kl( &trans );
          if( v == NULL )
          {
            break;
          }
          p = start_knoten( v );
          del_kante_verweis( h, v );
          // den Startknoten umsortieren
          if( knoten_umsortieren( h, p, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        // Änderung im Konstruktionsbaum speichern
        if( add_kante_mehr( b, m, j, mehrb ) != 0 )
        {
          return -1;
        }
        if( add_kante_mehr( b, m, k, mehrb ) != 0 )
        {
          return -1;
        }
        (*beschrift)[ m - *plus ] = 1; // parallele Vereinigung
        // neuen Knoten in entsprechende Liste schreiben
        if( knoten_eintragen( h, m, &inoutnull, &einzelkante,
                              &outnullineins, &innullouteins,
                              &inouteins, &outeins, &ineins,
                              &inoutgross, aktinliste ) != 0 )
        {
          return -1;
        }
        // alle Vorgänger und Nachfolger des neuen Knoten umsortieren
        reset_vorg( h, m );
        while( 1 )
        {
          k = next_vorg( h, m );
          if( k == 0 )
          {
            break;
          }
          if( knoten_umsortieren( h, k, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        reset_kind( h, m );
        while( 1 )
        {
          k = next_kind( h, m );
          if( k == 0 )
          {
            break;
          }
          if( knoten_umsortieren( h, k, &einzelkante, &outnullineins,
                                  &innullouteins, &inouteins, &outeins,
                                  &ineins, aktinliste ) != 0 )
          {
            return -1;
          }
        }
        continue;
      }
      goto neu_inoutgross;
    }
    // es sind nur noch Knoten in inoutnull
    break;
  }
  // vereinige restliche Knoten parallel (baumartig)
  i = anz_il( &inoutnull );
  j = verein_baum( b, &m, &inoutnull, i, *beschrift, *plus, mehrb );
  if( j < 0 )
  {
    return -1;
  }
  free( aktinliste );
  free_il( &inoutnull );
  free_il( &einzelkante );
  free_il( &outnullineins );
  free_il( &innullouteins );
  free_il( &inouteins );
  free_il( &outeins );
  free_il( &ineins );
  free_il( &inoutgross );
  free_kl( &trans );
#ifdef GRAPH_ERHALTEN
  free_graph( &h_h );
#endif
  return j;
}


// ------------------ Dateiorientierte Routinen ------------------------
#ifndef FORMEL

// berechnet die Zeiten za vom fiktiven Modul w;
// b ist der Konstruktionsbaum mit der Beschriftung beschrift;
// in typ wird die durchgeführte Vereinigung gespeichert
// Fehler: 1->zu wenig Speicher
int verein_modul_array(
  graph*  b,
  int     w,
  int*    beschrift,
  int     o,
  int*    typ,
  int     n,
  double* za,
  double* t )
{
  int i, j;
  double* zb;
  int x, y;
  if( w < o ) // Blatt erreicht; Modullaufzeiten zurückgeben
  {
    for( i=0; i<n; i++ )
    {
      za[i] = t[ (w-1) * n + i ];
    }
    return 0;
  }
  reset_kind( b, w );
  x = next_kind( b, w );
  y = next_kind( b, w );
  zb = (double*) malloc( sizeof( double ) * n );
  if( zb == NULL )
  {
    return 1;
  }
  if( verein_modul_array( b, x, beschrift, o, typ, n, za, t ) != 0 )
  {
    return 1;
  }
  if( verein_modul_array( b, y, beschrift, o, typ, n, zb, t ) != 0 )
  {
    return 1;
  }
  if( beschrift[ w-o ] == 2 ) // sequentielle Vereinigung
  {
    for( i=0; i<n; i++ )
    {
      // nur wenn beide Module mit der entsprechenden Prozessorenanzahl
      // ausführbar sind, ist auch die Vereinigung startbar
      if( za[i] != 0 )
      {
        if( zb[i] != 0 )
        {
          za[i] += zb[i];
        }
        else
        {
          za[i] = 0;
        }
      }
    }
  }
  else // parallele Vereinigung
  {
    // Dimension n wegen Bestimmung vom Wechsel extra betrachten
    za[n-1] += zb[n-1]; // mit n Prozessoren sind alle Module ausführbar
    typ[ (w-o) * n + n - 1 ] = 0;
    x = 0; // Stelle mit Wechsel der Größen von za und zb suchen
    y = n;
    while( x + 1 < y )
    {
      j = ( x + y ) / 2;
      if( za[j-1] == 0 )
      {
        x = j;
        continue;
      }
      if( zb[n-j-1] == 0 )
      {
        y = j;
        continue;
      }
      if( za[j-1] <= zb[n-j-1] ) // hier ist j immer >=1 und <=n-1
      {
        y = j;
      }
      else
      {
        x = j;
      }
    }
    // interpretiert man Nullen als unendlich, so weiß man:
    // wenn 1 <= x <= n-2 gilt za[x] <= zb[n-x-2] und
    //                         za[x-1] > zb[n-x-1]
    // wenn x == 0 gilt für alle i: za[i] <= za[0] <= zb[n-2]
    // wenn x == n-1 gilt für alle i: za[n-2] > zb[0] >= zb[i]
    // bzgl. Nullen weiß man
    // wenn 1 <= x <= n-2 ist für alle i>=x: za[i] != 0;
    //   ist zusätzlich za[x-1] != 0 so gilt zb[n-x-1] != 0
    //   (sonst wäre y=j gewählt wurden)
    if( x == 0 ) // za[x-1] existiert nicht
    {
      if( ( za[n-1] >= zb[n-2] ) && ( za[0] != 0 ) && ( zb[n-2] != 0 ) )
      { // parallele Ausführung besser ?
        za[n-1] = zb[n-2];
        typ[ (w-o) * n + n - 1 ] = 1;
      }
    }
    else // x >= 1
    {
      if( x == n-1 )
      {
        if( ( za[n-1] >= za[n-2] ) &&
            ( za[n-2] != 0 ) &&
            ( za[0] != 0 ) )
        {
          za[n-1] = za[n-2];
          typ[ (w-o) * n + n - 1 ] = n-1;
        }
        x--; // nächster Wert
      }
      else // 1 <= x <= n-2
      {
        // Minimum aus beiden Möglichkeiten suchen
        if( ( za[x-1] <= zb[n-x-2] ) || ( zb[n-x-2] == 0 ) )
        {
          if( ( za[n-1] >= za[x-1] ) && ( za[x-1] != 0 ) )
          {
            za[n-1] = za[x-1];
            typ[ (w-o) * n + n - 1 ] = x;
          }
          x--; // nächster Wert
        }
        else
        {
          if( za[n-1] >= zb[n-x-2] )
          {
            za[n-1] = zb[n-x-2];
            typ[ (w-o) * n + n - 1 ] = x + 1;
          }
        }
      }
    }
    for( i=n-2; i>=1; i-- )
    {
      if( za[i] == 0 ) // zu wenig Prozessoren
      {
        break;
      }
      if( zb[i] == 0 )
      {
        za[i] = 0;
        continue;
      }
      za[i] += zb[i];
      typ[ (w-o) * n + i ] = 0;
      if( x == 0 ) // za[x-1] existiert nicht
      {
        if( ( za[i] >= zb[i-1] ) && ( za[0] != 0 ) && ( zb[i-1] != 0 ) )
        { // parallele Ausführung besser ?
          za[i] = zb[i-1];
          typ[ (w-o) * n + i ] = 1;
        }
      }
      else // x >= 1
      {
        if( x == i )
        {
          if( ( za[i] >= za[i-1] ) &&
              ( za[i-1] != 0 ) &&
              ( zb[0] != 0 ) )
          {
            za[i] = za[i-1];
            typ[ (w-o) * n + i ] = i;
          }
          x--; // nächster Wert
        }
        else // 1 <= x <= i-1
        {
          // Minimum aus beiden Möglichkeiten suchen
          if( ( za[x-1] <= zb[i-x-1] ) || ( zb[i-x-1] == 0 ) )
          {
            if( ( za[i] >= za[x-1] ) && ( za[x-1] != 0 ) )
            {
              za[i] = za[x-1];
              typ[ (w-o) * n + i ] = x;
            }
            x--; // nächster Wert
          }
          else
          {
            if( za[i] >= zb[i-x-1] )
            {
              za[i] = zb[i-x-1];
              typ[ (w-o) * n + i ] = x + 1;
            }
          }
        }
      }
    }
    if( za[0] != 0 ) // sind beide sequentiell ausführbar
    {
      if( zb[0] == 0 )
      {
        za[0] = 0;
      }
      else
      {
        // bei einem Prozessor gibt es nur eine Möglichkeit
        za[0] += zb[0];
        typ[ (w-o) * n ] = 0;
      }
    }
  }
  free( zb );
  return 0;
}

// speichert das Ergebnis indem der rekursive Abstieg wie in
// verein_modul berechnet durchgeführt wird
// Rückgabe: Endzeit des fiktiven Knotens w bei Startzeit st und
//           Breite re-li und linken Prozessor li
//           -1->Schreibfehler
double setze_zeiten_array(
  graph*  b,
  int     w,
  int*    beschrift,
  int     o,
  int*    typ,
  int     n,
  double* t,
  FILE**  fi,
  int     li,
  int     re,
  double  st )
{
  int x, y;
  double sx;
  if( w < o ) // Blatt erreicht
  {
    if( save_text( fi, "Modul " ) != 0 )
    {
      return -1;
    }
    if( save_int( fi, w ) != 0 )
    {
      return -1;
    }
    if( save_text( fi, ": von " ) != 0 )
    {
      return -1;
    }
    if( save_double( fi, st ) != 0 )
    {
      return -1;
    }
    st += t[ (w-1) * n + re - li - 1 ];
    if( save_text( fi, " bis " ) != 0 )
    {
      return -1;
    }
    if( save_double( fi, st ) != 0 )
    {
      return -1;
    }
    if( save_text( fi, " auf Prozessoren " ) != 0 )
    {
      return -1;
    }
    if( save_int( fi, li+1 ) != 0 )
    {
      return -1;
    }
    if( save_text( fi, " bis " ) != 0 )
    {
      return -1;
    }
    if( save_int_r( fi, re ) != 0 )
    {
      return -1;
    }
    return st;
  }
  reset_kind( b, w );
  x = next_kind( b, w );
  y = next_kind( b, w );
  if( beschrift[ w-o ] == 2 ) // sequentielle Vereinigung
  {
    st = setze_zeiten_array( b, x, beschrift, o, typ, n, t, fi,
                             li, re, st );
    if( st < 0 )
    {
      return -1;
    }
    return setze_zeiten_array( b, y, beschrift, o, typ, n, t, fi,
                               li, re, st );
  }
  // parallele Vereinigung
  if( typ[ (w-o) * n + re - li - 1 ] == 0 )
  { // es wurde die sequentielle Vereinigung angewendet
    st = setze_zeiten_array( b, x, beschrift, o, typ, n, t, fi,
                             li, re, st );
    if( st < 0 )
    {
      return -1;
    }
    return setze_zeiten_array( b, y, beschrift, o, typ, n, t, fi,
                               li, re, st );
  }
  sx = setze_zeiten_array( b, x, beschrift, o, typ, n, t, fi,
                           li, li+typ[(w-o)*n+re-li-1], st );
  if( sx < 0 )
  {
    return -1;
  }
  st = setze_zeiten_array( b, y, beschrift, o, typ, n, t, fi,
                           li+typ[(w-o)*n+re-li-1], re, st );
  if( st < 0 )
  {
    return -1;
  }
  if( st < sx )
  {
    return sx;
  }
  return st;
}

// Optimierungsroutine siehe oben
int optimiere_datei(
  char* ergebnis,
  char* daten,
  char* graph_datei )
{
  graph g, b;
  int* beschrift;
  int w, m, o, i, j, n;
  int* typ;
  int pos;
  double d;
  double* za;
  double* t;
  FILE* fi;
  init_graph( &g );
  init_graph( &b );
  // Graph laden
  if( load_graph_rueck_mehr( graph_datei, &g, 2 ) != 0 )
  {
    return 3;
  }
  // Maximale Knotenanzahl ermitteln
  m = max_knoten( &g );
  // Konstruktionsbaum berechnen
  w = baue_sp_baum_pseudo( &g, &b, &beschrift, &o, 0, 0 );
  if( w < 0 )
  {
    return 1;
  }
  free_graph( &g );
  // Berechnen der Lösung
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 2;
  }
  // Bestimmen der Prozessorenanzahl
  pos = ftell( fi );
  n = 1;
  if( lese_next_double( &fi, &d ) != 0 )
  {
    return 2;
  }
  while( lese_next_double_in_line( &fi, &d ) == 0 )
  {
    n++;
  }
  fseek( fi, pos, SEEK_SET );
  // Hilfsvariablen
  typ = (int*) malloc( sizeof( int ) * m * n );
  if( typ == NULL )
  {
    return 1;
  }
  za = (double*) malloc( sizeof( double ) * n );
  if( za == NULL )
  {
    return 1;
  }
  t = (double*) malloc( sizeof( double ) * m * n );
  if( t == NULL )
  {
    return 1;
  }
  // Laufzeiten laden
  for( i=0; i<m; i++ )
  {
    for( j=0; j<n; j++ )
    {
      if( lese_next_double_in_line( &fi, &(t[i*n+j]) ) != 0 )
      {
        return 2;
      }
    }
    lese_next_ret( &fi );
  }
  fclose( fi );
  if( verein_modul_array( &b, w, beschrift, o, typ, n, za, t ) != 0 )
  {
    return 1;
  }
  free( za );
  fi = fopen( ergebnis, "w" );
  if( fi == NULL )
  {
    return 4;
  }
  if( setze_zeiten_array( &b, w, beschrift, o, typ, n, t, &fi,
                          0, n, 0.0 ) < 0 )
  {
    return 4;
  }
  fclose( fi );
  free( typ );
  free( t );
  free_graph( &b );
  free( beschrift );
  return 0;
}

#endif


// ---------------------- Formelorientierte Routinen -------------------
#ifdef FORMEL

// berechnet die Zeiten za vom fiktiven Modul w;
// b ist der Konstruktionsbaum mit der Beschriftung beschrift;
// in typ wird die durchgeführte Vereinigung gespeichert
// Fehler: 1->zu wenig Speicher
int verein_modul_formel(
  graph*  b,
  int     w,
  int*    beschrift,
  int     o,
  int*    typ,
  int     n,
  double* za,
  double  ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  int i, j;
  double* zb;
  int x, y;
  if( w < o ) // Blatt erreicht; Modullaufzeiten zurückgeben
  {
    for( i=0; i<n; i++ )
    {
      za[i] = laufzeit( w, i+1 );
    }
    return 0;
  }
  reset_kind( b, w );
  x = next_kind( b, w );
  y = next_kind( b, w );
  zb = (double*) malloc( sizeof( double ) * n );
  if( zb == NULL )
  {
    return 1;
  }
  if( verein_modul_formel( b, x, beschrift, o, typ, n, za,
                           laufzeit ) != 0 )
  {
    return 1;
  }
  if( verein_modul_formel( b, y, beschrift, o, typ, n, zb,
                           laufzeit ) != 0 )
  {
    return 1;
  }
  if( beschrift[ w-o ] == 2 ) // sequentielle Vereinigung
  {
    for( i=0; i<n; i++ )
    {
      // nur wenn beide Module mit der entsprechenden Prozessorenanzahl
      // ausführbar sind, ist auch die Vereinigung startbar
      if( za[i] != 0 )
      {
        if( zb[i] != 0 )
        {
          za[i] += zb[i];
        }
        else
        {
          za[i] = 0;
        }
      }
    }
  }
  else // parallele Vereinigung
  {
    // Dimension n wegen Bestimmung vom Wechsel extra betrachten
    za[n-1] += zb[n-1]; // mit n Prozessoren sind alle Module ausführbar
    typ[ (w-o) * n + n - 1 ] = 0;
    x = 0; // Stelle mit Wechsel der Größen von za und zb suchen
    y = n;
    while( x + 1 < y )
    {
      j = ( x + y ) / 2;
      if( za[j-1] == 0 )
      {
        x = j;
        continue;
      }
      if( zb[n-j-1] == 0 )
      {
        y = j;
        continue;
      }
      if( za[j-1] <= zb[n-j-1] ) // hier ist j immer >=1 und <=n-1
      {
        y = j;
      }
      else
      {
        x = j;
      }
    }
    // interpretiert man Nullen als unendlich, so weiß man:
    // wenn 1 <= x <= n-2 gilt za[x] <= zb[n-x-2] und
    //                         za[x-1] > zb[n-x-1]
    // wenn x == 0 gilt für alle i: za[i] <= za[0] <= zb[n-2]
    // wenn x == n-1 gilt für alle i: za[n-2] > zb[0] >= zb[i]
    // bzgl. Nullen weiß man
    // wenn 1 <= x <= n-2 ist für alle i>=x: za[i] != 0;
    //   ist zusätzlich za[x-1] != 0 so gilt zb[n-x-1] != 0
    //   (sonst wäre y=j gewählt wurden)
    if( x == 0 ) // za[x-1] existiert nicht
    {
      if( ( za[n-1] >= zb[n-2] ) && ( za[0] != 0 ) && ( zb[n-2] != 0 ) )
      { // parallele Ausführung besser ?
        za[n-1] = zb[n-2];
        typ[ (w-o) * n + n - 1 ] = 1;
      }
    }
    else // x >= 1
    {
      if( x == n-1 )
      {
        if( ( za[n-1] >= za[n-2] ) &&
            ( za[n-2] != 0 ) &&
            ( za[0] != 0 ) )
        {
          za[n-1] = za[n-2];
          typ[ (w-o) * n + n - 1 ] = n-1;
        }
        x--; // nächster Wert
      }
      else // 1 <= x <= n-2
      {
        // Minimum aus beiden Möglichkeiten suchen
        if( ( za[x-1] <= zb[n-x-2] ) || ( zb[n-x-2] == 0 ) )
        {
          if( ( za[n-1] >= za[x-1] ) && ( za[x-1] != 0 ) )
          {
            za[n-1] = za[x-1];
            typ[ (w-o) * n + n - 1 ] = x;
          }
          x--; // nächster Wert
        }
        else
        {
          if( za[n-1] >= zb[n-x-2] )
          {
            za[n-1] = zb[n-x-2];
            typ[ (w-o) * n + n - 1 ] = x + 1;
          }
        }
      }
    }
    for( i=n-2; i>=1; i-- )
    {
      if( za[i] == 0 ) // zu wenig Prozessoren
      {
        break;
      }
      if( zb[i] == 0 )
      {
        za[i] = 0;
        continue;
      }
      za[i] += zb[i];
      typ[ (w-o) * n + i ] = 0;
      if( x == 0 ) // za[x-1] existiert nicht
      {
        if( ( za[i] >= zb[i-1] ) && ( za[0] != 0 ) && ( zb[i-1] != 0 ) )
        { // parallele Ausführung besser ?
          za[i] = zb[i-1];
          typ[ (w-o) * n + i ] = 1;
        }
      }
      else // x >= 1
      {
        if( x == i )
        {
          if( ( za[i] >= za[i-1] ) &&
              ( za[i-1] != 0 ) &&
              ( zb[0] != 0 ) )
          {
            za[i] = za[i-1];
            typ[ (w-o) * n + i ] = i;
          }
          x--; // nächster Wert
        }
        else // 1 <= x <= i-1
        {
          // Minimum aus beiden Möglichkeiten suchen
          if( ( za[x-1] <= zb[i-x-1] ) || ( zb[i-x-1] == 0 ) )
          {
            if( ( za[i] >= za[x-1] ) && ( za[x-1] != 0 ) )
            {
              za[i] = za[x-1];
              typ[ (w-o) * n + i ] = x;
            }
            x--; // nächster Wert
          }
          else
          {
            if( za[i] >= zb[i-x-1] )
            {
              za[i] = zb[i-x-1];
              typ[ (w-o) * n + i ] = x + 1;
            }
          }
        }
      }
    }
    if( za[0] != 0 ) // sind beide sequentiell ausführbar
    {
      if( zb[0] == 0 )
      {
        za[0] = 0;
      }
      else
      {
        // bei einem Prozessor gibt es nur eine Möglichkeit
        za[0] += zb[0];
        typ[ (w-o) * n ] = 0;
      }
    }
  }
  free( zb );
  return 0;
}

// speichert das Ergebnis indem der rekursive Abstieg wie in
// verein_modul berechnet durchgeführt wird
// Rückgabe: >0 Endzeit des fiktiven Knotens w bei Startzeit st und
//              Breite re-li und linken Prozessor li
//           -1 Schreibfehler
double setze_zeiten_formel(
  graph*  b,
  int     w,
  int*    beschrift,
  int     o,
  int*    typ,
  int     n,
  double  ((*laufzeit)( int modulnummer, int prozessorenanzahl )),
  double* startzeit,
  int*    breite,
  int*    linker,
  int     li,
  int     re,
  double  st )
{
  int x, y;
  double sx;
  if( w < o ) // Blatt erreicht
  {
    startzeit[w-1] = st;
    st += laufzeit( w, re - li );
    breite[w-1] = re - li;
    linker[w-1] = li + 1;
    return st;
  }
  reset_kind( b, w );
  x = next_kind( b, w );
  y = next_kind( b, w );
  if( beschrift[ w-o ] == 2 ) // sequentielle Vereinigung
  {
    st = setze_zeiten_formel( b, x, beschrift, o, typ, n, laufzeit,
                              startzeit, breite, linker, li, re, st );
    if( st < 0 )
    {
      return -1;
    }
    return setze_zeiten_formel( b, y, beschrift, o, typ, n, laufzeit,
                                startzeit, breite, linker, li, re, st );
  }
  // parallele Vereinigung
  if( typ[ (w-o) * n + re - li - 1 ] == 0 )
  { // es wurde die sequentielle Vereinigung angewendet
    st = setze_zeiten_formel( b, x, beschrift, o, typ, n, laufzeit,
                              startzeit, breite, linker, li, re, st );
    if( st < 0 )
    {
      return -1;
    }
    return setze_zeiten_formel( b, y, beschrift, o, typ, n, laufzeit,
                                startzeit, breite, linker, li, re, st );
  }
  sx = setze_zeiten_formel( b, x, beschrift, o, typ, n, laufzeit,
                            startzeit, breite, linker, li,
                            li+typ[(w-o)*n+re-li-1], st );
  if( sx < 0 )
  {
    return -1;
  }
  st = setze_zeiten_formel( b, y, beschrift, o, typ, n, laufzeit,
                            startzeit, breite, linker,
                            li+typ[(w-o)*n+re-li-1], re, st );
  if( st < 0 )
  {
    return -1;
  }
  if( st < sx )
  {
    return sx;
  }
  return st;
}

// Optimierungsroutine siehe oben
int optimiere_formel(
  double** startzeit,
  int**    breite,
  int**    linker,
  graph*   g,
  int      n,
  double   ((*laufzeit)( int modulnummer, int prozessorenanzahl )) )
{
  graph b;
  int* beschrift;
  int w, m, o;
  int* typ;
  double* za;
  init_graph( &b );
  // Maximale Knotenanzahl ermitteln
  m = max_knoten( g );
  // Konstruktionsbaum berechnen
  w = baue_sp_baum_pseudo( g, &b, &beschrift, &o, 0, 0 );
  if( w < 0 )
  {
    return 1;
  }
  // Hilfsvariablen
  typ = (int*) malloc( sizeof( int ) * m * n );
  if( typ == NULL )
  {
    return 1;
  }
  za = (double*) malloc( sizeof( double ) * n );
  if( za == NULL )
  {
    return 1;
  }
  // Ergebnisvariablen
  *startzeit = (double*) malloc( sizeof( double ) * m );
  if( *startzeit == NULL )
  {
    return 1;
  }
  *breite = (int*) malloc( sizeof( int ) * m );
  if( *breite == NULL )
  {
    return 1;
  }
  *linker = (int*) malloc( sizeof( int ) * m );
  if( *linker == NULL )
  {
    return 1;
  }
  if( verein_modul_formel( &b, w, beschrift, o, typ,
                           n, za, laufzeit ) != 0 )
  {
    return 1;
  }
  free( za );
  if( setze_zeiten_formel( &b, w, beschrift, o, typ, n, laufzeit,
                           *startzeit, *breite, *linker,
                           0, n, 0.0 ) < 0 )
  {
    return 4;
  }
  free( typ );
  free_graph( &b );
  free( beschrift );
  return 0;
}

#endif
