// Testprogramm für die einzelnen Routinen
// Aufruf test();
// Fehler: 1->zu wenig Speicher zum Erstellen von Beispielen
//         2->Beispielgenerator konnte Ergebnisdatei nicht anlegen
//         3->Beispielgenerator hat nicht genug Einträge im Listenfeld /
//            Fehler in Datei
//         4->Beispielgenerator konnte Eingabedatei nicht finden
//         21->zu wenig Speicher für Optimierung
//         22->Eingabedatei mit Laufzeiten für Optimierung fehlerhaft /
//             nicht gefunden
//         23->Eingabedatei mit Modulabhängigkeitsgraph für Optimierung
//             fehlerhaft / nicht gefunden
//         24->Optimierungsalgorithmus kann Ergebnisdatei nicht anlegen
//         41->Analyseroutine hat Ergebnisdatei nicht gefunden
//         42->der Analyseroutine übergebene Ergebnisdatei ist
//             fehlerhaft
//         43->Analyseroutine hat Zeitendatei nicht gefunden
//         44->der Analyseroutine übergebene Zeitendatei ist fehlerhaft
//         45->Dateneintragung der Analyseroutine in LOGFILE hat nicht
//             geklappt
//         46->zu wenig Speicher um Ergebnis zu analysieren
//         60->kann LOGDATEI nicht anlegen / Eintragung vornehmen
//         62->nicht genug Speicher um Laufzeittabelle anzulegen
//         63->zu wenig Speicher um Rahmenzeit zu ermitteln
//         64->kann Hilfsergebnisdatei nicht anlegen
//         65->kann Modulverteilung nicht aus CONFIGDATEI entnehmen
//         81->zu wenig Speicher zum Laden der Configdatei
//         82->Configdatei nicht gefunden
//         83->Configdatei fehlerhaft

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "bsp_bau6.h"
#include "../prgx/load.h"
#include "../prgx/save.h"
#include "../prgx/listen.h"
#include "../prgx/zufall.h"

#ifdef GRAPHEN
#include "../prgx/graphen.h"
#endif

#ifdef DGRAPHEN
#include "../prgx/dgraphen.h"
#endif

#ifdef BAUM
#include "opti6b.h"
#endif

#ifdef KRITPFAD
#include "opti6j.h"
#endif

#ifdef SPGRAPH
#include "opti6r.h"
#endif

#ifdef SCHICHT_DYN
#include "opti6s.h"
#endif

#ifdef SCHICHT_TIEF
#include "opti6t.h"
#endif


// Formelorientiere Optimierung von zufälligen Beispielen
// ------------------------------------------------------
#ifdef FORMEL

// Feld zur Ablage der Laufzeitdaten und Modulgrößen
double **laufz;

// Laufzeitroutine
double laufzeit( int modulnummer, int prozessorenanzahl )
{
  return laufz[ modulnummer-1 ][ prozessorenanzahl-1 ];
}

// fügt Eintrag bzgl. der Mindestergebnisgüte ins LOGFILE hinzu
// Fehler: 5->Dateneintragung in LOGFILE hat nicht geklappt
//         6->zu wenig Speicher
int analyse(
  char*   logfile,
  graph*  g,
  int     n,
  double* startzeit,
  int*    breite )
{
  FILE* fi;
  double a, h, min_flae, max_hoehe;
  int i, j, k, m;
  intliste fertig;
  int* besucht;
  double* lang;
  m = max_knoten( g );
  besucht = (int*) malloc( m * sizeof( int ) );
  if( besucht == NULL )
  {
    return 6;
  }
  lang = (double*) malloc( m * sizeof( double ) );
  if( lang == NULL )
  {
    return 6;
  }
  init_il( &fertig );
  // Gesamthöhe und min_flae berechnen
  min_flae = 0.0;
  h = 0.0;
  max_hoehe = 0.0;
  for( i=0; i<m; i++ )
  {
    if( test_knoten( g, i+1 ) == 0 )
    {
      if( laufz[i][0] != 0 ) // malleable
      {
        min_flae += laufz[i][0];
      }
      else // nonmalleable
      {
        j = 1;
        while( laufz[i][j] == 0.0 )
        {
          j++;
        }
        a = laufz[i][j] * ( j + 1 );
        j++;
        while( j < n )
        {
          if( a > laufz[i][j] * ( j + 1 ) )
          {
            a = laufz[i][j] * ( j + 1 );
          }
          j++;
        }
        min_flae += a;
      }
      if( startzeit[i] + laufz[i][ breite[i]-1 ] > h )
      {
        h = startzeit[i] + laufz[i][ breite[i]-1 ];
      }
      // suchen des längsten Pfades vorbereiten
      besucht[i] = eing_grad( g, i+1 );
      if( besucht[i] == 0 )
      {
        lang[i] = laufz[i][n-1];
        if( lang[i] > max_hoehe )
        {
          max_hoehe = lang[i];
        }
        if( add_end_il( &fertig, i+1 ) != 0 )
        {
          return 6;
        }
      }
    }
  }
  min_flae /= ((double)(n));
  // längsten Pfad suchen
  while( 1 )
  {
    i = del_anf_il( &fertig );
    if( i == 0 )
    {
      break;
    }
    reset_kind( g, i );
    while( 1 )
    {
      j = next_kind( g, i );
      if( j == 0 )
      {
        break;
      }
      besucht[j-1]--;
      if( besucht[j-1] == 0 )
      {
        lang[j-1] = 0.0;
        reset_vorg( g, j );
        while( 1 )
        {
          k = next_vorg( g, j );
          if( k == 0 )
          {
            break;
          }
          if( lang[j-1] < lang[k-1] )
          {
            lang[j-1] = lang[k-1];
          }
        }
        lang[j-1] += laufz[j-1][n-1];
        if( lang[j-1] > max_hoehe )
        {
          max_hoehe = lang[j-1];
        }
        if( add_end_il( &fertig, j ) != 0 )
        {
          return 6;
        }
      }
    }
  }
  if( max_hoehe > min_flae )
  {
    min_flae = max_hoehe;
  }
  h /= min_flae;
  fi = fopen( logfile, "a" );
  if( fi == NULL )
  {
    return 5;
  }
  if( save_double_r( &fi, h ) != 0 )
  {
    return 5;
  }
  fclose( fi );
  free( lang );
  free( besucht );
  free_il( &fertig );
  return 0;
}

int test()
{
  int i, j;
  int n, fehler;
  double guete, max_guete, a;
  time_t anf, end, zeit;
#if ZEITMESS == 1
  time_t rahmen_anf, rahmen_end;
  char dummy_datei[] = "dummy.txt";
#endif
  FILE* fi;
  graph g;
  double* startzeit; // Variablen fürs Ergebnis
  int* breite;
  int* linker;
  verteilung_int vert_m; // Variablen für CONFIGDATEI
  verteilung_text vert_kanten;
  verteilung_text vert_seq;
  int anz_vert;
  verteilung_double_liste* liste_vert;
  zeit = 0;
  init_graph( &g );
  init_verteilung_text( &vert_kanten );
  init_verteilung_text( &vert_seq );
  fi = fopen( LOGFILE, "w" );
  if( fi == NULL )
  {
    return 60;
  }
  if( save_text( &fi, "Die einzelnen Gütefaktoren lauten:\n" ) != 0 )
  {
    return 60;
  }
  fclose( fi );
  fehler = load_configdatei_nonmoldable(
    CONFIGDATEI, &n, &vert_m, &vert_kanten, &vert_seq,
    &anz_vert, &liste_vert );
  if( fehler != 0 )
  {
    return 80 + fehler;
  }
  // Speicher für Laufzeiten reservieren
  if( init_laufzeit( &laufz, n, &vert_m ) != 0 )
  {
    return 62;
  }
#if ZEITMESS == 1
  time( &anf ); // Startzeit
#else
  zeit = 0;
#endif
  for( i=0; i<TESTANZAHL; i++ )
  {
    fehler = baue_bsp_array( laufz, &g, n, &vert_m, &vert_kanten,
                             &vert_seq, anz_vert, liste_vert );
    if( fehler != 0 )
    {
      return fehler;
    }
#if ZEITMESS == 2
    time( &anf ); // Startzeit
#endif
    for( j=1; j<OPTIMIERUNGEN; j++ )
    {
      fehler = optimiere_formel( &startzeit, &breite, &linker,
                                 &g, n, &(laufzeit) );
      if( fehler != 0 )
      {
        return fehler + 20;
      }
      free( startzeit ); // Ergebnisdaten löschen
      free( breite );
      free( linker );
    }
    fehler = optimiere_formel( &startzeit, &breite, &linker,
                               &g, n, &(laufzeit) );
    if( fehler != 0 )
    {
      return fehler + 20;
    }
#if ZEITMESS == 2
    time( &end ); // Endzeit
    zeit += end - anf;
#endif
    fehler = analyse( LOGFILE, &g, n, startzeit, breite );
    if( fehler != 0 )
    {
      return fehler + 40;
    }
    free( startzeit ); // Ergebnisdaten löschen
    free( breite );
    free( linker );
  }
#if ZEITMESS == 1
  time( &end ); // Endzeit
  startzeit = (double*) malloc(
    zufall_max_int( &vert_m ) * sizeof( double ) );
  if( startzeit == NULL )
  {
    return 63;
  }
  breite = (int*) malloc( zufall_max_int( &vert_m ) * sizeof( int ) );
  if( breite == NULL )
  {
    return 63;
  }
  for( i=0; i<zufall_max_int( &vert_m ); i++ )
  {
    breite[i] = 1;
  }
  time( &rahmen_anf );
  for( i=0; i<TESTANZAHL; i++ )
  {
    fehler = baue_bsp_array( laufz, &g, n, &vert_m, &vert_kanten,
                             &vert_seq, anz_vert, liste_vert );
    if( fehler != 0 )
    {
      return fehler;
    }
    fehler = analyse( dummy_datei, &g, n, startzeit, breite );
    if( fehler != 0 )
    {
      return fehler + 40;
    }
  }
  time( &rahmen_end );
  zeit = end - anf + rahmen_anf - rahmen_end;
  free( startzeit );
  free( breite );
  remove( dummy_datei );
#endif // ZEITMESS == 1
  frei_laufzeit( laufz, &vert_m );
  frei_configdatei( anz_vert, liste_vert );
  free_graph( &g );
  free_verteilung_text( &vert_kanten );
  free_verteilung_text( &vert_seq );
  fi = fopen( LOGFILE, "r" );
  if( fi == NULL )
  {
    return 60;
  }
  max_guete = 0.0;
  guete = 0.0;
  while( 1 )
  {
    if( lese_next_double_r( &fi, &a ) != 0 )
    {
      break;
    }
    guete += a;
    if( a > max_guete )
    {
      max_guete = a;
    }
  }
  fclose( fi );
  fi = fopen( LOGFILE, "a" );
  if( fi == NULL )
  {
    return 60;
  }
  if( save_text( &fi, "\nDurchschnittliche Zeit: " ) != 0 )
  {
    return 60;
  }
  if( save_double( &fi,
                   ((double)(zeit)) /
                   ((double)(OPTIMIERUNGEN)) /
                   ((double)(TESTANZAHL)) ) != 0 )
  {
    return 60;
  }
  if( save_text(
        &fi,
        " Sekunden\nDurchschnittliche Mindestergebnisgüte: " ) != 0 )
  {
    return 60;
  }
  if( save_double_r( &fi, guete / ((double)(TESTANZAHL)) ) != 0 )
  {
    return 60;
  }
  if( save_text( &fi, "Schlechteste Mindestgüte: " ) != 0 )
  {
    return 60;
  }
  if( save_double_r( &fi, max_guete ) != 0 )
  {
    return 60;
  }
  fclose( fi );
  printf( "\nDurchschnittliche Zeit: %lf Sekunden\n",
          ((double)(zeit)) /
          ((double)(OPTIMIERUNGEN)) /
          ((double)(TESTANZAHL)) );
  printf( "Durchschnittliche Mindestergebnisgüte: %lf\n",
          guete / ((double)(TESTANZAHL)) );
  printf( "Schlechteste Mindestgüte: %lf\n", max_guete );
  return 0;
}

#endif


// Dateiorientiere Optimierung von zufälligen Beispielen
// -----------------------------------------------------
#ifndef FORMEL

// fügt Eintrag bzgl. der Mindestergebnisgüte ins LOGFILE hinzu
// Fehler: 1->Ergebnisdatei nicht gefunden
//         2->Ergebnisdatei fehlerhaft
//         3->Zeitendatei nicht gefunden
//         4->Zeitendatei fehlerhaft
//         5->Dateneintragung in LOGFILE hat nicht geklappt
//         6->zu wenig Speicher
int analyse( char* logfile, char* ergeb, char* daten, char* abhgraph )
{
  FILE* fi;
  double h, min_flae, max_hoehe, d;
  int i, j, k, n, m;
  intliste fertig;
  int* besucht;
  double* lang;
  double* laufz;
  char c;
  graph g;
  init_graph( &g );
  if( load_graph_rueck( abhgraph, &g ) != 0 )
  {
    return 6;
  }
  m = max_knoten( &g );
  besucht = (int*) malloc( m * sizeof( int ) );
  if( besucht == NULL )
  {
    return 6;
  }
  lang = (double*) malloc( m * sizeof( double ) );
  if( lang == NULL )
  {
    return 6;
  }
  laufz = (double*) malloc( m * sizeof( double ) );
  if( laufz == NULL )
  {
    return 6;
  }
  init_il( &fertig );
  // Endzeit des letzten Moduls ermitteln
  fi = fopen( ergeb, "r" );
  if( fi == NULL )
  {
    return 1;
  }
  h = 0.0;
  for( i=0; i<m; i++ )
  {
    if( test_knoten( &g, i+1 ) == 0 )
    {
      if( lese_next_char( &fi, "i", &c ) != 0 )
      {
        return 2;
      }
      if( lese_next_double_r( &fi, &d ) != 0 )
      {
        return 2;
      }
      if( h < d )
      {
        h = d;
      }
    }
  }
  fclose( fi );
  // min_flae berechnen und Laufzeiten mit n Prozessoren ermitteln
  fi = fopen( daten, "r" );
  if( fi == NULL )
  {
    return 3;
  }
  max_hoehe = 0.0;
  i = 0; // wg. Bestimmung von n ersten Schleifendurchlauf extra
  while( test_knoten( &g, i+1 ) != 0 )
  {
    i++;
  }
  if( lese_next_double( &fi, &min_flae ) != 0 )
  {
    return 4;
  }
  n = 1;
  if( min_flae != 0 ) // moldable
  {
    while( lese_next_double_in_line( &fi, &d ) == 0 )
    {
      n++;
      laufz[i] = d;
    }
  }
  else // nonmoldable
  {
    do
    {
      if( lese_next_double_in_line( &fi, &min_flae ) != 0 )
      {
        return 4;
      }
      n++;
    } while( min_flae == 0 );
    laufz[i] = min_flae;
    min_flae *= n;
    while( lese_next_double_in_line( &fi, &d ) == 0 )
    {
      n++;
      laufz[i] = d;
      if( min_flae > d * n )
      {
        min_flae = d * n;
      }
    }
  }
  lese_next_ret( &fi ); // nächste Zeile
  // suchen des längsten Pfades vorbereiten
  besucht[i] = eing_grad( &g, i+1 );
  if( besucht[i] == 0 )
  {
    lang[i] = laufz[i];
    if( lang[i] > max_hoehe )
    {
      max_hoehe = lang[i];
    }
    if( add_end_il( &fertig, i+1 ) != 0 )
    {
      return 6;
    }
  }
  for( i++; i<m; i++ )
  {
    if( test_knoten( &g, i+1 ) == 0 )
    {
      if( lese_next_double( &fi, &d ) != 0 )
      {
        return 4;
      }
      if( d != 0 ) // moldable
      {
        min_flae += d;
        for( j=1; j<n; j++ ) // bis zur letzten Laufzeit lesen
        {
          if( lese_next_double_in_line( &fi, &(laufz[i]) ) != 0 )
          {
            return 4;
          }
        }
      }
      else // nonmoldable
      {
        j = 1;
        do
        {
          if( lese_next_double_in_line( &fi, &(laufz[i]) ) != 0 )
          {
            return 4;
          }
          j++;
        } while( laufz[i] == 0 );
        d = laufz[i] * j;
        for( ; j<n; j++ )
        {
          if( lese_next_double_in_line( &fi, &(laufz[i]) ) != 0 )
          {
            return 4;
          }
          if( laufz[i] * ( j + 1 ) < d )
          {
            d = laufz[i] * ( j + 1 );
          }
        }
        min_flae += d;
      }
      lese_next_ret( &fi ); // nächste Zeile
      // suchen des längsten Pfades vorbereiten
      besucht[i] = eing_grad( &g, i+1 );
      if( besucht[i] == 0 )
      {
        lang[i] = laufz[i];
        if( lang[i] > max_hoehe )
        {
          max_hoehe = lang[i];
        }
        if( add_end_il( &fertig, i+1 ) != 0 )
        {
          return 6;
        }
      }
    }
  }
  fclose( fi );
  min_flae /= ((double)(n));
  // längsten Pfad suchen
  while( 1 )
  {
    i = del_anf_il( &fertig );
    if( i == 0 )
    {
      break;
    }
    reset_kind( &g, i );
    while( 1 )
    {
      j = next_kind( &g, i );
      if( j == 0 )
      {
        break;
      }
      besucht[j-1]--;
      if( besucht[j-1] == 0 )
      {
        lang[j-1] = 0.0;
        reset_vorg( &g, j );
        while( 1 )
        {
          k = next_vorg( &g, j );
          if( k == 0 )
          {
            break;
          }
          if( lang[j-1] < lang[k-1] )
          {
            lang[j-1] = lang[k-1];
          }
        }
        lang[j-1] += laufz[j-1];
        if( lang[j-1] > max_hoehe )
        {
          max_hoehe = lang[j-1];
        }
        if( add_end_il( &fertig, j ) != 0 )
        {
          return 6;
        }
      }
    }
  }
  if( max_hoehe > min_flae )
  {
    min_flae = max_hoehe;
  }
  h /= min_flae;
  fi = fopen( logfile, "a" );
  if( fi == NULL )
  {
    return 5;
  }
  if( save_double_r( &fi, h ) != 0 )
  {
    return 5;
  }
  fclose( fi );
  free( lang );
  free( besucht );
  free( laufz );
  free_graph( &g );
  free_il( &fertig );
  return 0;
}

int test()
{
  int i, j;
  int fehler;
  double guete, max_guete, a;
  time_t anf, end, zeit;
#if ZEITMESS == 1
  time_t rahmen_anf, rahmen_end;
  verteilung_int vert_m;
  char dummy_datei[] = "dummy.txt";
#endif
  FILE* fi;
  zeit = 0;
  fi = fopen( LOGFILE, "w" );
  if( fi == NULL )
  {
    return 60;
  }
  if( save_text( &fi, "Die einzelnen Gütefaktoren lauten:\n" ) != 0 )
  {
    return 60;
  }
  fclose( fi );
#if ZEITMESS == 1
  time( &anf ); // Startzeit
#else
  zeit = 0;
#endif
  for( i=0; i<TESTANZAHL; i++ )
  {
    fehler = baue_bsp_datei( ZEITENDATEI, GRAPHDATEI, CONFIGDATEI );
    if( fehler != 0 )
    {
      return fehler;
    }
#if ZEITMESS == 2
    time( &anf ); // Startzeit
#endif
    for( j=0; j<OPTIMIERUNGEN; j++ )
    {
      fehler = optimiere_datei( ERGEBNISDATEI, ZEITENDATEI,
                                GRAPHDATEI );
      if( fehler != 0 )
      {
        return fehler + 20;
      }
    }
#if ZEITMESS == 2
    time( &end ); // Endzeit
    zeit += end - anf;
#endif
    fehler = analyse( LOGFILE, ERGEBNISDATEI, ZEITENDATEI, GRAPHDATEI );
    if( fehler != 0 )
    {
      return fehler + 40;
    }
  }
#if ZEITMESS == 1
  time( &end ); // Endzeit
  fi = fopen( CONFIGDATEI, "r" );
  if( fi == NULL )
  {
    return 65;
  }
  if( lese_next_int_r( &fi, &i ) != 0 )
  {
    return 65;
  }
  if( lese_next_verteilung_int_r( &fi, &vert_m ) != 0 )
  {
    return 65;
  }
  fclose( fi );
  fi = fopen( ERGEBNISDATEI, "w" );
  if( fi == NULL )
  {
    return 64;
  }
  for( i=0; i<zufall_max_int( &vert_m ); i++ )
  {
    if( save_text(
          &fi,
          "Modul 0000000001: von +0.00000000000000e+000 bis "
          "+1.93334715190320e-001 auf Prozessoren "
          "0000000001 bis 0000000010\n" ) != 0 )
    {
      return 64;
    }
  }
  fclose( fi );
  time( &rahmen_anf ); // Endzeit
  for( i=0; i<TESTANZAHL; i++ )
  {
    fehler = baue_bsp_datei( ZEITENDATEI, GRAPHDATEI, CONFIGDATEI );
    if( fehler != 0 )
    {
      return fehler;
    }
    fehler = analyse( dummy_datei, ERGEBNISDATEI, ZEITENDATEI,
                      GRAPHDATEI );
    if( fehler != 0 )
    {
      return fehler + 40;
    }
  }
  time( &rahmen_end );
  zeit += end - anf + rahmen_end - rahmen_anf;
  remove( dummy_datei );
#endif // ZEITMESS == 1
  remove( ERGEBNISDATEI );
  remove( ZEITENDATEI );
  remove( GRAPHDATEI );
  fi = fopen( LOGFILE, "r" );
  if( fi == NULL )
  {
    return 60;
  }
  max_guete = 0.0;
  guete = 0.0;
  while( 1 )
  {
    if( lese_next_double_r( &fi, &a ) != 0 )
    {
      break;
    }
    guete += a;
    if( a > max_guete )
    {
      max_guete = a;
    }
  }
  fclose( fi );
  fi = fopen( LOGFILE, "a" );
  if( fi == NULL )
  {
    return 60;
  }
  if( save_text( &fi, "\nDurchschnittliche Zeit: " ) != 0 )
  {
    return 60;
  }
  if( save_double( &fi,
                   ((double)(zeit)) /
                   ((double)(OPTIMIERUNGEN)) /
                   ((double)(TESTANZAHL)) ) != 0 )
  {
    return 60;
  }
  if( save_text(
        &fi,
        " Sekunden\nDurchschnittliche Mindestergebnisgüte: " ) != 0 )
  {
    return 60;
  }
  if( save_double_r( &fi, guete / ((double)(TESTANZAHL)) ) != 0 )
  {
    return 60;
  }
  if( save_text( &fi, "Schlechteste Mindestgüte: " ) != 0 )
  {
    return 60;
  }
  if( save_double_r( &fi, max_guete ) != 0 )
  {
    return 60;
  }
  fclose( fi );
  printf( "\nDurchschnittliche Zeit: %lf Sekunden\n",
          ((double)(zeit)) /
          ((double)(OPTIMIERUNGEN)) /
          ((double)(TESTANZAHL)) );
  printf( "Durchschnittliche Mindestergebnisgüte: %lf\n",
          guete / ((double)(TESTANZAHL)) );
  printf( "Schlechteste Mindestgüte: %lf\n", max_guete );
  return 0;
}
#endif

int main()
{
  int f;
  // nur wegen Fehlerausgabe; sonst könnte auch test
  // die main-Routine sein
  f = test();
  printf( "\nRückgabewert: %d\n", f );
  return f;
}
