#ifndef BAUM_OP

#define BAUM_OP

typedef struct ast
{
  int wert; // um Daten im Knoten abzulegen
  struct ast* left; // linkes Kind
  struct ast* right; // rechtes Kind
  struct ast* pred; // Vorgänger
} ast;

typedef struct baum
{
  ast* w; // Wurzel des Baumes
  int max; // Dimension des Puffers
  int akt; // Füllstand des Puffers
  ast *p; // Puffer
} baum;

// Konstruktor - Ersatz; Parameter max gibt die Größe des Puffers
// für ungenutzte Knoten an (Reduzierung von malloc/free)
void init_size_baum( baum* b, int max );

// Destruktor; Datenstruktur ist dann initialisiert,
// aber Puffer ist leer
void free_baum( baum* b );

// löscht a und alle Kinder von a
void del_all_baum( baum* b, ast* a );

// löscht Knoten a und den linken Teilbaum unterhalb von a;
// der rechte Unterbaum nimmt die Position von a ein
// Rückgabe: Knoten, der die Stelle von a eingenommen hat
//           (vorher a->right)
ast* del_l_baum( baum* b, ast* a );

// liefert die Wurzel des Baumes
ast* get_wurzel_baum( baum* b );

// liefert einen neuen Knoten, der keine Nachfolger hat
// Fehler: NULL->nicht genug Speicher
ast* new_baum( baum* b );

// fügt den Knoten c als linkes Kind von a ein;
// das linke Kind von a wird linkes Kind von c
// ist a == NULL so wird die Wurzel ersetzt
void insert_ast_ll_baum( baum* b, ast* a, ast* c );

// fügt einen Knoten als linkes Kind von a ein;
// das linke Kind von a wird linkes Kind vom neuen Knoten
// ist a == NULL so wird die Wurzel ersetzt
// Rückgabe: eingefügter Knoten
// Fehler: NULL->nicht genug Speicher; Baum ist unverändert
ast* insert_ll_baum( baum* b, ast* a );

// fügt den Knoten c als rechtes Kind von a ein;
// das rechte Kind von a wird rechtes Kind von c
// ist a == NULL so wird die Wurzel ersetzt
void insert_ast_rr_baum( baum* b, ast* a, ast* c );

// fügt einen Knoten als rechtes Kind von a ein;
// das rechte Kind von a wird rechtes Kind vom neuen Knoten
// ist a == NULL so wird die Wurzel ersetzt
// Rückgabe: eingefügter Knoten
// Fehler: NULL->nicht genug Speicher; Baum ist unverändert
ast* insert_rr_baum( baum* b, ast* a );

// fügt den Knoten c an die Stelle von a ein; das linke Kind von c
// wird a; a == NULL ist nicht zulässig
void insert_ast_l_baum( baum* b, ast* a, ast* c );

// fügt einen Knoten an die Stelle von a ein; das linke Kind vom neuen
// Knoten wird a; a == NULL ist nicht zulässig
// Rückgabe: eingefügter Knoten
// Fehler: NULL->nicht genug Speicher; Baum ist unverändert
ast* insert_l_baum( baum* b, ast* a );

// fügt den Knoten c an die Stelle von a ein; das rechte Kind von c
// wird a; a == NULL ist nicht zulässig
void insert_ast_r_baum( baum* b, ast* a, ast* c );

// fügt einen Knoten an die Stelle von a ein; das rechte Kind vom neuen
// Knoten wird a; a == NULL ist nicht zulässig
// Rückgabe: eingefügter Knoten
// Fehler: NULL->nicht genug Speicher; Baum ist unverändert
ast* insert_r_baum( baum* b, ast* a );

// hängt den Unterbaum mit Wurzel a rechts an den Knoten c
// !! c darf kein rechtes Kind haben !!
// !! c darf nicht unterhalb von a liegen !!
// !! c == NULL ist nicht zulässig !!
void move_r_baum( ast* a, ast* c );

// hängt den Unterbaum mit Wurzel a links an den Knoten c
// !! c darf kein linkes Kind haben !!
// !! c darf nicht unterhalb von a liegen !!
// !! c == NULL ist nicht zulässig !!
void move_l_baum( ast* a, ast* c );

// beschriftet den Knoten a mit dem Wert i
void set_baum( ast* a, int i );

// liefert den Wert des Knotens a
int get_baum( ast* a );

// liefert das linke Kind von a
// Fehler: NULL->kein linkes Kind vorhanden
ast* left_baum( ast* a );

// liefert das rechte Kind von a
// Fehler: NULL->kein rechtes Kind vorhanden
ast* right_baum( ast* a );

// liefert den Vater von a
// Fehler: NULL->a ist Wurzel
ast* pred_baum( ast* a );

#endif
