// Die Standardbeispiele werden mit
//   baue_standard_bsp(
//     double** Laufzeiten (Feld der Größe m*n, Speicher muß schon
//              reserviert sein),
//     int      Prozessorenanzahl,
//     graph*   Modulabhängigkeitsgraph (muß initialisiert sein)
//     char*    Filename Eingabedatei );
//
// erzeugt.
//
// Fehler: 1->zu wenig Speicher
//         3->nicht genug Einträge im Listenfeld / Fehler in Datei
//         4->konnte Eingabedatei nicht finden

#include <stdlib.h>
#include "../prgx/heap.h"
#include "../prgx/load.h"
#include "../prgx/suchsort.h"

#ifdef GRAPHEN
#include "../prgx/graphen.h"
#endif

#ifdef DGRAPHEN
#include "../prgx/dgraphen.h"
#endif


// berechnet x hoch y
int hoch( int x, int y )
{
  int z;
  if( y == 0 )
  {
    return 1;
  }
  if( y == 1 )
  {
    return x;
  }
  z = hoch( x, y/2 );
  if( (y/2)*2 == y ) // y ist gerade
  {
    return z * z;
  }
  return z * z * x;
}

// Aufrunden( Logarithmus zur Basis 2 von x )
int auf_log2( int x )
{
  int i, r;
  r = 0;
  i = 0;
  while( x > 1 )
  {
    if( (x/2)*2 != x )
    {
      r = 1;
    }
    x /= 2;
    i++;
  }
  return i + r;
}

// Abrunden( Logarithmus zur Basis 2 von x )
int ab_log2( int x )
{
  int i;
  i = 0;
  while( x > 1 )
  {
    x /= 2;
    i++;
  }
  return i;
}

// Erzeugt ein Standardbeispiel in Komplexität 1 auf Basis der
// Komplexität 3 bei Verwendung eines Greedy Algorithmus zur
// Optimierung
// Das Beispiel mit Komplexität 3 besteht aus dem Graph gr und den
// Laufzeiten lz und m Modulen
// Es werden die Modulmummern ab *m_nr benutzt und die Laufzeiten
// für n Prozessoren nach laufz geschrieben
// Fehler: 1->zu wenig Speicher
int standard1Greedy(
  int*     m_nr,
  int      n,
  double** laufz,
  graph*   gr,
  double** lz,
  int      m )
{
  int i, j, k, p;
  double* endzeit;
  int* ausg;
  intheap startbar;
  double* startmin;
  endzeit = (double*) malloc( n * sizeof( double ) );
  if( NULL == endzeit )
  {
    return 1;
  }
  ausg = (int*) malloc( m * sizeof( int ) );
  if( NULL == eing_grad )
  {
    return 1;
  }
  init_ih( &startbar );
  startmin = (double*) malloc( m * sizeof( double ) );
  if( NULL == startmin )
  {
    return 1;
  }
  for( k=1; k<=n; k++ )
  {
    for( i=0; i<k; i++ )
    {
      endzeit[i] = 0.0;
    }
    for( i=0; i<m; i++ )
    {
      ausg[i] = ausg_grad( gr, i+1 );
      startmin[i] = 0.0;
      if( 0 == ausg[i] )
      {
        // Zielknoten werden als erste ausgeführt (gedrehtes Schedule)
        if( 0 != add_min_indirect_ih( &startbar, i+1, startmin ) )
        {
          return 1;
        }
      }
    }
    while( 1 )
    {
      // nächstes Modul zur Ausführung
      i = del_min_wurzel_indirect_ih( &startbar, startmin );
      if( 0 == i )
      {
        break;
      }
      // Prozessor zur Modulausführung
      p = such_min_limit( endzeit, k, startmin[ i-1 ] );
      if( endzeit[p] < startmin[ i-1 ] )
      {
        endzeit[p] = startmin[ i-1 ] + lz[ i-1 ][0];
      }
      else
      {
        endzeit[p] += lz[ i-1 ][0];
      }
      // markiere Vorgängermodule als ausführbar
      reset_vorg( gr, i );
      while( 1 )
      {
        j = next_vorg( gr, i );
        if( 0 == j )
        {
          break;
        }
        ausg[j-1] --;
        if( endzeit[p] > startmin[j-1] )
        {
          startmin[j-1] = endzeit[p];
        }
        if( 0 == ausg[j-1] )
        {
          if( 0 != add_min_indirect_ih( &startbar, j, startmin ) )
          {
            return 1;
          }
        }
      }
    }
    // alle Module ausgeführt
    i = such_max( endzeit, k );
    laufz[*m_nr][k-1] = endzeit[i];
  }
  free( startmin );
  free_ih( &startbar );
  free( ausg );
  free( endzeit );
  (*m_nr)++;
  return 0;
}

// Erzeugt eine LR-Zerlegung nach Gauß der Dimension r
// in Komplexität 3
// Es werden die Modulmummern ab *m_nr benutzt und die Laufzeiten
// für n Prozessoren erstellt
// Fehler: 1->zu wenig Speicher
int gauss3( int* m_nr, int n, int r, graph* gr, double** laufz )
{
  int a, b, i, k, m, p, q, s, t, u, z;
  m = *m_nr;
  // Schicht r
  p = r * r;
  for( i=0; i<p; i++ )
  {
    for( k=0; k<n; k++ )
    {
      laufz[m+i][k] = DIVISION;
    }
  }
  m += p;
  if( 1 == r )
  { // Keine Kanten
    if( 0 != add_knoten( gr, m ) )
    {
      return 1;
    }
    *m_nr = m;
    return 0;
  }
  // Schicht s
  for( s=r-1; s>0; s-- )
  {
    // Knoten mit Subtraktion
    t = s * ( s + 1 );
    for( i=0; i<t; i++ )
    {
      for( k=0; k<n; k++ )
      {
        laufz[m+i][k] = 1.0;
      }
    }
    // Kanten von der Schicht s+1 zu Knonten mit Subtraktion der
    // Schicht s
    t = s + 1;
    u = t * t;
    for( a=0; a<s; a++ )
    {
      for( b=0; b<t; b++ )
      {
        p = m + a * t + b + 1;
        q = m - u + b + 1;
        if( 0 != add_kante( gr, q, p ) )
        {
          return 1;
        }
        q += t + t * a;
        if( 0 != add_kante( gr, q, p ) )
        {
          return 1;
        }
      }
    }
    // Knoten mit Division
    u = s * s;
    z = m + s * t; // erster Knoten mit Division
    for( i=0; i<u; i++ )
    {
      for( k=0; k<n; k++ )
      {
        laufz[z+i][k] = DIVISION;
      }
    }
    // Kanten innerhalb der Schicht s zu Knoten mit Division
    for( a=0; a<s; a++ )
    {
      for( b=0; b<s; b++ )
      {
        p = z + a * s + b + 1;
        q = m + a * t + 1;
        if( 0 != add_kante( gr, q, p ) )
        {
          return 1;
        }
        q += b + 1;
        if( 0 != add_kante( gr, q, p ) )
        {
          return 1;
        }
      }
    }
    m = z + u;
  }
  *m_nr = m;
  return 0;
}

// Erzeugt eine LR-Zerlegung nach Gauß der Dimension r
// in Komplexität 2
// Es werden die Modulmummern ab *m_nr benutzt und die Laufzeiten
// für n Prozessoren erstellt
// Fehler: 1->zu wenig Speicher
int gauss2( int* m_nr, int n, int r, graph* gr, double** laufz )
{
  int a, b, c, d, e, f, i, k, m, s, t;
  double laufzeit;
  m = *m_nr;
  // Schicht 1: r Module mit je r Divisionen
  for( k=0; k<n; k++ ) // Prozessorenanzahl
  {
    laufzeit =
      DIVISION * (double)(int)(( r + k ) / ( k + 1 ));
    for( i=0; i<r; i++ ) // Module in Schicht
    {
      laufz[m+i][k] = laufzeit;
    }
  }
  m += r;
  if( 1 == r )
  { // keine Kanten vorhanden; Knoten einfügen
    if( 0 != add_knoten( gr, m ) )
    {
      return 1;
    }
    *m_nr = m;
    return 0;
  }
  // Schicht r-s+1
  for( s=r-1; s>0; s-- )
  {
    for( k=0; k<n; k++ ) // Prozessorenanzahl
    {
      if( s <= k )
      { // alle Subtraktionen und anschließend alle Divisionen können
        // parallel ausgeführt werden
        laufzeit = DIVISION + 1.0;
      }
      else if( s == k + 1 )
      { // es gibt eine Subtraktion mehr als Prozessoren; aber genauso
        // viele Divisionen wie Prozessoren
        laufzeit = DIVISION + 2.0;
      }
      else
      {
        a = s / ( k + 1 ); // Mindestanzahl an Divisionen je Prozessor
        b = s - a * ( k + 1 ); // Restdivisionen
        if( 0 == b )
        {
          laufzeit = a * DIVISION + a + 1.0;
        }
        else
        {
          // verbleibende Anzahl an Subtraktionen nachdem jeder
          // Prozessor eine ausgeführt hat
          c = s - k;
          // maximale Anzahl an Subtraktionen, die parallel zu den
          // Restdivisionen ausgeführt werden kann
          d = ( k + 1 - b ) * ((int)DIVISION);
          if( c <= d )
          { // alle Subtraktionen sind parallel zu den Restdivisionen
            // ausführbar
            laufzeit = ( a + 1 ) * DIVISION + 1;
          }
          else
          {
            // Mindestanzahl aller Subtraktionen, die jeder Prozessor
            // ausführt
            e = ( c - d ) / ( k + 1 );
            // Restsubtraktionen
            f = c - d - e * ( k + 1 );
            if( 0 == f )
            {
              laufzeit = ( a + 1 ) * DIVISION + 1 + e;
            }
            else if( f < k - b )
            { // Restsubtraktionen sind parallel zu Restdivisionen
              // ausführbar
              laufzeit = a * DIVISION + ((int)DIVISION) + 2 + e;
            }
            else
            {
              laufzeit = ( a + 1 ) * DIVISION + 2 + e;
            }
          }
        }
      }
      for( i=0; i<s; i++ ) // Module in Schicht
      {
        laufz[m+i][k] = laufzeit;
      }
    }
    // Je eine Kante von jedem Knoten der Schicht zu
    // jedem Knoten der Vorgängerschicht
    t = s + 1;
    for( a=1; a<t; a++ )
    {
      for( b=0; b<t; b++ )
      {
        if( 0 != add_kante( gr, m-b, m+a ) )
        {
          return 1;
        }
      }
    }
    m += s;
  }
  *m_nr = m;
  return 0;
}

// Erzeugt eine LR-Zerlegung nach Gauß der Dimension r
// in Komplexität 1
// Es werden die Modulmummern ab *m_nr benutzt und die Laufzeiten
// für n Prozessoren erstellt
// Fehler: 1->zu wenig Speicher
int gauss1( int* m_nr, int n, int r, double** laufz )
{
  int i, m, z;
  double** lz;
  double* lz_help;
  graph gr;
  // Knotenanzahl
  z = (r-1) * r * (2*r-1) / 3 + (r-1) * r / 2 + r * r;
  init_graph( &gr );
  // Laufzeiten für n=1 reichen, da nur für alle n gleich
  lz = (double**) malloc( z * sizeof( double* ) );
  if( NULL == lz )
  {
    return 1;
  }
  lz_help = (double*) malloc( z * sizeof( double ) );
  if( NULL == lz_help )
  {
    return 1;
  }
  for( i=0; i<z; i++ )
  {
    lz[i] = & lz_help[i];
  }
  m = 0;
  if( 0 != gauss3( &m, 1, r, &gr, lz ) )
  {
    return 1;
  }
  if( 0 != standard1Greedy( m_nr, n, laufz, &gr, lz, m ) )
  {
    return 1;
  }
  free( lz_help );
  free( lz );
  free_graph( &gr );
  return 0;
}

// Erzeugt eine FFT der Dimension r in Komplexität 3
// Es werden die Modulmummern ab *m_nr benutzt und die Laufzeiten
// für n Prozessoren erstellt
// Fehler: 1->zu wenig Speicher
int fft3( int* m_nr, int n, int r, graph* gr, double** laufz )
{
  int j, k, lr, m, o, p, q, s, t, u;
  m = *m_nr;
  lr = auf_log2( r );
  t = 5 * r * lr - 3 * r; // Knotenanzahl
  if( r == 2 ) // nur 4 einzelne Knoten
  {
    for( j=0; j<4; j++ )
    {
      if( add_knoten( gr, m+1+j ) != 0 )
      {
        return 1;
      }
    }
    q = m;
  }
  else
  {
    q = 1; // erste Knotennummer Hauptebene
    o = 2 * r + 1; // erste Knotennummer Zwischenebene 1
    u = 4 * r + 1; // erste Knotennummer Zwischenebene 2
    for( j=4; j<=r; j*=2 )
    {
      for( s=0; s<r; s+=j )
      {
        // Knoten ohne komplexe Multiplikation
        for( p=0; p<j; p++ )
        {
          if( add_kante( gr, m+q, m+q+5*r ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+q, m+q+5*r+j ) != 0 )
          {
            return 1;
          }
          for( k=0; k<n; k++ )
          {
            laufz[m+q-1][k] = 1.0; // Addition
          }
          q++;
        }
        // Knoten mit komplexer Multiplikation
        for( p=0; p<j/2; p++ )
        {
          if( add_kante( gr, m+q, m+o ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+q, m+o+1 ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+q+1, m+o+2 ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+q+1, m+o+3 ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+o, m+u ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+o+2, m+u ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+o+1, m+u+1 ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+o+3, m+u+1 ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+u, m+q+5*r ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+u, m+q+5*r-j ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+u+1, m+q+5*r+1 ) != 0 )
          {
            return 1;
          }
          if( add_kante( gr, m+u+1, m+q+5*r-j+1 ) != 0 )
          {
            return 1;
          }
          for( k=0; k<n; k++ )
          {
            laufz[m+q-1][k] = 1.0; // Addition
            laufz[m+q][k]   = 1.0; // Addition
            laufz[m+o-1][k] = MULTIPLIKATION;
            laufz[m+o][k]   = MULTIPLIKATION;
            laufz[m+o+1][k] = MULTIPLIKATION;
            laufz[m+o+2][k] = MULTIPLIKATION;
            laufz[m+u-1][k] = 1.0; // Addition
            laufz[m+u][k]   = 1.0; // Addition
          }
          o += 4;
          u += 2;
          q += 2;
        }
      }
      o += 3 * r; // zur nächsten Zwischenschicht 1
      u += 4 * r; // zur nächsten Zwischenschicht 2
      q += 3 * r; // zur nächsten Hauptschicht
    }
    q = m + q - 1;
  }
  m += t;
  // Knoten mit Addition in letzter Schicht
  for( ; q<m; q++ )
  {
    for( k=0; k<n; k++ )
    {
      laufz[q][k] = 1.0; // Addition
    }
  }
  *m_nr = m;
  return 0;
}

// Erzeugt eine "halbe" FFT der Dimension r in Komplexität 2
// Es werden die Modulmummern ab *m_nr benutzt und die Laufzeiten
// für n Prozessoren erstellt
// Fehler: 1->zu wenig Speicher
int fft2( int* m_nr, int n, int r, graph* gr, double** laufz )
{
  int a, b, c, d, e, f, g, h, i, j, k, lr, m, v;
  m = *m_nr;
  lr = auf_log2( r );
  // aktueller Vorgängerknoten
  v = m;
  // Anzahl an normalen Additionen in Ebene 1 der aktuellen Schicht
  a = r;
  // Zielknoten des in-Baumes (Schicht lr)
  for( k=0; k<n; k++ )
  {
    laufz[m][k] = ( a + k ) / ( k + 1 );
  }
  if( 2 == r )
  { // Baum entartet zu einem einzelnen Knoten
    if( 0 != add_knoten( gr, m+1 ) )
    {
      return 1;
    }
    (*m_nr)++;
    return 0;
  }
  m++;
  b = 1; // 1/2 * Knoten pro Schicht
  // Schicht lr-1 bis 1; aktuelle Schicht ist lr-j
  for( j=1; j<lr; j++ )
  {
    a /= 2;
    // Laufzeiten für das erste Knotenpärchen der Schicht
    // (linker/rechter Vorgänger) berechnen
    for( k=0; k<n; k++ )
    {
      // linkes Kind (a/2 komplexe Additionen)
      laufz[m][k] = (int)(( a + k ) / ( k + 1 ));
      // rechtes Kind (a/2 komplexe Multiplikationen)
      if( k + 1 >= 2 * a )
      {
        // alle 2*a Multiplikationen können parallel ausgeführt werden
        laufz[m+1][k] = 2.0 + MULTIPLIKATION;
      }
      else if( k + 1 >= a )
      {
        // alle a Additionen können parallel ausgeführt werden, die
        // 2*a Multiplikationen brauchen aber mindestens einen
        // Prozessor zweimal
        laufz[m+1][k] = 2.0 + 2.0 * MULTIPLIKATION;
      }
      else
      {
        // Anzahl an Additionen der Ebene 1, die jeder Prozessor
        // ausführt
        d = a / ( k + 1 );
        // Restadditionen
        e = a - d * ( k + 1 );
        if( 0 == e )
        { // keine Restadditionen -> auch keine Restmultiplikationen
          // möglich
          laufz[m+1][k] = 2 * d + 2 * d * MULTIPLIKATION;
        }
        else if( 2 * e <= k + 1 )
        { // Restadditionen der Ebenen 1 und 3 können auf
          // unterschiedlichen Prozessoren ausgeführt werden;
          // Anzahl an Prozessoren ohne Restadditionen
          f = k + 1 - 2 * e;
          // Anzahl an verbleibenden Multiplikationen, nachdem die
          // Prozessoren ohne Restadditionen je eine Multiplikation
          // ausgeführt haben
          g = 2 * a - f;
          // Anzahl an Multiplikationen, die jeder Prozessor ausführt
          h = g / ( k + 1 );
          // Restmultiplikationen
          i = f - h * ( k + 1 );
          if( 0 == i )
          {
            laufz[m+1][k] = 2 * d + 1 + h * MULTIPLIKATION;
          }
          else if( h <= f )
          {
            laufz[m+1][k] = 2 * d + ( h + 2 ) * MULTIPLIKATION;
          }
          else
          {
            laufz[m+1][k] = 2 * d + 1 + ( h + 1 ) * MULTIPLIKATION;
          }
        }
        else
        { // Restadditionen der Ebenen 1 und 3 werden auf den
          // gleichen Prozessoren ausgeführt;
          // Anzahl an Prozessoren ohne Restadditionen
          f = k + 1 - e;
          // Anzahl an verbleibenden Multiplikationen, nachdem die
          // Prozessoren ohne Restadditionen je zwei Multiplikationen
          // ausgeführt haben
          g = 2 * a - 2 * f;
          // Anzahl an Multiplikationen, die jeder Prozessor ausführt
          h = g / ( k + 1 );
          // Restmultiplikationen
          i = f - h * ( k + 1 );
          if( 0 == i )
          {
            laufz[m+1][k] = 2 * d + 2 + h * MULTIPLIKATION;
          }
          else if( h <= f )
          {
            laufz[m+1][k] = 2 * d + ( h + 3 ) * MULTIPLIKATION;
          }
          else
          {
            laufz[m+1][k] = 2 * d + 2 + ( h + 1 ) * MULTIPLIKATION;
          }            
        }
      }
    }
    if( add_kante( gr, m+1, v+1 ) != 0 )
    {
      return 1;
    }
    if( add_kante( gr, m+2, v+1 ) != 0 )
    {
      return 1;
    }
    m += 2;
    v++;
    // für alle weiteren Knotenpärchen (linker/rechter Vorgänger)
    // die Laufzeiten übernehmen
    for( c=1; c<b; c++ )
    {
      for( k=0; k<n; k++ )
      {
        laufz[m][k]   = laufz[m-2][k];
        laufz[m+1][k] = laufz[m-1][k];
      }
      if( add_kante( gr, m+1, v+1 ) != 0 )
      {
        return 1;
      }
      if( add_kante( gr, m+2, v+1 ) != 0 )
      {
        return 1;
      }
      m += 2;
      v++;
    }
    b *= 2;
  }
  (*m_nr) += r - 1;
  return 0;
}

// Erzeugt eine FFT der Dimension r in Komplexität 1
// Es werden die Modulmummern ab *m_nr benutzt und die Laufzeiten
// für n Prozessoren erstellt
// Fehler: 1->zu wenig Speicher
int fft1( int* m_nr, int n, int r, double** laufz )
{
  int i, lr, m, z;
  double** lz;
  double* lz_help;
  graph gr;
  lr = auf_log2( r );
  // Knotenanzahl
  z = 5 * r * lr - 3 * r;
  init_graph( &gr );
  // Laufzeiten für n=1 reichen, da nur für alle n gleich
  lz = (double**) malloc( z * sizeof( double* ) );
  if( NULL == lz )
  {
    return 1;
  }
  lz_help = (double*) malloc( z * sizeof( double ) );
  if( NULL == lz_help )
  {
    return 1;
  }
  for( i=0; i<z; i++ )
  {
    lz[i] = & lz_help[i];
  }
  m = 0;
  if( 0 != fft3( &m, 1, r, &gr, lz ) )
  {
    return 1;
  }
  if( 0 != standard1Greedy( m_nr, n, laufz, &gr, lz, m ) )
  {
    return 1;
  }
  free( lz_help );
  free( lz );
  free_graph( &gr );
  return 0;
}

// Erzeugt den (r*t)-ten Teil einer Matrixmultiplikation der
// Dimension r * s * t in Komplexität 3
// Es werden die Modulmummern ab *m_nr benutzt und die Laufzeiten
// für n Prozessoren erstellt
// Fehler: 1->zu wenig Speicher
int matrix3( int* m_nr, int n, int s, graph* gr, double** laufz )
{
  int i, j, k, m;
  m = *m_nr;
  // innere Knoten sind Additionen
  j = m + s - 1;
  for( i=m; i<j; i++ )
  {
    for( k=0; k<n; k++ )
    {
      laufz[i][k] = 1.0;
    }
  }
  // Blätter sind Multiplikationen
  j += s;
  for( ; i<j; i++ )
  {
    for( k=0; k<n; k++ )
    {
      laufz[i][k] = MULTIPLIKATION;
    }
  }
  *(m_nr) = i;
  if( 1 == s )
  { // kein Baum sondern nur ein einzelner Knoten
    if( 0 != add_knoten( gr, m+1 ) )
    {
      return 1;
    }
  }
  else
  { // Binärbaum erstellen
    for( j=2; j<=2*s-1; j++ )
    {
      if( add_kante( gr, m+j, m+j/2 ) != 0 )
      {
        return 1;
      }
    }
  }
  return 0;
}

// Berechnet die Laufzeiten des Moduls m bei der Verwendung von
// 1 bis n Prozessoren bei einer Matrixmultiplikation der Dimension
// r * s * t. Der Wert rt steht für r * t.
void matrixSingle( int m, int n, int s, int rt, double** laufz )
{
  int a, b, c, d, e, i, k, ls, mr_auf, o;
  double mr;
  ls = auf_log2( s );
  for( k=0; k<n; k++ )
  {
    o = rt; // Anzahl Operationen in Schicht
    a = rt * (s - 1); // Anzahl an Additionen
    b = rt * s; // Anzahl an Multiplikationen
    laufz[m][k] = 0;
    // alle vollständigen Schichten
    for( i=0; i<ls; i++ )
    {
      if( o <= k + 1 )
      { // alle Operationen können gleichzeitig ausgeführt
        // werden
        laufz[m][k] += 1;
        if( a >= o )
        { // alles Additionen
          a -= o;
        }
        else
        {
          b -= o - a;
          a = 0;
        }
        o *= 2;
        continue;
      }
      // es können nicht alle Operationen gleichzeitig ausgeführt
      // werden      
      // Zeit für verbleibende Additionen auf allen k + 1 Prozessoren
      laufz[m][k] += (int)(a / ( k + 1 ));
      a %= ( k + 1 );
      if( a == 0 )
      { // die Additionen sind ein vielfaches der Prozessorenanzahl
        // Multiplikationen können alle parallel auf allen Prozessoren
        // ausgeführt werden
        laufz[m][k] += ((int)(( b + k ) / ( k + 1 ))) * MULTIPLIKATION;
        break;
      }
      c = b - 2 * a; // bereits ausführbare Multiplikationen
      mr = 1.0 / MULTIPLIKATION;
      if( mr != (double)(int)mr )
      {
        mr_auf = 1 + (int)mr;
      }
      else
      {
        mr_auf = (int)mr;
      }
      e = k + 1 - a; // freie Prozessoren parallel zu letzten a
      //                Additionen
      d = mr_auf * e; // parallel zu den a Additionen
      //                 ausführbare Multiplikationen
      if( c >= d )
      { // Platz kann vollständig genutzt werden
        b -= d; // restliche Multiplikationen zusammen mit denen
        //         in Schicht ls ausführen
        d = a; // freie Prozessoren nach dem Ende der Additionen
      }
      else
      {
        b -= c;
        d -= c;
        if( d >= e )
        { // kein Prozessor ist zum Ende der Additionen belegt
          d = k + 1;
        }
        else
        {
          d += a; // freie Prozessoren nach dem Ende der Additionen
        }
      }
      // Zeit für verbleibende Multiplikationen auf
      // allen k + 1 Prozessoren
      laufz[m][k] += ((int)(b / ( k + 1 ))) * MULTIPLIKATION;
      b %= k + 1;
      if( b <= d )
      { // restliche Multiplikationen können auf den Prozessoren
        // ausgeführt werden, die die letzten Additionen ausführen
        laufz[m][k] += 1.0 + MULTIPLIKATION;
      }
      else
      { // die Prozessoren, die parallel zu den letzten Additionen
        // schon Multiplikationen ausführen, dominieren die
        // Gesamtlaufzeit
        laufz[m][k] += ( mr_auf + 1 ) * MULTIPLIKATION;
      }
      break;
    }
    if( i != ls )
    { // bereits alles berechnet
      continue;
    }
    // Zeit für Multiplikationen fehlt noch
    laufz[m][k] += ( ( b + k ) / ( k + 1 ) ) * MULTIPLIKATION;
  }
}

// Erzeugt den (r*t)-ten Teil einer Matrixmultiplikation der
// Dimension r * s * t in Komplexität 2
// Es werden die Modulmummern ab *m_nr benutzt und die Laufzeiten
// für n Prozessoren erstellt
void matrix2( int* m_nr, int n, int s, double** laufz )
{
  matrixSingle( *m_nr, n, s, 1, laufz );
  (*m_nr)++;
}

// Erzeugt eine Matrixmultiplikation der Dimension r * s * t
// in Komplexität 1
// Es werden die Modulmummern ab *m_nr benutzt und die Laufzeiten
// für n Prozessoren erstellt
void matrix1( int* m_nr, int n, int r, int s, int t, double** laufz )
{
  matrixSingle( *m_nr, n, s, r * t, laufz );
  (*m_nr)++;
}

// siehe oben
int baue_standard_bsp( double** laufz, int n, graph* gr, char* fn )
{
  FILE* fi;
  char c;
  int i, r, s, t;
  int a, ko, m, m_old;
  fi = fopen( fn, "r" );
  if( fi == NULL )
  {
    return 4;
  }
  if( lese_next_char_r( &fi, "mgf", &c ) != 0 ) // Standardbeispielname
  {
    return 3;
  }
  if( lese_next_int_r( &fi, &ko ) != 0 ) // Komplexität
  {
    return 3;
  }
  // Prozessorenanzahl (von) -> überlesen
  if( lese_next_int( &fi, &a ) != 0 )
  {
    return 3;
  }
  // Prozessorenanzahl (bis) -> überlesen
  if( lese_next_int_r( &fi, &a ) != 0 )
  {
    return 3;
  }
  m = 0;
  switch( ko )
  {
    case 1 : // Komplexität 1
    {
      switch( c )
      {
        case 'm': // Matrixmultiplikation
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &r ) != 0 ) // 1. Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &s ) != 0 ) // 2. Dimension
            {
              return 3;
            }
            if( lese_next_int_in_line( &fi, &t ) != 0 ) // 3. Dimension
            {
              return 3;
            }
            if( lese_next_int_in_line( &fi, &a ) != 0 ) // Vielfachheit
            {
              a = 1; // keine Angabe => einfach
            }
            m_old = m;
            matrix1( &m, n, r, s, t, laufz );
            while( a > 1 )
            {
              a--;
              for( i=0; i<n; i++ )
              {
                laufz[m][i] = laufz[m_old][i];
              }
              m++;
            }
          }
          break;
        }
        case 'f' : // FFT
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &r ) != 0 ) // Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &a ) != 0 ) // Vielfachheit
            {
              a = 1; // keine Angabe => einfach
            }
            m_old = m;
            if( 0 != fft1( &m, n, r, laufz ) )
            {
              return 1;
            }
            while( a > 1 )
            {
              a--;
              for( i=0; i<n; i++ )
              {
                laufz[m][i] = laufz[m_old][i];
              }
              m++;
            }
          }
          break;
        }
        case 'g' : // LR - Zerlegung mittels Gauß
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &r ) != 0 ) // Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &a ) != 0 ) // Vielfachheit
            {
              a = 1; // keine Angabe => einfach
            }
            m_old = m;
            if( 0 != gauss1( &m, n, r, laufz ) )
            {
              return 1;
            }
            while( a > 1 )
            {
              a--;
              for( i=0; i<n; i++ )
              {
                laufz[m][i] = laufz[m_old][i];
              }
              m++;
            }
          }
          break;
        }
      }
      // Graph aus m Knoten erstellen
      if( create_knoten_graph( gr, m ) != 0 )
      {
        return 1;
      }
      break;
    }
    case 2 : // Komplexität 2
    {
      switch( c )
      {
        case 'm' : // Matrixmultiplikation
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &r ) != 0 ) // 1. Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &s ) != 0 ) // 2. Dimension
            {
              return 3;
            }
            if( lese_next_int_in_line( &fi, &t ) != 0 ) // 3. Dimension
            {
              return 3;
            }
            if( lese_next_int_in_line( &fi, &a ) != 0 ) // Vielfachheit
            {
              // keine Angabe => einfach; aber r*t Bäume
              a = r * t;
            }
            else
            {
              a *= r * t; // r*t Bäume
            }
            m_old = m;
            matrix2( &m, n, s, laufz );
            while( a > 1 )
            {
              a--;
              for( i=0; i<n; i++ )
              {
                laufz[m][i] = laufz[m_old][i];
              }
              m++;
            }
          }
          if( create_knoten_graph( gr, m ) != 0 )
          {
            return 1;
          }
          break;
        }
        case 'f' : // FFT
        {
          del_graph( gr );
          while( 1 )
          {
            if( lese_next_int( &fi, &r ) != 0 ) // Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &a ) != 0 ) // Vielfachheit
            {
              a = 1; // keine Angabe => einfach
            }
            a *= 2; // eine FFT zerfällt in 2 Bäume
            while( a > 0 )
            {
              a--;
              if( 0 != fft2( &m, n, r, gr, laufz ) )
              {
                return 1;
              }
            }
          }
          break;
        }
        case 'g' : // LR-Zerlegung nach Gauß
        {
          del_graph( gr );
          while( 1 )
          {
            if( lese_next_int( &fi, &r ) != 0 ) // Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &a ) != 0 ) // Vielfachheit
            {
              a = 1; // keine Angabe => einfach
            }
            while( a > 0 )
            {
              a--;
              if( 0 != gauss2( &m, n, r, gr, laufz ) )
              {
                return 1;
              }
            }
          }
          break;
        }
      }
      break;
    }
    case 3 : // Komplexität 3
    {
      del_graph( gr );
      switch( c )
      {
        case 'm' : // Matrixmultiplikation
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &r ) != 0 ) // 1. Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &s ) != 0 ) // 2. Dimension
            {
              return 3;
            }
            if( lese_next_int_in_line( &fi, &t ) != 0 ) // 3. Dimension
            {
              return 3;
            }
            if( lese_next_int_in_line( &fi, &a ) != 0 ) // Vielfachheit
            {
              a = r * t; // keine Angabe => einfach; aber r*t Bäume
            }
            else
            {
              a *= r * t; // je r*t Bäume
            }
            while( a > 0 )
            {
              a--;
              if( 0 != matrix3( &m, n, s, gr, laufz ) )
              {
                return 0;
              }
            }
          }
          break;
        }
        case 'f' : // FFT
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &r ) != 0 ) // Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &a ) != 0 ) // Vielfachheit
            {
              a = 1; // keine Angabe => einfach
            }
            while( a > 0 )
            {
              a--;
              if( 0 != fft3( &m, n, r, gr, laufz ) )
              {
                return 1;
              }
            }
          }
          break;
        }
        case 'g' : // LR - Zerlegung mittels Gauß
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &r ) != 0 ) // Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &a ) != 0 ) // Vielfachheit
            {
              a = 1; // keine Angabe => einfach
            }
            while( a > 0 )
            {
              a--;
              if( 0 != gauss3( &m, n, r, gr, laufz ) )
              {
                return 1;
              }
            }
          }
          break;
        }
      }
      break;
    }
  }
  fclose( fi );
  if( m == 0 )
  {
    return 3; // nichts erzeugt
  }
  return 0;
}

// bestimmt die Größe eines Standardbeispiels
// Fehler: 1->Standarddatei nicht gefunden
//         2->Standarddatei fehlerhaft
int bsp_groesse( int* m, int* na, int* ne, char* fn )
{
  FILE* fi;
  char c;
  int h, k;
  int i;
  fi = fopen( fn, "r" );
  if( fi == NULL )
  {
    return 1;
  }
  if( lese_next_char_r( &fi, "mgf", &c ) != 0 ) // Standardbeispielname
  {
    return 2;
  }
  if( lese_next_int_r( &fi, &k ) != 0 ) // Komplexität
  {
    return 2;
  }
  if( lese_next_int( &fi, na ) != 0 ) // Prozessorenanzahl (von)
  {
    return 2;
  }
  // Prozessorenanzahl (bis)
  if( lese_next_int_in_line_r( &fi, ne ) != 0 )
  {
    return 2;
  }
  *m = 0;
  switch( k )
  {
    case 1 : // Komplexität 1
    {
      if( c == 'm' ) // Matrixmultiplikation verlangt 3 Parameter
      {
        while( 1 )
        {
          if( lese_next_int( &fi, &h ) != 0 ) // 1. Dimension
          {
            break;
          }
          if( lese_next_int_in_line( &fi, &h ) != 0 ) // 2. Dimension
          {
            return 2;
          }
          if( lese_next_int_in_line( &fi, &h ) != 0 ) // 3. Dimension
          {
            return 2;
          }
          if( lese_next_int_in_line( &fi, &h ) != 0 ) // Vielfachheit
          {
            (*m)++; // keine Angabe => einfach
          }
          else
          {
            (*m) += h;
          }
        }
      }
      else
      {
        while( 1 )
        {
          if( lese_next_int( &fi, &h ) != 0 ) // Dimension
          {
            break;
          }
          if( lese_next_int_in_line( &fi, &h ) != 0 ) // Vielfachheit
          {
            (*m)++; // keine Angabe => einfach
          }
          else
          {
            (*m) += h;
          }
        }
      }
      break;
    }
    case 2 : // Komplexität 2
    {
      switch( c )
      {
        case 'm' : // Matrixmultiplikation
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &i ) != 0 ) // 1. Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &h ) != 0 ) // 2. Dimension
            {
              return 2;
            }
            if( lese_next_int_in_line( &fi, &h ) != 0 ) // 3. Dimension
            {
              return 2;
            }
            if( lese_next_int_in_line( &fi, &k ) != 0 ) // Vielfachheit
            {
              (*m) += i * h; // keine Angabe => einfach
            }
            else
            {
              (*m) += i * h * k;
            }
          }
          break;
        }
        case 'f' : // FFT
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &h ) != 0 ) // Dimension
            {
              break;
            }
            i = h - 1;
            if( lese_next_int_in_line( &fi, &h ) != 0 ) // Vielfachheit
            {
              (*m) += 2 * i; // keine Angabe => einfach
            }
            else
            {
              (*m) += 2 * i * h;
            }
          }
          break;
        }
        case 'g' : // LR-Zerlegung nach Gauß
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &h ) != 0 ) // Dimension
            {
              break;
            }
            i = ( h * ( h + 1 ) ) / 2;
            if( lese_next_int_in_line( &fi, &h ) != 0 ) // Vielfachheit
            {
              (*m) += i; // keine Angabe => einfach
            }
            else
            {
              (*m) += i * h;
            }
          }
          break;
        }
      }
    }
    case 3 : // Komplexität 3
    {
      switch( c )
      {
        case 'm' : // Matrixmultiplikation
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &k ) != 0 ) // 1. Dimension
            {
              break;
            }
            if( lese_next_int_in_line( &fi, &h ) != 0 ) // 2. Dimension
            {
              return 2;
            }
            k *= 2 * h - 1;
            if( lese_next_int_in_line( &fi, &h ) != 0 ) // 3. Dimension
            {
              return 2;
            }
            k *= h;
            if( lese_next_int_in_line( &fi, &h ) != 0 ) // Vielfachheit
            {
              (*m) += k; // keine Angabe => einfach
            }
            else
            {
              (*m) += h * k;
            }
          }
          break;
        }
        case 'f' : // FFT
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &k ) != 0 ) // Dimension
            {
              break;
            }
            i = auf_log2( k );
            i = 5 * k * i - 3 * k;
            if( lese_next_int_in_line( &fi, &h ) != 0 ) // Vielfachheit
            {
              (*m) += i; // keine Angabe => einfach
            }
            else
            {
              (*m) += h * i;
            }
          }
          break;
        }
        case 'g' : // LR-Zerlegung mittels Gauß
        {
          while( 1 )
          {
            if( lese_next_int( &fi, &k ) != 0 ) // Dimension
            {
              break;
            }
            // Knotenanzahl
            k = (k-1) * k * (2*k-1) / 3 + (k-1) * k / 2 + k * k;
            if( lese_next_int_in_line( &fi, &h ) != 0 ) // Vielfachheit
            {
              (*m) += k; // keine Angabe => einfach
            }
            else
            {
              (*m) += k * h;
            }
          }
          break;
        }
      }
      break;
    }
  }
  fclose( fi );
  if( m == 0 )
  {
    return 2; // nichts erzeugt
  }
  return 0;
}
