#include <limits.h>
#include <malloc.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "dgraphen.h"
#include "load.h"
#include "save.h"
#include "zufall.h"

// Konstruktor - Ersatz
void init_graph( graph* g )
{
  g->max = g->akt = 0;
  g->kanakt = g->knoakt = 0;
  g->dfsakt = 0;
  g->kanmax = INT_MAX;
  g->knomax = 10;
  g->knopuf = NULL;
  g->kanpuf = NULL;
  g->k = NULL;
}

// löscht (ohne Pufferung) alle mit a an einem Knoten hängenden Kanten;
// a muß die in der Liste erste Kante des Knotens sein
// interne Hilfsroutine
// übergebener Zeiger muß anschließend noch auf Null gesetzt werden
void free_graph_help( kante* a )
{
  kante* b;
  while( a != NULL )
  {
    b = a;
    a = a->next;
    free( b );
  }
}

// Destruktor - Ersatz; Graph ist dann wieder initialisiert,
// aber Puffer sind leer
void free_graph( graph* g )
{
  int i;
  for( i=0; i<g->akt; i++ )
  {
    if( (g->k)[i] != NULL )
    {
      free_graph_help( ((g->k)[i])->backward );
      free_graph_help( ((g->k)[i])->forward );
      free( (g->k)[i] );
    }
  }
  free( g->k );
  g->max = g->akt = 0;
  g->k = NULL;
  free_graph_help( g->kanpuf );
  g->kanpuf = NULL;
  g->kanakt = 0;
  if( g->knopuf != NULL )
  {
    for( i=0; i<g->knoakt; i++ )
    {
      free( (g->knopuf)[i] );
    }
    free( g->knopuf );
    g->knopuf = NULL;
    g->knoakt = 0;
  }
  if( g->dfsakt != 0 )
  {
    free( g->dfs );
    g->dfsakt = 0;
  }
}

// interne Hilfsroutine; löscht eine Kante (gepuffertes free)
void del_help_kante( graph* g, kante* k )
{
  if( g->kanakt != g->kanmax )
  {
    k->next = g->kanpuf;
    g->kanpuf = k;
    (g->kanakt)++;
    return;
  }
  free( k );
}

// interne Hilfsroutine; löscht eine Kantenverweiskette
// (gepuffertes free)
// der übergebene Zeiger k muß anschließend noch auf NULL gesetzt werden
void del_help_kantenverweise( graph* g, kante* k )
{
  kante* a;
  while( k != NULL )
  {
    a = k;
    k = k->next;
    del_help_kante( g, a );
  }
}

// interne Hilfsroutine; löscht den Knoten k (gepuffertes free)
// Verweise auf Kinder und Vorgänger müssen zuvor gelöscht sein
void del_help_knoten( graph* g, knoten* k )
{
  if( g->knoakt != g->knomax )
  {
    if( g->knopuf == NULL )
    {
      g->knopuf = (knoten* *)malloc( g->knomax * sizeof( knoten* ) );
      if( g->knopuf == NULL ) // zu wenig Speicher um Puffer anzulegen
      {
        free( k );
        return;
      }
    }
    (g->knopuf)[g->knoakt] = k;
    (g->knoakt)++;
    return;
  }
  free( k );
}

// löscht den Graph (ohne Puffer zu leeren)
void del_graph( graph* g )
{
  int i;
  for( i=0; i<g->akt; i++ )
  {
    if( (g->k)[i] != NULL )
    {
      del_help_kantenverweise( g, ((g->k)[i])->forward );
      del_help_kantenverweise( g, ((g->k)[i])->backward );
      del_help_knoten( g, (g->k)[i] );
    }
  }
  g->akt = 0;
}

// interne Hilfsroutine; reserviert Speicher für einen Knoten
// Fehler: NULL->zu wenig Speicher
knoten* neu_help_knoten( graph* g )
{
  if( g->knoakt != 0 )
  {
    (g->knoakt)--;
    return (g->knopuf)[g->knoakt];
  }
  return (knoten*)malloc( sizeof( knoten ) );
}

// interne Hilfsroutine; reserviert Speicher für eine Kante
kante* neu_help_kante( graph* g )
{
  kante* k;
  if( g->kanakt != 0 )
  {
    k = g->kanpuf;
    g->kanpuf = k->next;
    (g->kanakt)--;
    return k;
  }
  return (kante*)malloc( sizeof( kante ) );
}

// interne Hilfsroutine zum Anlegen des DFS-Feldes der Größe g->akt
// mehr gibt die maximale zu erwartende Größe an
// die aufrufende Routine verwendet die Werte g->dfsmax+1
// bis einschließlich g->dfsmax+diff im DFS-Feld
// durch die Verwendung immer neuer Werte baucht das Feld nicht
// nach jeder Benutzung neu initialisiert werden
// Rückgabe: 1->kein Speicher
int init_dfs_feld( graph* g, int diff, int mehr )
{
  int i;
  if( mehr < g->akt )
  {
    mehr = g->akt;
  }
  if( mehr == 0 )
  {
    mehr = 2;
  }
  if( g->dfsakt == 0 ) // noch nie benötigt
  {
    g->dfs = (int*)malloc( mehr * sizeof( int ) );
    if( g->dfs == NULL )
    {
      return 1;
    }
    g->dfsakt = mehr;
    for( mehr--; mehr>=0; mehr-- )
    {
      (g->dfs)[mehr] = 0;
    }
    g->dfsmax = 0;
    g->dfsdiff = diff;
    return 0;
  }
  g->dfsmax += g->dfsdiff;
  g->dfsdiff = diff;
  if( g->dfsakt < g->akt )
  {
    g->dfs = (int*)realloc( g->dfs, mehr * sizeof( int ) );
    if( g->dfs == NULL )
    {
      return 1;
    }
    if( INT_MAX - diff >= g->dfsmax )
    {
      for( i=g->dfsakt; i<mehr; i++ )
      {
        (g->dfs)[i] = 0;
      }
    }
    else
    {
      for( i=0; i<mehr; i++ )
      {
        (g->dfs)[i] = 0;
      }
      g->dfsmax = 0;
    }
    g->dfsakt = mehr;
    return 0;
  }
  if( INT_MAX - diff >= g->dfsmax )
  {
    return 0;
  }
  for( i=0; i<g->dfsakt; i++ )
  {
    (g->dfs)[i] = 0;
  }
  g->dfsmax = 0;
  return 0;
}

// interne Hilfsroutine zum Vergrößern des DFS-Feldes nach einer
// Vergrößerung des Graphen
// nur nach Aufruf von init_dfs_feld starten
// mehr gibt die maximale zu erwartende Größe an
// die aufrufende Routine kann die Werte g->dfsmax+1 bis
// g->dfsmax+diff weiter verwenden
// Fehler: 1->kein Speicher
int vergroessere_dfs_feld( graph* g, int mehr )
{
  int i;
  if( mehr < g->akt )
  {
    mehr = g->akt;
  }
  if( 2 * g->dfsakt > mehr )
  {
    mehr = 2 * g->dfsakt;
  }
  if( g->dfsakt < g->akt )
  {
    g->dfs = (int*)realloc( g->dfs, mehr * sizeof( int ) );
    if( g->dfs == NULL )
    {
      return 1;
    }
    for( i=g->dfsakt; i<mehr; i++ )
    {
      (g->dfs)[i] = 0;
    }
    g->dfsakt = mehr;
    return 0;
  }
  return 0;
}

// erzeugt einen Graphen der nur aus den Knoten 1 bis n besteht
// wenn g hier vergrößert wird, so g->max = mehr wählen,
// um nachfolgende Operationen optimal ausführen zu können
// Fehler: 1->zu wenig Speicher
int create_knoten_graph_mehr( graph* g, int n, int mehr )
{
  int i;
  if( mehr < n )
  {
    mehr = n;
  }
  if( g->max < n )
  {
    if( mehr < 2 * (g->max) )
    {
      mehr = 2 * (g->max);
    }
    if( g->k == NULL ) // noch kein Knotenfeld vorhanden
    {
      g->k = (knoten**) malloc( mehr * sizeof( knoten* ) );
    }
    else // nur vergrößern
    {
      g->k = (knoten**) realloc( g->k, mehr * sizeof( knoten* ) );
    }
    if( g->k == NULL )
    {
      return 1;
    }
    for( i=0; i<g->akt; i++ )
    {
      if( (g->k)[i] != NULL ) // Knoten existiert bereits
      {
         // alle Kanten löschen
        del_help_kantenverweise( g, ((g->k)[i])->forward );
        ((g->k)[i])->forward = NULL;
        del_help_kantenverweise( g, ((g->k)[i])->backward );
        ((g->k)[i])->backward = NULL;
      }
      else // Knoten nicht vorhanden
      {
        (g->k)[i] = neu_help_knoten( g );
        if( (g->k)[i] == NULL )
        {
          return 1;
        }
        ((g->k)[i])->forward = NULL;
        ((g->k)[i])->backward = NULL;
      }
      ((g->k)[i])->knotIdx = i;
    }
    for( ; i<n; i++ )
    {
      (g->k)[i] = neu_help_knoten( g );
      if( (g->k)[i] == NULL )
      {
        return 1;
      }
      ((g->k)[i])->forward = NULL;
      ((g->k)[i])->backward = NULL;
      ((g->k)[i])->knotIdx = i;
    }
    g->max = mehr;
    g->akt = n;
    return 0;
  }
  // n <= g->max
  while( g->akt > n ) // überflüssige Knoten löschen
  {
    (g->akt)--;
    if( (g->k)[g->akt] != NULL )
    {
      del_help_kantenverweise( g, ((g->k)[g->akt])->forward );
      del_help_kantenverweise( g, ((g->k)[g->akt])->backward );
      del_help_knoten( g, (g->k)[g->akt] );
    }
  }
  for( i=0; i<g->akt; i++ )
  { // vorhandene Knoten behalten, aber Kanten löschen
    if( (g->k)[i] != NULL )
    {
      del_help_kantenverweise( g, ((g->k)[i])->forward );
      ((g->k)[i])->forward = NULL;
      del_help_kantenverweise( g, ((g->k)[i])->backward );
      ((g->k)[i])->backward = NULL;
    }
    else
    {
      (g->k)[i] = neu_help_knoten( g );
      if( (g->k)[i] == NULL )
      {
        return 1;
      }
      ((g->k)[i])->forward = NULL;
      ((g->k)[i])->backward = NULL;
    }
    ((g->k)[i])->knotIdx = i;
  }
  while( g->akt < n ) // fehlende Knoten hinzufügen
  {
    (g->k)[g->akt] = neu_help_knoten( g );
    if( (g->k)[g->akt] == NULL )
    {
      return 1;
    }
    ((g->k)[g->akt])->forward = NULL;
    ((g->k)[g->akt])->backward = NULL;
    ((g->k)[g->akt])->knotIdx = g->akt;
    (g->akt)++;
  }
  return 0;
}

// erzeugt einen Graphen der nur aus den Knoten 1 bis n besteht
// Fehler: 1->zu wenig Speicher
int create_knoten_graph( graph* g, int n )
{
  return create_knoten_graph_mehr( g, n, n );
}

// fügt Knoten ein (sofern er noch nicht vorhanden ist);
// mehr gibt bei einer Vergrößerung die neue gewünschte Obergrenze
// Fehler: 1->zu wenig Speicher
int add_knoten_mehr( graph* g, int i, int mehr )
{
  int j;
  i--;
  if( i < g->akt )
  {
    if( (g->k)[i] != NULL )
    {
      return 0;
    }
    (g->k)[i] = neu_help_knoten( g );
    if( (g->k)[i] == NULL )
    {
      return 1;
    }
    ((g->k)[i])->forward = NULL;
    ((g->k)[i])->backward = NULL;
    ((g->k)[i])->knotIdx = i;
    return 0;
  }
  if( i < g->max )
  {
    for( ; g->akt<i; (g->akt)++ )
    {
      (g->k)[g->akt] = NULL;
    }
    (g->k)[i] = neu_help_knoten( g );
    if( (g->k)[i] == NULL )
    {
      return 1;
    }
    ((g->k)[i])->forward = NULL;
    ((g->k)[i])->backward = NULL;
    ((g->k)[i])->knotIdx = i;
    (g->akt)++;
    return 0;
  }
  if( mehr < 2 * (g->max) )
  {
    mehr = 2 * (g->max);
  }
  if( mehr <= i )
  {
    mehr = i + 1;
  }
  if( g->k == NULL )
  {
    g->k = (knoten**) malloc( mehr * sizeof( knoten* ) );
  }
  else
  {
    g->k = (knoten**) realloc( g->k, mehr * sizeof( knoten* ) );
  }
  if( g->k == NULL )
  {
    return 1;
  }
  for( j=g->akt; j<i; j++ )
  {
    (g->k)[j] = NULL;
  }
  (g->k)[i] = neu_help_knoten( g );
  if( (g->k)[i] == NULL )
  {
    return 1;
  }
  ((g->k)[i])->forward = NULL;
  ((g->k)[i])->backward = NULL;
  ((g->k)[i])->knotIdx = i;
  g->max = mehr;
  g->akt = i+1;
  return 0;
}

// fügt Knoten ein (sofern er noch nicht vorhanden ist)
// Fehler: 1->zu wenig Speicher
int add_knoten( graph* g, int i )
{
  return add_knoten_mehr( g, i, i );
}

// fügt eine Kante von i nach j ein; sollten Knoten i oder j nicht
// existieren, so werden sie erstellt
// kein Test ob Kante schon existiert,
// mehr ist die Vergrößerungsempfehlung
// Rückgabe über v: Verweis auf eingefügte Kante
// Fehler: 1->zu wenig Speicher (*v ist dann auch NULL)
int add_kante_mehr_verweis(
  graph*  g,
  int     i,
  int     j,
  kante** v,
  int     mehr )
{
  kante* w;
  if( i < j )
  {
    if( add_knoten_mehr( g, j, mehr ) != 0 )
    {
      *v = NULL;
      return 1;
    }
    if( add_knoten_mehr( g, i, mehr ) != 0 )
    {
      *v = NULL;
      return 1;
    }
  }
  else
  {
    if( add_knoten_mehr( g, i, mehr ) != 0 )
    {
      *v = NULL;
      return 1;
    }
    if( add_knoten_mehr( g, j, mehr ) != 0 )
    {
      *v = NULL;
      return 1;
    }
  }
  i--;
  j--;
  *v = neu_help_kante( g );
  if( *v == NULL )
  {
    return 1;
  }
  w = neu_help_kante( g );
  if( w == NULL )
  {
    *v = NULL;
    return 1;
  }
  (*v)->doppel = w;
  w->doppel = *v;
  (*v)->next = ((g->k)[i])->forward;
  if( (*v)->next != NULL )
  {
    (*v)->next->prev = *v;
  }
  (*v)->prev = NULL;
  (*v)->knot = (g->k)[j];
  ((g->k)[i])->forward = *v;
  w->next = ((g->k)[j])->backward;
  if( w->next != NULL )
  {
    w->next->prev = w;
  }
  w->prev = NULL;
  w->knot = (g->k)[i];
  ((g->k)[j])->backward = w;
  return 0;
}

// fügt eine Kante von i nach j ein; sollte Knoten i oder j nicht
// existieren, so wird er erstellt
// kein Test ob Kante schon existiert, mehr ist die
// Vergrößerungsempfehlung
// Fehler: 1->zu wenig Speicher
int add_kante_mehr( graph* g, int i, int j, int mehr )
{
  kante* v;
  return add_kante_mehr_verweis( g, i, j, &v, mehr );
}

// fügt eine Kante von i nach j ein; sollte Knoten i oder j nicht
// existieren, so wird er erstellt
// kein Test ob Kante schon existiert
// Fehler: 1->zu wenig Speicher
int add_kante( graph* g, int i, int j )
{
  kante* v;
  return add_kante_mehr_verweis( g, i, j, &v, j );
}

// lädt ein Graph g aus Datei fn und reserviert entsprechend Speicher
// (mehr steht als Empfehlung für die maximale
//  Knotenanzahl->Laufzeitminimierung)
// Dateiaufbau: der erste Eintrag in jeder Zeile stellt die aktuelle
// Knotennummer >= 1 dar; alle Knoten von denen eine Kante beim
// aktuellen Knoten eingehen folgen in der Zeile
// !! save_graph_rueck und anschließend load_graph_rueck ergibt nicht
//    die gleiche Reihenfolge in den Adjazenzlisten
// Fehler: 1->Datei nicht gefunden
//         2->zu wenig Speicher
int load_graph_rueck_mehr( char* fn, graph* g, int mehr )
{
  FILE* fi;
  int q, r;
  del_graph( g );
  fi = fopen( fn, "r" );
  if( fi == NULL )
  {
    return 1;
  }
  while( 1 )
  {
    if( lese_next_int( &fi, &q ) == 1 )
    {
      fclose( fi );
      return 0;
    }
    if( add_knoten_mehr( g, q, mehr ) != 0 )
    {
      return 2;
    }
    while( 1 )
    {
      if( lese_next_int_in_line( &fi, &r ) != 0 )
      {
        break;
      }
      if( add_kante_mehr( g, r, q, mehr ) != 0 )
      {
        return 2;
      }
    }
  }
}

// lädt ein Graph g aus Datei fn und reserviert entsprechend Speicher
// Dateiaufbau: der erste Eintrag in jeder Zeile stellt die aktuelle
// Knotennummer >= 1 dar; alle Knoten von denen eine Kante beim
// aktuellen Knoten eingehen folgen in der Zeile
// !! save_graph_rueck und anschließend load_graph_rueck ergibt
//    nicht die gleiche Reihenfolge in den Adjazenzlisten
// Fehler: 1->Datei nicht gefunden
//         2->zu wenig Speicher
int load_graph_rueck( char* fn, graph* g )
{
  return load_graph_rueck_mehr( fn, g, 0 );
}

// speichert Graph; hinter jedem Knoten stehen seine Vorgänger
// !! save_graph_rueck und anschließend load_graph_rueck ergibt
//    nicht die gleiche Reihenfolge in den Adjazenzlisten
// Fehler: 1->kann Datei nicht anlegen / Schreibfehler
int save_graph_rueck( graph* g, char* fn )
{
  int i;
  FILE* fi;
  kante* a;
  fi = fopen( fn, "w" );
  if( fi == NULL )
  {
    return 1;
  }
  for( i=0; i<g->akt; i++ )
  {
    if( (g->k)[i] != NULL )
    {
      if( save_int_s( &fi, i+1 ) != 0 )
      {
        return 1;
      }
      a = ((g->k)[i])->backward; // Vorgänger durchgehen
      while( a != NULL )
      {
        if( save_int_s( &fi, a->knot->knotIdx + 1 ) != 0 )
        {
          return 1;
        }
        a = a->next;
      }
      if( save_ret( &fi ) != 0 )
      {
        return 1;
      }
    }
  }
  fclose( fi );
  return 0;
}

// testet ob Knoten i existiert
// Rückgabe: 0->Knoten existiert
//           1->Knoten existiert nicht
int test_knoten( graph* g, int i )
{
  if( g->akt < i )
  {
    return 1;
  }
  i--;
  if( (g->k)[i] == NULL )
  {
    return 1;
  }
  return 0;
}

// testet ob eine Kante von i nach j geht
// Rückgabe über v: gefundene Kante
// Rückgabe: 0->Kante existiert
//           1->Startknoten i existiert nicht (*v ist NULL)
//           2->Endknoten j existiert nicht (*v ist NULL)
//           3->Knoten i und j existieren aber keine Kante (*v ist NULL)
int test_kante_verweis( graph* g, int i, int j, kante** v )
{
  if( test_knoten( g, i ) == 1 )
  {
    *v = NULL;
    return 1;
  }
  if( test_knoten( g, j ) == 1 )
  {
    *v = NULL;
    return 2;
  }
  i--;
  j--;
  *v = ((g->k)[i])->forward;
  while( *v != NULL )
  {
    if( (*v)->knot->knotIdx == j )
    {
      return 0;
    }
    *v = (*v)->next;
  }
  return 3;
}

// testet ob eine Kante von i nach j geht
// Rückgabe: 0->Kante existiert
//           1->Startknoten i existiert nicht
//           2->Endknoten j existiert nicht
//           3->Knoten i und j existieren aber keine Kante
int test_kante( graph* g, int i, int j )
{
  kante* v;
  return test_kante_verweis( g, i, j, &v );
}

// Vorbereitungsroutine zum Einfügen von mehr als einer Kante nach i
// Test ob Kante schon existiert, mehr ist die Vergrößerungsempfehlung
// !! zw. dem Initialisieren und dem Einfügen dürfen keine anderen
//    Operationen auf g gemacht werden
// Fehler: 1->zu wenig Speicher
int add_kanten_test_nach_init_mehr( graph* g, int i, int mehr )
{
  kante* a;
  if( init_dfs_feld( g, 1, mehr ) != 0 )
  {
    return 1;
  }
  a = ((g->k)[i-1])->backward;
  while( a != NULL )
  {
    (g->dfs)[ a->knot->knotIdx ] = g->dfsmax + 1;
    a = a->next;
  }
  return 0;
}

// fügt eine Kante von i nach j ein; sollte Knoten i oder j nicht
// existieren, so wird er erstellt
// Test ob Kante schon existiert, mehr ist die Vergrößerungsempfehlung
// !! zw. dem Initialisieren und dem Einfügen dürfen keine anderen
//    Operationen auf g gemacht werden
// Fehler: 1->zu wenig Speicher
int add_kanten_test_nach_mehr( graph* g, int i, int j, int mehr )
{
  if( i > g->akt )
  {
    if( add_kante_mehr( g, i, j, mehr ) != 0 )
    {
      return 1;
    }
    if( vergroessere_dfs_feld( g, mehr ) != 0 )
    {
      return 1;
    }
    (g->dfs)[i-1] = g->dfsmax + 1;
    return 0;
  }
  if( (g->dfs)[i-1] != g->dfsmax + 1 )
  {
    (g->dfs)[i-1] = g->dfsmax + 1;
    if( add_kante_mehr( g, i, j, mehr ) != 0 )
    {
      return 1;
    }
    return 0;
  }
  return 0;
}

// Vorbereitungsroutine zum Einfügen von mehr als einer Kante von i
// zu anderen Knoten
// Test ob Kante schon existiert, mehr ist die Vergrößerungsempfehlung
// !! zw. dem Initialisieren und dem Einfügen dürfen keine anderen
//    Operationen auf g gemacht werden
// Fehler: 1->zu wenig Speicher
int add_kanten_test_von_init_mehr( graph* g, int i, int mehr )
{
  kante* a;
  if( init_dfs_feld( g, 1, mehr ) != 0 )
  {
    return 1;
  }
  a = ((g->k)[i-1])->forward;
  while( a != NULL )
  {
    (g->dfs)[ a->knot->knotIdx ] = g->dfsmax + 1;
    a = a->next;
  }
  return 0;
}

// fügt eine Kante von i nach j ein; sollte Knoten i oder j nicht
// existieren, so wird er erstellt
// Test ob Kante schon existiert, mehr ist die Vergrößerungsempfehlung
// !! zw. dem Initialisieren und dem Einfügen dürfen keine anderen
//    Operationen auf g gemacht werden
// Fehler: 1->zu wenig Speicher
int add_kanten_test_von_mehr( graph* g, int i, int j, int mehr )
{
  if( j > g->akt )
  {
    if( add_kante_mehr( g, i, j, mehr ) != 0 )
    {
      return 1;
    }
    if( vergroessere_dfs_feld( g, mehr ) != 0 )
    {
      return 1;
    }
    (g->dfs)[j-1] = g->dfsmax + 1;
    return 0;
  }
  if( (g->dfs)[j-1] != g->dfsmax + 1 )
  {
    (g->dfs)[j-1] = g->dfsmax + 1;
    if( add_kante_mehr( g, i, j, mehr ) != 0 )
    {
      return 1;
    }
    return 0;
  }
  return 0;
}

// liefert den Startknoten zu einer Kante
int start_knoten( kante* v )
{
  return v->doppel->knot->knotIdx + 1;
}

// liefert den Endknoten zu einer Kante
int end_knoten( kante* v )
{
  return v->knot->knotIdx + 1;
}

// löscht Kante von i nach j
// Rückgabe: 0->Kante gelöscht
//           1->Anfangsknoten i nicht vorhanden
//           2->Endknoten j nicht vorhanden
//           3->keine Kante zwischen i und j
int del_kante( graph* g, int i, int j )
{
  kante* v;
  kante* w;
  if( test_knoten( g, i ) != 0 )
  {
    return 1;
  }
  if( test_knoten( g, j ) != 0 )
  {
    return 2;
  }
  i--;
  j--;
  v = ((g->k)[i])->forward;
  if( v == NULL )
  {
    return 3;
  }
  while( v->knot->knotIdx != j )
  {
    v = v->next;
    if( v == NULL )
    {
      return 3;
    }
  }
  if( v->prev != NULL )
  {
    v->prev->next = v->next;
  }
  else
  {
    ((g->k)[i])->forward = v->next;
  }
  if( v->next != NULL )
  {
    v->next->prev = v->prev;
  }
  w = v->doppel;
  del_help_kante( g, v );
  if( w->prev != NULL )
  {
    w->prev->next = w->next;
  }
  else
  {
    ((g->k)[j])->backward = w->next;
  }
  if( w->next != NULL )
  {
    w->next->prev = w->prev;
  }
  del_help_kante( g, w );
  return 0;
}

// löscht übergebene Kante
void del_kante_verweis( graph* g, kante* v )
{
  kante* w;
  w = v->doppel;
  if( v->prev != NULL )
  {
    v->prev->next = v->next;
  }
  else
  {
    ((g->k)[w->knot->knotIdx])->forward = v->next;
  }
  if( v->next != NULL )
  {
    v->next->prev = v->prev;
  }
  if( w->prev != NULL )
  {
    w->prev->next = w->next;
  }
  else
  {
    ((g->k)[v->knot->knotIdx])->backward = w->next;
  }
  if( w->next != NULL )
  {
    w->next->prev = w->prev;
  }
  del_help_kante( g, v );
  del_help_kante( g, w );
}

// löscht einen Knoten und alle seine Kanten
// Rückgabe: 0->Knoten gelöscht
//           1->Knoten nicht vorhanden
int del_knoten( graph* g, int i )
{
  kante* v;
  kante* w;
  if( test_knoten( g, i ) != 0 )
  {
    return 1;
  }
  i--;
  v = ((g->k)[i])->forward;
  while( v != NULL )
  {
    w = v->doppel;
    if( w->prev == NULL )
    {
      v->knot->backward = w->next;
    }
    else
    {
      w->prev->next = w->next;
    }
    if( w->next != NULL )
    {
      w->next->prev = w->prev;
    }
    del_help_kante( g, w );
    w = v;
    v = v->next;
    del_help_kante( g, w );
  }
  v = ((g->k)[i])->backward;
  while( v != NULL )
  {
    w = v->doppel;
    if( w->prev == NULL )
    {
      v->knot->forward = w->next;
    }
    else
    {
      w->prev->next = w->next;
    }
    if( w->next != NULL )
    {
      w->next->prev = w->prev;
    }
    del_help_kante( g, w );
    w = v;
    v = v->next;
    del_help_kante( g, w );
  }
  del_help_knoten( g, (g->k)[i] );
  if( g->akt == i + 1 )
  {
    g->akt--;
    while( g->akt > 0 )
    {
      if( (g->k)[ g->akt - 1 ] != NULL )
      {
        break;
      }
      (g->akt)--;
    }
    return 0;
  }
  (g->k)[i] = NULL;
  return 0;
}

// gibt den Knoten mit der Nummer i die Nummer j
// mehr ist die Vergrößerungsempfehlung
// Fehler: 1->zu wenig Speicher
//         2->Knoten i existiert nicht
//         3->Knoten j existiert schon
int move_knoten_mehr( graph* g, int i, int j, int mehr )
{
  if( test_knoten( g, i ) != 0 )
  {
    return 2;
  }
  if( test_knoten( g, j ) == 0 )
  {
    return 3;
  }
  if( g->max < j )
  {
    if( add_knoten_mehr( g, j, mehr ) != 0 )
    {
      return 1;
    }
    del_help_knoten( g, (g->k)[j-1] );
  }
  else
  {
    for( ; g->akt<j-1; (g->akt)++ )
    {
      (g->k)[ g->akt ] = NULL;
    }
    if( g->akt < j )
    {
      g->akt = j;
    }
  }
  (g->k)[j-1] = (g->k)[i-1];
  (g->k)[i-1] = NULL;
  ((g->k)[j-1])->knotIdx = j-1;
  return 0;
}

// liefert eine Obergrenze für die Knotenanzahl von g
int max_knoten( graph* g )
{
  return g->akt;
}

// testet ob i ein Quellknoten ist
// Rückgabe: 0->i ist Quelle
//           1->i ist keine Quelle
//           2->i ist nicht vorhanden
int quell_knoten( graph* g, int i )
{
  if( test_knoten( g, i ) != 0 )
  {
    return 2;
  }
  if( ((g->k)[i-1])->backward == NULL )
  {
    return 0;
  }
  return 1;
}

// testet ob i ein Zielknoten (Senke) ist
// Rückgabe: 0->i ist Senke
//           1->i ist keine Senke
//           2->i ist nicht vorhanden
int ziel_knoten( graph* g, int i )
{
  if( test_knoten( g, i ) != 0 )
  {
    return 2;
  }
  if( ((g->k)[i-1])->forward == NULL )
  {
    return 0;
  }
  return 1;
}

// vor Verwendung von next_kind / get_kind starten
// Rückgabe: 1->Knoten i nicht vorhanden
int reset_kind( graph* g, int i )
{
  if( test_knoten( g, i ) != 0 )
  {
    return 1;
  }
  i--;
  ((g->k)[i])->currForward = ((g->k)[i])->forward;
  return 0;
}

// gibt das nächste Kind des Knotens i; nach jeder Änderung
// des Graphen an diesem Knoten erst reset_kind starten!
// Rückgabe: 0->kein weiteres Kind vorhanden; Knoten i nicht vorhanden
int next_kind( graph* g, int i )
{
  int j;
  if( test_knoten( g, i ) != 0 )
  {
    return 0;
  }
  i--;
  if( ((g->k)[i])->currForward == NULL )
  {
    return 0;
  }
  j = ((g->k)[i])->currForward->knot->knotIdx + 1;
  ((g->k)[i])->currForward = ((g->k)[i])->currForward->next;
  return j;
}

// vor Verwendung von next_vorg / get_vorg starten
// Rückgabe: 1->Knoten i nicht vorhanden
int reset_vorg( graph* g, int i )
{
  if( test_knoten( g, i ) != 0 )
  {
    return 1;
  }
  i--;
  ((g->k)[i])->currBackward = ((g->k)[i])->backward;
  return 0;
}

// gibt den nächsten Vorgänger des Knotens i; nach jeder Änderung
// des Graphen an diesem Knoten erst reset_vorg starten!
// Rückgabe: 0->kein weiterer Vorgänger vorhanden;
//              Knoten i nicht vorhanden
int next_vorg( graph* g, int i )
{
  int j;
  if( test_knoten( g, i ) != 0 )
  {
    return 0;
  }
  i--;
  if( ((g->k)[i])->currBackward == NULL )
  {
    return 0;
  }
  j = ((g->k)[i])->currBackward->knot->knotIdx + 1;
  ((g->k)[i])->currBackward = ((g->k)[i])->currBackward->next;
  return j;
}

// liefert den Eingangsgrad von Knoten i
// Rückgabe: -1->Knoten nicht vorhanden
int eing_grad( graph* g, int i )
{
  kante* a;
  if( test_knoten( g, i ) != 0 )
  {
    return -1;
  }
  a = ((g->k)[i-1])->backward;
  i = 0;
  while( a != NULL )
  {
    i++;
    a = a->next;
  }
  return i;
}

// liefert den Ausgangsgrad von Knoten i
// Rückgabe: -1->Knoten nicht vorhanden
int ausg_grad( graph* g, int i )
{
  kante* a;
  if( test_knoten( g, i ) != 0 )
  {
    return -1;
  }
  a = ((g->k)[i-1])->forward;
  i = 0;
  while( a != NULL )
  {
    i++;
    a = a->next;
  }
  return i;
}

// testet, ob der Eingangsgrad von Knoten i gleich x ist
// Rückgabe:  0->Eingangsgrad ist x
//            1->Eingangsgrad ist nicht x
//            2->Knoten nicht vorhanden
int vgl_eing_grad( graph* g, int i, int x )
{
  kante* a;
  if( test_knoten( g, i ) != 0 )
  {
    return 2;
  }
  a = ((g->k)[i-1])->backward;
  while( ( a != NULL ) && ( x != 0 ) )
  {
    a = a->next;
    x--;
  }
  if( ( a == NULL ) && ( x == 0 ) )
  {
    return 0;
  }
  return 1;
}

// testet, ob der Ausgangsgrad von Knoten i gleich x ist
// Rückgabe:  0->Ausgangsgrad ist x
//            1->Ausgangsgrad ist nicht x
//            2->Knoten nicht vorhanden
int vgl_ausg_grad( graph* g, int i, int x )
{
  kante* a;
  if( test_knoten( g, i ) != 0 )
  {
    return 2;
  }
  a = ((g->k)[i-1])->forward;
  while( ( a != NULL ) && ( x != 0 ) )
  {
    a = a->next;
    x--;
  }
  if( ( a == NULL ) && ( x == 0 ) )
  {
    return 0;
  }
  return 1;
}

// erzeugt einen zufälligen Graphen mit knot Knoten und kanten Kanten
// es gibt nur Kanten von i nach i+1, ..., i+seq
// sollte Kantenanzahl größer als die Anzahl der möglichen
// Kanten sein, wird sie reduziert
// Fehler: 1->zu wenig Speicher
int rnd_graph( graph* g, int kanten, int seq, int knot, int mehr )
{
  int v, w, i, j;
  char * kantFlag;
  if( create_knoten_graph_mehr( g, knot, mehr ) != 0 )
  {
    return 1;
  }
  if( seq > knot )
  {
    seq = knot;
  }
  // Anzahl an Kanten aus Knoten mit Ausgangsgrad seq
  v = ( knot - seq ) * seq;
  // maximale Kantenanzahl berechnen
  w = v + ( seq * ( seq - 1 ) ) / 2;
  if( kanten > 99 * (w / 100) )
  {
    // alle möglichen Kanten einfügen
    for( i=1; i<=knot-seq; i++ )
    {
      for( j=i+1; j<=i+seq; j++ )
      {
        if( add_kante( g, i, j ) != 0 )
        {
          return 1;
        }
      }
    }
    for( ; i<knot; i++ )
    {
      for( j=i+1; j<=knot; j++ )
      {
        if( add_kante( g, i, j ) != 0 )
        {
          return 1;
        }
      }
    }
    // zuviel eingefügte Kanten löschen
    while( kanten < w )
    {
      i = gleich_int( 0, w-1 );
      if( i < v )
      {
        j = i % seq;
        i = i / seq;
        if( del_kante( g, i+1, i+j+2 ) == 0 )
        {
          kanten++;
        }
      }
      else
      {
        i -= v;
        j = (int)(0.5 + ((double)(seq)) -
                  sqrt( 0.25 - ((double)(seq)) +
                        ((double)(seq)) * ((double)(seq)) -
                        2 * ((double)(i)) ));
        i += ( ( j - 1 ) * j ) / 2 - ( j - 1 ) * seq;
        if( del_kante( g, knot-seq+j, knot-seq+j+i+1 ) == 0 )
        {
          kanten++;
        }
      }
    }
  }
  else
  {
    kantFlag = (char*) malloc( sizeof( char ) * knot * knot );
    if( NULL == kantFlag )
    {
      return 1;
    }
    memset( kantFlag, 0x00, sizeof( char ) * knot * knot );
    while( kanten > 0 )
    {
      do
      {
        i = gleich_int( 0, w-1 );
      } while( 0 != kantFlag[i] );
      kantFlag[i] = 1;
      if( i < v )
      {
        j = i % seq;
        i = i / seq;
        if( add_kante_mehr( g, i+1, i+j+2, i+j+2 ) != 0 )
        {
          return 1;
        }
      }
      else
      {
        i -= v;
        j = (int)(0.5 + ((double)(seq)) -
                  sqrt( 0.25 - ((double)(seq)) +
                        ((double)(seq)) * ((double)(seq)) -
                        2 * ((double)(i)) ));
        i += ( ( j - 1 ) * j ) / 2 - ( j - 1 ) * seq;
        if( add_kante_mehr( g, knot-seq+j, knot-seq+j+i+1,
                            knot-seq+j+i+1 ) != 0 )
        {
          return 1;
        }
      }
      kanten--;
    }
  }
  return 0;
}

// kopiert g nach h; h muß initialisiert sein !!
// mehr gibt die Mindestanzahl an möglichen Knoten in h an
// Fehler: 1->zu wenig Speicher
int copy_graph_mehr( graph* g, graph* h, int mehr )
{
  int i;
  kante* k;
  i = g->akt;
  // Grundgerüst anlegen
  if( create_knoten_graph_mehr( h, i, mehr ) != 0 )
  {
    return 1;
  }
  while( i > 0 )
  {
    if( test_knoten( g, i ) == 0 )
    {
      k = ((g->k)[i-1])->forward;
      while( k != NULL )
      {
        if( add_kante_mehr( h, i, k->knot->knotIdx + 1, mehr ) != 0 )
        {
          return 1;
        }
        k = k->next;
      }
    }
    else
    {
      del_knoten( h, i ); // überflüssiger Knoten
    }
    i--;
  }
  return 0;
}

// rekursive Hilfsroutine von transitiv_kante (DFS-Lauf)
void transitiv_kante_aus_help( graph* g, int i )
{
  kante* a;
  if( ( (g->dfs)[i] == g->dfsmax + 1 ) ||
      ( (g->dfs)[i] == g->dfsmax + 3 ) ) // Knoten schon besucht
  {
    return;
  }
  if( (g->dfs)[i] == g->dfsmax + 2 ) // transitive Kante gefunden
  {
    // damit sie nicht noch mal gefunden wird
    (g->dfs)[i] = g->dfsmax + 3;
    return;
  }
  (g->dfs)[i] = g->dfsmax + 1;
  a = ((g->k)[i])->forward;
  while( a != NULL )
  {
    transitiv_kante_aus_help( g, a->knot->knotIdx );
    a = a->next;
  }
}

// testet, ob eine aus i ausgehende Kante transitiv ist
// Kante (i,j) ist transitiv, wenn es einen Pfad von i nach j gibt
// !! nur kreisfreie Graphen
// mehr gibt die Mindestanzahl an möglichen Knoten in g an
// Rückgabe: trans enthält alle transitiven Kanten
// Fehler: 1->zu wenig Speicher
//         2->Knoten i existiert nicht
int transitiv_kante_aus_mehr(
  graph*       g,
  int          i,
  kantenliste* trans,
  int          mehr )
{
  kante* a;
  kante* b;
  int k;
  if( test_knoten( g, i ) != 0 )
  {
    return 2;
  }
  if( init_dfs_feld( g, 3, mehr ) != 0 )
  {
    return 1;
  }
  a = ((g->k)[i-1])->forward;
  k = 0;
  while( a != NULL )
  {
    k++;
    (g->dfs)[ a->knot->knotIdx ] = g->dfsmax + 2;
    a = a->next;
  }
  del_all_kl( trans );
  if( k < 2 )
  {
    return 0;
  }
  a = ((g->k)[i-1])->forward;
  while( a != NULL )
  {
    b = a->knot->forward;
    while( b != NULL )
    {
      transitiv_kante_aus_help( g, b->knot->knotIdx );
      b = b->next;
    }
    a = a->next;
  }
  a = ((g->k)[i-1])->forward;
  while( a != NULL )
  {
    if( (g->dfs)[ a->knot->knotIdx ] == g->dfsmax + 3 )
    {
      if( add_anf_kl( trans, a ) != 0 )
      {
        return 1;
      }
    }
    a = a->next;
  }
  return 0;
}

// rekursive Hilfsroutine von transitiv_kante_ein (DFS-Lauf)
void transitiv_kante_ein_help( graph* g, int i )
{
  kante* a;
  if( ( (g->dfs)[i] == g->dfsmax + 1 ) ||
      ( (g->dfs)[i] == g->dfsmax + 3 ) ) // Knoten schon besucht
  {
    return;
  }
  if( (g->dfs)[i] == g->dfsmax + 2 ) // transitive Kante gefunden
  {
    // damit sie nicht noch einmal aufgenommen wird
    (g->dfs)[i] = g->dfsmax + 3;
    return;
  }
  (g->dfs)[i] = g->dfsmax + 1;
  a = ((g->k)[i])->backward;
  while( a != NULL )
  {
    transitiv_kante_ein_help( g, a->knot->knotIdx );
    a = a->next;
  }
  return;
}

// testet, ob eine in i eingehende Kante transitiv ist
// Kante (i,j) ist transitiv, wenn es einen Pfad von i nach j gibt
// !! nur kreisfreie Graphen
// mehr gibt die Mindestanzahl an möglichen Knoten in g an
// Rückgabe: trans enthält alle transitiven Kanten
// Fehler: 1->zu wenig Speicher
//         2->Knoten i existiert nicht
int transitiv_kante_ein_mehr(
  graph*       g,
  int          i,
  kantenliste* trans,
  int          mehr )
{
  kante* a;
  kante* b;
  int k;
  if( test_knoten( g, i ) != 0 )
  {
    return 2;
  }
  if( init_dfs_feld( g, 3, mehr ) != 0 )
  {
    return 1;
  }
  a = ((g->k)[i-1])->backward;
  k = 0;
  while( a != NULL )
  {
    k++;
    (g->dfs)[ a->knot->knotIdx ] = g->dfsmax + 2;
    a = a->next;
  }
  del_all_kl( trans );
  if( k < 2 )
  {
    return 0;
  }
  a = ((g->k)[i-1])->backward;
  while( a != NULL )
  {
    b = a->knot->backward;
    while( b != NULL )
    {
      transitiv_kante_ein_help( g, b->knot->knotIdx );
      b = b->next;
    }
    a = a->next;
  }
  a = ((g->k)[i-1])->backward;
  while( a != NULL )
  {
    if( (g->dfs)[ a->knot->knotIdx ] == g->dfsmax + 3 )
    {
      if( add_anf_kl( trans, a->doppel ) != 0 )
      {
        return 1;
      }
    }
    a = a->next;
  }
  return 0;
}

// Hilfsroutine zu transitiv_kante_ueber
// markiert alle Knoten, von denen ein Pfad zu i+1 führt
void transitiv_kante_ueber_help1( graph* g, int i )
{
  kante* a;
  a = ((g->k)[i])->backward;
  while( a != NULL )
  {
    if( (g->dfs)[ a->knot->knotIdx ] != g->dfsmax + 1 )
    {
      (g->dfs)[ a->knot->knotIdx ] = g->dfsmax + 1;
      transitiv_kante_ueber_help1( g, a->knot->knotIdx );
    }
    a = a->next;
  }
}

// Hilfsroutine zu transitiv_kante_ueber
// testet alle Knoten, zu denen ein Pfad von i+1 führt, ob sie einen
// markierten Vorgänger haben; wenn ja wird die Kante nach trans
// geschrieben
// Fehler: 1->zu wenig Speicher
int transitiv_kante_ueber_help2( graph* g, int i, kantenliste* trans )
{
  kante* a;
  kante* b;
  a = ((g->k)[i])->forward;
  while( a != NULL )
  {
    if( (g->dfs)[ a->knot->knotIdx ] != g->dfsmax + 2 )
    {
      (g->dfs)[ a->knot->knotIdx ] = g->dfsmax + 2;
      b = a->knot->backward;
      while( b != NULL )
      {
        if( (g->dfs)[ b->knot->knotIdx ] == g->dfsmax + 1 )
        {
          if( add_anf_kl( trans, b->doppel ) != 0 )
          {
            return 1;
          }
        }
        b = b->next;
      }
      if( transitiv_kante_ueber_help2( g, a->knot->knotIdx,
                                       trans ) != 0 )
      {
        return 1;
      }
    }
    a = a->next;
  }
  return 0;
}

// schreibt Kante (j, k) nach trans, wenn es einen Pfad von j nach k
// über i gibt
// mehr ist die Verbreiterungsempfehlung
// !! nur kreisfreie Graphen
// Fehler: 1->zu wenig Speicher
//         2->Knoten i existiert nicht
int transitiv_kante_ueber_mehr(
  graph*       g,
  int          i,
  kantenliste* trans,
  int          mehr )
{
  if( test_knoten( g, i ) != 0 )
  {
    return 2;
  }
  if( init_dfs_feld( g, 2, mehr ) != 0 )
  {
    return 1;
  }
  // alle Knoten, von denen ein Pfad zu i führt, markieren
  transitiv_kante_ueber_help1( g, i-1 );
  // über alle Knoten, zu denen ein Pfad von i führt,
  // laufen und testen ob sie einen markierten Vorgänger haben
  return transitiv_kante_ueber_help2( g, i-1, trans );
}

// rekursive Hilfsroutine zu get_mitte_pfad
// sucht Knoten j+1 von i+1 aus
int get_mitte_pfad_help(
  graph* g,
  int    i,
  int    j,
  int    tief,
  int*   tief_max )
{
  kante* a;
  int k;
  if( i == j )
  {
    *tief_max = tief;
    return -1;
  }
  if( (g->dfs)[i] == g->dfsmax + 1 )
  {
    return -1;
  }
  (g->dfs)[i] = g->dfsmax + 1;
  a = ((g->k)[i])->forward;
  while( a != NULL )
  {
    k = get_mitte_pfad_help( g, a->knot->knotIdx, j, tief+1, tief_max );
    if( *tief_max / 2 == tief )
    {
      return i;
    }
    if( *tief_max != -2 ) // j gefunden -> Suche abbrechen
    {
      return k;
    }
    a = a->next;
  }
  return -1;
}

// Rückgabe: 0 wenn kein Pfad von i nach j existiert
//           k>0 es gibt einen Pfad von i nach k und einen von k nach j;
//               der Pfad von i nach k ist höchstens einen Knoten
//               kürzer als der von k nach j; ist k = i, so gibt es
//               eine Kante von i nach j oder es ist i = j
// mehr gibt die Mindestanzahl an möglichen Knoten in g an
// Fehler: -1->zu wenig Speicher
//         -2->Knoten i existiert nicht
//         -3->Knoten j existiert nicht
int get_mitte_pfad_mehr( graph* g, int i, int j, int mehr )
{
  int tief_max;
  if( test_knoten( g, i ) != 0 )
  {
    return -2;
  }
  if( test_knoten( g, j ) != 0 )
  {
    return -3;
  }
  if( init_dfs_feld( g, 1, mehr ) != 0 )
  {
    return -1;
  }
  tief_max = -2;
  j = get_mitte_pfad_help( g, i-1, j-1, 0, &tief_max );
  if( tief_max / 2 == 0 )
  {
    return i;
  }
  return j+1;
}

//----------------------------------------------------------------------

// Initialisierung einer kantenliste
void init_kl( kantenliste* x )
{
  x->next = x->last = NULL;
  x->maxpuf = INT_MAX;
  x->aktpuf = 0;
  x->puf = NULL;
}

// löscht kantenliste, ist anschließend wieder initialisiert, aber
// Puffer sind leer
void free_kl( kantenliste* x )
{
  kl_element* v;
  while( x->next != NULL )
  {
    v = x->next;
    x->next = x->next->next;
    free( v );
  }
  while( x->puf != NULL )
  {
    v = x->puf;
    x->puf = x->puf->next;
    free( v );
  }
  x->last = NULL;
  x->aktpuf = 0;
}

// interne Hilfsroutine; liefert neues kl_element (gepuffert)
// Fehler: NULL->zu wenig Speicher
kl_element* help_neu_kl_element( kantenliste* x )
{
  kl_element* v;
  if( x->aktpuf == 0 )
  {
    return (kl_element*)malloc( sizeof( kl_element ) );
  }
  v = x->puf;
  x->puf = v->next;
  (x->aktpuf)--;
  return v;
}

// interne Hilfsroutine; löscht kl_element v (gepuffert)
void help_del_kl_element( kantenliste* x, kl_element* v )
{
  if( x->aktpuf == x->maxpuf )
  {
    free( v );
    return;
  }
  v->next = x->puf;
  x->puf = v;
  (x->aktpuf)++;
}

// löscht alle Elemente von x, jedoch nicht den Puffer
void del_all_kl( kantenliste* x )
{
  while( x->next != NULL )
  {
    x->last = x->next;
    x->next = x->next->next;
    help_del_kl_element( x, x->last );
  }
  x->last = NULL;
}

// fügt ein Element am Anfang ein
// Fehler: 1->zu wenig Speicher
int add_anf_kl( kantenliste* x, kante* v )
{
  kl_element* a;
  a = help_neu_kl_element( x );
  if( a == NULL )
  {
    return 1;
  }
  a->v = v;
  if( x->next == NULL )
  {
    a->next = NULL;
    x->next = x->last = a;
    return 0;
  }
  a->next = x->next;
  x->next = a;
  return 0;
}

// löscht das erste Element von x und gibt es zurück
// Fehler: NULL->Liste ist leer
kante* del_anf_kl( kantenliste* x )
{
  kl_element* a;
  kante* v;
  if( x->next == NULL )
  {
    return NULL;
  }
  a = x->next;
  v = a->v;
  x->next = a->next;
  if( x->next == NULL )
  {
    x->last = NULL;
  }
  help_del_kl_element( x, a );
  return v;
}
