#ifndef DGRAPHEN_OP

#define DGRAPHEN_OP

// IM VERGLEICH ZU GRAPHEN GEÄNDERTE ROUTINEN:
// add_kante; test_kante; add_kante_test (weitere Versionen)
// add_kanten_test_von; add_kanten_test_next (weitere Versionen)
// start_knoten; end_knoten (neu)
// Struktur "kantenliste" (neu)
//
// aus "knotenliste" wurde "kante"
//
// im Gegensatz zu graphen ist das Löschen von Kanten und Knoten
// schneller; auch können Kanten als Element behandelt werden

#ifdef GRAPHEN_OP
#error "Entweder graphen.h oder dgraphen.h verwenden."
#endif

typedef struct kante
{
  struct knoten* knot; // Endknoten der Kante
  struct kante* doppel; // Partnerkante mit Gegenrichtung
  struct kante* next; // nächste Kante mit gleichem Startknoten
  struct kante* prev; // vorherige Kante mit gleichem Startknoten
} kante;

typedef struct knoten
{
  int knotIdx; // Knotennummer
  kante* forward; // Kanten zu Nachfolgerknoten
  kante* backward; // Kanten zu Vorgängerknoten
  kante* currForward; // nächster zu bearbeitender Nachfolgerknoten
  kante* currBackward; // nächster zu bearbeitender Vorgängerknoten
} knoten;

typedef struct graph
{
  knoten* *k; // Verweis auf einzelne Knoten (Feld)
  int max; // aktuelle Dimension der Feldes k
  int akt; // Anzahl der belegten Felder (größtes Feld)
  int* dfs; // Feld zum Speichern von Besucht-Informationen bei
  //           DFS-Läufen
  int dfsakt; // 0  DFS-Feld nicht angelegt;
  //             >0 max. Größe des DFS-Feldes
  int dfsmax; // größter im DFS-Feld gespeicherter Zahlenwert
  int dfsdiff; // genutzte Zahlenwerte bei der letzten Verwendung
  //              (dfsmax+1 bis dfsmax+dfsdiff)
  kante* kanpuf; // Puffer für Kanten
  int kanmax; // Maximalgröße des Kantenpuffers
  int kanakt; // aktueller Füllstand des Kantenpuffers
  knoten* *knopuf; // Puffer für Knoten
  int knomax; // Maximalgröße des Knotenpuffers
  int knoakt; // aktueller Füllstand des Knotenpuffers
} graph;

typedef struct kl_element
{
  kante* v; // Verweis auf Kante
  struct kl_element* next; // nächstes Element
} kl_element;

typedef struct kantenliste
{
  kl_element* next; // erstes Element
  kl_element* last; // letztes Element
  int maxpuf; // maximale Puffergröße
  int aktpuf; // aktueller Pufferfüllstand
  kl_element* puf; // Puffer
} kantenliste;

// Konstruktor - Ersatz
void init_graph( graph* g );

// Destruktor - Ersatz; Graph ist dann wieder initialisiert;
// Puffer sind leer
void free_graph( graph* g );

// löscht den Graph (ohne Puffer zu leeren)
void del_graph( graph* g );

// lädt ein Graph g aus Datei fn und reserviert entsprechend Speicher
// (mehr steht als Empfehlung für die maximale
//  Knotenanzahl->Laufzeitminimierung)
// Dateiaufbau: der erste Eintrag in jeder Zeile stellt die
// aktuelle Knotennummer >= 1 dar; alle Knoten von denen eine
// Kante beim aktuellen Knoten eingehen folgen in der Zeile
// !! save_graph_rueck und anschließend load_graph_rueck ergibt
//    nicht die gleiche Reihenfolge in den Adjazenzlisten
// Fehler: 1->Datei nicht gefunden
//         2->zu wenig Speicher
int load_graph_rueck_mehr( char* fn, graph* g, int mehr );
int load_graph_rueck( char* fn, graph* g );

// speichert Graph; hinter jedem Knoten stehen seine Vorgänger
// !! save_graph_rueck und anschließend load_graph_rueck ergibt
//    nicht die gleiche Reihenfolge in den Adjazenzlisten
// Fehler: 1->kann Datei nicht anlegen / Schreibfehler
int save_graph_rueck( graph* g, char* fn );

// erzeugt einen Graphen der nur aus den Knoten 1 bis n besteht
// (wenn g hier vergrößert wird, so g->max = mehr wählen, um
//  nachfolgende Operationen optimal ausführen zu können)
// Fehler: 1->zu wenig Speicher
int create_knoten_graph_mehr( graph* g, int n, int mehr );
int create_knoten_graph( graph* g, int n );

// fügt Knoten ein (sofern er noch nicht vorhanden ist);
// mehr gibt bei einer Vergrößerung die neue gewünschte Obergrenze
// Fehler: 1->zu wenig Speicher
int add_knoten_mehr( graph* g, int i, int mehr );
int add_knoten( graph* g, int i );

// fügt eine Kante von i nach j ein; sollte Knoten i oder j
// nicht existieren, so wird er erstellt
// kein Test ob Kante schon existiert,
// mehr ist die Vergrößerungsempfehlung
// Rückgabe über v: Verweis auf eingefügte Kante
// Fehler: 1->zu wenig Speicher (*v ist dann auch NULL)
int add_kante_mehr( graph* g, int i, int j, int mehr );
int add_kante( graph* g, int i, int j );
int add_kante_mehr_verweis(
  graph*  g,
  int     i,
  int     j,
  kante** v,
  int     mehr );

// Vorbereitungsroutine zum Einfügen von mehr als einer Kante nach i
// Test ob Kante schon existiert, mehr ist die Vergrößerungsempfehlung
// !! zw. dem Initialisieren und dem Einfügen dürfen keine anderen
//    Operationen auf g gemacht werden
// Fehler: 1->zu wenig Speicher
int add_kanten_test_nach_init_mehr( graph* g, int i, int mehr );

// fügt eine Kante von i nach j ein; sollte Knoten i oder j nicht
// existieren, so wird er erstellt
// Test ob Kante schon existiert, mehr ist die Vergrößerungsempfehlung
// !! zw. dem Initialisieren und dem Einfügen dürfen keine anderen
//    Operationen auf g gemacht werden
// Rückgabe über v: eingefügte Kante; NULL->Kante existiert schon
// Fehler: 1->zu wenig Speicher
int add_kanten_test_nach_mehr( graph* g, int i, int j, int mehr );

// Vorbereitungsroutine zum Einfügen von mehr als einer Kante
// von i zu anderen Knoten
// Test ob Kante schon existiert, mehr ist die Vergrößerungsempfehlung
// !! zw. dem Initialisieren und dem Einfügen dürfen keine anderen
//    Operationen auf g gemacht werden
// Fehler: 1->zu wenig Speicher
int add_kanten_test_von_init_mehr( graph* g, int i, int mehr );

// fügt eine Kante von i nach j ein; sollte Knoten i oder j nicht
// existieren, so wird er erstellt
// Test ob Kante schon existiert, mehr ist die Vergrößerungsempfehlung
// !! zw. dem Initialisieren und dem Einfügen dürfen keine anderen
//    Operationen auf g gemacht werden
// Rückgabe über v: eingefügte Kante; NULL->Kante existiert schon
// Fehler: 1->zu wenig Speicher
int add_kanten_test_von_mehr( graph* g, int i, int j, int mehr );

// testet ob eine Kante von i nach j geht
// Rückgabe über v: gefundene Kante
// Rückgabe: 0->Kante existiert
//           1->Startknoten i existiert nicht (*v ist NULL)
//           2->Endknoten j existiert nicht (*v ist NULL)
//           3->Knoten i und j existieren aber keine Kante (*v ist NULL)
int test_kante( graph* g, int i, int j );
int test_kante_verweis( graph* g, int i, int j, kante** v );

// testet ob Knoten i existiert
// Rückgabe: 0->Knoten existiert
//           1->Knoten existiert nicht
int test_knoten( graph* g, int i );

// liefert den Startknoten zu einer Kante
int start_knoten( kante* v );

// liefert den Endknoten zu einer Kante
int end_knoten( kante* v );

// löscht Kante von i nach j
// Rückgabe: 0->Kante gelöscht
//           1->Anfangsknoten i nicht vorhanden
//           2->Endknoten j nicht vorhanden
//           3->keine Kante zwischen i und j
int del_kante( graph* g, int i, int j );

// löscht übergebene Kante
void del_kante_verweis( graph* g, kante* v );

// löscht einen Knoten und alle seine Kanten
// Rückgabe: 0->Knoten gelöscht
//           1->Knoten nicht vorhanden
int del_knoten( graph* g, int i );

// gibt den Knoten mit der Nummer i die Nummer j
// mehr ist die Vergrößerungsempfehlung
// Fehler: 1->zu wenig Speicher
//         2->Knoten i existiert nicht
//         3->Knoten j existiert schon
int move_knoten_mehr( graph* g, int i, int j, int mehr );

// reduziert das Knotenfeld, jedoch nicht unter mehr
// nur Reduzierungen die mindestens 50% des Speichers einsparen
// (Routine zur Speicherfreigabe nach vielen Knotenlöschungen)
// Fehler: 1->zu wenig Speicher

// liefert eine Obergrenze für die Knotenanzahl von g
int max_knoten( graph* g );

// testet ob i ein Quellknoten ist
// Rückgabe: 0->i ist Quelle
//           1->i ist keine Quelle
//           2->i ist nicht vorhanden
int quell_knoten( graph* g, int i );

// testet ob i ein Zielknoten (Senke) ist
// Rückgabe: 0->i ist Senke
//           1->i ist keine Senke
//           2->i ist nicht vorhanden
int ziel_knoten( graph* g, int i );

// liefert den Eingangsgrad von Knoten i
// Rückgabe: -1->Knoten nicht vorhanden
int eing_grad( graph* g, int i );

// liefert den Ausgangsgrad von Knoten i
// Rückgabe: -1->Knoten nicht vorhanden
int ausg_grad( graph* g, int i );

// testet, ob der Eingangsgrad von Knoten i gleich x ist
// Rückgabe:  0->Eingangsgrad ist x
//            1->Eingangsgrad ist nicht x
//            2->Knoten nicht vorhanden
int vgl_eing_grad( graph* g, int i, int x );

// testet, ob der Ausgangsgrad von Knoten i gleich x ist
// Rückgabe:  0->Ausgangsgrad ist x
//            1->Ausgangsgrad ist nicht x
//            2->Knoten nicht vorhanden
int vgl_ausg_grad( graph* g, int i, int x );

// vor Verwendung von next_kind starten
// Rückgabe: 1->Knoten i nicht vorhanden
int reset_kind( graph* g, int i );

// gibt das nächste Kind des Knotens i;
// nach jeder Änderung des Graphen an diesem Knoten erst
// reset_kind starten!
// Rückgabe: 0->kein weiteres Kind vorhanden; Knoten i nicht vorhanden
int next_kind( graph* g, int i );

// vor Verwendung von next_vorg starten
// Rückgabe: 1->Knoten i nicht vorhanden
int reset_vorg( graph* g, int i );

// gibt den nächsten Vorgänger des Knotens i;
// nach jeder Änderung des Graphen an diesem Knoten erst
// reset_vorg starten!
// Rückgabe: 0->kein weiterer Vorgänger vorhanden; Knoten i nicht
//              vorhanden
int next_vorg( graph* g, int i );

// erzeugt einen zufälligen Graphen mit knoten Knoten und kanten Kanten
// es gibt nur Kanten von i nach i+1, ..., i+seq
// sollte Kantenanzahl größer als die Anzahl der möglichen Kanten sein,
// wird sie reduziert
// Fehler: 1->zu wenig Speicher
int rnd_graph( graph* g, int kanten, int seq, int knoten, int mehr );

// kopiert g nach h; h muß initialisiert sein !!
// mehr gibt die Mindestanzahl an möglichen Knoten in h an
// Fehler: 1->zu wenig Speicher
int copy_graph_mehr( graph* g, graph* h, int mehr );

// testet, ob eine aus i ausgehende Kante transitiv ist
// Kante (i,j) ist transitiv, wenn es einen Pfad von i nach j gibt
// !! nur kreisfreie Graphen
// mehr gibt die Mindestanzahl an möglichen Knoten in g an
// Rückgabe: trans enthält alle transitiven Kanten
// Fehler: 1->zu wenig Speicher
//         2->Knoten i existiert nicht
int transitiv_kante_aus_mehr(
  graph*       g,
  int          i,
  kantenliste* trans,
  int          mehr );

// testet, ob eine in i eingehende Kante transitiv ist
// Kante (i,j) ist transitiv, wenn es einen Pfad von i nach j gibt
// !! nur kreisfreie Graphen
// mehr gibt die Mindestanzahl an möglichen Knoten in g an
// Rückgabe: trans enthält alle transitiven Kanten
// Fehler: 1->zu wenig Speicher
//         2->Knoten i existiert nicht
int transitiv_kante_ein_mehr(
  graph*       g,
  int          i,
  kantenliste* trans,
  int          mehr );

// schreibt Kante (j, k) nach trans, wenn es einen Pfad von j nach k
// über i gibt
// mehr ist die Verbreiterungsempfehlung
// !! nur kreisfreie Graphen
// Fehler: 1->zu wenig Speicher
//         2->Knoten i existiert nicht
int transitiv_kante_ueber_mehr(
  graph*       g,
  int          i,
  kantenliste* trans,
  int          mehr );

// Rückgabe: 0 wenn kein Pfad von i nach j existiert
//           k>0 es gibt einen Pfad von i nach k und einen von k nach j;
//               der Pfad von i nach k ist höchstens einen Knoten kürzer
//               als der von k nach j; ist k = i, so gibt es eine Kante
//               von i nach j oder es ist i = j
// mehr gibt die Mindestanzahl an möglichen Knoten in g an
// Fehler: -1->zu wenig Speicher
//         -2->Knoten i existiert nicht
//         -3->Knoten j existiert nicht
int get_mitte_pfad_mehr( graph* g, int i, int j, int mehr );

//----------------------------------------------------------------------

// Initialisierung einer kantenliste
void init_kl( kantenliste* x ); // unbegrenzter Puffer

// löscht kantenliste, ist anschließend wieder initialisiert, aber
// Puffer sind leer
void free_kl( kantenliste* x );

// löscht alle Elemente von x, jedoch nicht den Puffer
void del_all_kl( kantenliste* x );

// fügt ein Element am Anfang ein
// Fehler: 1->zu wenig Speicher
int add_anf_kl( kantenliste* x, kante* v );

// löscht das erste Element von x und gibt es zurück
// Fehler: NULL->Liste ist leer
kante* del_anf_kl( kantenliste* x );

#endif
