#include <limits.h>
#include <malloc.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "graphen.h"
#include "listen.h"
#include "load.h"
#include "save.h"
#include "zufall.h"

// Konstruktor - Ersatz
void init_graph( graph* g )
{
  g->max = g->akt = 0;
  g->kanakt = g->knoakt = 0;
  g->dfsakt = 0;
  g->kanmax = INT_MAX;
  g->knomax = 10;
  g->knopuf = NULL;
  g->kanpuf = NULL;
  g->k = NULL;
}

// löscht (ohne Pufferung) alle in v stehenden Knotenverweise;
// interne Hilfsroutine
// übergebener Zeiger muß anschließend noch auf Null gesetzt werden
void free_graph_help( knotenliste* a )
{
  knotenliste* b;
  while( a != NULL )
  {
    b = a;
    a = a->next;
    free( b );
  }
}

// Destruktor - Ersatz; Graph ist dann wieder initialisiert,
// aber Puffer sind leer
void free_graph( graph* g )
{
  int i;
  for( i=0; i<g->akt; i++ )
  {
    if( (g->k)[i] != NULL )
    {
      free_graph_help( ((g->k)[i])->backward );
      free_graph_help( ((g->k)[i])->forward );
      free( (g->k)[i] );
    }
  }
  free( g->k );
  g->max = g->akt = 0;
  g->k = NULL;
  free_graph_help( g->kanpuf );
  g->kanpuf = NULL;
  g->kanakt = 0;
  if( g->knopuf != NULL )
  {
    for( i=0; i<g->knoakt; i++ )
    {
      free( (g->knopuf)[i] );
    }
    free( g->knopuf );
    g->knopuf = NULL;
    g->knoakt = 0;
  }
  if( g->dfsakt != 0 )
  {
    free( g->dfs );
    g->dfsakt = 0;
  }
}

// interne Hilfsroutine;
// löscht ein Knotenlistenelement (gepuffertes free)
void del_help_kante( graph* g, knotenliste* k )
{
  if( g->kanakt != g->kanmax )
  {
    k->next = g->kanpuf;
    g->kanpuf = k;
    (g->kanakt)++;
    return;
  }
  free( k );
}

// interne Hilfsroutine; löscht eine Knotenliste (gepuffertes free)
// der übergebene Zeiger k muß anschließend noch auf NULL gesetzt werden
void del_help_knotenliste( graph* g, knotenliste* k )
{
  knotenliste* a;
  while( k != NULL )
  {
    a = k;
    k = k->next;
    del_help_kante( g, a );
  }
}

// interne Hilfsroutine; löscht den Knoten k (gepuffertes free)
// Verweise auf Kinder und Vorgänger müssen zuvor gelöscht sein
void del_help_knoten( graph* g, knoten* k )
{
  if( g->knoakt != g->knomax )
  {
    if( g->knopuf == NULL )
    {
      g->knopuf = (knoten* *)malloc( g->knomax * sizeof( knoten* ) );
      if( g->knopuf == NULL ) // zu wenig Speicher um Puffer anzulegen
      {
        free( k );
        return;
      }
    }
    (g->knopuf)[g->knoakt] = k;
    (g->knoakt)++;
    return;
  }
  free( k );
}

// löscht den Graph (ohne Puffer zu leeren)
void del_graph( graph* g )
{
  int i;
  for( i=0; i<g->akt; i++ )
  {
    if( (g->k)[i] != NULL )
    {
      del_help_knotenliste( g, ((g->k)[i])->forward );
      del_help_knotenliste( g, ((g->k)[i])->backward );
      del_help_knoten( g, (g->k)[i] );
    }
  }
  g->akt = 0;
}

// interne Hilfsroutine; reserviert Speicher für einen Knoten
// Fehler: NULL->zu wenig Speicher
knoten* neu_help_knoten( graph* g )
{
  if( g->knoakt != 0 )
  {
    (g->knoakt)--;
    return (g->knopuf)[g->knoakt];
  }
  return (knoten*)malloc( sizeof( knoten ) );
}

// interne Hilfsroutine; reserviert Speicher für eine Kante
knotenliste* neu_help_kante( graph* g )
{
  knotenliste* k;
  if( g->kanakt != 0 )
  {
    k = g->kanpuf;
    g->kanpuf = k->next;
    (g->kanakt)--;
    return k;
  }
  return (knotenliste*)malloc( sizeof( knotenliste ) );
}

// erzeugt einen Graphen der nur aus den Knoten 1 bis n besteht
// wenn g hier vergrößert wird, so g->max = mehr wählen, um
// nachfolgende Operationen optimal ausführen zu können
// Fehler: 1->zu wenig Speicher
int create_knoten_graph_mehr( graph* g, int n, int mehr )
{
  int i;
  if( mehr < n )
  {
    mehr = n;
  }
  if( g->max < n )
  {
    if( mehr < 2 * (g->max) )
    {
      mehr = 2 * (g->max);
    }
    if( g->k == NULL ) // noch kein Knotenfeld vorhanden
    {
      g->k = (knoten**) malloc( mehr * sizeof( knoten* ) );
    }
    else // nur vergrößern
    {
      g->k = (knoten**) realloc( g->k, mehr * sizeof( knoten* ) );
    }
    if( g->k == NULL )
    {
      return 1;
    }
    for( i=0; i<g->akt; i++ )
    {
      if( (g->k)[i] != NULL ) // Knoten existiert bereits
      { // alle Kanten löschen
        del_help_knotenliste( g, ((g->k)[i])->forward );
        ((g->k)[i])->forward = NULL;
        del_help_knotenliste( g, ((g->k)[i])->backward );
        ((g->k)[i])->backward = NULL;
      }
      else // Knoten nicht vorhanden
      {
        (g->k)[i] = neu_help_knoten( g );
        if( (g->k)[i] == NULL )
        {
          return 1;
        }
        ((g->k)[i])->forward = NULL;
        ((g->k)[i])->backward = NULL;
      }
      ((g->k)[i])->knotIdx = i;
    }
    for( ; i<n; i++ )
    {
      (g->k)[i] = neu_help_knoten( g );
      if( (g->k)[i] == NULL )
      {
        return 1;
      }
      ((g->k)[i])->forward = NULL;
      ((g->k)[i])->backward = NULL;
      ((g->k)[i])->knotIdx = i;
    }
    g->max = mehr;
    g->akt = n;
    return 0;
  }
  // n <= g->max
  while( g->akt > n ) // überflüssige Knoten löschen
  {
    (g->akt)--;
    if( (g->k)[g->akt] != NULL )
    {
      del_help_knotenliste( g, ((g->k)[g->akt])->forward );
      del_help_knotenliste( g, ((g->k)[g->akt])->backward );
      del_help_knoten( g, (g->k)[g->akt] );
    }
  }
  for( i=0; i<g->akt; i++ )
  { // vorhandene Knoten behalten, aber Kanten löschen
    if( (g->k)[i] != NULL )
    {
      del_help_knotenliste( g, ((g->k)[i])->forward );
      ((g->k)[i])->forward = NULL;
      del_help_knotenliste( g, ((g->k)[i])->backward );
      ((g->k)[i])->backward = NULL;
    }
    else
    {
      (g->k)[i] = neu_help_knoten( g );
      if( (g->k)[i] == NULL )
      {
        return 1;
      }
      ((g->k)[i])->forward = NULL;
      ((g->k)[i])->backward = NULL;
    }
    ((g->k)[i])->knotIdx = i;
  }
  while( g->akt < n ) // fehlende Knoten hinzufügen
  {
    (g->k)[g->akt] = neu_help_knoten( g );
    if( (g->k)[g->akt] == NULL )
    {
      return 1;
    }
    ((g->k)[g->akt])->forward = NULL;
    ((g->k)[g->akt])->backward = NULL;
    ((g->k)[g->akt])->knotIdx = g->akt;
    (g->akt)++;
  }
  return 0;
}

// erzeugt einen Graphen der nur aus den Knoten 1 bis n besteht
// Fehler: 1->zu wenig Speicher
int create_knoten_graph( graph* g, int n )
{
  return create_knoten_graph_mehr( g, n, n );
}

// fügt Knoten ein (sofern er noch nicht vorhanden ist);
// mehr gibt bei einer Vergrößerung die neue gewünschte Obergrenze
// Fehler: 1->zu wenig Speicher
int add_knoten_mehr( graph* g, int i, int mehr )
{
  int j;
  i--;
  if( i < g->akt )
  {
    if( (g->k)[i] != NULL )
    {
      return 0;
    }
    (g->k)[i] = neu_help_knoten( g );
    if( (g->k)[i] == NULL )
    {
      return 1;
    }
    ((g->k)[i])->forward = NULL;
    ((g->k)[i])->backward = NULL;
    ((g->k)[i])->knotIdx = i;
    return 0;
  }
  if( i < g->max )
  {
    for( ; g->akt<i; (g->akt)++ )
    {
      (g->k)[g->akt] = NULL;
    }
    (g->k)[i] = neu_help_knoten( g );
    if( (g->k)[i] == NULL )
    {
      return 1;
    }
    ((g->k)[i])->forward = NULL;
    ((g->k)[i])->backward = NULL;
    ((g->k)[i])->knotIdx = i;
    (g->akt)++;
    return 0;
  }
  if( mehr < 2 * (g->max) )
  {
    mehr = 2 * (g->max);
  }
  if( mehr <= i )
  {
    mehr = i + 1;
  }
  if( g->k == NULL )
  {
    g->k = (knoten**) malloc( mehr * sizeof( knoten* ) );
  }
  else
  {
    g->k = (knoten**) realloc( g->k, mehr * sizeof( knoten* ) );
  }
  if( g->k == NULL )
  {
    return 1;
  }
  for( j=g->akt; j<i; j++ )
  {
    (g->k)[j] = NULL;
  }
  (g->k)[i] = neu_help_knoten( g );
  if( (g->k)[i] == NULL )
  {
    return 1;
  }
  ((g->k)[i])->forward = NULL;
  ((g->k)[i])->backward = NULL;
  ((g->k)[i])->knotIdx = i;
  g->max = mehr;
  g->akt = i+1;
  return 0;
}

// fügt Knoten ein (sofern er noch nicht vorhanden ist)
// Fehler: 1->zu wenig Speicher
int add_knoten( graph* g, int i )
{
  return add_knoten_mehr( g, i, i );
}

// fügt eine Kante von i nach j ein; sollte Knoten i oder j nicht
// existieren, so wird er erstellt
// kein Test ob Kante schon existiert,
// mehr ist die Vergrößerungsempfehlung
// Fehler: 1->zu wenig Speicher
int add_kante_mehr( graph* g, int i, int j, int mehr )
{
  knotenliste* v;
  if( i < j )
  {
    if( add_knoten_mehr( g, j, mehr ) != 0 )
    {
      return 1;
    }
    if( add_knoten_mehr( g, i, mehr ) != 0 )
    {
      return 1;
    }
  }
  else
  {
    if( add_knoten_mehr( g, i, mehr ) != 0 )
    {
      return 1;
    }
    if( add_knoten_mehr( g, j, mehr ) != 0 )
    {
      return 1;
    }
  }
  i--;
  j--;
  v = neu_help_kante( g );
  if( v == NULL )
  {
    return 1;
  }
  v->next = ((g->k)[i])->forward;
  v->knot = (g->k)[j];
  ((g->k)[i])->forward = v;
  v = neu_help_kante( g );
  if( v == NULL )
  {
    return 1;
  }
  v->next = ((g->k)[j])->backward;
  v->knot = (g->k)[i];
  ((g->k)[j])->backward = v;
  return 0;
}

// fügt eine Kante von i nach j ein; sollte Knoten i oder j nicht
// existieren, so wird er erstellt
// kein Test ob Kante schon existiert
// Fehler: 1->zu wenig Speicher
int add_kante( graph* g, int i, int j )
{
  return add_kante_mehr( g, i, j, j );
}

// lädt ein Graph g aus Datei fn und reserviert entsprechend Speicher
// (mehr steht als Empfehlung für die maximale
//  Knotenanzahl->Laufzeitminimierung)
// Dateiaufbau: der erste Eintrag in jeder Zeile stellt die aktuelle
// Knotennummer >= 1 dar; alle Knoten von denen eine Kante beim
// aktuellen Knoten eingehen folgen in der Zeile
// !! save_graph_rueck und anschließend load_graph_rueck ergibt nicht
//    die gleiche Reihenfolge in den Adjazenzlisten
// Fehler: 1->Datei nicht gefunden
//         2->zu wenig Speicher
int load_graph_rueck_mehr( char* fn, graph* g, int mehr )
{
  FILE* fi;
  int q, r;
  del_graph( g );
  fi = fopen( fn, "r" );
  if( fi == NULL )
  {
    return 1;
  }
  while( 1 )
  {
    if( lese_next_int( &fi, &q ) == 1 )
    {
      fclose( fi );
      return 0;
    }
    if( add_knoten_mehr( g, q, mehr ) != 0 )
    {
      return 2;
    }
    while( 1 )
    {
      if( lese_next_int_in_line( &fi, &r ) != 0 )
      {
        break;
      }
      if( add_kante_mehr( g, r, q, mehr ) != 0 )
      {
        return 2;
      }
    }
  }
}

// lädt ein Graph g aus Datei fn und reserviert entsprechend Speicher
// Dateiaufbau: der erste Eintrag in jeder Zeile stellt die aktuelle
// Knotennummer >= 1 dar; alle Knoten von denen eine Kante beim
// aktuellen Knoten eingehen folgen in der Zeile
// !! save_graph_rueck und anschließend load_graph_rueck ergibt
//    nicht die gleiche Reihenfolge in den Adjazenzlisten
// Fehler: 1->Datei nicht gefunden
//         2->zu wenig Speicher
int load_graph_rueck( char* fn, graph* g )
{
  return load_graph_rueck_mehr( fn, g, 0 );
}

// speichert Graph; hinter jedem Knoten stehen seine Vorgänger
// !! save_graph_rueck und anschließend load_graph_rueck ergibt nicht
//    die gleiche Reihenfolge in den Adjazenzlisten
// Fehler: 1->kann Datei nicht anlegen / Schreibfehler
int save_graph_rueck( graph* g, char* fn )
{
  int i;
  FILE* fi;
  knotenliste* a;
  fi = fopen( fn, "w" );
  if( fi == NULL )
  {
    return 1;
  }
  for( i=0; i<g->akt; i++ )
  {
    if( (g->k)[i] != NULL )
    {
      if( save_int_s( &fi, i+1 ) != 0 )
      {
        return 1;
      }
      a = ((g->k)[i])->backward; // Vorgänger durchgehen
      while( a != NULL )
      {
        if( save_int_s( &fi, a->knot->knotIdx + 1 ) != 0 )
        {
          return 1;
        }
        a = a->next;
      }
      if( save_ret( &fi ) != 0 )
      {
        return 1;
      }
    }
  }
  fclose( fi );
  return 0;
}

// testet ob Knoten i existiert
// Rückgabe: 0->Knoten existiert
//           1->Knoten existiert nicht
int test_knoten( graph* g, int i )
{
  if( g->akt < i )
  {
    return 1;
  }
  i--;
  if( (g->k)[i] == NULL )
  {
    return 1;
  }
  return 0;
}

// testet ob eine Kante von i nach j geht
// Rückgabe: 0->Kante existiert
//           1->Startknoten i existiert nicht
//           2->Endknoten j existiert nicht
//           3->Knoten i und j existieren aber keine Kante
int test_kante( graph* g, int i, int j )
{
  knotenliste* v;
  if( test_knoten( g, i ) == 1 )
  {
    return 1;
  }
  if( test_knoten( g, j ) == 1 )
  {
    return 2;
  }
  i--;
  j--;
  v = ((g->k)[i])->forward;
  while( v != NULL )
  {
    if( v->knot->knotIdx == j )
    {
      return 0;
    }
    v = v->next;
  }
  return 3;
}

// löscht Kante von i nach j
// Rückgabe: 0->Kante gelöscht
//           1->Anfangsknoten i nicht vorhanden
//           2->Endknoten j nicht vorhanden
//           3->keine Kante zwischen i und j
int del_kante( graph* g, int i, int j )
{
  knotenliste* v;
  knotenliste* w;
  if( test_knoten( g, i ) != 0 )
  {
    return 1;
  }
  if( test_knoten( g, j ) != 0 )
  {
    return 2;
  }
  i--;
  j--;
  v = ((g->k)[i])->forward;
  if( v == NULL )
  {
    return 3;
  }
  w = NULL;
  while( v->knot->knotIdx != j )
  {
    w = v;
    v = v->next;
    if( v == NULL )
    {
      return 3;
    }
  }
  if( w != NULL )
  {
    w->next = v->next;
  }
  else
  {
    ((g->k)[i])->forward = v->next;
  }
  del_help_kante( g, v );
  v = ((g->k)[j])->backward;
  w = NULL;
  while( v->knot->knotIdx != i )
  {
    w = v;
    v = v->next;
  }
  if( w != NULL )
  {
    w->next = v->next;
  }
  else
  {
    ((g->k)[j])->backward = v->next;
  }
  del_help_kante( g, v );
  return 0;
}

// löscht einen Knoten und alle seine Kanten
// Rückgabe: 0->Knoten gelöscht
//           1->Knoten nicht vorhanden
int del_knoten( graph* g, int i )
{
  knotenliste* v;
  knotenliste* w;
  knotenliste* x;
  if( test_knoten( g, i ) != 0 )
  {
    return 1;
  }
  i--;
  v = ((g->k)[i])->forward;
  while( v != NULL )
  {
    w = v->knot->backward;
    x = NULL;
    while( w->knot->knotIdx != i )
    {
      x = w;
      w = w->next;
    }
    if( x != NULL )
    {
      x->next = w->next;
    }
    else
    {
      v->knot->backward = w->next;
    }
    del_help_kante( g, w );
    w = v;
    v = v->next;
    del_help_kante( g, w );
  }
  v = ((g->k)[i])->backward;
  while( v != NULL )
  {
    w = v->knot->forward;
    x = NULL;
    while( w->knot->knotIdx != i )
    {
      x = w;
      w = w->next;
    }
    if( x != NULL )
    {
      x->next = w->next;
    }
    else
    {
      v->knot->forward = w->next;
    }
    del_help_kante( g, w );
    w = v;
    v = v->next;
    del_help_kante( g, w );
  }
  del_help_knoten( g, (g->k)[i] );
  if( g->akt == i + 1 )
  {
    g->akt--;
    while( g->akt > 0 )
    {
      if( (g->k)[ g->akt - 1 ] != NULL )
      {
        break;
      }
      (g->akt)--;
    }
    return 0;
  }
  (g->k)[i] = NULL;
  return 0;
}

// liefert eine Obergrenze für die Knotenanzahl von g
int max_knoten( graph* g )
{
  return g->akt;
}

// testet ob i ein Zielknoten (Senke) ist
// Rückgabe: 0->i ist Senke
//           1->i ist keine Senke
//           2->i ist nicht vorhanden
int ziel_knoten( graph* g, int i )
{
  if( test_knoten( g, i ) != 0 )
  {
    return 2;
  }
  if( ((g->k)[i-1])->forward == NULL )
  {
    return 0;
  }
  return 1;
}

// vor Verwendung von next_kind / get_kind starten
// Rückgabe: 1->Knoten i nicht vorhanden
int reset_kind( graph* g, int i )
{
  if( test_knoten( g, i ) != 0 )
  {
    return 1;
  }
  i--;
  ((g->k)[i])->currForward = ((g->k)[i])->forward;
  return 0;
}

// stellt einen zuvor geretteten Zeiger k auf das nächste Kind des
// Knotens i wieder her
// Rückgabe: der aktuelle Zeiger auf das nächste Kind des Knotens i
knotenliste* swap_kind( graph* g, int i, knotenliste* k )
{
  knotenliste* kk;
  i--;
  kk = ((g->k)[i])->currForward;
  ((g->k)[i])->currForward = k;
  return kk;
}

// liefert den aktuellen Zeiger auf das nächste Kind des Knotens i
knotenliste* swap_kind_get( graph* g, int i )
{
  return ((g->k)[i-1])->currForward;
}

// gibt das nächste Kind des Knotens i; nach jeder Änderung des Graphen
// an diesem Knoten erst reset_kind starten!
// Rückgabe: 0 ->kein weiteres Kind vorhanden; Knoten i nicht vorhanden
//           >0->Nummer des Kindes
int next_kind( graph* g, int i )
{
  int j;
  if( test_knoten( g, i ) != 0 )
  {
    return 0;
  }
  i--;
  if( ((g->k)[i])->currForward == NULL )
  {
    return 0;
  }
  j = ((g->k)[i])->currForward->knot->knotIdx + 1;
  ((g->k)[i])->currForward = ((g->k)[i])->currForward->next;
  return j;
}

// vor Verwendung von next_vorg / get_vorg starten
// Rückgabe: 1->Knoten i nicht vorhanden
int reset_vorg( graph* g, int i )
{
  if( test_knoten( g, i ) != 0 )
  {
    return 1;
  }
  i--;
  ((g->k)[i])->currBackward = ((g->k)[i])->backward;
  return 0;
}

// gibt den nächsten Vorgänger des Knotens i;
// nach jeder Änderung des Graphen an diesem Knoten erst reset_vorg
// starten!
// Rückgabe: 0 ->kein weiterer Vorgänger vorhanden;
//               Knoten i nicht vorhanden
//           >0->Nummer des Vorgängers
int next_vorg( graph* g, int i )
{
  int j;
  if( test_knoten( g, i ) != 0 )
  {
    return 0;
  }
  i--;
  if( ((g->k)[i])->currBackward == NULL )
  {
    return 0;
  }
  j = ((g->k)[i])->currBackward->knot->knotIdx + 1;
  ((g->k)[i])->currBackward = ((g->k)[i])->currBackward->next;
  return j;
}

// liefert den Eingangsgrad von Knoten i
// Rückgabe: -1->Knoten nicht vorhanden
int eing_grad( graph* g, int i )
{
  knotenliste* a;
  if( test_knoten( g, i ) != 0 )
  {
    return -1;
  }
  a = ((g->k)[i-1])->backward;
  i = 0;
  while( a != NULL )
  {
    i++;
    a = a->next;
  }
  return i;
}

// liefert den Ausgangsgrad von Knoten i
// Rückgabe: -1->Knoten nicht vorhanden
int ausg_grad( graph* g, int i )
{
  knotenliste* a;
  if( test_knoten( g, i ) != 0 )
  {
    return -1;
  }
  a = ((g->k)[i-1])->forward;
  i = 0;
  while( a != NULL )
  {
    i++;
    a = a->next;
  }
  return i;
}

// erzeugt einen zufälligen Graphen mit knot Knoten und kanten Kanten
// es gibt nur Kanten von i nach i+1, ..., i+seq
// sollte Kantenanzahl größer als die Anzahl der möglichen Kanten sein,
// wird sie reduziert
// Fehler: 1->zu wenig Speicher
int rnd_graph( graph* g, int kanten, int seq, int knot, int mehr )
{
  int v, w, i, j;
  char * kantFlag;
  if( create_knoten_graph_mehr( g, knot, mehr ) != 0 )
  {
    return 1;
  }
  if( seq > knot )
  {
    seq = knot;
  }
  // Anzahl an Kanten aus Knoten mit Ausgangsgrad seq
  v = ( knot - seq ) * seq;
  // maximale Kantenanzahl berechnen
  w = v + ( seq * ( seq - 1 ) ) / 2;
  if( kanten > 99 * (w / 100) )
  {
    // alle möglichen Kanten einfügen
    for( i=1; i<=knot-seq; i++ )
    {
      for( j=i+1; j<=i+seq; j++ )
      {
        if( add_kante( g, i, j ) != 0 )
        {
          return 1;
        }
      }
    }
    for( ; i<knot; i++ )
    {
      for( j=i+1; j<=knot; j++ )
      {
        if( add_kante( g, i, j ) != 0 )
        {
          return 1;
        }
      }
    }
    // zuviel eingefügte Kanten löschen
    while( kanten < w )
    {
      i = gleich_int( 0, w-1 );
      if( i < v )
      {
        j = i % seq;
        i = i / seq;
        if( del_kante( g, i+1, i+j+2 ) == 0 )
        {
          kanten++;
        }
      }
      else
      {
        i -= v;
        j = (int)(0.5 + ((double)(seq)) -
                  sqrt( 0.25 - ((double)(seq)) +
                        ((double)(seq)) * ((double)(seq)) -
                        2 * ((double)(i)) ));
        i += ( ( j - 1 ) * j ) / 2 - ( j - 1 ) * seq;
        if( del_kante( g, knot-seq+j, knot-seq+j+i+1 ) == 0 )
        {
          kanten++;
        }
      }
    }
  }
  else
  {
    kantFlag = (char*) malloc( sizeof( char ) * knot * knot );
    if( NULL == kantFlag )
    {
      return 1;
    }
    memset( kantFlag, 0x00, sizeof( char ) * knot * knot );
    while( kanten > 0 )
    {
      do
      {
        i = gleich_int( 0, w-1 );
      } while( 0 != kantFlag[i] );
      kantFlag[i] = 1;
      if( i < v )
      {
        j = i % seq;
        i = i / seq;
        if( add_kante_mehr( g, i+1, i+j+2, i+j+2 ) != 0 )
        {
          return 1;
        }
      }
      else
      {
        i -= v;
        j = (int)(0.5 + ((double)(seq)) -
                  sqrt( 0.25 - ((double)(seq)) +
                        ((double)(seq)) * ((double)(seq)) -
                        2 * ((double)(i)) ));
        i += ( ( j - 1 ) * j ) / 2 - ( j - 1 ) * seq;
        if( add_kante_mehr( g, knot-seq+j, knot-seq+j+i+1,
                            knot-seq+j+i+1 ) != 0 )
        {
          return 1;
        }
      }
      kanten--;
    }
    free( kantFlag );
  }
  return 0;
}

// Hilfsroutine zu Tiefe
int tiefe_help( graph* g, int* tief_num, int i )
{
  knotenliste* v;
  v = ((g->k)[i-1])->backward;
  while( v != NULL )
  {
    if( tief_num[v->knot->knotIdx + 1] == 0 ) // Schleife ?
    {
      return 1;
    }
    if( tief_num[v->knot->knotIdx + 1] == -1 ) // noch nicht bestimmt
    {
      tief_num[v->knot->knotIdx + 1] = 0;
      if( tiefe_help( g, tief_num, v->knot->knotIdx + 1 ) != 0 )
      {
        return 1;
      }
    }
    if( tief_num[v->knot->knotIdx + 1] > tief_num[i] )
    {
      tief_num[i] = tief_num[v->knot->knotIdx + 1];
    }
    v = v->next;
  }
  (tief_num[i])++;
  if( tief_num[i] > tief_num[0] )
  {
    tief_num[0] = tief_num[i];
  }
  return 0;
}

// bestimmt die Tiefe der einzelnen Knoten;
// Speicher für tief_num (max_knoten+1 Einträge) muß reserviert sein
// in tief_num[0] steht die größte vergebene tief_nummer;
// in tief_num[i] die von Knoten i
// Fehler: 1->Graph enthält Zyklus
int tiefe( graph* g, int* tief_num )
{
  int i, m;
  m = max_knoten( g );
  tief_num[0] = 0;
  for( i=1; i<=m; i++ )
  {
    tief_num[i] = -1;
  }
  i = 1;
  while( i<=m )
  {
    if( tief_num[i] == -1 )
    {
      tief_num[i] = 0; // in Bearbeitung (Erkennung von Schleifen)
      if( test_knoten( g, i ) == 0 )
      {
        if( tiefe_help( g, tief_num, i ) != 0 )
        {
          return 1;
        }
      }
    }
    i++;
  }
  return 0;
}

// Hilfsroutine zu hoehe
int hoehe_help( graph* g, int* hoch_num, int i )
{
  knotenliste* v;
  v = ((g->k)[i-1])->forward;
  while( v != NULL )
  {
    if( hoch_num[v->knot->knotIdx + 1] == 0 ) // Schleife ?
    {
      return 1;
    }
    if( hoch_num[v->knot->knotIdx + 1] == -1 ) // noch nicht bestimmt
    {
      hoch_num[v->knot->knotIdx + 1] = 0;
      if( hoehe_help( g, hoch_num, v->knot->knotIdx + 1 ) != 0 )
      {
        return 1;
      }
    }
    if( hoch_num[v->knot->knotIdx + 1] > hoch_num[i] )
    {
      hoch_num[i] = hoch_num[v->knot->knotIdx + 1];
    }
    v = v->next;
  }
  (hoch_num[i])++;
  if( hoch_num[i] > hoch_num[0] )
  {
    hoch_num[0] = hoch_num[i];
  }
  return 0;
}

// bestimmt die Höhe der einzelnen Knoten, also die Tiefen vom
// Graph mit umgedrehten Kanten; hoch_num[Zielknoten] = 1
// Speicher für hoch_num (max_knoten+1 Einträge) muß reserviert sein
// in hoch_num[0] steht die größte vergebene hoch_nummer;
// in hoch_num[i] die von Knoten i
// Fehler: 1->Graph enthält Zyklus
int hoehe( graph* g, int* hoch_num )
{
  int i, m;
  m = max_knoten( g );
  hoch_num[0] = 0;
  for( i=1; i<=m; i++ )
  {
    hoch_num[i] = -1;
  }
  i = 1;
  while( i<=m )
  {
    if( hoch_num[i] == -1 )
    {
      hoch_num[i] = 0; // in Bearbeitung (Erkennung von Schleifen)
      if( test_knoten( g, i ) == 0 )
      {
        if( hoehe_help( g, hoch_num, i ) != 0 )
        {
          return 1;
        }
      }
    }
    i++;
  }
  return 0;
}
