#include <limits.h>
#include <malloc.h>
#include <stdlib.h>
#include "heap.h"

// initialisiert einen Heap
void init_ih( intheap* x )
{
  x->root = NULL;
  x->pufakt = 0;
  x->pufmax = INT_MAX;
}

// interne Hilfsroutine; liefert ein neues Heapelement
// Fehler: NULL->zu wenig Speicher
ih_element* neu_help_ih( intheap* x )
{
  ih_element* a;
  if( x->pufakt != 0 )
  {
    a = x->puf;
    x->puf = x->puf->left;
    (x->pufakt)--;
    return a;
  }
  return (ih_element*)malloc( sizeof( ih_element ) );
}

// interne Hilfsroutine; löscht ein neues Heapelement (gepuffert)
void frei_help_ih( intheap* x, ih_element* a )
{
  if( x->pufakt != x->pufmax )
  {
    (x->pufakt)++;
    a->left = x->puf;
    x->puf = a;
    return;
  }
  free( a );
}

// interne Hilfsroutine; rekursives Löschen; !! a muß != NULL
void help_free_ih( ih_element* a )
{
  if( a->left != NULL )
  {
    help_free_ih( a->left );
    if( a->right != NULL )
    {
      help_free_ih( a->right );
    }
  }
  free( a );
}

// löscht einen Heap (Destruktor); x ist anschließend initialisiert,
// aber der Puffer ist leer
void free_ih( intheap* x )
{
  ih_element* a;
  // Puffer löschen
  while( x->pufakt != 0 )
  {
    (x->pufakt)--;
    a = x->puf;
    x->puf = x->puf->left;
    free( a );
  }
  if( x->root == NULL )
  {
    return;
  }
  // rekursiv den Heap löschen
  help_free_ih( x->root );
}

// fügt i in den Heap ein
// in der Wurzel steht das Element j mit größtem Wert j
// von allen Heapelementen
// Fehler: 1->zu wenig Speicher
int add_max_ih( intheap* x, int i )
{
  ih_element* a;
  ih_element* b;
  int j;
  a = neu_help_ih( x );
  if( a == NULL )
  {
    return 1;
  }
  a->left = NULL;
  a->right = NULL;
  a->bal = 0;
  if( x->root == NULL ) // Heap bis jetzt leer
  {
    x->root = a;
    a->i = i;
    return 0;
  }
  b = x->root;
  while( b->i >= i ) // weiter unten einfügen
  {
    if( b->bal == 0 )
    {
      // nach links
      b->bal = 1;
      if( b->left == NULL )
      {
        b->left = a;
        a->i = i;
        return 0;
      }
      b = b->left;
      continue;
    }
    // nach rechts
    b->bal = 0;
    if( b->right == NULL )
    {
      b->right = a;
      a->i = i;
      return 0;
    }
    b = b->right;
  }
  // andere Werte nach unten ziehen
  while( 1 )
  {
    j = b->i;
    b->i = i;
    if( b->bal == 0 )
    {
      // nach links
      b->bal = 1;
      if( b->left == NULL )
      {
        b->left = a;
        a->i = j;
        return 0;
      }
      i = j;
      b = b->left;
      continue;
    }
    // nach rechts
    b->bal = 0;
    if( b->right == NULL )
    {
      b->right = a;
      a->i = j;
      return 0;
    }
    i = j;
    b = b->right;
  }
}

// fügt i in den Heap ein
// in der Wurzel steht das Element j mit größtem Wert in f[j-1]
// Fehler: 1->zu wenig Speicher
int add_max_indirect_ih( intheap* x, int i, double* f )
{
  ih_element* a;
  ih_element* b;
  int j;
  a = neu_help_ih( x );
  if( a == NULL )
  {
    return 1;
  }
  a->left = NULL;
  a->right = NULL;
  a->bal = 0;
  if( x->root == NULL ) // Heap bis jetzt leer
  {
    x->root = a;
    a->i = i;
    return 0;
  }
  b = x->root;
  while( f[ b->i - 1 ] >= f[ i-1 ] ) // weiter unten einfügen
  {
    if( b->bal == 0 )
    {
      // nach links
      b->bal = 1;
      if( b->left == NULL )
      {
        b->left = a;
        a->i = i;
        return 0;
      }
      b = b->left;
      continue;
    }
    // nach rechts
    b->bal = 0;
    if( b->right == NULL )
    {
      b->right = a;
      a->i = i;
      return 0;
    }
    b = b->right;
  }
  // andere Werte nach unten ziehen
  while( 1 )
  {
    j = b->i;
    b->i = i;
    if( b->bal == 0 )
    {
      // nach links
      b->bal = 1;
      if( b->left == NULL )
      {
        b->left = a;
        a->i = j;
        return 0;
      }
      i = j;
      b = b->left;
      continue;
    }
    // nach rechts
    b->bal = 0;
    if( b->right == NULL )
    {
      b->right = a;
      a->i = j;
      return 0;
    }
    i = j;
    b = b->right;
  }
}

// fügt i in den Heap ein
// in der Wurzel steht das Element j mit kleinstem Wert in f[j-1]
// Fehler: 1->zu wenig Speicher
int add_min_indirect_ih( intheap* x, int i, double* f )
{
  ih_element* a;
  ih_element* b;
  int j;
  a = neu_help_ih( x );
  if( a == NULL )
  {
    return 1;
  }
  a->left = NULL;
  a->right = NULL;
  a->bal = 0;
  if( x->root == NULL ) // Heap bis jetzt leer
  {
    x->root = a;
    a->i = i;
    return 0;
  }
  b = x->root;
  while( f[ b->i - 1 ] <= f[ i-1 ] ) // weiter unten einfügen
  {
    if( b->bal == 0 )
    {
      // nach links
      b->bal = 1;
      if( b->left == NULL )
      {
        b->left = a;
        a->i = i;
        return 0;
      }
      b = b->left;
      continue;
    }
    // nach rechts
    b->bal = 0;
    if( b->right == NULL )
    {
      b->right = a;
      a->i = i;
      return 0;
    }
    b = b->right;
  }
  // andere Werte nach unten ziehen
  while( 1 )
  {
    j = b->i;
    b->i = i;
    if( b->bal == 0 )
    {
      // nach links
      b->bal = 1;
      if( b->left == NULL )
      {
        b->left = a;
        a->i = j;
        return 0;
      }
      i = j;
      b = b->left;
      continue;
    }
    // nach rechts
    b->bal = 0;
    if( b->right == NULL )
    {
      b->right = a;
      a->i = j;
      return 0;
    }
    i = j;
    b = b->right;
  }
}

// liefert die Wurzel von x
// Rückgabe: 0->Heap ist leer
int get_wurzel_ih( intheap* x )
{
  if( x->root == NULL )
  {
    return 0;
  }
  return x->root->i;
}

// löscht die Wurzel des Heaps und gibt das Element zurück
// in der Wurzel steht i, wenn i maximal in allen Heapelementen ist
// Rückgabe: 0->Heap ist bereits leer
int del_max_wurzel_ih( intheap* x )
{
  ih_element* b;
  int i, j;
  if( x->root == NULL ) // Heap leer
  {
    return 0; // nichts zu löschen
  }
  b = x->root;
  i = b->i; // Rückgabe retten
  if( b->left == NULL ) // nur ein Element
  {
    frei_help_ih( x, b );
    x->root = NULL;
    return i;
  }
  // letztes Element suchen
  while( 1 )
  {
    if( b->bal == 0 ) // nach rechts
    {
      b->bal = 1;
      if( b->right->left == NULL ) // unten angekommen
      {
        j = b->right->i; // nach oben mitnehmen
        frei_help_ih( x, b->right );
        b->right = NULL;
        break;
      }
      b = b->right; // absteigen
      continue;
    }
    // nach links
    b->bal = 0;
    if( b->left->left == NULL ) // unten angekommen
    {
      j = b->left->i; // nach oben mitnehmen
      frei_help_ih( x, b->left );
      b->left = NULL;
      break;
    }
    b = b->left; // absteigen
  }
  b = x->root;
  // geholtes Element absteigen lassen
  while( 1 )
  {
    if( b->left == NULL ) // unten angekommen
    {
      b->i = j;
      return i;
    }
    if( b->right == NULL ) // nur noch ein Kind
    {
      if( b->left->i >= j )
      {
        b->i = b->left->i;
        b->left->i = j;
        return i;
      }
      b->i = j;
      return i;
    }
    if( b->left->i >= j )
    {
      if( b->left->i >= b->right->i )
      {
        // links hochziehen
        b->i = b->left->i;
        b = b->left;
        continue;
      }
      // rechts hochziehen
      b->i = b->right->i;
      b = b->right;
      continue;
    }
    if( b->right->i >= j )
    {
      // rechts hochziehen
      b->i = b->right->i;
      b = b->right;
      continue;
    }
    // Position erreicht
    b->i = j;
    return i;
  }
}

// löscht die Wurzel des Heaps und gibt das Element zurück
// in der Wurzel steht i, wenn f[i-1] maximal in allen Heapelementen ist
// Rückgabe: 0->Heap ist bereits leer
int del_max_wurzel_indirect_ih( intheap* x, double* f )
{
  ih_element* b;
  int i, j;
  if( x->root == NULL ) // Heap leer
  {
    return 0; // nichts zu löschen
  }
  b = x->root;
  i = b->i; // Rückgabe retten
  if( b->left == NULL ) // nur ein Element
  {
    frei_help_ih( x, b );
    x->root = NULL;
    return i;
  }
  // letztes Element suchen
  while( 1 )
  {
    if( b->bal == 0 ) // nach rechts
    {
      b->bal = 1;
      if( b->right->left == NULL ) // unten angekommen
      {
        j = b->right->i; // nach oben mitnehmen
        frei_help_ih( x, b->right );
        b->right = NULL;
        break;
      }
      b = b->right; // absteigen
      continue;
    }
    // nach links
    b->bal = 0;
    if( b->left->left == NULL ) // unten angekommen
    {
      j = b->left->i; // nach oben mitnehmen
      frei_help_ih( x, b->left );
      b->left = NULL;
      break;
    }
    b = b->left; // absteigen
  }
  b = x->root;
  // geholtes Element absteigen lassen
  while( 1 )
  {
    if( b->left == NULL ) // unten angekommen
    {
      b->i = j;
      return i;
    }
    if( b->right == NULL ) // nur noch ein Kind
    {
      if( f[ b->left->i - 1 ] >= f[ j - 1 ] )
      {
        b->i = b->left->i;
        b->left->i = j;
        return i;
      }
      b->i = j;
      return i;
    }
    if( f[ b->left->i - 1 ] >= f[ j - 1 ] )
    {
      if( f[ b->left->i - 1 ] >= f[ b->right->i - 1 ] )
      {
        // links hochziehen
        b->i = b->left->i;
        b = b->left;
        continue;
      }
      // rechts hochziehen
      b->i = b->right->i;
      b = b->right;
      continue;
    }
    if( f[ b->right->i - 1 ] >= f[ j - 1 ] )
    {
      // rechts hochziehen
      b->i = b->right->i;
      b = b->right;
      continue;
    }
    // Position erreicht
    b->i = j;
    return i;
  }
}

// löscht die Wurzel des Heaps und gibt das Element zurück
// in der Wurzel steht i, wenn f[i-1] minimal in allen Heapelementen ist
// Rückgabe: 0->Heap ist bereits leer
int del_min_wurzel_indirect_ih( intheap* x, double* f )
{
  ih_element* b;
  int i, j;
  if( x->root == NULL ) // Heap leer
  {
    return 0; // nichts zu löschen
  }
  b = x->root;
  i = b->i; // Rückgabe retten
  if( b->left == NULL ) // nur ein Element
  {
    frei_help_ih( x, b );
    x->root = NULL;
    return i;
  }
  // letztes Element suchen
  while( 1 )
  {
    if( b->bal == 0 ) // nach rechts
    {
      b->bal = 1;
      if( b->right->left == NULL ) // unten angekommen
      {
        j = b->right->i; // nach oben mitnehmen
        frei_help_ih( x, b->right );
        b->right = NULL;
        break;
      }
      b = b->right; // absteigen
      continue;
    }
    // nach links
    b->bal = 0;
    if( b->left->left == NULL ) // unten angekommen
    {
      j = b->left->i; // nach oben mitnehmen
      frei_help_ih( x, b->left );
      b->left = NULL;
      break;
    }
    b = b->left; // absteigen
  }
  b = x->root;
  // geholtes Element absteigen lassen
  while( 1 )
  {
    if( b->left == NULL ) // unten angekommen
    {
      b->i = j;
      return i;
    }
    if( b->right == NULL ) // nur noch ein Kind
    {
      if( f[ b->left->i - 1 ] <= f[ j - 1 ] )
      {
        b->i = b->left->i;
        b->left->i = j;
        return i;
      }
      b->i = j;
      return i;
    }
    if( f[ b->left->i - 1 ] <= f[ j - 1 ] )
    {
      if( f[ b->left->i - 1 ] <= f[ b->right->i - 1 ] )
      {
        // links hochziehen
        b->i = b->left->i;
        b = b->left;
        continue;
      }
      // rechts hochziehen
      b->i = b->right->i;
      b = b->right;
      continue;
    }
    if( f[ b->right->i - 1 ] <= f[ j - 1 ] )
    {
      // rechts hochziehen
      b->i = b->right->i;
      b = b->right;
      continue;
    }
    // Position erreicht
    b->i = j;
    return i;
  }
}

// liefert 0 wenn der Heap leer ist
int empty_ih( intheap* x )
{
  if( x->root == NULL )
  {
    return 0;
  }
  return 1;
}
