#ifndef HEAP_OP

#define HEAP_OP

typedef struct ih_element
{
  int i;
  int bal; // 0->rechts und links gleich viel Nachfolger;
  //          1->links ein Nachfolger mehr
  struct ih_element* left; // linkes Kind
  struct ih_element* right; // rechtes Kind
} ih_element;

typedef struct intheap
{
  ih_element* root; // Wurzel
  ih_element* puf; // Puffer
  int pufmax; // max. zulässige Anzahl an Einträgen im Puffer
  int pufakt; // aktuelle Anzahl an Einträgen im Puffer
} intheap;


// initialisiert einen Heap
void init_ih( intheap* x ); // unendlicher Puffer

// löscht einen Heap (Destruktor); x ist anschließend
// initialisiert, aber der Puffer ist leer
void free_ih( intheap* x );

// fügt i in den Heap ein
// in der Wurzel steht das Element j mit größtem Wert j
// von allen Heapelementen
// Fehler: 1->zu wenig Speicher
int add_max_ih( intheap* x, int i );

// fügt i in den Heap ein
// in der Wurzel steht das Element j mit größtem Wert in f[j-1]
// Fehler: 1->zu wenig Speicher
int add_max_indirect_ih( intheap* x, int i, double* f );

// fügt i in den Heap ein
// in der Wurzel steht das Element j mit kleinstem Wert in f[j-1]
// Fehler: 1->zu wenig Speicher
int add_min_indirect_ih( intheap* x, int i, double* f );

// liefert die Wurzel von x
// Rückgabe: 0->Heap ist leer
int get_wurzel_ih( intheap* x );

// löscht die Wurzel des Heaps und gibt das Element zurück
// in der Wurzel steht i, wenn i maximal in allen Heapelementen ist
// Rückgabe: 0->Heap ist bereits leer
int del_max_wurzel_ih( intheap* x );

// löscht die Wurzel des Heaps und gibt das Element zurück
// in der Wurzel steht i, wenn f[i-1] maximal in allen Heapelementen ist
// Rückgabe: 0->Heap ist bereits leer
int del_max_wurzel_indirect_ih( intheap* x, double* f );

// löscht die Wurzel des Heaps und gibt das Element zurück
// in der Wurzel steht i, wenn f[i-1] minimal in allen Heapelementen ist
// Rückgabe: 0->Heap ist bereits leer
int del_min_wurzel_indirect_ih( intheap* x, double* f );

// liefert 0 wenn der Heap leer ist
int empty_ih( intheap* x );

#endif
