#include <limits.h>
#include <malloc.h>
#include <stdlib.h>
#include "listen.h"

//---------------------- D O P P E L I N T L I S T E -------------------
// Initialisiert doppelintliste
void init_dil( doppelintliste* x )
{
  x->first = NULL;
  x->last = NULL;
  x->pmax = INT_MAX;
  x->pakt = 0;
}

// Destruktor - Ersatz; Liste ist anschließend initialisiert,
// der Puffer ist aber leer
void free_dil( doppelintliste* x )
{
  dil_element* a;
  while( x->first != NULL )
  {
    a = x->first;
    x->first = x->first->next;
    free( a );
  }
  while( x->pakt != 0 )
  {
    a = x->p;
    x->p = a->next;
    free( a );
    (x->pakt)--;
  }
}

// interne Hilfsroutine; löscht das Element a (gepuffertes free)
void frei_help_dil( doppelintliste* x, dil_element* a )
{
  if( x->pakt != x->pmax )
  {
    a->next = x->p;
    x->p = a;
    (x->pakt)++;
    return;
  }
  free( a );
}

// interne Hilfsroutine; erzeugt ein neues Element
// Fehler: NULL->zu wenig Speicher
dil_element* neu_help_dil( doppelintliste* x )
{
  dil_element* a;
  if( x->pakt == 0 )
  {
    return (dil_element*) malloc( sizeof( dil_element ) );
  }
  a = x->p;
  x->p = a->next;
  (x->pakt)--;
  return a;
}

// fügt Wert i am Ende von x ein
// Fehler: 1->zu wenig Speicher
int insert_end_dil( doppelintliste* x, int i )
{
  dil_element* a;
  a = neu_help_dil( x );
  if( a == NULL )
  {
    return 1;
  }
  a->i = i;
  a->next = NULL;
  a->prev = x->last;
  if( x->last != NULL )
  {
    x->last->next = a;
  }
  else
  {
    x->first = a;
  }
  x->last = a;
  return 0;
}

// Fügt alle Elemente die in y hinter a stehen zu x hinzu
// (a nicht mit!!); a ist anschließend das letzte Element von y
// keine Überprüfung ob a zu y gehört; wenn a == NULL wird
// ganz y angehangen
void insert_list_end_dil(
  doppelintliste* x,
  doppelintliste* y,
  dil_element*    a )
{
  if( y->first == NULL ) // y leer
  {
    return;
  }
  if( a == NULL ) // gesamte Liste anhängen
  {
    if( x->first == NULL ) // x leer
    {
      x->first = y->first;
      x->last = y->last;
      y->first = y->last = NULL;
      return;
    }
    x->last->next = y->first;
    y->first->prev = x->last;
    x->last = y->last;
    y->first = y->last = NULL;
    return;
  }
  if( a->next == NULL ) // nichts anzuhängen
  {
    return;
  }
  if( x->first == NULL ) // x leer
  {
    x->first = a->next;
    a->next->prev = NULL;
    x->last = y->last;
    y->last = a;
    a->next = NULL;
    return;
  }
  x->last->next = a->next;
  a->next->prev = x->last;
  x->last = y->last;
  y->last = a;
  a->next = NULL;
  return;
}

// fügt z hinter Position y in x ein; z ist anschließend leer
// (!keine Überprüfung ob y in x liegt!)
// y == NULL bedeutet einfügen am Anfang von x
void insert_pos_dil(
  doppelintliste* x,
  dil_element*    y,
  doppelintliste* z )
{
  if( z->first == NULL ) // wenn z leer ist -> nichts zu tun
  {
    return;
  }
  if( y == NULL ) // am Anfang einfügen
  {
    if( x->first == NULL ) // x ist leer, d.h. z wird zu x
    {
      x->last = z->last;
    }
    else
    {
      x->first->prev = z->last;
      z->last->next = x->first;
    }
    x->first = z->first;
  }
  else // nicht am Anfang einfügen
  {
    if( y->next == NULL ) // am Ende einfügen
    {
      y->next = z->first;
      z->first->prev = y;
      x->last = z->last;
    }
    else // in der Mitte einfügen
    {
      z->last->next = y->next;
      y->next->prev = z->last;
      y->next = z->first;
      z->first->prev = y;
    }
  }
  // z als leer markieren
  z->first = z->last = NULL;
}

// liefert das letzte Element von x und löscht es
// Rückgabe: 0->x ist leer
int del_end_dil( doppelintliste* x )
{
  int i;
  dil_element* a;
  if( x->last == NULL )
  {
    return 0;
  }
  a = x->last;
  i = a->i;
  x->last = a->prev;
  if( x->last != NULL )
  {
    x->last->next = NULL;
  }
  else
  {
    x->first = NULL;
  }
  frei_help_dil( x, a );
  return i;
}

// liefert einen Zeiger auf das letzte Symbol der Liste
// Rückgabe: NULL->Liste ist leer
dil_element* get_endpos_dil( doppelintliste* x )
{
  return x->last;
}

// liefert einen Zeiger auf das Symbol unter dem Lesezeiger
// Rückgabe: kein sinnvoller Wert wenn Lesezeiger nicht
//           initialisiert wurde
dil_element* get_pos_dil( doppelintliste* x )
{
  return x->akt;
}

// setzt den Lesezeiger auf das letzte Element
void reset_end_dil( doppelintliste* x )
{
  x->akt = x->last;
}

// setzt den Lesezeiger auf das angegebenen Element;
// keine Überprüfung ob a zu x gehört
void reset_pos_dil( doppelintliste* x, dil_element* a )
{
  x->akt = a;
}

// liefert das Zeichen an der Stelle des Lesezeigers
// Rückgabe: 0->Ende oder Anfang von x überschritten
int get_dil( doppelintliste* x )
{
  if( x->akt != NULL )
  {
    return x->akt->i;
  }
  return 0;
}

// schreibt i an die aktuelle Stelle des Lesezeigers
// Rückgabe: 1->Ende oder Anfang von x überschritten
int set_dil( doppelintliste* x, int i )
{
  if( x->akt == NULL )
  {
    return 1;
  }
  x->akt->i = i;
  return 0;
}

// löscht das Element, auf welches der Lesezeiger zeigt;
// der Lesezeiger wird ein Element nach hinten gesetzt
// Rückgabe: das gelöschte Element oder
//           0->Ende oder Anfang von x überschritten
int del_dil( doppelintliste* x )
{
  int i;
  dil_element* a;
  if( x->akt == NULL )
  {
    return 0;
  }
  i = x->akt->i;
  if( x->akt->prev == NULL ) // erstes Element
  {
    if( x->akt->next == NULL ) // nur ein Element
    {
      x->first = x->last = NULL;
      frei_help_dil( x, x->akt );
      x->akt = NULL;
      return i;
    }
    x->first = x->akt->next;
    x->akt->next->prev = NULL;
    frei_help_dil( x, x->akt );
    x->akt = x->first;
    return i;
  }
  if( x->akt->next == NULL ) // letztes Element
  {
    x->akt->prev->next = NULL;
    x->last = x->akt->prev;
    frei_help_dil( x, x->akt );
    x->akt = NULL;
    return i;
  }
  // in der Mitte
  a = x->akt->prev->next = x->akt->next;
  x->akt->next->prev = x->akt->prev;
  frei_help_dil( x, x->akt );
  x->akt = a;
  return i;
}

// setzt den Lesezeiger ein Element zurück
// Rückgabe: Element unter dem Lesezeiger vor dem Versetzen
// Fehler: 0->Lesezeiger hat den Anfang überschritten
int last_dil( doppelintliste* x )
{
  int i;
  if( x->akt == NULL )
  {
    return 0;
  }
  i = x->akt->i;
  x->akt = x->akt->prev;
  return i;
}

// liefert 0 wenn x keine Einträge enthält; sonst 1
int empty_dil( doppelintliste* x )
{
  if( x->first == NULL )
  {
    return 0;
  }
  return 1;
}

//--------------------------- I N T L I S T E --------------------------

// Konstruktor - Ersatz; maximale Puffergröße
void init_il( intliste* x )
{
  x->akt = x->first = x->last = NULL;
  x->pakt = 0;
  x->pmax = INT_MAX;
}

// Konstruktor - Ersatz; Puffer hat die Größe i
void init_size_il( intliste* x, int i )
{
  x->akt = x->first = x->last = NULL;
  x->pakt = 0;
  x->pmax = i;
}

// Destruktor - Ersatz; Liste ist anschließend initialisiert;
// Puffer ist aber leer
void free_il( intliste* x )
{
  il_element* a;
  while( x->first != NULL )
  {
    a = x->first;
    x->first = x->first->next;
    free( a );
  }
  x->last = NULL;
  while( x->pakt != 0 )
  {
    a = x->p;
    x->p = a->next;
    free( a );
    (x->pakt)--;
  }
}

// interne Hilfsroutine; löscht das Element a (gepuffertes free)
void frei_help_il( intliste* x, il_element* a )
{
  if( x->pakt != x->pmax )
  {
    a->next = x->p;
    x->p = a;
    (x->pakt)++;
    return;
  }
  free( a );
}

// interne Hilfsroutine; liefert ein neues Element
il_element* neu_help_il( intliste* x )
{
  il_element* a;
  if( x->pakt == 0 )
  {
    return (il_element*) malloc( sizeof( il_element ) );
  }
  a = x->p;
  x->p = a->next;
  (x->pakt)--;
  return a;
}

// fügt i am Anfang von x ein
// Fehler: 1->zu wenig Speicher
int add_anf_il( intliste* x, int i )
{
  il_element* a;
  a = neu_help_il( x );
  if( a == NULL )
  {
    return 1;
  }
  if( x->first == NULL )
  {
    x->last = a; // dieser ist der letzte
  }
  a->next = x->first;
  a->i = i;
  x->first = a;
  if( a->next == x->akt ) // der Lesezeiger stand auf dem ersten Element
  {
    x->old = a;
  }
  return 0;
}

// fügt i am Ende von x ein
// Fehler: 1->zu wenig Speicher
int add_end_il( intliste* x, int i )
{
  il_element* a;
  a = neu_help_il( x );
  if( a == NULL )
  {
    return 1;
  }
  if( x->last == NULL )
  {
    x->first = a; // dieser ist der erste
  }
  else
  {
    x->last->next = a;
  }
  a->i = i;
  x->last = a;
  a->next = NULL;
  return 0;
}

// hängt y ans Ende von x; y ist dann leer
void add_list_end_il( intliste* x, intliste* y )
{
  if( y->first == NULL ) // y leer
  {
    return;
  }
  if( x->first == NULL ) // x leer
  {
    x->first = y->first;
    x->last = y->last;
    y->first = y->last = NULL;
    return;
  }
  x->last->next = y->first;
  x->last = y->last;
  y->first = y->last = NULL;
  return;
}

// setzt den Lesezeiger auf das erste Element
void reset_il( intliste* x )
{
  x->akt = x->first;
  x->old = NULL;
}

// liefert in y den aktuellen Lesezeiger
void get_pos_il( intliste* x, il_pos* y )
{
  y->akt = x->akt;
  y->old = x->old;
}

// setzt den Lesezeiger auf den zuvor geretteten Wert y
// füllt y mit der vorher aktuellen Lesezeigerposition
// !keine Überprüfung ob y zu x gehört!
void swap_pos_il( intliste* x, il_pos* y )
{
  il_pos p;
  p.akt = x->akt;
  p.old = x->old;
  x->akt = y->akt;
  x->old = y->old;
  y->akt = p.akt;
  y->old = p.old;
}

// setzt den Lesezeiger ein Element weiter
// Rückgabe: Element unter dem Lesezeiger vor dem Versetzen
// Fehler: 0->Lesezeiger hat das Ende überschritten
int next_il( intliste* x )
{
  if( x->akt == NULL )
  {
    return 0;
  }
  x->old = x->akt;
  x->akt = x->akt->next;
  return x->old->i;
}

// liefert das Element unter dem Lesezeiger
// Rückgabe: 0->Ende überschritten
int get_il( intliste* x )
{
  if( x->akt == NULL )
  {
    return 0;
  }
  return x->akt->i;
}

// löscht das Element unter dem Lesezeiger und gibt seinen Inhalt zurück
// der Lesezeiger rückt ein Element weiter
// Fehler: 0->Lesezeiger hat das Ende der Liste schon im letzten Schritt
//            erreicht
int del_il( intliste* x )
{
  int i;
  if( x->akt == NULL )
  {
    return 0;
  }
  if( x->old == NULL ) // erstes Element löschen
  {
    x->first = x->first->next;
    i = x->akt->i;
    frei_help_il( x, x->akt );
    x->akt = x->first;
  }
  else
  {
    x->old->next = x->akt->next;
    i = x->akt->i;
    frei_help_il( x, x->akt );
    x->akt = x->old->next;
  }
  if( x->akt == NULL ) // letzes Element gelöscht?
  {
    x->last = x->old;
  }
  return i;
}

// löscht das erste Element und gibt es zurück
// Rückgabe: 0->Liste leer
int del_anf_il( intliste* x )
{
  il_element* a;
  int i;
  if( x->first == NULL )
  {
    return 0;
  }
  a = x->first;
  x->first = x->first->next;
  i = a->i;
  frei_help_il( x, a );
  if( x->first == NULL )
  {
    x->last = NULL;
  }
  return i;
}

// bestimmt die Anzahl der Listenelemente
int anz_il( intliste* x )
{
  il_element* a;
  int i;
  a = x->first;
  i = 0;
  while( a != NULL )
  {
    a = a->next;
    i++;
  }
  return i;
}

// testet ob x leer ist
// Rückgabe: 0->x ist leer
//           1->sonst
int empty_il( intliste* x )
{
  if( x->first == NULL )
  {
    return 0;
  }
  return 1;
}
