#ifndef LISTEN_OP

#define LISTEN_OP

//---------------- D O P P E L I N T L I S T E -------------------------

typedef struct dil_element
{
  int i;
  struct dil_element* next;
  struct dil_element* prev;
} dil_element;

typedef struct doppelintliste
{
  dil_element* first; // erstes Element
  dil_element* last; // letztes Element
  dil_element* akt; // Lesezeiger
  dil_element* p; // Pufferliste
  int pmax; // maximale Puffergröße
  int pakt; // aktuelle Puffergröße
} doppelintliste;

// Initialisiert doppelintliste
void init_dil( doppelintliste* x );

// Destruktor - Ersatz; Liste ist anschließend initialisiert,
// der Puffer ist aber leer
void free_dil( doppelintliste* x );

// fügt Wert i am Ende von x ein
// Fehler: 1->zu wenig Speicher
int insert_end_dil( doppelintliste* x, int i );

// Fügt alle Elemente die in y hinter a stehen zu x hinzu
// (a nicht mit!!); a ist anschließend das letzte Element von y
// keine Überprüfung ob a zu y gehört;
// wenn a == NULL wird ganz y angehangen
void insert_list_end_dil(
  doppelintliste* x,
  doppelintliste* y,
  dil_element*    a );

// fügt z ab Position y in x ein; z ist anschließend leer
// (!keine Überprüfung ob y in x liegt!)
// y == NULL bedeutet einfügen am Anfang von x
void insert_pos_dil(
  doppelintliste* x,
  dil_element*    y,
  doppelintliste* z );

// liefert das letzte Element von x und löscht es
// Rückgabe: 0->x ist leer
int del_end_dil( doppelintliste* x );

// liefert einen Zeiger auf das letzte Symbol der Liste
// Rückgabe: NULL->Liste ist leer
dil_element* get_endpos_dil( doppelintliste* x );

// liefert einen Zeiger auf das Symbol unter dem Lesezeiger
// Rückgabe: kein sinnvoller Wert wenn Lesezeiger nicht
// initialisiert wurde
dil_element* get_pos_dil( doppelintliste* x );

// setzt den Lesezeiger auf das letzte Element
void reset_end_dil( doppelintliste* x );

// setzt den Lesezeiger auf das angegebenen Element;
// keine Überprüfung ob a zu x gehört
void reset_pos_dil( doppelintliste* x, dil_element* a );

// liefert das Zeichen an der Stelle des Lesezeigers
// Rückgabe: 0->Ende oder Anfang von x überschritten
int get_dil( doppelintliste* x );

// löscht das Element, auf welches der Lesezeiger zeigt;
// der Lesezeiger wird ein Element nach hinten gesetzt
// Rückgabe: das gelöschte Element oder
//           0->Ende oder Anfang von x überschritten
int del_dil( doppelintliste* x );

// schreibt i an die aktuelle Stelle des Lesezeigers
// Rückgabe: 1->Ende oder Anfang von x überschritten
int set_dil( doppelintliste* x, int i );

// setzt den Lesezeiger ein Element zurück
// Rückgabe: Element unter dem Lesezeiger vor dem Versetzen
// Fehler: 0->Lesezeiger hat den Anfang überschritten
int last_dil( doppelintliste* x );

// liefert 0 wenn x keine Einträge enthält; sonst 1
int empty_dil( doppelintliste* x );

//------------------------- I N T L I S T E ----------------------------

typedef struct il_element
{
  int i;
  struct il_element* next;
} il_element;

typedef struct intliste
{
  il_element* first; // erstes Element
  il_element* last; // letztes Element
  il_element* akt; // Lesezeiger
  il_element* old; // Element vor dem Lesezeiger
  il_element* p; // Pufferliste
  int pmax; // maximale Größe des Puffers
  int pakt; // aktuelle Größe des Puffers
} intliste;

typedef struct il_pos
{
  il_element* akt;
  il_element* old;
} il_pos;

// Konstruktor - Ersatz; Puffer hat die Größe i
void init_size_il( intliste* x, int i );
void init_il( intliste* x ); // unbegrenzter Puffer

// Destruktor - Ersatz
void free_il( intliste* x );

// fügt i am Anfang von x ein
// Fehler: 1->zu wenig Speicher
int add_anf_il( intliste* x, int i );

// fügt i am Ende von x ein
// Fehler: 1->zu wenig Speicher
int add_end_il( intliste* x, int i );

// hängt y ans Ende von x; y ist dann leer
void add_list_end_il( intliste* x, intliste* y );

// setzt den Lesezeiger auf das erste Element
void reset_il( intliste* x );

// liefert in y den aktuellen Lesezeiger
void get_pos_il( intliste* x, il_pos* y );

// setzt den Lesezeiger auf den zuvor geretteten Wert y
// füllt y mit der vorher aktuellen Lesezeigerposition
// !keine Überprüfung ob y zu x gehört!
void swap_pos_il( intliste* x, il_pos* y );

// setzt den Lesezeiger ein Element weiter
// Rückgabe: Element unter dem Lesezeiger vor dem Versetzen
// Fehler: 0->Lesezeiger hat das Ende überschritten
int next_il( intliste* x );

// liefert das Element unter dem Lesezeiger
// Rückgabe: 0->Ende überschritten
int get_il( intliste* x );

// löscht das Element unter dem Lesezeiger und gibt seinen Inhalt zurück
// der Lesezeiger rückt ein Element weiter
// Fehler: 0->Lesezeiger hat das Ende der Liste schon im letzten Schritt
//            erreicht
int del_il( intliste* x );

// löscht das erste Element und gibt es zurück
// Rückgabe: 0->Liste leer
int del_anf_il( intliste* x );

// bestimmt die Anzahl der Listenelemente
int anz_il( intliste* x );

// testet ob x leer ist
// Rückgabe: 0->x ist leer
//           1->sonst
int empty_il( intliste* x );

#endif
