#include <malloc.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zufall.h"

// liest bis über das nächste '\n'
// Fehler: 1->Dateiende erreicht
int lese_next_ret( FILE* *fi )
{
  char c;
  do
  {
    c = (char)fgetc( *fi );
    if( feof( *fi ) )
    {
      return 1;
    }
  } while( c != '\n' );
  return 0;
}

// liest nächste ganze Zahl in aktueller Zeile
// Fehler: 1->Zeilenende erreicht bevor eine Zahl gelesen wurde
//         2->Dateiende erreicht bevor eine Zahl gelesen wurde
int lese_next_int_in_line( FILE* *fi, int* a )
{
  char c = ' ';
  char c_last;
  do
  {
    c_last = c;
    c = (char)fgetc( *fi );
    if( feof( *fi ) )
    {
      return 2;
    }
    if( c == '\n' )
    {
      ungetc( c, *fi );
      return 1;
    }
  } while( ( c > '9' ) || ( c < '0' ) );
  *a = c - '0';
  while( 1 )
  {
    c = (char)fgetc( *fi );
    if( feof( *fi ) )
    {
      if( c_last == '-' )
        (*a) = -(*a);

      return 0;
    }
    if( ( c > '9' ) || ( c < '0' ) )
    {
      ungetc( c, *fi );

      if( c_last == '-' )
        (*a) = -(*a);

      return 0;
    }
    *a = 10 * (*a) + c - '0';
  }
}

// liest nächste ganze Zahl in aktueller Zeile und
// einen anschließenden Zeilenumbruch
// Fehler: 1->Zeilenende erreicht bevor eine Zahl gelesen wurde
//         2->Dateiende erreicht bevor eine Zahl gelesen wurde
int lese_next_int_in_line_r( FILE* *fi, int* a )
{
  int rc;
  rc = lese_next_int_in_line( fi, a );
  if( rc != 0 )
  {
    return rc;
  }
  lese_next_ret( fi );
  return 0;
}

// liest nächste ganze Zahl (egal in welcher Zeile)
// Fehler: 1->Dateiende erreicht bevor eine Zahl gelesen wurde
int lese_next_int( FILE* *fi, int* a )
{
  int rc;
  while( 1 )
  {
    rc = lese_next_int_in_line( fi, a );
    switch( rc )
    {
      case 0:
        return 0;
      case 2:
        return 1;
    }
    if( lese_next_ret( fi ) != 0 )
    {
      return 1;
    }
  }
}

// liest nächste ganze Zahl
// und überliest alle Zeichen bis zum Zeilenumbruch
// Fehler: 1->Dateiende erreicht bevor eine Zahl gelesen wurde
int lese_next_int_r( FILE* *fi, int* a )
{
  if( lese_next_int( fi, a ) != 0 )
  {
    return 1;
  }
  lese_next_ret( fi );
  return 0;
}

// liest nächste gebrochene Zahl mit Vorzeichen (! bei "ab.de"
// wird der "." als 0.0 gedeutet; ".." liefert sogar bei 2 Aufrufen
// den Wert 0; bei ".e+vg" wird ".e+" weggelesen !)
// Fehler: 1->Zeilenende erreicht
//         2->Dateiende erreicht
int lese_next_double_in_line( FILE* *fi, double* a )
{
  char c;
  int vz, vze;
  double pot, expo;
  vz = 0;
  pot = 0.0;
  while( 1 )
  {
    c = (char)fgetc( *fi );
    if( feof( *fi ) )
    {
      return 2;
    }
    if( c == '\n' )
    {
      ungetc( c, *fi );
      return 1;
    }
    if( ( c <= '9' ) && ( c >= '0' ) )
    {
      *a = c - '0';
      break;
    }
    if( c == '.' )
    {
      pot = 1.0;
      *a = 0.0;
      break;
    }
    if( c == '-' )
    {
      vz = 1;
    }
    else
    {
      if( c != ' ' )
      {
        vz = 0;
      }
    }
  }
  while( 1 )
  {
    c = (char)fgetc( *fi );
    if( feof( *fi ) )
    {
      return 0;
    }
    if( ( c <= '9' ) && ( c >= '0' ) )
    {
      pot *= 10;
      *a = 10 * (*a) + ( c - '0' );
    }
    else
    {
      if( ( c == '.' ) && ( pot == 0.0 ) )
      {
        pot = 1.0;
      }
      else
      {
        if( c == 'e' ) // Exponent einlesen
        {
          c = (char)fgetc( *fi );
          if( !feof( *fi ) )
          {
            if( c == '-' )
            {
              vze = 1;
              c = (char)fgetc( *fi );
            }
            else
            {
              vze = 0;
              if( c == '+' )
              {
                c = (char)fgetc( *fi );
              }
            }
            if( !feof( *fi ) )
            {
              expo = 0.0;
              while( ( c >= '0' ) && ( c <= '9' ) )
              {
                expo = 10.0 * expo + ( c - '0' );
                c = (char)fgetc( *fi );
                if( feof( *fi ) )
                {
                  break;
                }
              }
              if( vze == 0 )
              {
                expo = -expo;
              }
              if( pot == 0.0 )
              {
                pot = pow( 10, expo );
              }
              else
              {
                pot *= pow( 10, expo );
              }
            }
          }
        }
        if( !feof( *fi ) )
        {
          ungetc( c, *fi );
        }
        if( vz == 1 )
        {
          *a = -(*a);
        }
        if( pot != 0.0 )
        {
          *a /= pot;
        }
        return 0;
      }
    }
  }
}

// liest nächste gebrochene Zahl mit Vorzeichen (! bei "ab.de"
// wird der "." als 0.0 gedeutet; ".." liefert sogar bei 2 Aufrufen
// den Wert 0; bei ".e+vg" wird ".e+" weggelesen !)
// und einen Zeilenumbruch
// Fehler: 1->Zeilenende erreicht bevor Zahl gelesen wurde
//         2->Dateiende erreicht
int lese_next_double_in_line_r( FILE* *fi, double* a )
{
  int i;
  i = lese_next_double_in_line( fi, a );
  if( i != 0 )
  {
    return i;
  }
  lese_next_ret( fi );
  return 0;
}

// liest nächste gebrochene Zahl mit Vorzeichen, egal in welcher Zeile
// (! bei "ab.de" wird der "." als 0.0 gedeutet; ".." liefert sogar
// bei 2 Aufrufen den Wert 0; bei ".e+vg" wird ".e+" weggelesen !)
// Fehler: 1->Dateiende erreicht
int lese_next_double( FILE* *fi, double* a )
{
  int i;
  while( 1 )
  {
    i = lese_next_double_in_line( fi, a );
    if( i == 0 )
    {
      return 0;
    }
    if( i == 2 )
    {
      return 1;
    }
    if( lese_next_ret( fi ) != 0 )
    {
      return 1;
    }
  }
}

// liest nächste gebrochene Zahl mit Vorzeichen, egal in welcher Zeile
// (! bei "ab.de" wird der "." als 0.0 gedeutet; ".." liefert sogar
// bei 2 Aufrufen den Wert 0; bei ".e+vg" wird ".e+" weggelesen !)
// überliest den nächsten Zeilenwechsel
// Fehler: 1->Dateiende erreicht
int lese_next_double_r( FILE* *fi, double* a )
{
  if( lese_next_double( fi, a ) != 0 )
  {
    return 1;
  }
  lese_next_ret( fi );
  return 0;
}

// liest solange weiter, bis ein in a[] enthaltenes Zeichen auftritt
// oder das Zeilenende erreicht ist
// Zeichen wird nach b gegeben; a[] muß mit '\0' abgeschlossen sein
// Fehler: 1->Zeilenende erreicht
//         2->Dateiende erreicht
int lese_next_char_in_line( FILE* *fi, char* a, char *b )
{
  int i;
  while( 1 )
  {
    *b = (char)fgetc( *fi );
    if( feof( *fi ) )
    {
      return 2;
    }
    i = 0;
    while( a[i] != '\0' )
    {
      if( a[i] == *b )
      {
        return 0;
      }
      i++;
    }
    if( *b == '\n' )
    {
      ungetc( *b, *fi );
      return 1;
    }
  }
}

// liest solange weiter, bis ein in a[] enthaltenes Zeichen auftritt
// (egal in welcher Zeile)
// Zeichen wird nach b gegeben; a[] muß mit '\0' abgeschlossen sein
// Fehler: 1->Dateiende erreicht
int lese_next_char( FILE* *fi, char* a, char *b )
{
  int i;
  while( 1 )
  {
    i = lese_next_char_in_line( fi, a, b );
    if( i == 0 )
    {
      return 0;
    }
    if( i == 2 )
    {
      return 1;
    }
    if( lese_next_ret( fi ) != 0 )
    {
      return 1;
    }
  }
}

// liest solange weiter, bis ein in a[] enthaltenes Zeichen auftritt;
// dieses wird nach b gegeben; a[] muß mit '\0' abgeschlossen sein
// überliest den nächsten Zeilenwechsel
// Fehler: 1->Dateiende erreicht
int lese_next_char_r( FILE* *fi, char* a, char *b )
{
  int i;
  while( 1 )
  {
    i = lese_next_char( fi, a, b );
    if( i == 0 )
    {
      return 0;
    }
    if( i == 2 )
    {
      return 1;
    }
    lese_next_ret( fi );
  }
}

// liest nächsten String in der aktuellen Zeile
// bis ein in a enthaltenes Zeichen auftritt;
// der Speicher für b wird reserviert, wenn *b != NULL, so wird der
// Speicher wiederverwendet
// in c wird das in a aufgetretene Zeichen zurück gegeben
// Fehler: 1->leeren String gelesen
//         2->Dateiende erreicht (b enthält alle bis dahin gelesenen
//            Zeichen)
//         3->zu wenig Speicher
int lese_next_string_in_line( FILE* *fi, char* a, char** b )
{
  int i, j, max;
  if( *b == NULL )
  {
    max = 20;
    *b = (char*) malloc( max * sizeof( char ) );
    if( *b == NULL )
    {
      return 3;
    }
  }
  else
  {
    max = strlen( *b ) + 1;
  }
  i = 0;
  while( 1 )
  {
    for( ; i<max; i++ )
    {
      (*b)[i] = (char)fgetc( *fi );
      if( feof( *fi ) )
      {
        (*b)[i] = '\0';
        if( max - 1 != i )
        {
          *b = (char*) realloc( *b, ( i + 1 ) * sizeof( char ) );
          if( *b == NULL )
          {
            return 3;
          }
        }
        return 2;
      }
      if( (*b)[i] == '\n' )
      {
        ungetc( (*b)[i], *fi );
        (*b)[i] = '\0';
        if( max - 1 != i )
        {
          *b = (char*) realloc( *b, ( i + 1 ) * sizeof( char ) );
          if( *b == NULL )
          {
            return 3;
          }
        }
        if( i == 0 )
        {
          return 1;
        }
        return 0;
      }
      j = 0;
      while( a[j] != '\0' )
      {
        if( (*b)[i] == a[j] )
        {
          (*b)[i] = '\0';
          if( max - 1 != i )
          {
            *b = (char*) realloc( *b, ( i + 1 ) * sizeof( char ) );
            if( *b == NULL )
            {
              return 3;
            }
          }
          if( i == 0 )
          {
            return 1;
          }
          return 0;
        }
        j++;
      }
    }
    max *= 2;
    *b = (char*) realloc( *b, max * sizeof( char ) );
  }
}

// liest nächste ganzzahlige Verteilung
// Fehler: 1->Zeilenende erreicht
//         2->Dateiende erreicht
int lese_next_verteilung_int_in_line( FILE* *fi, verteilung_int* v )
{
  char a;
  if( lese_next_char_in_line( fi, "kgen", &a ) != 0 )
  {
    return 1;
  }
  if( lese_next_int_in_line( fi, &(v->a) ) != 0 )
  {
    return 1;
  }
  if( a == 'k' )
  {
    v->typ = konstant;
    return 0;
  }
  if( lese_next_int_in_line( fi, &(v->b) ) != 0 )
  {
    return 1;
  }
  if( a == 'g' )
  {
    v->typ = gleich;
    return 0;
  }
  if( lese_next_double_in_line( fi, &(v->p1) ) != 0 )
  {
    return 1;
  }
  if( a == 'e' )
  {
    v->typ = exponential;
    return 0;
  }
  if( lese_next_double_in_line( fi, &(v->p2) ) != 0 )
  {
    return 1;
  }
  v->typ = normal;
  return 0;
}

// liest nächste ganzzahlige Verteilung (egal in welcher Zeile)
// Fehler: 1->Dateiende erreicht oder keine gültige Verteilung gefunden
int lese_next_verteilung_int( FILE* *fi, verteilung_int* v )
{
  int i;
  while( 1 )
  {
    i = lese_next_verteilung_int_in_line( fi, v );
    if( i == 0 )
    {
      return 0;
    }
    if( i == 2 )
    {
      return 1;
    }
    if( lese_next_ret( fi ) != 0 )
    {
      return 1;
    }
  }
}

// liest nächste ganzzahlige Verteilung (egal in welcher Zeile)
// und einen anschließenden Zeilenwechsel
// Fehler: 1->Dateiende erreicht oder keine gültige Verteilung gefunden
int lese_next_verteilung_int_r( FILE* *fi, verteilung_int* v )
{
  if( lese_next_verteilung_int( fi, v ) != 0 )
  {
    return 1;
  }
  lese_next_ret( fi );
  return 0;
}

// liest nächste gebrochene Verteilung in aktueller Zeile
// Fehler: 1->Zeilenende erreicht
//         2->Dateiende erreicht oder keine gültige Verteilung gefunden
int lese_next_verteilung_double_in_line(
  FILE**             fi,
  verteilung_double* v )
{
  char a;
  if( lese_next_char_in_line( fi, "kgen", &a ) != 0 )
  {
    return 1;
  }
  if( lese_next_double_in_line( fi, &(v->a) ) != 0 )
  {
    return 2;
  }
  if( a == 'k' )
  {
    v->typ = konstant;
    return 0;
  }
  if( lese_next_double_in_line( fi, &(v->b) ) != 0 )
  {
    return 2;
  }
  if( a == 'g' )
  {
    v->typ = gleich;
    return 0;
  }
  if( lese_next_double_in_line( fi, &(v->p1) ) != 0 )
  {
    return 2;
  }
  if( a == 'e' )
  {
    v->typ = exponential;
    return 0;
  }
  if( lese_next_double_in_line( fi, &(v->p2) ) != 0 )
  {
    return 2;
  }
  v->typ = normal;
  return 0;
}

// liest nächste gebrochene Verteilung (egal in welcher Zeile)
// Fehler: 1->Dateiende erreicht oder keine gültige Verteilung gefunden
int lese_next_verteilung_double( FILE* *fi, verteilung_double* v )
{
  int i;
  while( 1 )
  {
    i = lese_next_verteilung_double_in_line( fi, v );
    if( i == 0 )
    {
      return 0;
    }
    if( i == 2 )
    {
      return 1;
    }
    if( lese_next_ret( fi ) != 0 )
    {
      return 1;
    }
  }
}

// liest nächste gebrochene Verteilung (egal in welcher Zeile)
// und einen anschließenden Zeilenumbruch
// Fehler: 1->Dateiende erreicht oder keine gültige Verteilung gefunden
int lese_next_verteilung_double_r( FILE* *fi, verteilung_double* v )
{
  int i;
  i = lese_next_verteilung_double( fi, v );
  if( i != 0 )
  {
    return i;
  }
  i = lese_next_ret( fi );
  return i;
}

// liest nächste Verteilung (Parameter werden als String abgelegt)
// Fehler: 1->Verteilung unvollständig
//         2->Dateiende erreicht
//         3->zu wenig Speicher
int lese_next_verteilung_text( FILE* *fi, verteilung_text* v )
{
  char a, b;
  int i;
  i = lese_next_char( fi, "kgen", &a );
  if( i != 0 )
  {
    return i;
  }
  i = lese_next_char_in_line( fi, " ", &b );
  if( i != 0 )
  {
    return i;
  }
  i = lese_next_string_in_line( fi, " ", &(v->a) );
  if( i != 0 )
  {
    return i;
  }
  if( a == 'k' )
  {
    v->typ = konstant;
    return 0;
  }
  i = lese_next_string_in_line( fi, " ", &(v->b) );
  if( i != 0 )
  {
    return i;
  }
  if( a == 'g' )
  {
    v->typ = gleich;
    return 0;
  }
  i = lese_next_string_in_line( fi, " ", &(v->p1) );
  if( i != 0 )
  {
    return i;
  }
  if( a == 'e' )
  {
    v->typ = exponential;
    return 0;
  }
  i = lese_next_string_in_line( fi, " \n", &(v->p2) );
  if( i != 0 )
  {
    return i;
  }
  v->typ = normal;
  return 0;
}

// liest nächste Verteilung (Parameter werden als String abgelegt)
// und einen anschließenden Zeilenumbruch
// Fehler: 2->Dateiende erreicht
//         3->zu wenig Speicher
int lese_next_verteilung_text_r( FILE* *fi, verteilung_text* v )
{
  int i;
  i = lese_next_verteilung_text( fi, v );
  if( i != 0 )
  {
    return i;
  }
  lese_next_ret( fi );
  return 0;
}

// liest das (spalte + 1)-te Element in der (zeile + 1)-ten
// Zeile ab aktueller Position; Zeilen ohne Zahleneintrag werden nicht
// als solche gezählt
// Fehler: 1->nicht genug Einträge in der Zeile
//         2->Dateiende erreicht (nicht genug Zeilen)
int lese_pos_double( FILE* *fi, int zeile, int spalte, double *d )
{
  while( zeile > 0 )
  {
    if( lese_next_double_r( fi, d ) != 0 )
    {
      return 2;
    }
    zeile--;
  }
  if( lese_next_double( fi, d ) != 0 )
  {
    return 2;
  }
  while( spalte > 0 )
  {
    if( lese_next_double_in_line( fi, d ) != 0 )
    {
      return 1;
    }
    spalte--;
  }
  return 0;
}

