#ifndef LOAD_OP

#define LOAD_OP

#include <stdio.h>
#include "zufall.h"

// liest bis über das nächste '\n'
// Fehler: 1->Dateiende erreicht
int lese_next_ret( FILE* *fi );

// liest nächste ganze Zahl in der aktuellen Zeile
// Fehler: 1->Zeilenende erreicht bevor eine Zahl gelesen wurde
//         2->Dateiende erreicht bevor eine Zahl gelesen wurde
int lese_next_int_in_line( FILE* *fi, int* a );

// liest nächste ganze Zahl in aktueller Zeile und
// einen anschließenden Zeilenumbruch
// Fehler: 1->Zeilenende erreicht bevor eine Zahl gelesen wurde
//         2->Dateiende erreicht bevor eine Zahl gelesen wurde
int lese_next_int_in_line_r( FILE* *fi, int* a );

// liest nächste ganze Zahl (egal in welcher Zeile)
// Fehler: 1->Dateiende erreicht bevor eine Zahl gelesen wurde
int lese_next_int( FILE* *fi, int* a );

// liest nächste ganze Zahl
// und überliest alle Zeichen bis zum Zeilenumbruch
// Fehler: 1->Dateiende erreicht bevor eine Zahl gelesen wurde
int lese_next_int_r( FILE* *fi, int* a );

// liest nächste gebrochene Zahl mit Vorzeichen (! bei "ab.de"
// wird der "." als 0.0 gedeutet; ".." liefert sogar bei 2 Aufrufen
// den Wert 0; bei ".e+vg" wird ".e+" weggelesen !)
// Fehler: 1->Zeilenende erreicht
//         2->Dateiende erreicht
int lese_next_double_in_line( FILE* *fi, double* a );

// liest nächste gebrochene Zahl mit Vorzeichen (! bei "ab.de"
// wird der "." als 0.0 gedeutet; ".." liefert sogar bei 2 Aufrufen
// den Wert 0; bei ".e+vg" wird ".e+" weggelesen !)
// und einen Zeilenumbruch
// Fehler: 1->Zeilenende erreicht bevor Zahl gelesen wurde
//         2->Dateiende erreicht
int lese_next_double_in_line_r( FILE* *fi, double* a );

// liest nächste gebrochene Zahl mit Vorzeichen (! bei "ab.de"
// wird der "." als 0.0 gedeutet; ".." liefert sogar bei 2 Aufrufen
// den Wert 0; bei ".e+vg" wird ".e+" weggelesen !)
// Fehler: 1->Zeilenende erreicht
//         2->Dateiende erreicht
int lese_next_double( FILE* *fi, double* a );

// liest nächste gebrochene Zahl mit Vorzeichen (! bei "ab.de"
// wird der "." als 0.0 gedeutet; ".." liefert sogar bei 2 Aufrufen
// den Wert 0; bei ".e+vg" wird ".e+" weggelesen !)
// und überliest alle Zeichen bis zum Zeilenumbruch
// Fehler: 1->Dateiende erreicht
int lese_next_double_r( FILE* *fi, double* a );

// liest solange weiter, bis ein in a[] enthaltenes Zeichen
// auftritt; dieses wird nach b gegeben; a[] muß mit '\0'
// abgeschlossen sein
// Fehler: 1->Zeilenende erreicht
//         2->Dateiende erreicht
int lese_next_char( FILE* *fi, char* a, char *b );

// liest solange weiter, bis ein in a[] enthaltenes Zeichen
// auftritt; dieses wird nach b gegeben; a[] muß mit '\0'
// abgeschlossen sein und überliest alle Zeichen bis zum Zeilenumbruch
// Fehler: 1->Dateiende erreicht
int lese_next_char_r( FILE* *fi, char* a, char *b );

// liest nächsten String in der aktuellen Zeile
// bis ein in a enthaltenes Zeichen auftritt;
// der Speicher für b wird reserviert, wenn *b != NULL, so wird der
// Speicher wiederverwendet
// Fehler: 1->leeren String gelesen
//         2->Dateiende erreicht (b enthält alle bis dahin gelesenen
//            Zeichen)
//         3->zu wenig Speicher
int lese_next_string_in_line( FILE* *fi, char* a, char** b );

// liest nächste ganzzahlige Verteilung
// Fehler: 1->Zeilenende erreicht
//         2->Dateiende erreicht
int lese_next_verteilung_int_in_line( FILE* *fi, verteilung_int* v );

// liest nächste ganzzahlige Verteilung (egal in welcher Zeile)
// und einen anschließenden Zeilenwechsel
// Fehler: 1->Dateiende erreicht oder keine gültige Verteilung gefunden
int lese_next_verteilung_int( FILE* *fi, verteilung_int* v );

// liest nächste ganzzahlige Verteilung (egal in welcher Zeile)
// und einen anschließenden Zeilenwechsel
// Fehler: 1->Dateiende erreicht oder keine gültige Verteilung gefunden
int lese_next_verteilung_int_r( FILE* *fi, verteilung_int* v );

// liest nächste gebrochene Verteilung in aktueller Zeile
// Fehler: 1->Zeilenende erreicht
//         2->Dateiende erreicht oder keine gültige Verteilung gefunden
int lese_next_verteilung_double_in_line(
  FILE**             fi,
  verteilung_double* v );

// liest nächste gebrochene Verteilung
// Fehler: 1->Zeilenende erreicht
//         2->Dateiende erreicht
int lese_next_verteilung_double( FILE* *fi, verteilung_double* v );

// liest nächste gebrochene Verteilung (egal in welcher Zeile)
// und einen anschließenden Zeilenumbruch
// Fehler: 1->Dateiende erreicht oder keine gültige Verteilung gefunden
int lese_next_verteilung_double_r( FILE* *fi, verteilung_double* v );

// liest nächste Verteilung (Parameter werden als String abgelegt)
// Fehler: 2->Dateiende erreicht
//         3->zu wenig Speicher
int lese_next_verteilung_text( FILE* *fi, verteilung_text* v );

// liest nächste Verteilung (Parameter werden als String abgelegt)
// und überliest alle Zeichen bis zum Zeilenumbruch
// Fehler: 2->Dateiende erreicht
//         3->zu wenig Speicher
int lese_next_verteilung_text_r( FILE* *fi, verteilung_text* v );

// liest das (spalte + 1)-te Element in der (zeile + 1)-ten Zeile
// ab aktueller Position; Zeilen ohne Zahleneintrag werden nicht
// als solche gezählt
// Fehler: 1->nicht genug Einträge in der Zeile
//         2->Dateiende erreicht (nicht genug Zeilen)
int lese_pos_double( FILE* *fi, int zeile, int spalte, double *d );

#endif
