// Speichert Zahlen, dabei wird immer die gleiche Anzahl an Stellen
// in der Datei belegt (double = 22; int = 11)

#include <stdio.h>
#include <math.h>

// speichert den int in die angegebene Datei
// ohne abschließendes '\n'
// Fehler: 1->Schreibvorgang negativ
int save_int( FILE* *fi, int a )
{
  char c[11]; // ein int hat maximal 10 Stellen und ein Vorzeichen
  int i;
  if( a < 0 )
  {
    c[10] = '-';
    a = -a;
  }
  else
  {
    c[10] = ' ';
  }
  for( i=0; i<9; i++ )
  {
    c[i] = (char)('0' + a % 10);
    a /= 10;
  }
  c[9] = (char)('0' + a);
  for( i=10; i>=0; i-- )
  {
    if( fputc( c[i], *fi ) == EOF )
    {
      return 1;
    }
  }
  return 0;
}

// speichert den int in die angegebene Datei
// mit abschließenden '\n'
// Fehler: 1->Schreibvorgang negativ
int save_int_r( FILE* *fi, int a )
{
  if( save_int( fi, a ) != 0 )
  {
    return 1;
  }
  if( fputc( '\n', *fi ) == EOF )
  {
    return 1;
  }
  return 0;
}

// speichert den int in die angegebene Datei
// mit abschließenden ' '
// Fehler: 1->Schreibvorgang negativ
int save_int_s( FILE* *fi, int a )
{
  if( save_int( fi, a ) != 0 )
  {
    return 1;
  }
  if( fputc( ' ', *fi ) == EOF )
  {
    return 1;
  }
  return 0;
}

// speichert einen double in die angegebene Datei
// ohne abschließendes '\n'
// Fehler: 1->Schreibvorgang negativ
int save_double( FILE* *fi, double a )
{
  char c[22]; // 15 Stellen; einen Punkt, zwei Vorzeichen,
  //             ein 'e', drei Stellen Exponent
  int e;
  c[2] = '.';
  c[17] = 'e';
  if( a < 0.0 )
  {
    c[0] = '-';
    a = -a;
  }
  else
  {
    c[0] = '+';
  }
  if( a != 0.0 )
  {
    e = ((int)(log10( a ))); // Exponent ermitteln
    if( a < 1 )
    {
      e--; // eigentlich hätte von Null weg gerundet werden müssen
      c[18] = '-';
      e = -e;
      a *= pow( 10.0, ((double)(e)) );
      if( a >= 10.0 )
      { // galt zufällig log10( a ) = ((int)(log10( a ))) ?
        a /= 10.0;
        e--;
      }
    }
    else
    {
      a /= pow( 10.0, ((double)(e)) );
      c[18] = '+';
    }
  }
  else
  {
    e = 0;
    c[18] = '+';
  }
  c[21] = (char)('0' + e % 10);
  e /= 10;
  c[20] = (char)('0' + e % 10);
  e /= 10;
  c[19] = (char)('0' + e);
  c[1] = (char)('0' + ((int)(a)));
  for( e=3; e<16; e++ )
  {
    a -= ((int)(a));
    a *= 10;
    c[e] = (char)('0' + ((int)(a)));
  }
  a -= ((int)(a));
  a *= 10;
  a += 0.5; // letzte Stelle muß gerundet werden
  c[16] = (char)('0' + ((int)(a)));
  e = 16;
  while( c[e] > '9' ) // eventuell auftretenden Überlauf verfolgen
  {
    c[e] = '0';
    e--;
    if( e < 3 )
    {
      c[1]++;
      if( c[1] > '9' ) // auch an erster Stelle Überlauf
      {
        c[1] = '1';
        // Exponent ändern
        if( c[18] == '+' )
        {
          c[21]++;
          if( c[21] > '9' )
          {
            c[21] = '0';
            c[20]++;
            if( c[20] > '9' )
            {
              c[20] = '0';
              c[19]++; // hier kann kein Überlauf mehr auftreten
            }
          }
        }
        else
        {
          c[21]--;
          if( c[21] < '0' )
          {
            c[21] = '9';
            c[20]--;
            if( c[20] < '0' )
            {
              c[20] = '9';
              c[19]--; // hier kann kein Unterlauf mehr auftreten
            }
          }
          else // ist Exponent zufällig Null geworden?
          {
            if( ( c[21] == '0' ) &&
                ( c[20] == '0' ) &&
                ( c[19] == '0' ) )
            {
              c[18] = '+';
            }
          }
        }
      }
      break;
    }
    else
    {
      c[e]++;
    }
  }
  for( e=0; e<22; e++ )
  {
    if( fputc( c[e], *fi ) == EOF )
    {
      return 1;
    }
  }
  return 0;
}

// speichert einen double in die angegebene Datei
// mit abschließenden '\n'
// Fehler: 1->Schreibvorgang negativ
int save_double_r( FILE* *fi, double a )
{
  if( save_double( fi, a ) != 0 )
  {
    return 1;
  }
  if( fputc( '\n', *fi ) == EOF )
  {
    return 1;
  }
  return 0;
}

// speichert einen double in die angegebene Datei
// mit abschließenden ' '
// Fehler: 1->Schreibvorgang negativ
int save_double_s( FILE* *fi, double a )
{
  if( save_double( fi, a ) != 0 )
  {
    return 1;
  }
  if( fputc( ' ', *fi ) == EOF )
  {
    return 1;
  }
  return 0;
}

// speichert den in a übergebenen Text;
// das letzte Zeichen von a muß '\0' sein!
// Fehler: 1->Schreibvorgang nicht erfolgreich
int save_text( FILE* *fi, char* a )
{
  int i;
  i = 0;
  while( a[i] != '\0' )
  {
    if( fputc( a[i], *fi ) == EOF )
    {
      return 1;
    }
    i++;
  }
  return 0;
}

// speichert ein '\n'
// Fehler: 1->Schreibvorgang nicht erfolgreich
int save_ret( FILE* *fi )
{
  return save_text( fi, "\n" );
}
