#include <malloc.h>
#include <stdlib.h>
#include "listen.h"
#include "zufall.h"

#ifdef GRAPHEN
#include "../prgx/graphen.h"
#endif

#ifdef DGRAPHEN
#include "../prgx/dgraphen.h"
#endif

// berechnet den Konstruktionsbaum b zu einem sp-Graphen g;
// linearer paralleler Vereinigungsbaum
// Info für Wahl von "mehr" bei vorhergehenden Routinen:
// baue_sp_baum legt in g zwei temporäre Knoten an
// mehrg gibt die zu erwartende Knotenanzahl für g an, mehrb die für b
// in beschrift (Speicher wird reserviert!) liegt unter i die
// Beschriftung (1..parallel; 2..sequentiell) für Knoten i+plus
// die Knoten mit einer Nummer kleiner als plus sind Blätter in b
// und entsprechen den Knoten von g
// Rückgabe: -1->zu wenig Speicher
//           -2->g ist garantiert kein sp-Graph (es kann aber sein,
//               daß g keiner ist und der Alg trotzdem einen
//               Baum berechnet!!)
//           >0 Nummer der Wurzel von b
int baue_sp_baum_linear(
  graph* g,
  graph* b,
  int**  beschrift,
  int*   plus,
  int    mehrg,
  int    mehrb )
{
  int m, i, j, k, x, y, z, ref, r, akt;
  doppelintliste* merke;
  dil_element** merke_start;
  dil_element* tupel;
  doppelintliste s1;
  doppelintliste s2;
  int* wert;
  int* besucht;
  del_graph( b ); // b löschen
  m = max_knoten( g );
  if( m == 0 ) // ist g leer ?
  {
    return 0;
  }
  // benötigte Felder
  merke = (doppelintliste*) malloc( ( m + 2 ) *
                                    sizeof( doppelintliste ) );
  if( merke == NULL )
  {
    return -1;
  }
  merke_start = (dil_element**) malloc( ( m + 2 ) *
                                        sizeof( dil_element* ) );
  if( merke_start == NULL )
  {
    return -1;
  }
  wert = (int*) malloc( ( m + 2 ) * sizeof( int ) );
  if( wert == NULL )
  {
    return -1;
  }
  besucht = (int*) malloc( ( m + 2 ) * sizeof( int ) );
  if( besucht == NULL )
  {
    return -1;
  }
  (*beschrift) = (int*) malloc( ( 2 + m ) * sizeof( int ) );
  if( *beschrift == NULL )
  {
    return -1;
  }
  // füge in g zwei Dummy-Knoten ein
  if( add_knoten_mehr( g, m+1, mehrg ) != 0 )
  {
    return -1;
  }
  if( add_knoten_mehr( g, m+2, mehrg ) != 0 )
  {
    return -1;
  }
  k = 0; // zur Eingangsgradbestimmung unterer Dummyknoten
  for( i=0; i<m; i++ )
  {
    besucht[i] = eing_grad( g, i+1 );
    if( besucht[i] == 0 )
    {
      if( add_kante_mehr( g, m+1, i+1, mehrg ) != 0 )
      {
        return -1;
      }
      wert[i] = besucht[i] = 1;
    }
    else
    {
      wert[i] = besucht[i];
    }
    init_dil( &(merke[i]) );
    merke_start[i] = NULL;
    if( ziel_knoten( g, i+1 ) == 0 )
    {
      if( add_kante_mehr( g, i+1, m+2, mehrg ) != 0 )
      {
        return -1;
      }
      k++; // wegen Eingangsgrad
    }
  }
  wert[m+1] = besucht[m+1] = k;
  init_dil( &(merke[m+1]) );
  merke_start[m+1] = NULL;
  init_dil( &s1 );
  init_dil( &s2 );
  // oberen Dummyknoten auf s1
  if( insert_end_dil( &s1, m+1 ) != 0 )
  {
    return -1;
  }
  tupel = NULL; // kein Tupel vorhanden
  ref = -1; // zuletzt beschriebener Merke
  akt = m + 3; // nächster freier Knoten
  while( 1 )
  {
    x = del_end_dil( &s1 );
    if( x == 0 ) // s1 leer?
    {
      break;
    }
    if( x < 0 ) // Tupel ?
    {
      tupel = NULL; // kein Verweis auf Tupel vorhanden
      y = del_end_dil( &s1 );
      r = 0; // von merke[ref-1] oder s2 lesen
      if( ref == -1 )
      {
        r = 1; // die letzten Knoten von s2 vereinigen
        ref = 1; // zur Vermeidung von Fehlern; beliebigen Stack wählen
        i = del_end_dil( &s2 );
      }
      else
      {
        wert[ref-1] += x + 1; // Wert des Nachfolgeknotens anpassen
        if( wert[ref-1] < 1 )
        {
          return -2;
        }
        i = del_end_dil( &(merke[ref-1]) );
        if( i == 0 )
        {
          r = 1; // die letzten Knoten von s2 vereinigen
          i = del_end_dil( &s2 );
        }
      }
      while( x < -1 ) // erster Ast
      {
        x++;
        if( r == 0 )
        {
          j = del_end_dil( &(merke[ref-1]) );
          if( j == 0 )
          {
            r = 1; // von s2 weiter lesen
            j = del_end_dil( &s2 );
          }
        }
        else
        {
          j = del_end_dil( &s2 );
        }
        if( add_kante_mehr( b, akt, i, mehrb ) != 0 )
        {
          return -1;
        }
        if( add_kante_mehr( b, akt, j, mehrb ) != 0 )
        {
          return -1;
        }
        i = akt;
        (*beschrift)[ akt-m-3 ] = 1; // parallele Vereinigung
        akt++; // nächster freier Knoten
      }
      if( r == 0 )
      {
        k = del_end_dil( &(merke[ref-1]) );
        if( k == 0 )
        {
          r = 1; // von s2 weiter lesen
          k = del_end_dil( &s2 );
        }
      }
      else
      {
        k = del_end_dil( &s2 );
      }
      while( y < -1 ) // zweiter Ast
      {
        y++;
        if( r == 0 )
        {
          j = del_end_dil( &(merke[ref-1]) );
          if( j == 0 )
          {
            r = 1; // von s2 weiter lesen
            j = del_end_dil( &s2 );
          }
        }
        else
        {
          j = del_end_dil( &s2 );
        }
        if( add_kante_mehr( b, akt, k, mehrb ) != 0 )
        {
          return -1;
        }
        if( add_kante_mehr( b, akt, j, mehrb ) != 0 )
        {
          return -1;
        }
        k = akt;
        (*beschrift)[ akt-m-3 ] = 1; // parallele Vereinigung
        akt++; // nächster freier Knoten
      }
      if( r != 0 )
      {
        merke_start[ref-1] = get_endpos_dil( &s2 ); // Stackende merken
      }
      // rechte Kante zuerst
      if( add_kante_mehr( b, akt, i, mehrb ) != 0 )
      {
        return -1;
      }
      if( add_kante_mehr( b, akt, k, mehrb ) != 0 ) // linke Kante
      {
        return -1;
      }
      (*beschrift)[ akt-m-3 ] = 2; // serielle Vereinigung
      insert_end_dil( &(merke[ref-1]), akt ); // neuen Knoten ablegen
      akt++;
      continue;
    }
    reset_kind( g, x );
    j = i = next_kind( g, x );
    if( i != 0 )
    {
      z = 1;
      while( 1 )
      {
        i = next_kind( g, x );
        if( i == 0 )
        {
          break;
        }
        if( i > j ) // größtes Kind suchen
        {
          j = i;
        }
        z++; // Kinder zählen
      }
      besucht[j-1]--;
      if( besucht[j-1] == 0 )
      {
        if( insert_end_dil( &s2, x ) != 0 ) // x auf s2 schreiben
        {
          return -1;
        }
        // Hilfsstack zurückschreiben
        insert_pos_dil( &s2, merke_start[j-1], &(merke[j-1]) );
        while( 1 )
        {
          if( tupel == NULL )
          {
            reset_end_dil( &s1 ); // von hinten nach vorne in s1 lesen
            i = get_dil( &s1 ); // oberstes Tupel suchen
            while( i > 0 )
            {
              last_dil( &s1 );
              i = get_dil( &s1 );
            }
            if( i == 0 ) // kein Tupel vorhanden
            {
              break;
            }
          }
          else // Pointer auf Tupel vorhanden
          {
            reset_pos_dil( &s1, tupel );
            i = get_dil( &s1 );
          }
          if( wert[j-1] < -i )
          {
            set_dil( &s1, i + wert[j-1] - 1 );
            tupel = get_pos_dil( &s1 );
            break;
          }
          tupel = NULL; // oberstes Tupel wird gelöscht
          wert[j-1] += i + 1;
          last_dil( &s1 );
          k = del_dil( &s1 ); // vorderen Wert des Tupels lesen
          del_dil( &s1 ); // Tupel löschen
          // Tupel ausführen
          y = del_end_dil( &s2 );
          while( i < -1 ) // erster Ast
          {
            i++;
            r = del_end_dil( &s2 );
            if( add_kante_mehr( b, akt, y, mehrb ) != 0 )
            {
              return -1;
            }
            if( add_kante_mehr( b, akt, r, mehrb ) != 0 )
            {
              return -1;
            }
            y = akt;
            (*beschrift)[ akt-m-3 ] = 1; // parallele Vereinigung
            akt++;
          }
          i = del_end_dil( &s2 );
          while( k < -1 ) // zweiter Ast
          {
            k++;
            r = del_end_dil( &s2 );
            if( add_kante_mehr( b, akt, i, mehrb ) != 0 )
            {
              return -1;
            }
            if( add_kante_mehr( b, akt, r, mehrb ) != 0 )
            {
              return -1;
            }
            i = akt;
            (*beschrift)[ akt-m-3 ] = 1; // parallele Vereinigung
            akt++;
          }
          if( add_kante_mehr( b, akt, y, mehrb ) != 0 )
          {
            return -1;
          }
          if( add_kante_mehr( b, akt, i, mehrb ) != 0 )
          {
            return -1;
          }
          (*beschrift)[akt-m-3 ] = 2; // sequentielle Vereinigung
          if( insert_end_dil( &s2, akt ) != 0 )
          {
            return -1;
          }
          akt++;
        } // Ende Tupelbearbeitung
        // neues Tupel schreiben
        if( insert_end_dil( &s1, -wert[j-1] ) != 0 )
        {
          return -1;
        }
        if( insert_end_dil( &s1, -z ) != 0 )
        {
          return -1;
        }
        tupel = get_endpos_dil( &s1 );
        reset_kind( g, x );
        while( 1 )
        {
          i = next_kind( g, x );
          if( i == 0 )
          {
            break;
          }
          if( insert_end_dil( &s1, i ) != 0 )
          {
            return -1;
          }
        }
        continue;
      } // endif alle Vorgängerknoten besucht
      // nicht alle Vorgängerknoten besucht
      if( empty_dil( &(merke[j-1]) ) == 0 )
      { // noch nie etwas auf den Hilfsstack geschrieben
        merke_start[j-1] = get_endpos_dil( &s2 );
        // aktuellen Knoten nicht auf s2 sondern auf den
        // Hilfsstack schreiben
        if( insert_end_dil( &(merke[j-1]), x ) != 0 )
        {
          return -1;
        }
        ref = j;
        continue;
      }
      // alles was ab Marke steht auf anderen Stack schieben
      insert_list_end_dil( &(merke[j-1]), &s2, merke_start[j-1] );
      // aktuellen Knoten nicht auf s2 sondern auf den
      // Hilfsstack schreiben
      if( insert_end_dil( &(merke[j-1]), x ) != 0 )
      {
        return -1;
      }
      ref = j;
      continue;
    } // endif Kinder vorhanden
    if( insert_end_dil( &s2, x ) != 0 ) // x auf s2 schreiben
    {
      return -1;
    }
  }
  free_dil( &s1 );
  free_dil( &s2 );
  for( i=0; i<m; i++ )
  {
    free_dil( &(merke[i]) );
  }
  free_dil( &(merke[m+1]) );
  free( merke );
  free( merke_start );
  free( wert );
  free( besucht );
  // Dummyknoten in g löschen
  del_knoten( g, m+1 );
  del_knoten( g, m+2 );
  // Dummyknoten aus b entfernen
  akt--; // Nummer der Wurzel
  reset_vorg( b, m+1 ); // erster Dummy
  i = next_vorg( b, m+1 );
  reset_vorg( b, i );
  j = next_vorg( b, i );
  // Nachbar von m+1 bestimmen
  reset_kind( b, i );
  k = next_kind( b, i );
  if( k == m+1 )
  {
    k = next_kind( b, i );
  }
  if( j == 0 )
  {
    // i war Wurzel, einfach m+1 und i löschen und neue Wurzel
    // definieren
    del_knoten( b, m+1 );
    del_knoten( b, i );
    akt = k;
  }
  else
  {
    // k anstelle von i an j hängen (Reihenfolge beachten!!)
    reset_kind( b, j );
    x = next_kind( b, j );
    if( x == i ) // linkes Kind ersetzen
    {
      del_kante( b, j, i );
      if( add_kante_mehr( b, j, k, mehrb ) != 0 )
      {
        return -1;
      }
    }
    else // rechtes Kind ersetzen
    {
      del_kante( b, j, x );
      del_kante( b, j, i );
      if( add_kante_mehr( b, j, k, mehrb ) != 0 )
      {
        return -1;
      }
      if( add_kante_mehr( b, j, x, mehrb ) != 0 )
      {
        return -1;
      }
    }
    // überflüssige Knoten m+1 und i löschen
    del_knoten( b, m+1 );
    del_knoten( b, i );
  }
  reset_vorg( b, m+2 ); // zweiter Dummy
  i = next_vorg( b, m+2 );
  reset_vorg( b, i );
  j = next_vorg( b, i );
  // Nachbar von m+2 bestimmen
  reset_kind( b, i );
  k = next_kind( b, i );
  if( k == m+2 )
  {
    k = next_kind( b, i );
  }
  if( j == 0 )
  {
    // i war Wurzel, einfach m+2 und i löschen und neue Wurzel
    // definieren
    del_knoten( b, m+2 );
    del_knoten( b, i );
    akt = k;
  }
  else
  {
    // k anstelle von i an j hängen (Reihenfolge beachten!!)
    reset_kind( b, j );
    x = next_kind( b, j );
    if( x == i ) // linkes Kind ersetzen
    {
      del_kante( b, j, i );
      if( add_kante_mehr( b, j, k, mehrb ) != 0 )
      {
        return -1;
      }
    }
    else // rechtes Kind ersetzen
    {
      del_kante( b, j, x );
      del_kante( b, j, i );
      if( add_kante_mehr( b, j, k, mehrb ) != 0 )
      {
        return -1;
      }
      if( add_kante_mehr( b, j, x, mehrb ) != 0 )
      {
        return -1;
      }
    }
    // überflüssige Knoten m+2 und i löschen
    del_knoten( b, m+2 );
    del_knoten( b, i );
  }
  *plus = m + 3;
  return akt;
}

// Hilfsroutine zu baue_sp_baum_binaer;
// balancierten parallelen Vereinigungsbaum erstellen
// Fehler: -1->zu wenig Speicher
int baue_sp_baum_binaer_help2(
  graph*          b,
  doppelintliste* s2,
  doppelintliste* merke,
  int**           beschrift,
  int             m,
  int             mehrb,
  int*            r,
  int*            akt,
  int             x )
{
  int i, j;
  if( x == -1 )
  {
    if( *r == 0 )
    {
      i = del_end_dil( merke );
      if( i != 0 )
      {
        return i;
      }
      *r = 1;
    }
    return del_end_dil( s2 );
  }
  // größere Hälfte
  i = baue_sp_baum_binaer_help2( b, s2, merke, beschrift, m,
                                 mehrb, r, akt, x - x/2 );
  if( i < 0 )
  {
    return -1;
  }
  // kleinere Hälfte
  j = baue_sp_baum_binaer_help2( b, s2, merke, beschrift, m,
                                 mehrb, r, akt, x/2 );
  if( j < 0 )
  {
    return -1;
  }
  if( add_kante_mehr( b, *akt, j, mehrb ) != 0 )
  {
    return -1;
  }
  if( add_kante_mehr( b, *akt, i, mehrb ) != 0 )
  {
    return -1;
  }
  (*beschrift)[ (*akt)-m-3 ] = 1; // parallele Vereinigung
  i = *akt;
  (*akt)++;
  return i;
}

// Hilfsroutine zu baue_sp_baum_binaer;
// balancierten parallelen Vereinigungsbaum erstellen
// Fehler: -1->zu wenig Speicher
int baue_sp_baum_binaer_help1(
  graph*          b,
  doppelintliste* s2,
  int**           beschrift,
  int             m,
  int             mehrb,
  int*            akt,
  int             x )
{
  int i, j;
  if( x == -1 )
  {
    return del_end_dil( s2 );
  }
  // größere Hälfte
  i = baue_sp_baum_binaer_help1( b, s2, beschrift, m,
                                 mehrb, akt, x - x/2 );
  if( i < 0 )
  {
    return -1;
  }
  // kleinere Hälfte
  j = baue_sp_baum_binaer_help1( b, s2, beschrift, m,
                                 mehrb, akt, x/2 );
  if( j < 0 )
  {
    return -1;
  }
  if( add_kante_mehr( b, *akt, j, mehrb ) != 0 )
  {
    return -1;
  }
  if( add_kante_mehr( b, *akt, i, mehrb ) != 0 )
  {
    return -1;
  }
  (*beschrift)[ (*akt)-m-3 ] = 1; // parallele Vereinigung
  i = *akt;
  (*akt)++;
  return i;
}

// berechnet den Konstruktionsbaum b zu einem sp-Graphen g;
// balancierten parallelen Vereinigungsbaum erstellen
// Info für Wahl von "mehr" bei vorhergehenden Routinen:
// baue_sp_baum legt in g zwei temporäre Knoten an
// mehrg gibt die zu erwartende Knotenanzahl für g an, mehrb die für b
// in beschrift (Speicher wird reserviert!) liegt unter i die
// Beschriftung (1..parallel; 2..sequentiell) für Knoten i+plus
// die Knoten mit einer Nummer kleiner als plus sind Blätter in b
// und entsprechen den Knoten von g
// Rückgabe: -1->zu wenig Speicher
//           -2->g ist garantiert kein sp-Graph (es kann aber sein,
//               daß g keiner ist und der Alg trotzdem einen
//               Baum berechnet!!)
//           >0 Nummer der Wurzel von b
int baue_sp_baum_binaer(
  graph* g,
  graph* b,
  int**  beschrift,
  int*   plus,
  int    mehrg,
  int    mehrb )
{
  int m, i, j, k, x, y, z, ref, r, akt;
  doppelintliste* merke;
  dil_element** merke_start;
  dil_element* tupel;
  doppelintliste s1;
  doppelintliste s2;
  int* wert;
  int* besucht;
  del_graph( b ); // b löschen
  m = max_knoten( g );
  if( m == 0 ) // ist g leer ?
  {
    return 0;
  }
  // benötigte Felder
  merke = (doppelintliste*) malloc( ( m + 2 ) *
                                    sizeof( doppelintliste ) );
  if( merke == NULL )
  {
    return -1;
  }
  merke_start = (dil_element**) malloc( ( m + 2 ) *
                                        sizeof( dil_element* ) );
  if( merke_start == NULL )
  {
    return -1;
  }
  wert = (int*) malloc( ( m + 2 ) * sizeof( int ) );
  if( wert == NULL )
  {
    return -1;
  }
  besucht = (int*) malloc( ( m + 2 ) * sizeof( int ) );
  if( besucht == NULL )
  {
    return -1;
  }
  (*beschrift) = (int*) malloc( ( 2 + m ) * sizeof( int ) );
  if( *beschrift == NULL )
  {
    return -1;
  }
  // füge in g zwei Dummy-Knoten ein
  if( add_knoten_mehr( g, m+1, mehrg ) != 0 )
  {
    return -1;
  }
  if( add_knoten_mehr( g, m+2, mehrg ) != 0 )
  {
    return -1;
  }
  k = 0; // zur Eingangsgradbestimmung unterer Dummyknoten
  for( i=0; i<m; i++ )
  {
    besucht[i] = eing_grad( g, i+1 );
    if( besucht[i] == 0 )
    {
      if( add_kante_mehr( g, m+1, i+1, mehrg ) != 0 )
      {
        return -1;
      }
      wert[i] = besucht[i] = 1;
    }
    else
    {
      wert[i] = besucht[i];
    }
    init_dil( &(merke[i]) );
    merke_start[i] = NULL;
    if( ziel_knoten( g, i+1 ) == 0 )
    {
      if( add_kante_mehr( g, i+1, m+2, mehrg ) != 0 )
      {
        return -1;
      }
      k++; // wegen Eingangsgrad
    }
  }
  wert[m+1] = besucht[m+1] = k;
  init_dil( &(merke[m+1]) );
  merke_start[m+1] = NULL;
  init_dil( &s1 );
  init_dil( &s2 );
  // oberen Dummyknoten auf s1
  if( insert_end_dil( &s1, m+1 ) != 0 )
  {
    return -1;
  }
  tupel = NULL; // kein Tupel vorhanden
  ref = -1; // zuletzt beschriebener Merke
  akt = m + 3; // nächster freier Knoten
  while( 1 )
  {
    x = del_end_dil( &s1 );
    if( x == 0 ) // s1 leer?
    {
      break;
    }
    if( x < 0 ) // Tupel ?
    {
      tupel = NULL; // kein Verweis auf Tupel vorhanden
      y = del_end_dil( &s1 );
      r = 0; // von merke[ref-1] oder s2 lesen
      if( ref == -1 )
      {
        r = 1; // die letzten Knoten von s2 vereinigen
        ref = 1; // zur Vermeidung von Fehlern; beliebigen Stack wählen
      }
      else
      {
        wert[ref-1] += x + 1; // Wert des Nachfolgeknotens anpassen
        if( wert[ref-1] < 1 )
        {
          return -2;
        }
      }
      i = baue_sp_baum_binaer_help2( b, &s2, &(merke[ref-1]),
                                     beschrift, m, mehrb, &r, &akt, x );
      if( i < 0 )
      {
        return -1;
      }
      k = baue_sp_baum_binaer_help2( b, &s2, &(merke[ref-1]),
                                     beschrift, m, mehrb, &r, &akt, y );
      if( k < 0 )
      {
        return -1;
      }
      if( r != 0 )
      {
        merke_start[ref-1] = get_endpos_dil( &s2 ); // Stackende merken
      }
      // rechte Kante zuerst
      if( add_kante_mehr( b, akt, i, mehrb ) != 0 )
      {
        return -1;
      }
      if( add_kante_mehr( b, akt, k, mehrb ) != 0 ) // linke Kante
      {
        return -1;
      }
      (*beschrift)[ akt-m-3 ] = 2; // serielle Vereinigung
      insert_end_dil( &(merke[ref-1]), akt ); // neuen Knoten ablegen
      akt++;
      continue;
    }
    reset_kind( g, x );
    j = i = next_kind( g, x );
    if( i != 0 )
    {
      z = 1;
      while( 1 )
      {
        i = next_kind( g, x );
        if( i == 0 )
        {
          break;
        }
        if( i > j ) // größtes Kind suchen
        {
          j = i;
        }
        z++; // Kinder zählen
      }
      besucht[j-1]--;
      if( besucht[j-1] == 0 )
      {
        if( insert_end_dil( &s2, x ) != 0 ) // x auf s2 schreiben
        {
          return -1;
        }
        // Hilfsstack zurückschreiben
        insert_pos_dil( &s2, merke_start[j-1], &(merke[j-1]) );
        while( 1 )
        {
          if( tupel == NULL )
          {
            reset_end_dil( &s1 ); // von hinten nach vorne in s1 lesen
            i = get_dil( &s1 ); // oberstes Tupel suchen
            while( i > 0 )
            {
              last_dil( &s1 );
              i = get_dil( &s1 );
            }
            if( i == 0 ) // kein Tupel vorhanden
            {
              break;
            }
          }
          else // Pointer auf Tupel vorhanden
          {
            reset_pos_dil( &s1, tupel );
            i = get_dil( &s1 );
          }
          if( wert[j-1] < -i )
          {
            set_dil( &s1, i + wert[j-1] - 1 );
            tupel = get_pos_dil( &s1 );
            break;
          }
          tupel = NULL; // oberstes Tupel wird gelöscht
          wert[j-1] += i + 1;
          last_dil( &s1 );
          k = del_dil( &s1 ); // vorderen Wert des Tupels lesen
          del_dil( &s1 ); // Tupel löschen
          // Tupel ausführen
          y = baue_sp_baum_binaer_help1( b, &s2, beschrift,
                                         m, mehrb, &akt, i );
          if( y < 0 )
          {
            return -1;
          }
          i = baue_sp_baum_binaer_help1( b, &s2, beschrift,
                                         m, mehrb, &akt, k );
          if( i < 0 )
          {
            return -1;
          }
          if( add_kante_mehr( b, akt, y, mehrb ) != 0 )
          {
            return -1;
          }
          if( add_kante_mehr( b, akt, i, mehrb ) != 0 )
          {
            return -1;
          }
          (*beschrift)[akt-m-3 ] = 2; // sequentielle Vereinigung
          if( insert_end_dil( &s2, akt ) != 0 )
          {
            return -1;
          }
          akt++;
        } // Ende Tupelbearbeitung
        // neues Tupel schreiben
        if( insert_end_dil( &s1, -wert[j-1] ) != 0 )
        {
          return -1;
        }
        if( insert_end_dil( &s1, -z ) != 0 )
        {
          return -1;
        }
        tupel = get_endpos_dil( &s1 );
        reset_kind( g, x );
        while( 1 )
        {
          i = next_kind( g, x );
          if( i == 0 )
          {
            break;
          }
          if( insert_end_dil( &s1, i ) != 0 )
          {
            return -1;
          }
        }
        continue;
      } // endif alle Vorgängerknoten besucht
      // nicht alle Vorgängerknoten besucht
      if( empty_dil( &(merke[j-1]) ) == 0 )
      { // noch nie etwas auf den Hilfsstack geschrieben
        merke_start[j-1] = get_endpos_dil( &s2 );
        if( insert_end_dil( &(merke[j-1]), x ) != 0 )
        { // aktuellen Knoten nicht auf s2 sondern auf
          // den Hilfsstack schreiben
          return -1;
        }
        ref = j;
        continue;
      }
      // alles was ab Marke steht auf anderen Stack schieben
      insert_list_end_dil( &(merke[j-1]), &s2, merke_start[j-1] );
      // aktuellen Knoten nicht auf s2 sondern auf den
      // Hilfsstack schreiben
      if( insert_end_dil( &(merke[j-1]), x ) != 0 )
      {
        return -1;
      }
      ref = j;
      continue;
    } // endif Kinder vorhanden
    if( insert_end_dil( &s2, x ) != 0 ) // x auf s2 schreiben
    {
      return -1;
    }
  }
  free_dil( &s1 );
  free_dil( &s2 );
  for( i=0; i<m; i++ )
  {
    free_dil( &(merke[i]) );
  }
  free_dil( &(merke[m+1]) );
  free( merke );
  free( merke_start );
  free( wert );
  free( besucht );
  // Dummyknoten in g löschen
  del_knoten( g, m+1 );
  del_knoten( g, m+2 );
  // Dummyknoten aus b entfernen
  akt--; // Nummer der Wurzel
  reset_vorg( b, m+1 ); // erster Dummy
  i = next_vorg( b, m+1 );
  reset_vorg( b, i );
  j = next_vorg( b, i );
  // Nachbar von m+1 bestimmen
  reset_kind( b, i );
  k = next_kind( b, i );
  if( k == m+1 )
  {
    k = next_kind( b, i );
  }
  if( j == 0 )
  {
    // i war Wurzel, einfach m+1 und i löschen und neue
    // Wurzel definieren
    del_knoten( b, m+1 );
    del_knoten( b, i );
    akt = k;
  }
  else
  {
    // k anstelle von i an j hängen (Reihenfolge beachten!!)
    reset_kind( b, j );
    x = next_kind( b, j );
    if( x == i ) // linkes Kind ersetzen
    {
      del_kante( b, j, i );
      if( add_kante_mehr( b, j, k, mehrb ) != 0 )
      {
        return -1;
      }
    }
    else // rechtes Kind ersetzen
    {
      del_kante( b, j, x );
      del_kante( b, j, i );
      if( add_kante_mehr( b, j, k, mehrb ) != 0 )
      {
        return -1;
      }
      if( add_kante_mehr( b, j, x, mehrb ) != 0 )
      {
        return -1;
      }
    }
    // überflüssige Knoten m+1 und i löschen
    del_knoten( b, m+1 );
    del_knoten( b, i );
  }
  reset_vorg( b, m+2 ); // zweiter Dummy
  i = next_vorg( b, m+2 );
  reset_vorg( b, i );
  j = next_vorg( b, i );
  // Nachbar von m+2 bestimmen
  reset_kind( b, i );
  k = next_kind( b, i );
  if( k == m+2 )
  {
    k = next_kind( b, i );
  }
  if( j == 0 )
  {
    // i war Wurzel, einfach m+2 und i löschen und neue
    // Wurzel definieren
    del_knoten( b, m+2 );
    del_knoten( b, i );
    akt = k;
  }
  else
  {
    // k anstelle von i an j hängen (Reihenfolge beachten!!)
    reset_kind( b, j );
    x = next_kind( b, j );
    if( x == i ) // linkes Kind ersetzen
    {
      del_kante( b, j, i );
      if( add_kante_mehr( b, j, k, mehrb ) != 0 )
      {
        return -1;
      }
    }
    else // rechtes Kind ersetzen
    {
      del_kante( b, j, x );
      del_kante( b, j, i );
      if( add_kante_mehr( b, j, k, mehrb ) != 0 )
      {
        return -1;
      }
      if( add_kante_mehr( b, j, x, mehrb ) != 0 )
      {
        return -1;
      }
    }
    // überflüssige Knoten m+2 und i löschen
    del_knoten( b, m+2 );
    del_knoten( b, i );
  }
  *plus = m + 3;
  return akt;
}

// rekursive Hilfsroutine zu rnd_sp_graph
// Fehler: 1->zu wenig Speicher
int rnd_sp_graph_hilf(
  graph*    g,
  double    verh,
  int       m,
  int       nr,
  double*   baum,
  int       mehrg,
  intliste* quell,
  intliste* ziel )
{
  int f, x, y, i;
  double b, c;
  intliste ziel1, quell2;
  if( m == 0 ) // Blatt erreicht
  {
    if( add_anf_il( quell, nr ) != 0 )
    {
      return 1;
    }
    return add_anf_il( ziel, nr );
  }
  // Anzahl der Module in den Unterbäumen (rechts / links) bestimmen
  b = gleich_double( 0, 1 );
  b *= baum[ m ];
  i = m / 2;
  c = baum[i] * baum[ m - 1 - i ];
  while( c < b )
  {
    i++;
    c += baum[i] * baum[ m - 1 - i ];
  }
  init_il( &ziel1 );
  init_il( &quell2 );
  f = rnd_sp_graph_hilf( g, verh, i, nr, baum, mehrg, quell, &ziel1 );
  if( f != 0 )
  {
    return f;
  }
  f = rnd_sp_graph_hilf( g, verh, m-1-i, nr+1+i, baum, mehrg,
                         &quell2, ziel );
  if( f != 0 )
  {
    return f;
  }
  if( gleich_double( 0, 1 ) >= verh ) // parallele Vereinigung
  {
    // Mengen der Eingangsknoten und Ausgangsknoten vereinigen
    add_list_end_il( quell, &quell2 );
    add_list_end_il( ziel, &ziel1 );
    return 0;
  }
  // sequentielle Vereinigung
  // Ausgangs- mit Eingangsknoten verbinden
  while( 1 )
  {
    x = del_anf_il( &ziel1 );
    if( x == 0 )
    {
      break;
    }
    reset_il( &quell2 );
    while( 1 )
    {
      y = next_il( &quell2 );
      if( y == 0 )
      {
        break;
      }
      if( add_kante_mehr( g, x, y, mehrg ) != 0 )
      {
        return 1;
      }
    }
  }
  free_il( &quell2 );
  free_il( &ziel1 );
  return 0;
}

// erzeugt zufälligen serien-parallelen Graphen g mit m Knoten
// in verh steht das Verhältnis von paralleler zu serieller
// Vereinigung (0..nur parallel; 1..nur seriell)
// die Gesamtanzahl der Knoten wird mittels m übergeben
// mit mehrg kann die zu erwartende Knotenanzahl für g angegeben werden
// Fehler: 1->zu wenig Speicher
int rnd_sp_graph(
  graph* g,
  double verh,
  int    m,
  int    mehrg )
{
  intliste quell, ziel;
  int f, i, j;
  double* baum;
  init_il( &quell );
  init_il( &ziel );
  // Hilfsdaten für die Bestimmung der Unterbaumgrößen
  baum = (double*) malloc( ( m + 1 ) * sizeof( double ) );
  if( baum == NULL )
  {
    return 1;
  }
  baum[0] = 1.0;
  for( i=1; i<m; i++ )
  {
    baum[i] = 0.0;
    for( j=i-1; j>=i/2; j-- )
    {
      baum[i] += baum[j] * baum[ i - 1 - j ];
    }
  }
  if( create_knoten_graph_mehr( g, m, mehrg ) != 0 )
  {
    return 1;
  }
  f = rnd_sp_graph_hilf( g, verh, m-1, 1, baum, mehrg, &quell, &ziel );
  free_il( &quell );
  free_il( &ziel );
  free( baum );
  return f;
}
