#ifndef SP_GRAPH_OP

#define SP_GRAPH_OP

#ifdef GRAPHEN
#include "../prgx/graphen.h"
#endif

#ifdef DGRAPHEN
#include "../prgx/dgraphen.h"
#endif


// berechnet den Konstruktionsbaum b zu einem sp-Graphen g;
// linearer paralleler Vereinigungsbaum
// Info für Wahl von "mehr" bei vorhergehenden Routinen:
// baue_sp_baum legt in g zwei temporäre Knoten an
// mehrg gibt die zu erwartende Knotenanzahl für g an, mehrb die für b
// in beschrift (Speicher wird reserviert!) liegt unter i die
// Beschriftung (1..parallel; 2..sequentiell) für Knoten i+plus
// die Knoten mit einer Nummer kleiner als plus sind Blätter in b
// und entsprechen den Knoten von g
// Rückgabe: -1->zu wenig Speicher
//           -2->g ist garantiert kein sp-Graph (es kann aber sein,
//               daß g keiner ist und der Alg trotzdem einen
//               Baum berechnet!!)
int baue_sp_baum_linear(
  graph* g,
  graph* b,
  int**  beschrift,
  int*   plus,
  int    mehrg,
  int    mehrb );

// berechnet den Konstruktionsbaum b zu einem sp-Graphen g;
// balancierten parallelen Vereinigungsbaum erstellen
// Info für Wahl von "mehr" bei vorhergehenden Routinen:
// baue_sp_baum legt in g zwei temporäre Knoten an
// mehrg gibt die zu erwartende Knotenanzahl für g an, mehrb die für b
// in beschrift (Speicher wird reserviert!) liegt unter i die
// Beschriftung (1..parallel; 2..sequentiell) für Knoten i+plus
// die Knoten mit einer Nummer kleiner als plus sind Blätter in b
// und entsprechen den Knoten von g
// Rückgabe: -1->zu wenig Speicher
//           -2->g ist garantiert kein sp-Graph (es kann aber sein,
//               daß g keiner ist und der Alg trotzdem einen
//               Baum berechnet!!)
//           >0 Nummer der Wurzel von b
int baue_sp_baum_binaer(
  graph* g,
  graph* b,
  int**  beschrift,
  int*   plus,
  int    mehrg,
  int    mehrb );

// erzeugt zufälligen serien-parallelen Graphen g mit m Knoten
// in verh steht das Verhältnis von paralleler zu serieller
// Vereinigung (0..nur parallel; 1..nur seriell)
// die Gesamtanzahl der Knoten wird mittels m übergeben
// mit mehrg kann die zu erwartende Knotenanzahl für g angegeben werden
// Fehler: 1->zu wenig Speicher
int rnd_sp_graph(
  graph* g,
  double verh,
  int    m,
  int    mehrg );

#endif
