#include <float.h>
#include <limits.h>
#include <stdio.h>
#include "baum.h"
#include "load.h"
#include "save.h"


// sortiert Einträge von nr[] im Intervall [a; b) so, daß nr[a]
// den größten Eintrag aller nr[] in [a; b) in zeit[] hat;
// nrh ist Hilfsfeld
void merge_modul_zeit(
  double* zeit,
  int*    nr,
  int*    nrh,
  int     a,
  int     b )
{
  int c, i, j, k, h;
  if( a + 2 < b )
  {
    c = ( a + b + 1 ) / 2;
    merge_modul_zeit( zeit, nr, nrh, a, c );
    merge_modul_zeit( zeit, nr, nrh, c, b );
    for( i=a; i<c; i++ )
    {
      nrh[i-a] = nr[i];
    }
    i = 0;
    j = c;
    k = a;
    while( ( j < b ) && ( i < c-a ) ) // Pointer durchrücken
    {
      if( zeit[ nrh[i] ] < zeit[ nr[j] ] )
      {
        nr[k] = nr[j];
        j++;
      }
      else
      {
        nr[k] = nrh[i];
        i++;
      }
      k++;
    }
    if( j >= b )
    {
      while( i < c-a ) // Rest anhängen
      {
        nr[k] = nrh[i];
        i++;
        k++;
      }
    }
  }
  else // höchstens 2 Zahlen
  {
    if( a + 1 < b )
    {
      if( zeit[ nr[a] ] < zeit[ nr[ a + 1 ] ] )
      {
        h = nr[ a + 1 ];
        nr[ a + 1 ] = nr[a];
        nr[a] = h;
      }
    }
  }
}
/*
// sortiert Einträge von nr[] im Intervall [a; b) so, daß nr[a] den
// größten Eintrag aller nr[] in [a; b) in zeit[] hat, bei gleicher
// Zeit, entscheidet die Breite (breite[rf[nr[]]]); nrh ist Hilfsfeld
void merge_indirekt_modul_zeit_breit(
  double* zeit,
  int*    breite,
  int*    nr,
  int*    nrh,
  int*    rf,
  int     a,
  int     b )
{
  int c, i, j, k, h;
  if( a + 2 < b )
  {
    c = ( a + b + 1 ) / 2;
    merge_modul( zeit, breite, nr, nrh, rf, a, c );
    merge_modul( zeit, breite, nr, nrh, rf, c, b );
    for( i=a; i<c; i++ )
    {
      nrh[i-a] = nr[i];
    }
    i = 0;
    j = c;
    k = a;
    while( ( j < b ) && ( i < c-a ) ) // Pointer durchrücken
    {
      if( ( zeit[ nrh[i] ] < zeit[ nr[j] ] ) ||
          ( ( zeit[ nrh[i] ] == zeit[ nr[j] ] ) &&
            ( breite[ rf[ nrh[i] ] ] < breite[ rf[ nr[j] ] ] ) ) )
      {
        nr[k] = nr[j];
        j++;
      }
      else
      {
        nr[k] = nrh[i];
        i++;
      }
      k++;
    }
    if( j >= b )
    {
      while( i < c-a ) // Rest anhängen
      {
        nr[k] = nrh[i];
        i++;
        k++;
      }
    }
  }
  else // höchstens 2 Zahlen
  {
    if( a + 1 < b )
    {
      if( ( zeit[ nr[a] ] < zeit[ nr[ a + 1 ] ] ) ||
          ( ( zeit[ nr[a] ] == zeit[ nr[ a + 1 ] ] ) &&
            ( breite[ rf[ nr[a] ] ] < breite[ rf[ nr[ a + 1 ] ] ] ) ) )
      {
        h = nr[ a + 1 ];
        nr[ a + 1 ] = nr[a];
        nr[a] = h;
      }
    }
  }
}
*/
// sortiert Einträge von nr[] im Intervall [a; b) so, daß nr[a] den
// größten Eintrag aller nr[] in [a; b) in zeit[] hat, bei gleicher
// Zeit, entscheidet die Breite; nrh ist Hilfsfeld
void merge_modul_zeit_breit(
  double* zeit,
  int*    breite,
  int*    nr,
  int*    nrh,
  int     a,
  int     b )
{
  int c, i, j, k, h;
  if( a + 2 < b )
  {
    c = ( a + b + 1 ) / 2;
    merge_modul_zeit_breit( zeit, breite, nr, nrh, a, c );
    merge_modul_zeit_breit( zeit, breite, nr, nrh, c, b );
    for( i=a; i<c; i++ )
    {
      nrh[i-a] = nr[i];
    }
    i = 0;
    j = c;
    k = a;
    while( ( j < b ) && ( i < c-a ) ) // Pointer durchrücken
    {
      if( ( zeit[ nrh[i] ] < zeit[ nr[j] ] ) ||
          ( ( zeit[ nrh[i] ] == zeit[ nr[j] ] ) &&
            ( breite[ nrh[i] ] < breite[ nr[j] ] ) ) )
      {
        nr[k] = nr[j];
        j++;
      }
      else
      {
        nr[k] = nrh[i];
        i++;
      }
      k++;
    }
    if( j >= b )
    {
      while( i < c-a ) // Rest anhängen
      {
        nr[k] = nrh[i];
        i++;
        k++;
      }
    }
  }
  else // höchstens 2 Zahlen
  {
    if( a + 1 < b )
    {
      if( ( zeit[ nr[a] ] < zeit[ nr[ a + 1 ] ] ) ||
          ( ( zeit[ nr[a] ] == zeit[ nr[ a + 1 ] ] ) &&
            ( breite[ nr[a] ] < breite[ nr[ a + 1 ] ] ) ) )
      {
        h = nr[ a + 1 ];
        nr[ a + 1 ] = nr[a];
        nr[a] = h;
      }
    }
  }
}

// sortiert Einträge von rf[] im Intervall [a; b) so, daß rf[a] den
// größten Eintrag aller rf[] in [a; b) in breite[] hat; in zweiter
// Instanz wird zeit[nr[rf[]]] betrachtet; nrh ist Hilfsfeld
void merge_doppel_indirekt_modul_breit_zeit(
  double* zeit,
  int*    breite,
  int*    nr,
  int*    rf,
  int*    nrh,
  int     a,
  int     b )
{
  int c, i, j, k, h;
  if( a + 2 < b )
  {
    c = ( a + b + 1 ) / 2;
    merge_doppel_indirekt_modul_breit_zeit( zeit, breite, nr, rf,
                                            nrh, a, c );
    merge_doppel_indirekt_modul_breit_zeit( zeit, breite, nr, rf,
                                            nrh, c, b );
    for( i=a; i<c; i++ )
    {
      nrh[i-a] = rf[i];
    }
    i = 0;
    j = c;
    k = a;
    while( ( j < b ) && ( i < c-a ) ) // Pointer durchrücken
    {
      if( ( breite[ nrh[i] ] < breite[ rf[j] ] ) ||
          ( ( breite[ nrh[i] ] == breite[ rf[j] ] ) &&
            ( zeit[ nr[ nrh[i] ] ] < zeit[ nr[ rf[j] ] ] ) ) )
      {
        rf[k] = rf[j];
        j++;
      }
      else
      {
        rf[k] = nrh[i];
        i++;
      }
      k++;
    }
    if( j >= b )
    {
      while( i < c-a ) // Rest anhängen
      {
        rf[k] = nrh[i];
        i++;
        k++;
      }
    }
  }
  else // höchstens 2 Zahlen
  {
    if( a + 1 < b )
    {
      if( ( breite[ rf[a] ] < breite[ rf[ a+1 ] ] ) ||
          ( ( breite[ rf[a] ] == breite[ rf[ a+1 ] ] ) &&
            ( zeit[ nr[ rf[a] ] ] < zeit[ nr[ rf[ a+1 ] ] ] ) ) )
      {
        h = rf[ a + 1 ];
        rf[ a + 1 ] = rf[a];
        rf[a] = h;
      }
    }
  }
}

// sortiert Einträge von nr[] im Intervall [a; b) so, daß nr[a] den
// größten Eintrag aller nr[] in [a; b) in breite[] hat; als zweite
// Instanz wird zeit[] verwendet; nrh ist Hilfsfeld
void merge_modul_breit_zeit(
  double* zeit,
  int*    breite,
  int*    nr,
  int*    nrh,
  int     a,
  int     b )
{
  int c, i, j, k, h;
  if( a + 2 < b )
  {
    c = ( a + b + 1 ) / 2;
    merge_modul_breit_zeit( zeit, breite, nr, nrh, a, c );
    merge_modul_breit_zeit( zeit, breite, nr, nrh, c, b );
    for( i=a; i<c; i++ )
    {
      nrh[i-a] = nr[i];
    }
    i = 0;
    j = c;
    k = a;
    while( ( j < b ) && ( i < c-a ) ) // Pointer durchrücken
    {
      if( ( breite[ nrh[i] ] < breite[ nr[j] ] ) ||
          ( ( breite[ nrh[i] ] == breite[ nr[j] ] ) &&
            ( zeit[ nrh[i] ] < zeit[ nr[j] ] ) ) )
      {
        nr[k] = nr[j];
        j++;
      }
      else
      {
        nr[k] = nrh[i];
        i++;
      }
      k++;
    }
    if( j >= b )
    {
      while( i < c-a ) // Rest anhängen
      {
        nr[k] = nrh[i];
        i++;
        k++;
      }
    }
  }
  else // höchstens 2 Zahlen
  {
    if( a + 1 < b )
    {
      if( ( breite[ nr[a] ] < breite[ nr[ a + 1 ] ] ) ||
          ( ( breite[ nr[a] ] == breite[ nr[ a + 1 ] ] ) &&
            ( zeit[ nr[a] ] < zeit[ nr[ a + 1 ] ] ) ) )
      {
        h = nr[ a + 1 ];
        nr[ a + 1 ] = nr[a];
        nr[a] = h;
      }
    }
  }
}

// mischt Einträge von nr[] im Intervall [a; b) mit denen
// in [b, c) so, daß nr[a] den größten Eintrag aller nr[]
// in [a; b) in zeit[] hat; nrh ist Hilfsfeld
void misch_modul_zeit(
  double* zeit,
  int*    nr,
  int*    nrh,
  int     a,
  int     b,
  int     c )
{
  int i, j, k;
  for( i=a; i<b; i++ )
  {
    nrh[i-a] = nr[i];
  }
  j = b;
  i = 0;
  k = a;
  while( ( j < c ) && ( i < b-a ) ) // Pointer durchrücken
  {
    if( zeit[ nrh[i] ] < zeit[ nr[j] ] )
    {
      nr[k] = nr[j];
      j++;
    }
    else
    {
      nr[k] = nrh[i];
      i++;
    }
    k++;
  }
  if( j >= c )
  {
    while( i < b-a ) // Rest anhängen
    {
      nr[k] = nrh[i];
      i++;
      k++;
    }
  }
}

// mischt Einträge von rf[] im Intervall [a; b) mit denen in [b, c) so,
// daß rf[a] den größten Eintrag aller rf[] in [a; b) in breite[] hat;
// in zweiter Instanz wird zeit[nr[rf[]]] betrachtet; nrh ist Hilfsfeld
void misch_doppel_indirekt_modul_breit_zeit(
  double* zeit,
  int*    breite,
  int*    nr,
  int*    rf,
  int*    nrh,
  int     a,
  int     b,
  int     c )
{
  int i, j, k;
  for( i=a; i<b; i++ )
  {
    nrh[i-a] = rf[i];
  }
  j = b;
  i = 0;
  k = a;
  while( ( j < c ) && ( i < b-a ) ) // Pointer durchrücken
  {
    if( ( breite[ nrh[i] ] < breite[ rf[j] ] ) ||
        ( ( breite[ nrh[i] ] == breite[ rf[j] ] ) &&
          ( zeit[ nr[ nrh[i] ] ] < zeit[ nr[ rf[j] ] ] ) ) )
    {
      rf[k] = rf[j];
      j++;
    }
    else
    {
      rf[k] = nrh[i];
      i++;
    }
    k++;
  }
  if( j >= c )
  {
    while( i < b-a ) // Rest anhängen
    {
      rf[k] = nrh[i];
      i++;
      k++;
    }
  }
}

// mischt Einträge von nr[] im Intervall [a; b) mit denen in [b, c) so,
// daß nr[a] den größten Eintrag aller nr[] in [a; b) in breite[] hat;
// in zweiter Instanz wird zeit[] verwendet; nrh ist Hilfsfeld
void misch_modul_breit_zeit(
  double* zeit,
  int*    breite,
  int*    nr,
  int*    nrh,
  int     a,
  int     b,
  int     c )
{
  int i, j, k;
  for( i=a; i<b; i++ )
  {
    nrh[i-a] = nr[i];
  }
  j = b;
  i = 0;
  k = a;
  while( ( j < c ) && ( i < b-a ) ) // Pointer durchrücken
  {
    if( ( breite[ nrh[i] ] < breite[ nr[j] ] ) ||
        ( ( breite[ nrh[i] ] == breite[ nr[j] ] ) &&
          ( zeit[ nrh[i] ] < zeit[ nr[j] ] ) ) )
    {
      nr[k] = nr[j];
      j++;
    }
    else
    {
      nr[k] = nrh[i];
      i++;
    }
    k++;
  }
  if( j >= c )
  {
    while( i < b-a ) // Rest anhängen
    {
      nr[k] = nrh[i];
      i++;
      k++;
    }
  }
}

// mischt Einträge von nr[] im Intervall [a; b) mit denen in [b, c) so,
// daß nr[a] den größten Eintrag aller nr[] in [a; b) in breite[] hat;
// nrh ist Hilfsfeld
void misch_modul_breit(
  int* breite,
  int* nr,
  int* nrh,
  int  a,
  int  b,
  int  c )
{
  int i, j, k;
  for( i=a; i<b; i++ )
  {
    nrh[i-a] = nr[i];
  }
  j = b;
  i = 0;
  k = a;
  while( ( j < c ) && ( i < b-a ) ) // Pointer durchrücken
  {
    if( breite[ nrh[i] ] < breite[ nr[j] ] )
    {
      nr[k] = nr[j];
      j++;
    }
    else
    {
      nr[k] = nrh[i];
      i++;
    }
    k++;
  }
  if( j >= c )
  {
    while( i < b-a ) // Rest anhängen
    {
      nr[k] = nrh[i];
      i++;
      k++;
    }
  }
}

// sortiert Einträge von nr[] im Intervall [a; b) so, daß nr[a] den
// größten Eintrag aller nr[] in [a; b) in breite[] hat; nrhh ist
// zusätzliches Hilfsfeld der Dimension n = max{ breite[ nr[] ] } + 1
void bucket_modul_breit(
  int* breite,
  int* nr,
  int* nrh,
  int* nrhh,
  int  n,
  int  a,
  int  b )
{
  int i;
  int k, w;
  for( i=0; i<n; i++ )
  {
    nrhh[i] = 0;
  }
  for( k=a; k<b; k++ )
  {
    nrhh[ breite[ nr[k] ] ]++;
    nrh[k] = nr[k];
  }
  w = 0;
  for( i=0; i<n; i++ )
  {
    w += nrhh[i];
    nrhh[i] = b - w; // große nach vorne
  }
  for( k=a; k<b; k++ )
  {
    nr[ nrhh[ breite[ nrh[k] ] ] ] = nrh[k];
    nrhh[ breite[ nrh[k] ] ]++;
  }
}

// sucht das Minimum aus a[] der Dimension n; gibt Position zurück
int such_min( double* a, int n )
{
  double b;
  int i, j;
  b = a[0];
  i = 0;
  for( j=1; j<n; j++ )
  {
    if( a[j] < b )
    {
      b = a[j];
      i = j;
    }
  }
  return i;
}

// sucht das Maximum aus a[] der Dimension n; gibt Position zurück
int such_max( double* a, int n )
{
  double b;
  int i, j;
  b = a[0];
  i = 0;
  for( j=1; j<n; j++ )
  {
    if( a[j] > b )
    {
      b = a[j];
      i = j;
    }
  }
  return i;
}

// sucht die am weitesten links liegende Position in a[] mit einem
// Wert kleiner gleich limit; sind alle Werte in a[] größer als
// limit, so wird die Position des Minimums zurück gegeben
int such_min_limit( double* a, int n, double limit )
{
  double b;
  int i, j;
  if( a[0] <= limit )
  {
    return 0;
  }
  b = a[0];
  i = 0;
  for( j=1; j<n; j++ )
  {
    if( a[j] <= limit )
    {
      return j;
    }
    if( a[j] < b )
    {
      i = j;
      b = a[j];
    }
  }
  return i;
}

// sucht das Maximum aus a[] wobei nur die Elemente b[0] bis b[n-1]
// betrachtet werden; gibt Position zurück
int such_max_indirekt( double* a, int* b, int n )
{
  double c;
  int i, j;
  c = a[ b[0] ];
  i = 0;
  for( j=1; j<n; j++ )
  {
    if( a[ b[j] ] > c )
    {
      c = a[ b[j] ];
      i = j;
    }
  }
  return i;
}

// sucht das Maximum aus a[] wobei nur die Elemente b[m] bis b[n-1]
// betrachtet werden; gibt Position zurück
int such_max_indirekt_range(
  double* a,
  int*    b,
  int     m,
  int     n )
{
  double c;
  int i, j;
  c = a[ b[m] ];
  i = m;
  for( j=m+1; j<n; j++ )
  {
    if( a[ b[j] ] > c )
    {
      c = a[ b[j] ];
      i = j;
    }
  }
  return i;
}

// sucht das Maximum aus a[] wobei nur die Elemente b[c[0]] bis
// b[c[n-1]] betrachtet werden; gibt Position zurück
int such_max_doppel_indirekt(
  double* a,
  int*    b,
  int*    c,
  int     n )
{
  double d;
  int i, j;
  d = a[ b[ c[0] ] ];
  i = 0;
  for( j=1; j<n; j++ )
  {
    if( a[ b[ c[j] ] ] > d )
    {
      d = a[ b[ c[j] ] ];
      i = j;
    }
  }
  return i;
}

// sucht das Minimum aus den Maxima von d+1 nebeneinander liegenden
// Einträgen in a[]; in b wird die Höhe zurückgegeben
int such_min_breit(
  double* a,
  int     n,
  int     d,
  double* b )
{
  double c;
  int h, i, j, k;
  i = such_max( a, d+1 );
  *b = a[i];
  j = 0;
  h = 0;
  i++;
  while( i + d < n )
  {
    k = 0;
    c = -DBL_MAX;
    while( 1 )
    {
      if( a[ i + k ] >= *b )
      {
        i += k + 1;
        break;
      }
      if( a[ i + k ] > c )
      {
        c = a[ i + k ];
        h = i + k;
      }
      if( k == d )
      {
        *b = c;
        j = i;
        i = h + 1;
        break;
      }
      k++;
    }
  }
  return j;
}

// sucht das Minimum aus den Maxima von d+1 nebeneinander liegenden
// Einträgen in a[]; in b wird die Höhe zurückgegeben;
// ist das Ergebnis in b kleiner als limit, so ist *b = limit und
// es wird die am weitesten links liegende Möglichkeit zurückgegeben
int such_min_breit_limit(
  double* a,
  int     n,
  int     d,
  double  limit,
  double* b )
{
  double c;
  int h, i, j, k;
  h = 0;
  i = such_max( a, d+1 );
  *b = a[i];
  if( *b <= limit )
  {
    *b = limit;
    return 0;
  }
  j = 0;
  i++;
  while( i + d < n )
  {
    k = 0;
    c = -DBL_MAX;
    while( 1 )
    {
      if( a[ i + k ] >= *b )
      {
        i += k + 1;
        break;
      }
      if( a[ i + k ] > c )
      {
        c = a[ i + k ];
        h = i + k;
      }
      if( k == d )
      {
        if( c <= limit )
        {
          *b = limit;
          return i;
        }
        *b = c;
        j = i;
        i = h + 1;
        break;
      }
      k++;
    }
  }
  return j;
}

// liefert das Minimum zweier Zahlen
int min( int x, int y )
{
  if( x < y )
  {
    return x;
  }
  return y;
}

// liefert das Maximum zweier Zahlen
int max( int x, int y )
{
  if( x > y )
  {
    return x;
  }
  return y;
}

// Hilfsroutine zu such_heap_min_init (noch n Knoten an a hängen;
// Numerierung ab i)
int such_heap_min_init_help(
  baum*   b,
  double* f,
  ast*    a,
  int     n,
  int     i )
{
  ast* c;
  int j;
  c = insert_ll_baum( b, a );
  if( c == NULL )
  {
    return 1;
  }
  set_baum( c, i );
  f[i] = 0.0;
  if( n == 1 )
  {
    return 0;
  }
  a = insert_rr_baum( b, a );
  if( a == NULL )
  {
    return 1;
  }
  set_baum( a, i+1 );
  f[i+1] = 0.0;
  if( n == 2 )
  {
    return 0;
  }
  j = ( n - 1 ) / 2; // eigentlich j = aufrunden( (n-2)/2 )
  if( such_heap_min_init_help( b, f, c, j, i+2 ) != 0 )
  {
    return 1;
  }
  if( n == 3 )
  {
    return 0;
  }
  return such_heap_min_init_help( b, f, a, n-2-j, i+2+j );
}

// erstellt einen Heap für n Werte
// baum muß initialisiert sein (kein Puffer für Einfügeoperationen
// notwendig)
// Feld f wird auf 0 gesetzt
// Fehler: 1->zu wenig Speicher
int such_heap_min_init(
  baum*   b,
  double* f,
  int     n )
{
  ast* a;
  del_all_baum( b, get_wurzel_baum( b ) );
  a = insert_ll_baum( b, NULL );
  if( a == NULL )
  {
    return 1;
  }
  set_baum( a, 0 );
  f[0] = 0.0;
  if( n == 1 )
  {
    return 0;
  }
  // Heapstruktur anlegen (balancierter Baum)
  return such_heap_min_init_help( b, f, a, n-1, 1 );
}

// Hilfsroutine zu such_heap_min_init (noch n Knoten an a hängen;
// Numerierung ab i; f auf s setzen)
int such_heap_min_init_set_werte_help(
  baum*   b,
  double* f,
  ast*    a,
  int     n,
  int     i,
  double  s )
{
  ast* c;
  int j;
  c = insert_ll_baum( b, a );
  if( c == NULL )
  {
    return 1;
  }
  set_baum( c, i );
  f[i] = s;
  if( n == 1 )
  {
    return 0;
  }
  a = insert_rr_baum( b, a );
  if( a == NULL )
  {
    return 1;
  }
  set_baum( a, i+1 );
  f[i+1] = s;
  if( n == 2 )
  {
    return 0;
  }
  j = ( n - 1 ) / 2; // eigentlich j = aufrunden( (n-2)/2 )
  if( such_heap_min_init_set_werte_help( b, f, c, j, i+2, s ) != 0 )
  {
    return 1;
  }
  if( n == 3 )
  {
    return 0;
  }
  return such_heap_min_init_set_werte_help( b, f, a, n-2-j, i+2+j, s );
}

// erstellt einen Heap für n Werte
// baum muß initialisiert sein (kein Puffer für Einfügeoperationen
// notwendig)
// Feld f wird auf s gesetzt
// Fehler: 1->zu wenig Speicher
int such_heap_min_init_set_werte(
  baum*   b,
  double* f,
  int     n,
  double  s )
{
  ast* a;
  del_all_baum( b, get_wurzel_baum( b ) );
  a = insert_ll_baum( b, NULL );
  if( a == NULL )
  {
    return 1;
  }
  set_baum( a, 0 );
  f[0] = s;
  if( n == 1 )
  {
    return 0;
  }
  // Heapstruktur anlegen (balancierter Baum)
  return such_heap_min_init_set_werte_help( b, f, a, n-1, 1, s );
}

// Hilfsroutine zu such_heap_min_werte_init (noch n Knoten an a
// hängen; Numerierung ab i)
int such_heap_min_init_werte_help(
  baum* b,
  ast*  a,
  int   n,
  int   i )
{
  ast* c;
  int j;
  c = insert_ll_baum( b, a );
  if( c == NULL )
  {
    return 1;
  }
  set_baum( c, i );
  if( n == 1 )
  {
    return 0;
  }
  a = insert_rr_baum( b, a );
  if( a == NULL )
  {
    return 1;
  }
  set_baum( a, i+1 );
  if( n == 2 )
  {
    return 0;
  }
  j = ( n - 1 ) / 2; // eigentlich j = aufrunden( (n-2)/2 )
  if( such_heap_min_init_werte_help( b, c, j, i+2 ) != 0 )
  {
    return 1;
  }
  if( n == 3 )
  {
    return 0;
  }
  return such_heap_min_init_werte_help( b, a, n-2-j, i+2+j );
}

// erstellt einen Heap für n Werte
// baum muß initialisiert sein (kein Puffer für Einfügeoperationen
// notwendig)
// Feld f bleibt unverändert, Heap wird entsprechend aufgebaut
// Fehler: 1->zu wenig Speicher
int such_heap_min_init_werte(
  baum*   b,
  double* f,
  int     n )
{
  ast* a;
  ast* c;
  ast* d;
  int i, j, k;
  del_all_baum( b, get_wurzel_baum( b ) );
  a = insert_ll_baum( b, NULL );
  if( a == NULL )
  {
    return 1;
  }
  set_baum( a, 0 );
  if( n == 1 )
  {
    return 0;
  }
  // Heapstruktur anlegen (balancierter Baum)
  if( such_heap_min_init_werte_help( b, a, n-1, 1 ) != 0 )
  {
    return 1;
  }
  for( i=0; i<n; i++ )
  {
    a = get_wurzel_baum( b );
    while( 1 )
    {
      c = left_baum( a );
      if( c == NULL ) // Blatt erreicht
      {
        set_baum( a, i ); // Wert schreiben
        break; // Baum ist links maximal tiefer als rechts
      }
      j = get_baum( c );
      if( ( j > i ) || ( f[j] <= f[i] ) )
      {
        // c ist kleiner als a
        d = right_baum( a );
        if( d == NULL ) // nur einseitig belegt
        {
          // a und c tauschen
          set_baum( a, j );
          set_baum( c, i );
          // da a kein rechtes Kind hat, ist c Blatt
          break;
        }
        k = get_baum( d );
        if( ( ( j > i ) && ( ( j < k ) || ( k < i ) ) ) ||
            ( ( j < i ) && ( k < i ) &&
              ( ( f[j] < f[k] ) ||
                ( ( f[j] == f[k] ) && ( j < k ) ) ) ) )
        {
          // c ist kleiner als d
          set_baum( a, j ); // Position von c nach oben ziehen
          a = c; // in Richtung c absteigen
          continue;
        }
        // d ist kleiner als c
        set_baum( a, k ); // Position von d nach oben ziehen
        a = d; // in Richtung d absteigen
        continue;
      }
      // a ist kleiner als c
      d = right_baum( a );
      if( d == NULL ) // nur einseitig belegt
      {
        // kein weiterer Abstieg notwendig
        set_baum( a, i );
        break;
      }
      k = get_baum( d );
      if( ( k < i ) && ( f[k] > f[i] ) )
      {
        // a ist kleiner als d
        // kein weiterer Abstieg notwendig
        set_baum( a, i );
        break;
      }
      // d ist kleiner als a
      set_baum( a, k ); // Position von d nach oben ziehen
      a = d; // in Richtung d absteigen
      continue;
    }
  }
  return 0;
}

// liefert die Stelle mit dem Minimum; bei mehreren gleichen wird
// der kleinste Index gewählt
int such_heap_min( baum* b )
{
  return get_baum( get_wurzel_baum( b ) );
}

// setzt den bisher kleinsten Wert im Heap auf h
void set_heap_min( baum* b, double* f, double h )
{
  ast* a;
  ast* c;
  ast* d;
  int i, j, k;
  a = get_wurzel_baum( b );
  i = get_baum( a );
  f[i] = h;
  while( 1 )
  {
    c = left_baum( a );
    if( c == NULL ) // Blatt erreicht
    {
      set_baum( a, i ); // Wert schreiben
      return; // Baum ist links maximal tiefer als rechts
    }
    j = get_baum( c );
    if( ( f[j] < h ) || ( ( f[j] == h ) && ( j < i ) ) )
    {
      // c ist kleiner als a
      d = right_baum( a );
      if( d == NULL ) // nur einseitig belegt
      {
        // a und c tauschen
        set_baum( a, j );
        set_baum( c, i );
        // da a kein rechtes Kind hat, ist c Blatt
        return;
      }
      k = get_baum( d );
      if( ( f[j] < f[k] ) || ( ( f[j] == f[k] ) && ( j < k ) ) )
      {
        // c ist kleiner als d
        set_baum( a, j ); // Position von c nach oben ziehen
        a = c; // in Richtung c absteigen
        continue;
      }
      // d ist kleiner als c
      set_baum( a, k ); // Position von d nach oben ziehen
      a = d; // in Richtung d absteigen
      continue;
    }
    // a ist kleiner als c
    d = right_baum( a );
    if( d == NULL ) // nur einseitig belegt
    {
      // kein weiterer Abstieg notwendig
      set_baum( a, i );
      return;
    }
    k = get_baum( d );
    if( ( h < f[k] ) || ( ( h == f[k] ) && ( i < k ) ) )
    {
      // a ist kleiner als d
      // kein weiterer Abstieg notwendig
      set_baum( a, i );
      return;
    }
    // d ist kleiner als a
    set_baum( a, k ); // Position von d nach oben ziehen
    a = d; // in Richtung d absteigen
    continue;
  }
}
