#ifndef SUCHSORT_OP

#define SUCHSORT_OP

#include <stdio.h>
#include "baum.h"

// sortiert Einträge von nr[] im Intervall [a; b) so, daß nr[a]
// den größten Eintrag aller nr[] in [a; b) in zeit[] hat;
// nrh ist Hilfsfeld
void merge_modul_zeit(
  double* zeit,
  int*    nr,
  int*    nrh,
  int     a,
  int     b );

// sortiert Einträge von rf[] im Intervall [a; b) so, daß rf[a] den
// größten Eintrag aller rf[] in [a; b) in breite[] hat; in zweiter
// Instanz wird zeit[nr[rf[]]] betrachtet; nrh ist Hilfsfeld
void merge_doppel_indirekt_modul_breit_zeit(
  double* zeit,
  int*    breite,
  int*    nr,
  int*    rf,
  int*    nrh,
  int     a,
  int     b );

// sortiert Einträge von nr[] im Intervall [a; b) so, daß nr[a] den
// größten Eintrag aller nr[] in [a; b) in breite[] hat; als zweite
// Instanz wird zeit[] verwendet; nrh ist Hilfsfeld
void merge_modul_breit_zeit(
  double* zeit,
  int*    breite,
  int*    nr,
  int*    nrh,
  int     a,
  int     b );

// mischt Einträge von nr[] im Intervall [a; b) mit denen
// in [b, c) so, daß nr[a] den größten Eintrag aller nr[]
// in [a; b) in zeit[] hat; nrh ist Hilfsfeld
void misch_modul_zeit(
  double* zeit,
  int*    nr,
  int*    nrh,
  int     a,
  int     b,
  int     c );

// mischt Einträge von rf[] im Intervall [a; b) mit denen in [b, c) so,
// daß rf[a] den größten Eintrag aller rf[] in [a; b) in breite[] hat;
// in zweiter Instanz wird zeit[nr[rf[]]] betrachtet; nrh ist Hilfsfeld
void misch_doppel_indirekt_modul_breit_zeit(
  double* zeit,
  int*    breite,
  int*    nr,
  int*    rf,
  int*    nrh,
  int     a,
  int     b,
  int     c );

// mischt Einträge von nr[] im Intervall [a; b) mit denen in [b, c) so,
// daß nr[a] den größten Eintrag aller nr[] in [a; b) in breite[] hat;
// in zweiter Instanz wird zeit[] verwendet; nrh ist Hilfsfeld
void misch_modul_breit_zeit(
  double* zeit,
  int*    breite,
  int*    nr,
  int*    nrh,
  int     a,
  int     b,
  int     c );

// mischt Einträge von nr[] im Intervall [a; b) mit denen in [b, c) so,
// daß nr[a] den größten Eintrag aller nr[] in [a; b) in breite[] hat;
// nrh ist Hilfsfeld
void misch_modul_breit(
  int* breite,
  int* nr,
  int* nrh,
  int  a,
  int  b,
  int  c );

// sortiert Einträge von nr[] im Intervall [a; b) so, daß nr[a] den
// größten Eintrag aller nr[] in [a; b) in breite[] hat; nrhh ist
// zusätzliches Hilfsfeld der Dimension n = max{ breite[ nr[] ] } + 1
void bucket_modul_breit(
  int* breite,
  int* nr,
  int* nrh,
  int* nrhh,
  int  n,
  int  a,
  int  b );

// sucht das Minimum aus a[] der Dimension n; gibt Position zurück
int such_min( double* a, int n );

// sucht das Maximum aus a[] der Dimension n; gibt Position zurück
int such_max( double* a, int n );

// sucht die am weitesten links liegende Position in a[] mit einem
// Wert kleiner gleich limit; sind alle Werte in a[] größer als
// limit, so wird die Position des Minimums zurück gegeben
int such_min_limit( double* a, int n, double limit );

// sucht das Maximum aus a[] wobei nur die Elemente b[0] bis b[n-1]
// betrachtet werden; gibt Position zurück
int such_max_indirekt( double* a, int* b, int n );

// sucht das Maximum aus a[] wobei nur die Elemente b[m] bis b[n-1]
// betrachtet werden; gibt Position zurück
int such_max_indirekt_range(
  double* a,
  int*    b,
  int     m,
  int     n );

// sucht das Maximum aus a[] wobei nur die Elemente b[c[0]] bis
// b[c[n-1]] betrachtet werden; gibt Position zurück
int such_max_doppel_indirekt(
  double* a,
  int*    b,
  int*    c,
  int     n );

// sucht das Minimum aus den Maxima von d+1 nebeneinander liegenden
// Einträgen in a[]; in b wird die Höhe zurückgegeben
int such_min_breit(
  double* a,
  int     n,
  int     d,
  double* b );

// sucht das Minimum aus den Maxima von d+1 nebeneinander liegenden
// Einträgen in a[]; in b wird die Höhe zurückgegeben;
// ist das Ergebnis in b kleiner als limit, so ist *b = limit und
// es wird die am weitesten links liegende Möglichkeit zurückgegeben
int such_min_breit_limit(
  double* a,
  int     n,
  int     d,
  double  limit,
  double* b );

// liefert das Minimum zweier Zahlen
int min( int x, int y );

// liefert das Maximum zweier Zahlen
int max( int x, int y );

// erstellt einen Heap für n Werte
// baum muß initialisiert sein (kein Puffer für Einfügeoperationen
// notwendig)
// Feld f wird auf 0 gesetzt
// Fehler: 1->zu wenig Speicher
int such_heap_min_init(
  baum*   b,
  double* f,
  int     n );

// erstellt einen Heap für n Werte
// baum muß initialisiert sein (kein Puffer für Einfügeoperationen
// notwendig)
// Feld f wird auf s gesetzt
// Fehler: 1->zu wenig Speicher
int such_heap_min_init_set_werte(
  baum*   b,
  double* f,
  int     n,
  double  s );

// erstellt einen Heap für n Werte
// baum muß initialisiert sein (kein Puffer für Einfügeoperationen
// notwendig)
// Feld f bleibt unverändert, Heap wird entsprechend aufgebaut
// Fehler: 1->zu wenig Speicher
int such_heap_min_init_werte(
  baum*   b,
  double* f,
  int     n );

// liefert die Stelle mit dem Minimum; bei mehreren gleichen wird
// der kleinste Index gewählt
int such_heap_min( baum* b );

// setzt den bisher kleinsten Wert im Heap auf h
void set_heap_min( baum* b, double* f, double h );

#endif
