#ifndef ZUFALL_OP

#define ZUFALL_OP

typedef enum
{
  konstant,
  gleich,
  exponential,
  normal
} verteilungsart;

typedef struct verteilung_int
{
  verteilungsart typ; // Verteilung
  int a; // Intervallanfang (bei konstant der Rückgabewert)
  int b; // Intervallende
  double p1; // Verteilungsparameter 1 (entweder alpha oder lambda)
  double p2; // Verteilungsparameter 2 (sigma)
} verteilung_int;

typedef struct verteilung_double
{
  verteilungsart typ; // Verteilung
  double a; // Intervallanfang
  double b; // Intervallende
  double p1; // Verteilungsparameter 1 (entweder alpha oder lambda)
  double p2; // Verteilungsparameter 2 (sigma)
} verteilung_double;

typedef struct verteilung_text
{
  verteilungsart typ; // Verteilung
  char* a; // Intervallanfang (bei konstant der Rückgabewert)
  char* b; // Intervallende
  char* p1; // Verteilungsparameter 1 (entweder alpha oder lambda)
  char* p2; // Verteilungsparameter 2 (sigma)
} verteilung_text;

typedef struct verteilung_double_liste
{
  verteilung_double v;
  int bis;
  struct verteilung_double_liste* next;
} verteilung_double_liste;

// erzeugt gleichverteilte ganze Zahl im Intervall [a, b]
int gleich_int( int a, int b );

// erzeugt gleichverteilte gebrochene Zahl im Intervall [a, b]
double gleich_double( double a, double b );

// erzeugt exponentialverteilte gebrochene Zahl im Intervall [a, b]
// mit Parameter lambda (ist lambda < 0, so entsteht eine
// gespiegelte Kurve)
double exponential_double(
  double a,
  double b,
  double lambda );

// erzeugt exponentialverteilte ganze Zahl im Intervall [a, b] mit
// Parameter lambda (ist lambda < 0, so entsteht eine gespiegelte Kurve)
int exponential_int(
  int    a,
  int    b,
  double lambda );

// erzeugt normalverteilte gebrochene Zahl im Intervall [a, b] mit
// Parametern alpha und sigma
double normal_double(
  double a,
  double b,
  double alpha,
  double sigma );

// erzeugt normalverteilte ganze Zahl im Intervall [a, b] mit
// Parametern alpha und sigma
int normal_int(
  int    a,
  int    b,
  double alpha,
  double sigma );

// liefert entsprechend der Verteilung eine ganze Zufallszahl
int zufall_int( verteilung_int* v );

// liefert entsprechend der Verteilung eine ganze Zufallszahl
// die in der double-verteilung angegebenen Werte sollten ca.
// ganze Zahlen (0.5 bis 1.4999999 wird als 1 gelesen) sein
int zufall_int_from_double( verteilung_double* v );

// liefert entsprechend der Verteilung eine ganze Zufallszahl
// Variablen werden nur am 1. Buchstaben erkannt
// (Unterscheidung von groß/klein)
// in var_wert[i] steht die Belegung von var_name[i]
// Fehler: 1->Syntax im String falsch
//         2->verwendete Variable unbekannt
int zufall_int_var(
  verteilung_text* v,
  char*            var_name,
  int*             var_wert,
  int*             erg );

// liefert entsprechend der Verteilung eine gebrochene Zufallszahl
double zufall_double( verteilung_double* v );

// gibt die größtmögliche Zufallszahl einer Verteilung zurück
int zufall_max_int( verteilung_int* v );

// Konstruktor für Variablen vom Typ verteilung_text
void init_verteilung_text( verteilung_text* v );

// Destruktor für Variablen vom Typ verteilung_text;
// v ist dann wieder initialisiert
void free_verteilung_text( verteilung_text* v );

#endif
