#include <float.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include "../prgx/load.h"
#include "../prgx/save.h"

// Maximale Anzahl an Eingabedateien
#define FILENAME_CNT 31


// Berechnet den Abstand zwischen den Kurven i und j
double berechne_abstand(
  int    gueltig[FILENAME_CNT],
  double mehraufwand[FILENAME_CNT][N_MAX],
  int    i,
  int    j )
{
  double abstand = 0;
  double a;
  int k;

  if( gueltig[i] != i || gueltig[j] != j )
  { // keine Originaldaten
    return DBL_MAX;
  }
  for( k=0; k<N_MAX; k++ )
  {
    a = fabs( mehraufwand[i][k] - mehraufwand[j][k] );
    abstand += a * a;
  }
  return abstand;
}


// Fehler: 1 = Eingabedatei fehlerhaft
//         2 = Fehler beim Schreiben der Ausgabedateien
int main()
{
  // Namen aller maximal möglichen Eingabedateien
  char filenames[FILENAME_CNT][30] =
    { "test1b_komplex1.txt",
      "test1b_komplex2.txt",
      "test1h_komplex1.txt",
      "test1h_komplex2.txt",
      "test3_komplex1.txt",
      "test3_komplex2.txt",
      "test3_komplex3.txt",
      "test5b_komplex1.txt",
      "test5b_komplex2.txt",
      "test5b_komplex3.txt",
      "test5jGuet1MF_komplex1.txt",
      "test5jGuet1MF_komplex2.txt",
      "test5jGuet1MF_komplex3.txt",
      "test5jGuet1P1_komplex1.txt",
      "test5jGuet1P1_komplex2.txt",
      "test5jGuet1P1_komplex3.txt",
      "test5jGuet2MF_komplex1.txt",
      "test5jGuet2MF_komplex2.txt",
      "test5jGuet2MF_komplex3.txt",
      "test5jGuet2P1_komplex1.txt",
      "test5jGuet2P1_komplex2.txt",
      "test5jGuet2P1_komplex3.txt",
      "test5r_komplex1.txt",
      "test5r_komplex2.txt",
      "test5r_komplex3.txt",
      "test5s_komplex1.txt",
      "test5s_komplex2.txt",
      "test5s_komplex3.txt",
      "test5t_komplex1.txt",
      "test5t_komplex2.txt",
      "test5t_komplex3.txt" };
  // Meßdaten (Spalte Mehraufwand)
  double mehraufwand[FILENAME_CNT][N_MAX];
  // Ist Eintrag gültig? -1: ungültig; >=0: Index in Mehraufwand, d.h.
  // gleiche Werte zeigen auf bereits vereinigte Messungen
  int gueltig[FILENAME_CNT];
  // Abstand zwischen zwei Meßreihen; es sind nur die Einträge [i][j]
  // mit i<j gefüllt
  double abstand[FILENAME_CNT][FILENAME_CNT];
  // Anzahl an noch vorhandenen verschiedenen Kurven
  int verschieden = 0;
  // Anzahl an bereits zu dieser Meßreihe vereinigte elementare
  // Meßreihen (Gewicht einer Meßreihe)
  int gewicht[FILENAME_CNT];

  size_t verzeichnisLen = strlen( VERZEICHNIS );
  char filename[100];
  size_t filenameLen;
  int g, h, i, j, k;
  double a;
  FILE* fi;
  double min_abstand = 0;
  int min_i = 0;
  int min_j = 0;

  memcpy( filename, VERZEICHNIS, verzeichnisLen );
  // Lese Eingabedaten
  for( i=0; i<FILENAME_CNT; i++ )
  {
    filenameLen = strlen( filenames[i] );
    memcpy( filename + verzeichnisLen, filenames[i], filenameLen + 1 );
    fi = fopen( filename, "r" );
    if( NULL == fi )
    {
      gueltig[i] = -1;
      continue;
    }
    gueltig[i] = i;
    gewicht[i] = 1;
    verschieden++;
    // Überlesen der Spaltenüberschriften
    if( 0 != lese_next_ret( &fi ) )
    {
      return 1;
    }
    for( j=0; j<N_MAX; j++ )
    {
      // überlese Prozessoranzahl
      if( 0 != lese_next_int_in_line( &fi, &h ) )
      {
        return 1;
      }
      // überlese erzielte Höhe
      if( 0 != lese_next_double_in_line( &fi, &a ) )
      {
        return 1;
      }
      // überlese Gütefaktor
      if( 0 != lese_next_double_in_line( &fi, &a ) )
      {
        return 1;
      }
      // lese Mehraufwand
      if( 0 != lese_next_double_in_line_r( &fi, &mehraufwand[i][j] ) )
      {
        return 1;
      }
    }
    fclose( fi );
  }

  // Berechne paarweise Abstände
  for( i = 0; i < FILENAME_CNT - 1; i++ )
  {
    for( j = i + 1; j < FILENAME_CNT; j++ )
    {
      abstand[i][j] = berechne_abstand( gueltig, mehraufwand, i, j );
    }
  }

  while( verschieden > MAX_VERSCHIEDEN ||
         ( min_abstand < 0.1 && verschieden > 1 ) )
  {
    verschieden--;
    // suche Meßreihen mit geringstem Abstand
    min_abstand = DBL_MAX;
    for( i = 0; i < FILENAME_CNT - 1; i++ )
    {
      for( j = i + 1; j < FILENAME_CNT; j++ )
      {
        if( abstand[i][j] < min_abstand )
        {
          min_abstand = abstand[i][j];
          min_i = i;
          min_j = j;
        }
      }
    }
    // Vereinige Meßreihen min_i und min_j nach min_i
    g = gewicht[min_i] + gewicht[min_j];
    for( i=0; i<N_MAX; i++ )
    {
      mehraufwand[min_i][i] =
        ( mehraufwand[min_i][i] * gewicht[min_i] +
          mehraufwand[min_j][i] * gewicht[min_j] ) / g;
    }
    gewicht[min_i] = g;
    gueltig[min_j] = min_i; // Verweis zur vereinigten Meßreihe
    // Invalidiere alle Abstände zu min_j
    for( i = 0; i < min_j; i++ )
    {
      abstand[i][min_j] = DBL_MAX;
    }
    for( i = min_j + 1; i < FILENAME_CNT; i++ )
    {
      abstand[min_j][i] = DBL_MAX;
    }
    // Berechne Abstände zu min_i neu
    for( i = 0; i < min_i; i++ )
    {
      abstand[i][min_i] =
        berechne_abstand( gueltig, mehraufwand, i, min_i );
    }
    for( i = min_i + 1; i < FILENAME_CNT; i++ )
    {
      abstand[min_i][i] =
        berechne_abstand( gueltig, mehraufwand, min_i, i );
    }
  }
  // Auflösen der Gültig-Verweise (bei mehrstufigen Vereinigungen)
  for( i=0; i<FILENAME_CNT; i++ )
  {
    if( -1 == gueltig[i] ||
        i  == gueltig[i] )
    {
      continue;
    }
    // Da immer zum kleineren Eintrag vereinigt wird, und jetzt von
    // vorne nach hinten durchgegangen wird, kann der Zielwert durch
    // eine Indirektion bestimmt werden
    gueltig[i] = gueltig[ gueltig[i] ];
  }
  // Ausgabe des Ergebnisses
  memcpy( filename, "vereinigung00_a.txt", 20 );
  for( i=0; i<FILENAME_CNT; i++ )
  {
    if( -1 == gueltig[i] )
    {
      continue;
    }
    // Ausgabe der Liste mit enthaltenen Meßreihen
    printf( "%s:\n", filename );
    for( j=i; j<FILENAME_CNT; j++ )
    {
      if( i == gueltig[j] )
      {
        printf( "  %s\n", filenames[j] );
        gueltig[j] = -1;
      }
    }
    // Ausgabe der Meßreihe; sie wird mehrmals ausgegeben um eine
    // Verdeckung unterschiedlicher Meßreihen auszugleichen
    for( j=0; j<verschieden; j++ )
    {
      fi = fopen( filename, "w" );
      if( NULL == fi )
      {
        return 2;
      }
      for( k=j; k<N_MAX; k+=verschieden )
      {
        if( 0 != save_int_s( &fi, k+1 ) )
        {
          return 2;
        }
        if( 0 != save_double_r( &fi, mehraufwand[i][k] ) )
        {
          return 2;
        }
      }
      fclose( fi );
      // Buchstabe weiterzählen
      filename[14]++;
    }
    filename[14] = 'a';
    // Ziffer weiterzählen
    if( '9' == filename[12] )
    {
      filename[11]++;
      filename[12] = '0';
    }
    else
    {
      filename[12]++;
    }
  }
  return 0;
}
