#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "../prgx/baum.h"
#include "../prgx/suchsort.h"

// initialisieren einer Struktur für min_max_suche der Dimension n
// die Elemente von f werden auf 0 gesetzt
// Baum muß initialisiert sein
// Speicher für Feld f der Dimension n muß reserviert sein
// Fehler: 1->zu wenig Speicher
int such_min_max_init( baum* b, double* f, int n )
{
  ast* a;
  int i;
  del_all_baum( b, get_wurzel_baum( b ) );
  for( i=0; i<n; i++ )
  {
    f[i] = 0.0;
    a = insert_ll_baum( b, NULL );
    if( a == NULL )
    {
      return 1;
    }
    set_baum( a, i );
  }
  return 0;
}

// Hilfsroutine zu such_min_max
int such_min_max_help(
  ast*    c,
  double* f,
  int     x,
  int     y,
  int     i,
  ast**   a,
  double* h )
{
  int j, k, p, q;
  ast* d;
  double hh;
  k = get_baum( c );
  if( i > k - x )
  {
    j = 1;
    p = -1; // Avoid warning
  }
  else
  {
    j = 0;
    p = such_min_max_help( left_baum( c ), f, x, k-1, i, a, h );
  }
  if( i > y - k )
  {
    if( j == 1 )
    {
      *a = c;
      *h = f[k];
      return x;
    }
    return p;
  }
  else
  {
    if( j == 1 )
    {
      return such_min_max_help( right_baum( c ), f, k+1, y, i, a, h );
    }
    q = such_min_max_help( right_baum( c ), f, k+1, y, i, &d, &hh );
    if( *h <= hh )
    {
      return p;
    }
    *h = hh;
    *a = d;
    return q;
  }
}

// sucht das Minimum aus den Maxima von i nebeneinander liegenden
// Werten in f mit Dimension n
// Rückgabe: - Position zwischen 0 und n-i
//           - höchste Stelle h
//           - der Wert ast** a wird bei der Variierung der
//             Datenstruktur benötigt
int such_min_max( baum* b, double* f, int n, int i, ast** a, double* h )
{
  return such_min_max_help( get_wurzel_baum( b ), f, 0, n-1, i, a, h );
}

// Hilfsroutine zu set_min_max (Löschen im Baum)
void set_min_max_help( baum* b, ast* a, int* i )
{
  ast* d;
  d = left_baum( a );
  if( d != NULL )
  {
    set_min_max_help( b, d, i );
  }
  if( *i != 0 )
  {
    (*i)--;
    a = del_l_baum( b, a );
    if( a != NULL )
    {
      set_min_max_help( b, a, i );
    }
  }
}

// Verändert min_max_Struktur indem ein gefundenes Minima (daher der
// Wert für a) der Breite i ab x (Rückgabewert der Suchroutine) auf die
// Höhe h gesetzt wird
// Fehler: 1->zu wenig Speicher
int set_min_max( baum* b, double* f, int x, int i, ast* a, double h )
{
  int j;
  ast* c;
  ast* d;
  j = i;
  set_min_max_help( b, a, &j );
  a = get_wurzel_baum( b );
  if( a == NULL ) // es wurde die Breite n eingefügt
  {
    for( j=0; j<i; j++ )
    {
      a = insert_ll_baum( b, NULL );
      if( a == NULL )
      {
        return 1;
      }
      set_baum( a, j );
      f[j] = h;
    }
    return 0;
  }
  while( 1 ) // Einfügestelle suchen
  {
    j = get_baum( a );
    if( x < j ) // Abstiegsrichtung
    {
      if( f[j] >= h ) // Abstiegsbedingung
      {
        c = left_baum( a );
        if( c == NULL )
        {
          // links an a anfügen
          for( j=x; j<x+i; j++ )
          {
            c = insert_ll_baum( b, a );
            if( c == NULL )
            {
              return 1;
            }
            set_baum( c, j );
            f[j] = h;
          }
          return 0;
        }
        a = c;
        continue;
      }
      // anstelle von a einfügen
      // a rechts an linksgeneigte Kette hängen
      c = insert_r_baum( b, a );
      if( c == NULL )
      {
        return 1;
      }
      set_baum( c, x+i-1 );
      f[x+i-1] = h;
      for( j=x+i-2; j>=x; j-- )
      {
        c = insert_ll_baum( b, c );
        if( c == NULL )
        {
          return 1;
        }
        set_baum( c, j );
        f[j] = h;
      }
      // Unterbaum aufteilen
      while( 1 )
      {
        a = left_baum( a );
        if( a == NULL )
        {
          return 0;
        }
        if( get_baum( a ) < x )
        {
          break;
        }
      }
      d = pred_baum( a );
      move_l_baum( a, c );
      while( 1 )
      {
        while( 1 )
        {
          a = right_baum( a );
          if( a == NULL )
          {
            return 0;
          }
          if( get_baum( a ) > x )
          {
            break;
          }
        }
        c = pred_baum( a );
        move_l_baum( a, d );
        while( 1 )
        {
          a = left_baum( a );
          if( a == NULL )
          {
            return 0;
          }
          if( get_baum( a ) < x )
          {
            break;
          }
        }
        d = pred_baum( a );
        move_r_baum( a, c );
      }
      return 0;
    }
    // j < x
    if( f[j] > h ) // Abstiegsbedingung
    {
      c = right_baum( a );
      if( c == NULL )
      {
        // linksgeneigte Kette rechts an a anfügen
        c = insert_rr_baum( b, a );
        if( c == NULL )
        {
          return 1;
        }
        set_baum( c, x+i-1 );
        f[i+x-1] = h;
        for( j=x; j<x+i-1; j++ )
        {
          a = insert_ll_baum( b, c );
          if( a == NULL )
          {
            return 1;
          }
          set_baum( a, j );
          f[j] = h;
        }
        return 0;
      }
      a = c;
      continue;
    }
    // anstelle von a einfügen die gesamte Kette einfügen
    // a hängt links unten
    // größere Knoten als x unterhalb von a werden rechts oben an
    // Kette gehängt
    c = a;
    for( j=x; j<x+i; j++ )
    {
      a = insert_l_baum( b, a );
      if( a == NULL )
      {
        return 1;
      }
      set_baum( a, j );
      f[j] = h;
    }
    // Unterbaum aufteilen
    while( 1 )
    {
      c = right_baum( c );
      if( c == NULL )
      {
        return 0;
      }
      if( get_baum( c ) > x )
      {
        break;
      }
    }
    d = pred_baum( c );
    move_r_baum( c, a );
    while( 1 )
    {
      while( 1 )
      {
        c = left_baum( c );
        if( c == NULL )
        {
          return 0;
        }
        if( get_baum( c ) < x )
        {
          break;
        }
      }
      a = pred_baum( c );
      move_r_baum( c, d );
      while( 1 )
      {
        c = right_baum( c );
        if( c == NULL )
        {
          return 0;
        }
        if( get_baum( c ) > x )
        {
          break;
        }
      }
      d = pred_baum( c );
      move_l_baum( c, a );
    }
  }
}

int main()
{
  baum b;
  ast* a;
  double h;
  double f[N_MAX];
  int i, k;
  int j, p1, p2;
  int n, m;
  time_t t1, t2;
  int durchlaeufe_m_n;
  for( n=10; n<=N_MAX; n*=10 )
  {
    int breite_max = n / BREITE_DIV;
    for( m=10; m<=M_MAX; m*=10 )
    {
      durchlaeufe_m_n = M_MAX / m * N_MAX / n;
      printf( "N=%d, M=%d\n", n, m );
      srand( SEED );
      time( &t1 );
      init_size_baum( &b, n );
      for( p1=0; p1<durchlaeufe_m_n; p1++ )
      {
        for( p2=0; p2<DURCHLAEUFE; p2++ )
        {
          such_min_max_init( &b, f, n );
          for( j=0; j<m; j++ )
          {
            k = rand() % breite_max + 1;
            i = such_min_max( &b, f, n, k, &a, &h );
            h += rand() % 10 + 1;
            set_min_max( &b, f, i, k, a, h );
          }
        }
      }
      free_baum( &b );
      time( &t2 );
      printf( "O(N * M) Algorithmus: %lf s\n",
              ((double)(t2 - t1)) /
              ((double)durchlaeufe_m_n) /
              ((double)DURCHLAEUFE) );
      srand( SEED );
      time( &t1 );
      for( p1=0; p1<durchlaeufe_m_n; p1++ )
      {
        for( p2=0; p2<DURCHLAEUFE; p2++ )
        {
          for( j=0; j<n; j++ )
          {
            f[j] = 0.0;
          }
          for( j=0; j<m; j++ )
          {
            k = rand() % breite_max;
            i = such_min_breit( f, n, k, &h );
            h += rand() % 10 + 1;
            k += i;
            while( i <= k )
            {
              f[i] = h;
              i++;
            }
          }
        }
      }
      time( &t2 );
      printf( "O(N^2 * M) Algorithmus: %lf s\n",
              ((double)(t2 - t1)) /
              ((double)durchlaeufe_m_n) /
              ((double)DURCHLAEUFE) );
    }
  }
  return 0;
}
